# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: continual/rpc/management/v1/management_types.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from continual.rpc.graphql import options_pb2 as continual_dot_rpc_dot_graphql_dot_options__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='continual/rpc/management/v1/management_types.proto',
  package='continual.rpc.management.v1',
  syntax='proto3',
  serialized_options=b'ZJgithub.com/continual-ai/continual/continual/rpc/management/v1;managementv1',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n2continual/rpc/management/v1/management_types.proto\x12\x1b\x63ontinual.rpc.management.v1\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1egoogle/protobuf/duration.proto\x1a#continual/rpc/graphql/options.proto\"\xe0\x02\n\x04User\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05\x65mail\x18\x02 \x01(\t\x12\x16\n\x0e\x65mail_verified\x18\x03 \x01(\x08\x12\x11\n\tfull_name\x18\x04 \x01(\t\x12\x0b\n\x03\x62io\x18\x05 \x01(\t\x12\x10\n\x08location\x18\x06 \x01(\t\x12\x15\n\x08password\x18\x07 \x01(\tB\x03\xb8\x45\x00\x12\x1c\n\x0fservice_account\x18\x08 \x01(\x08\x42\x03\xb8\x45\x01\x12\x34\n\x0bupdate_time\x18\t \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x34\n\x0b\x63reate_time\x18\n \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x10\n\x08\x64isabled\x18\x0b \x01(\x08\x12\x17\n\x0ftrial_available\x18\x0c \x01(\x08\x12\x12\n\nfirst_name\x18\r \x01(\t\x12\x11\n\tlast_name\x18\x0e \x01(\t\"\x9a\x01\n\tPrincipal\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x39\n\x04type\x18\x02 \x01(\x0e\x32+.continual.rpc.management.v1.Principal.Type\x12\x14\n\x0c\x64isplay_name\x18\x03 \x01(\t\".\n\x04Type\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x08\n\x04USER\x10\x01\x12\x0b\n\x07\x41PI_KEY\x10\x02\"\xe3\x02\n\x07Session\x12\x11\n\x04name\x18\x01 \x01(\tB\x03\xb8\x45\x01\x12\x14\n\x07subject\x18\x02 \x01(\tB\x03\xb8\x45\x01\x12\x1b\n\x0elast_access_ip\x18\x03 \x01(\tB\x03\xb8\x45\x01\x12\x39\n\x10last_access_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x15\n\x08login_ip\x18\x05 \x01(\tB\x03\xb8\x45\x01\x12\x33\n\nlogin_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x1d\n\x10login_user_agent\x18\x07 \x01(\tB\x03\xb8\x45\x01\x12\x39\n\x05state\x18\x08 \x01(\x0e\x32*.continual.rpc.management.v1.Session.State\"1\n\x05State\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x41\x43TIVE\x10\x01\x12\x0b\n\x07\x45XPIRED\x10\x02\"d\n\x04Role\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05title\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x15\n\rresource_type\x18\x04 \x01(\t\x12\x13\n\x0bpermissions\x18\x05 \x03(\t\"\x90\x01\n\x0c\x41\x63\x63\x65ssPolicy\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x18\n\x07subject\x18\x02 \x01(\tB\x07\xc2\x45\x04User\x12\x10\n\x08resource\x18\x03 \x01(\t\x12\x15\n\x04role\x18\x04 \x01(\tB\x07\xc2\x45\x04Role\x12/\n\x0b\x63reate_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\xc4\x01\n\x06Invite\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1a\n\x06sender\x18\x02 \x01(\tB\n\xc2\x45\x04User\xb8\x45\x01\x12\r\n\x05\x65mail\x18\x03 \x01(\t\x12\x15\n\x04role\x18\x04 \x01(\tB\x07\xc2\x45\x04Role\x12\x34\n\x0b\x63reate_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x34\n\x0b\x65xpire_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\"\xd9\x04\n\x0cOrganization\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x14\n\x0c\x64isplay_name\x18\x02 \x01(\t\x12\x34\n\x0bupdate_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x34\n\x0b\x63reate_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x45\n\x06status\x18\x06 \x01(\x0e\x32\x30.continual.rpc.management.v1.Organization.StatusB\x03\xb8\x45\x01\x12\x37\n\x0etrial_end_date\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x1a\n\rtrial_credits\x18\x08 \x01(\x03\x42\x03\xb8\x45\x01\x12\x1f\n\x12trial_credits_used\x18\t \x01(\x03\x42\x03\xb8\x45\x01\x12\x11\n\tshow_plan\x18\n \x01(\x08\x12\x18\n\x0bsso_enabled\x18\x0b \x01(\x08\x42\x03\xb8\x45\x01\x12\x14\n\x0crequires_sso\x18\x0c \x01(\x08\x12\x13\n\x0bsso_domains\x18\r \x03(\t\x12\x1c\n\x14\x61llow_external_users\x18\x0e \x01(\x08\x12\x1b\n\x0esso_configured\x18\x0f \x01(\x08\x42\x03\xb8\x45\x01\x12&\n\x19\x64irectory_sync_configured\x18\x10 \x01(\x08\x42\x03\xb8\x45\x01\"A\n\x06Status\x12\x0b\n\x07\x45XPIRED\x10\x00\x12\x10\n\x0c\x41\x43TIVE_TRIAL\x10\x01\x12\n\n\x06\x41\x43TIVE\x10\x02\x12\x0c\n\x08\x44ISABLED\x10\x03\"\xfd\x01\n\nOnboarding\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x34\n\x0b\x63reate_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x34\n\x0bupdate_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x17\n\x0fproject_created\x18\x04 \x01(\x08\x12\x15\n\rmodel_created\x18\x05 \x01(\x08\x12\x1b\n\x13\x66\x65\x61ture_set_created\x18\x06 \x01(\x08\x12\x15\n\rbilling_setup\x18\x07 \x01(\x08\x12\x11\n\tdismissed\x18\x08 \x01(\x08\"\x82\x04\n\x07Project\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x14\n\x0c\x64isplay_name\x18\x02 \x01(\t\x12%\n\x0corganization\x18\x03 \x01(\tB\x0f\xc2\x45\x0cOrganization\x12\x41\n\x07summary\x18\x04 \x01(\x0b\x32+.continual.rpc.management.v1.ProjectSummaryB\x03\xb8\x45\x01\x12\x34\n\x0bupdate_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x34\n\x0b\x63reate_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x1b\n\x13\x64\x65\x66\x61ult_environment\x18\x08 \x01(\t\x12Q\n\x12\x63reate_source_type\x18\t \x01(\x0e\x32\x35.continual.rpc.management.v1.Project.CreateSourceType\x12\x15\n\rcreate_source\x18\n \x01(\t\x12\x31\n\x0fgit_integration\x18\x0b \x01(\tB\x18\xc2\x45\x15ProjectGitIntegration\"C\n\x10\x43reateSourceType\x12\t\n\x05\x42LANK\x10\x00\x12\x11\n\rREPO_EXISTING\x10\x01\x12\x11\n\rREPO_TEMPLATE\x10\x02\"\xbf\x04\n\x0eProjectSummary\x12\x19\n\x11\x66\x65\x61ture_set_count\x18\x01 \x01(\x05\x12\x1d\n\x15\x66\x65\x61ture_set_row_count\x18\x02 \x01(\x03\x12\x19\n\x11\x66\x65\x61ture_set_bytes\x18\x03 \x01(\x03\x12\x15\n\rfeature_count\x18\x04 \x01(\x05\x12\x18\n\x10\x63onnection_count\x18\x05 \x01(\x05\x12\x13\n\x0bmodel_count\x18\x06 \x01(\x05\x12\x1b\n\x13model_version_count\x18\x07 \x01(\x05\x12\x18\n\x10\x65xperiment_count\x18\x08 \x01(\x03\x12\x18\n\x10prediction_count\x18\t \x01(\x03\x12\x45\n\x12\x66\x65\x61ture_set_health\x18\n \x01(\x0b\x32).continual.rpc.management.v1.HealthCounts\x12?\n\x0cmodel_health\x18\x0b \x01(\x0b\x32).continual.rpc.management.v1.HealthCounts\x12\x38\n\x14last_deployment_time\x18\x0c \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x36\n\x12last_training_time\x18\r \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x11\n\trun_count\x18\x0e \x01(\x03\x12\x15\n\rdataset_count\x18\x0f \x01(\x03\x12\x1d\n\x15\x64\x61taset_version_count\x18\x10 \x01(\x03\"\xd6\x0b\n\tDataStore\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x44\n\tbig_query\x18\x02 \x01(\x0b\x32/.continual.rpc.management.v1.DataStore.BigQueryH\x00\x12\x45\n\tsnowflake\x18\x03 \x01(\x0b\x32\x30.continual.rpc.management.v1.DataStore.SnowflakeH\x00\x12\x43\n\x08postgres\x18\x04 \x01(\x0b\x32/.continual.rpc.management.v1.DataStore.PostgresH\x00\x12\x43\n\x08redshift\x18\x05 \x01(\x0b\x32/.continual.rpc.management.v1.DataStore.RedshiftH\x00\x12G\n\ndatabricks\x18\x07 \x01(\x0b\x32\x31.continual.rpc.management.v1.DataStore.DatabricksH\x00\x12\x1a\n\x12is_demo_data_store\x18\x06 \x01(\x08\x1aX\n\x08\x42igQuery\x12\x11\n\tauth_file\x18\x01 \x01(\t\x12\x16\n\x0e\x61uth_file_name\x18\x02 \x01(\t\x12\x0f\n\x07\x64\x61taset\x18\x03 \x01(\t\x12\x10\n\x08location\x18\x04 \x01(\t\x1a\x8e\x02\n\tSnowflake\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x03\x12\x10\n\x08username\x18\x03 \x01(\t\x12\x10\n\x08password\x18\x04 \x01(\t\x12\x10\n\x08\x64\x61tabase\x18\x05 \x01(\t\x12\x11\n\twarehouse\x18\x06 \x01(\t\x12\x0c\n\x04role\x18\x07 \x01(\t\x12L\n\x06params\x18\x08 \x03(\x0b\x32<.continual.rpc.management.v1.DataStore.Snowflake.ParamsEntry\x12\x11\n\tdb_schema\x18\t \x01(\t\x1a-\n\x0bParamsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\xeb\x01\n\x08Postgres\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x03\x12\x10\n\x08username\x18\x03 \x01(\t\x12\x10\n\x08password\x18\x04 \x01(\t\x12\x10\n\x08\x64\x61tabase\x18\x05 \x01(\t\x12K\n\x06params\x18\x06 \x03(\x0b\x32;.continual.rpc.management.v1.DataStore.Postgres.ParamsEntry\x12\x11\n\tdb_schema\x18\x07 \x01(\t\x1a-\n\x0bParamsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\xeb\x01\n\x08Redshift\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x03\x12\x10\n\x08username\x18\x03 \x01(\t\x12\x10\n\x08password\x18\x04 \x01(\t\x12\x10\n\x08\x64\x61tabase\x18\x05 \x01(\t\x12K\n\x06params\x18\x06 \x03(\x0b\x32;.continual.rpc.management.v1.DataStore.Redshift.ParamsEntry\x12\x11\n\tdb_schema\x18\x07 \x01(\t\x1a-\n\x0bParamsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\xed\x01\n\nDatabricks\x12\x0c\n\x04host\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x03\x12\r\n\x05token\x18\x03 \x01(\t\x12\x11\n\thttp_path\x18\x04 \x01(\t\x12\x10\n\x08\x64\x61tabase\x18\x05 \x01(\t\x12M\n\x06params\x18\x06 \x03(\x0b\x32=.continual.rpc.management.v1.DataStore.Databricks.ParamsEntry\x12\x11\n\tdb_schema\x18\x07 \x01(\t\x1a-\n\x0bParamsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x42\x08\n\x06\x63onfig\"V\n\x0cHealthCounts\x12\x15\n\rhealthy_count\x18\x01 \x01(\x05\x12\x17\n\x0funhealthy_count\x18\x02 \x01(\x05\x12\x16\n\x0e\x63ritical_count\x18\x03 \x01(\x05\"\xf0\x02\n\x0b\x45nvironment\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x41\n\x07summary\x18\x02 \x01(\x0b\x32+.continual.rpc.management.v1.ProjectSummaryB\x03\xb8\x45\x01\x12:\n\ndata_store\x18\x03 \x01(\x0b\x32&.continual.rpc.management.v1.DataStore\x12\x34\n\x0bupdate_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x34\n\x0b\x63reate_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x1a\n\x12scheduling_enabled\x18\x06 \x01(\x08\x12\x0e\n\x06source\x18\x07 \x01(\t\x12\x10\n\x08\x65nv_type\x18\x08 \x01(\t\x12\x14\n\x0cis_protected\x18\t \x01(\x08\x12\x14\n\x0cis_ephemeral\x18\n \x01(\x08\"\xa5\x01\n\nFeatureSet\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0bschema_text\x18\x03 \x01(\t\x12/\n\x0b\x63reate_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12/\n\x0bupdate_time\x18\t \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x12\n\x05table\x18\n \x01(\tB\x03\xb8\x45\x01\"\xce\x02\n\x05\x45vent\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x1d\n\x07subject\x18\x02 \x01(\tB\x0c\xc2\x45\tPrincipal\x12\r\n\x05\x65vent\x18\x03 \x01(\t\x12\x10\n\x08resource\x18\x04 \x01(\t\x12\x10\n\x08severity\x18\x05 \x01(\t\x12\x0f\n\x07message\x18\x06 \x01(\t\x12:\n\x04\x64\x61ta\x18\x07 \x03(\x0b\x32,.continual.rpc.management.v1.Event.DataEntry\x12/\n\x0b\x63reate_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0e\n\x06notify\x18\t \x01(\x08\x12\x0b\n\x03run\x18\n \x01(\t\x12\x0e\n\x06\x62\x65\x66ore\x18\x0b \x01(\t\x12\r\n\x05\x61\x66ter\x18\x0c \x01(\t\x1a+\n\tDataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xef\x03\n\rServingMetric\x12\r\n\x05model\x18\x01 \x01(\t\x12\x0e\n\x06parent\x18\x02 \x01(\t\x12\x0b\n\x03tag\x18\x03 \x01(\t\x12\x36\n\x12request_start_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x34\n\x10model_start_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x33\n\x10request_duration\x18\x06 \x01(\x0b\x32\x19.google.protobuf.Duration\x12\x31\n\x0eroute_duration\x18\x07 \x01(\x0b\x32\x19.google.protobuf.Duration\x12\x33\n\x10\x66\x65\x61ture_duration\x18\x08 \x01(\x0b\x32\x19.google.protobuf.Duration\x12\x33\n\x10predict_duration\x18\t \x01(\x0b\x32\x19.google.protobuf.Duration\x12\x31\n\x0emodel_duration\x18\n \x01(\x0b\x32\x19.google.protobuf.Duration\x12\x15\n\rmodel_version\x18\x0b \x01(\t\x12\x13\n\x0bresult_code\x18\x0c \x01(\t\x12\x13\n\x0bresult_data\x18\r \x01(\t\"\xaa\x03\n\x06\x41piKey\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x03key\x18\x02 \x01(\tB\x03\xb8\x45\x01\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x1b\n\ncreated_by\x18\x04 \x01(\tB\x07\xc2\x45\x04User\x12\x34\n\x0b\x63reate_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x1b\n\x0elast_access_ip\x18\x06 \x01(\tB\x03\xb8\x45\x01\x12\x39\n\x10last_access_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x38\n\x05state\x18\x08 \x01(\x0e\x32).continual.rpc.management.v1.ApiKey.State\x12\x35\n\x0crevoked_time\x18\t \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x1b\n\nrevoked_by\x18\n \x01(\tB\x07\xc2\x45\x04User\"2\n\x05State\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06\x41\x43TIVE\x10\x01\x12\x0c\n\x08INACTIVE\x10\x02\"\xe6\x02\n\x05Model\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12\r\n\x05state\x18\x03 \x01(\t\x12\x0e\n\x06\x61uthor\x18\x04 \x01(\t\x12\x14\n\x0c\x64isplay_name\x18\x05 \x01(\t\x12(\n\x0f\x63urrent_version\x18\x06 \x01(\tB\x0f\xc2\x45\x0cModelVersion\x12\x34\n\x0b\x63reate_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x34\n\x0bupdate_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12-\n\x14latest_model_version\x18\t \x01(\tB\x0f\xc2\x45\x0cModelVersion\x12\x33\n\x17latest_batch_prediction\x18\n \x01(\tB\x12\xc2\x45\x0f\x42\x61tchPrediction\x12\x0b\n\x03run\x18\x0b \x01(\t\"\x95\x06\n\tPromotion\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x34\n\x0b\x63reate_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x36\n\rpromoted_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x35\n\x0c\x64\x65moted_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12&\n\rmodel_version\x18\x05 \x01(\tB\x0f\xc2\x45\x0cModelVersion\x12+\n\x12\x62\x61se_model_version\x18\x06 \x01(\tB\x0f\xc2\x45\x0cModelVersion\x12@\n\x05state\x18\x07 \x01(\x0e\x32,.continual.rpc.management.v1.Promotion.StateB\x03\xb8\x45\x01\x12\x1a\n\x12improvement_metric\x18\x08 \x01(\t\x12 \n\x18improvement_metric_value\x18\t \x01(\x01\x12%\n\x1d\x62\x61se_improvement_metric_value\x18\n \x01(\x01\x12\x1f\n\x17improvement_metric_diff\x18\x0b \x01(\x01\x12\x0e\n\x06reason\x18\x0c \x01(\t\x12\x0f\n\x07message\x18\r \x01(\t\x12\x1d\n\x07\x61\x63\x63ount\x18\x0e \x01(\tB\x0c\xc2\x45\tPrincipal\x12\x10\n\x08run_name\x18\x0f \x01(\t\"\xaa\x01\n\x06Reason\x12\x16\n\x12REASON_UNSPECIFIED\x10\x00\x12\n\n\x06UPLIFT\x10\x01\x12$\n NO_PROMOTED_MODEL_VERSION_EXISTS\x10\x02\x12\x11\n\rLATEST_POLICY\x10\x03\x12\x12\n\x0eUNDERPERFORMED\x10\x04\x12\n\n\x06MANUAL\x10\x05\x12\n\n\x06\x46ORCED\x10\x06\x12\x17\n\x13INCOMPATIBLE_SCHEMA\x10\x07\"9\n\x05State\x12\x15\n\x11STATE_UNSPECIFIED\x10\x00\x12\r\n\tSUCCEEDED\x10\x01\x12\n\n\x06\x46\x41ILED\x10\x02\"\xb3\x06\n\x0cModelVersion\x12\x11\n\x04name\x18\x01 \x01(\tB\x03\xb8\x45\x01\x12\x10\n\x08run_name\x18\x02 \x01(\t\x12&\n\x0f\x65xperiment_name\x18\x04 \x01(\tB\r\xc2\x45\nExperiment\x12\x43\n\x05state\x18\x05 \x01(\x0e\x32/.continual.rpc.management.v1.ModelVersion.StateB\x03\xb8\x45\x01\x12\x1a\n\rerror_message\x18\x06 \x01(\tB\x03\xb8\x45\x01\x12\x18\n\x0bstack_trace\x18\x07 \x01(\tB\x03\xb8\x45\x01\x12\x34\n\x0b\x63reate_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x34\n\x0bupdate_time\x18\t \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12.\n\nstart_time\x18\n \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x0b \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x1f\n\x12training_row_count\x18\x0c \x01(\x04\x42\x03\xb8\x45\x01\x12!\n\x14validation_row_count\x18\r \x01(\x04\x42\x03\xb8\x45\x01\x12\x1b\n\x0etest_row_count\x18\x0e \x01(\x04\x42\x03\xb8\x45\x01\x12\x1f\n\x12performance_metric\x18\x0f \x01(\tB\x03\xb8\x45\x01\x12#\n\x16performance_metric_val\x18\x10 \x01(\x01\x42\x03\xb8\x45\x01\x12\x1f\n\tpromotion\x18\x11 \x01(\tB\x0c\xc2\x45\tPromotion\x12\x10\n\x08promoted\x18\x12 \x01(\x08\"\xb6\x01\n\x05State\x12\x15\n\x11STATE_UNSPECIFIED\x10\x00\x12\x0b\n\x07\x43REATED\x10\x01\x12\x0b\n\x07PENDING\x10\x02\x12\x11\n\rPREPROCESSING\x10\x03\x12\x0c\n\x08TRAINING\x10\x04\x12\x12\n\x0ePOSTPROCESSING\x10\x05\x12\x0e\n\nCANCELLING\x10\x06\x12\r\n\tSUCCEEDED\x10\x07\x12\n\n\x06\x46\x41ILED\x10\x08\x12\r\n\tCANCELLED\x10\t\x12\r\n\tPROFILING\x10\n\"\x84\x01\n\x0f\x43onfusionMatrix\x12\x14\n\x0c\x64isplay_name\x18\x03 \x03(\t\x12=\n\x03row\x18\x02 \x03(\x0b\x32\x30.continual.rpc.management.v1.ConfusionMatrix.Row\x1a\x1c\n\x03Row\x12\x15\n\rexample_count\x18\x01 \x03(\x05\"\xd5\x01\n\x18\x42ooleanEvaluationMetrics\x12\x10\n\x08log_loss\x18\x01 \x01(\x01\x12\x10\n\x08\x61\x63\x63uracy\x18\x02 \x01(\x01\x12\x0f\n\x07roc_auc\x18\x03 \x01(\x01\x12\r\n\x05\x65rror\x18\x04 \x01(\x01\x12\x46\n\x10\x63onfusion_matrix\x18\x05 \x01(\x0b\x32,.continual.rpc.management.v1.ConfusionMatrix\x12\x11\n\tprecision\x18\x06 \x01(\x01\x12\x0e\n\x06recall\x18\x07 \x01(\x01\x12\n\n\x02\x66\x31\x18\x08 \x01(\x01\"\x86\x01\n\x15TextEvaluationMetrics\x12\x10\n\x08log_loss\x18\x01 \x01(\x01\x12\n\n\x02\x66\x31\x18\x02 \x01(\x01\x12\x12\n\nperplexity\x18\x03 \x01(\x01\x12\x15\n\redit_distance\x18\x04 \x01(\x01\x12\x16\n\x0etoken_accuracy\x18\x05 \x01(\x01\x12\x0c\n\x04\x62leu\x18\x06 \x01(\x01\"\xe6\x03\n\x1c\x43\x61tegoricalEvaluationMetrics\x12\x10\n\x08log_loss\x18\x01 \x01(\x01\x12\x10\n\x08\x61\x63\x63uracy\x18\x02 \x01(\x01\x12\x0f\n\x07roc_auc\x18\x03 \x01(\x01\x12\x18\n\x10multiclass_error\x18\x04 \x01(\x01\x12\x46\n\x10\x63onfusion_matrix\x18\x05 \x01(\x0b\x32,.continual.rpc.management.v1.ConfusionMatrix\x12\x11\n\tprecision\x18\x06 \x01(\x01\x12\x17\n\x0fprecision_micro\x18\x07 \x01(\x01\x12\x17\n\x0fprecision_macro\x18\x08 \x01(\x01\x12\x1a\n\x12precision_weighted\x18\t \x01(\x01\x12\x0e\n\x06recall\x18\n \x01(\x01\x12\x14\n\x0crecall_micro\x18\x0b \x01(\x01\x12\x14\n\x0crecall_macro\x18\x0c \x01(\x01\x12\x17\n\x0frecall_weighted\x18\r \x01(\x01\x12\n\n\x02\x66\x31\x18\x0e \x01(\x01\x12\x10\n\x08\x66\x31_micro\x18\x0f \x01(\x01\x12\x10\n\x08\x66\x31_macro\x18\x10 \x01(\x01\x12\x13\n\x0b\x66\x31_weighted\x18\x11 \x01(\x01\x12\x19\n\x11\x62\x61lanced_accuracy\x18\x12 \x01(\x01\x12\x19\n\x11\x61verage_precision\x18\x13 \x01(\x01\"\xbb\x01\n\x17NumberEvaluationMetrics\x12\x0c\n\x04rmse\x18\x01 \x01(\x01\x12\x0b\n\x03mae\x18\x02 \x01(\x01\x12\n\n\x02r2\x18\x03 \x01(\x01\x12\x11\n\tprecision\x18\x04 \x01(\x01\x12\x0e\n\x06recall\x18\x05 \x01(\x01\x12\x0c\n\x04mape\x18\x06 \x01(\x01\x12\r\n\x05smape\x18\x07 \x01(\x01\x12\x1a\n\x12mean_squared_error\x18\x08 \x01(\x01\x12\x1d\n\x15median_absolute_error\x18\t \x01(\x01\"\xc6\x02\n\x11\x45valuationMetrics\x12H\n\x07\x62oolean\x18\x01 \x01(\x0b\x32\x35.continual.rpc.management.v1.BooleanEvaluationMetricsH\x00\x12\x46\n\x06number\x18\x02 \x01(\x0b\x32\x34.continual.rpc.management.v1.NumberEvaluationMetricsH\x00\x12\x42\n\x04text\x18\x03 \x01(\x0b\x32\x32.continual.rpc.management.v1.TextEvaluationMetricsH\x00\x12P\n\x0b\x63\x61tegorical\x18\x04 \x01(\x0b\x32\x39.continual.rpc.management.v1.CategoricalEvaluationMetricsH\x00\x42\t\n\x07metrics\"\xcd\x04\n\nExperiment\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08run_name\x18\x02 \x01(\t\x12<\n\x05state\x18\x03 \x01(\x0e\x32-.continual.rpc.management.v1.Experiment.State\x12/\n\x0b\x63reate_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12/\n\x0bupdate_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12.\n\nstart_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x1f\n\x17performance_metric_name\x18\x08 \x01(\t\x12\x1e\n\x16performance_metric_val\x18\t \x01(\x01\x12\x17\n\x0ftraining_config\x18\n \x01(\t\x12\x15\n\rerror_message\x18\x0b \x01(\t\x12\x13\n\x0bstack_trace\x18\x0c \x01(\t\"\x9a\x01\n\x05State\x12\x15\n\x11STATE_UNSPECIFIED\x10\x00\x12\x0b\n\x07PENDING\x10\x01\x12\x11\n\rPREPROCESSING\x10\x02\x12\x0c\n\x08TRAINING\x10\x03\x12\x12\n\x0ePOSTPROCESSING\x10\x04\x12\x0e\n\nCANCELLING\x10\x05\x12\r\n\tSUCCEEDED\x10\x06\x12\n\n\x06\x46\x41ILED\x10\x07\x12\r\n\tCANCELLED\x10\x08\"\x1d\n\x0cTextSequence\x12\r\n\x05value\x18\x01 \x03(\t\"\x1f\n\x0eNumberSequence\x12\r\n\x05value\x18\x01 \x03(\x01\"`\n\x10SignatureFeature\x12\x16\n\x0equalified_name\x18\x01 \x01(\t\x12\x34\n\x04type\x18\x02 \x01(\x0e\x32&.continual.rpc.management.v1.FieldType\"\x90\x02\n\x0cQueryFeature\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x16\n\x0equalified_name\x18\x02 \x01(\t\x12\x34\n\x08relation\x18\x03 \x01(\x0b\x32\".continual.rpc.management.v1.Query\x12\r\n\x05value\x18\x04 \x01(\t\x12\x34\n\x04type\x18\x05 \x01(\x0e\x32&.continual.rpc.management.v1.FieldType\x12\r\n\x05\x64type\x18\x06 \x01(\t\x12\r\n\x05model\x18\x07 \x01(\x08\x12\r\n\x05index\x18\x08 \x01(\x08\x12\x12\n\ntime_index\x18\t \x01(\x08\x12\x0f\n\x07\x65xclude\x18\n \x01(\x08\x12\r\n\x05split\x18\x0b \x01(\x08\"\xac\x01\n\x05Query\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\tentity_id\x18\x02 \x01(\t\x12\x12\n\ntable_name\x18\x03 \x01(\t\x12\x39\n\x06\x66ields\x18\x04 \x03(\x0b\x32).continual.rpc.management.v1.QueryFeature\x12\x33\n\x07related\x18\x05 \x03(\x0b\x32\".continual.rpc.management.v1.Query\"\x9d\x04\n\x0f\x42\x61tchPrediction\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x41\n\x05state\x18\x02 \x01(\x0e\x32\x32.continual.rpc.management.v1.BatchPrediction.State\x12&\n\rmodel_version\x18\x03 \x01(\tB\x0f\xc2\x45\x0cModelVersion\x12/\n\x0b\x63reate_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12.\n\nstart_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x18\n\x10prediction_count\x18\x07 \x01(\x03\x12\x15\n\rerror_message\x18\x08 \x01(\t\x12\x13\n\x0bstack_trace\x18\t \x01(\t\x12\x10\n\x08run_name\x18\n \x01(\t\"\xa9\x01\n\x05State\x12\x15\n\x11STATE_UNSPECIFIED\x10\x00\x12\x0b\n\x07\x43REATED\x10\x01\x12\x0b\n\x07PENDING\x10\x02\x12\x11\n\rPREPROCESSING\x10\x03\x12\x0e\n\nPROCESSING\x10\x04\x12\x12\n\x0ePOSTPROCESSING\x10\x05\x12\x0e\n\nCANCELLING\x10\x06\x12\r\n\tSUCCEEDED\x10\x07\x12\n\n\x06\x46\x41ILED\x10\x08\x12\r\n\tCANCELLED\x10\t\"\x1f\n\tSignedURL\x12\x12\n\nsigned_url\x18\x01 \x01(\t\"<\n\x06\x45ntity\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06models\x18\x02 \x03(\t\x12\x14\n\x0c\x66\x65\x61ture_sets\x18\x03 \x03(\t\"\x86\x02\n\x07\x44\x61taset\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12\r\n\x05state\x18\x03 \x01(\t\x12\x0e\n\x06\x61uthor\x18\x04 \x01(\t\x12\x14\n\x0c\x64isplay_name\x18\x05 \x01(\t\x12*\n\x0f\x63urrent_version\x18\x06 \x01(\tB\x11\xc2\x45\x0e\x44\x61tasetVersion\x12\x34\n\x0b\x63reate_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x34\n\x0bupdate_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x0b\n\x03run\x18\t \x01(\t\"\x9c\x01\n\x0e\x44\x61tasetVersion\x12\x11\n\x04name\x18\x01 \x01(\tB\x03\xb8\x45\x01\x12\x0b\n\x03run\x18\x02 \x01(\t\x12\x34\n\x0b\x63reate_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x34\n\x0bupdate_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\"p\n\x18\x44\x61tasetVersionAssignment\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x15\n\rresource_name\x18\x02 \x01(\t\x12/\n\x0b\x63reate_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\x93\x01\n\x0b\x44\x61taProfile\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x45\n\rdataset_stats\x18\x02 \x03(\x0b\x32..continual.rpc.management.v1.DatasetStatsEntry\x12/\n\x0b\x63reate_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\x91\x04\n\x11\x44\x61tasetStatsEntry\x12\x12\n\nentry_name\x18\x01 \x01(\t\x12\x14\n\x0cnum_examples\x18\x02 \x01(\x04\x12J\n\x12\x63orrelation_matrix\x18\x03 \x01(\x0b\x32..continual.rpc.management.v1.CorrelationMatrix\x12>\n\x0cstring_stats\x18\x04 \x03(\x0b\x32(.continual.rpc.management.v1.StringStats\x12@\n\rnumeric_stats\x18\x05 \x03(\x0b\x32).continual.rpc.management.v1.NumericStats\x12\x44\n\x0ftimestamp_stats\x18\x06 \x03(\x0b\x32+.continual.rpc.management.v1.TimestampStats\x12H\n\x11\x63\x61tegorical_stats\x18\x07 \x03(\x0b\x32-.continual.rpc.management.v1.CategoricalStats\x12\x16\n\x0e\x64uplicate_rows\x18\x08 \x01(\x04\x12\x19\n\x11\x64uplicate_indexes\x18\t \x01(\x04\x12\x41\n\x0e\x64\x61taset_ranges\x18\n \x03(\x0b\x32).continual.rpc.management.v1.DatasetRange\"z\n\x0c\x44\x61tasetRange\x12\x0c\n\x04name\x18\x01 \x01(\t\x12.\n\nstart_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\x89\x01\n\x11\x43orrelationMatrix\x12\x15\n\rdisplay_names\x18\x01 \x03(\t\x12@\n\x04rows\x18\x02 \x03(\x0b\x32\x32.continual.rpc.management.v1.CorrelationMatrix.Row\x1a\x1b\n\x03Row\x12\x14\n\x0c\x63orrelations\x18\x01 \x03(\x02\"\xba\x01\n\tHistogram\x12\x0f\n\x07num_nan\x18\x01 \x01(\x04\x12\x15\n\rnum_undefined\x18\x02 \x01(\x04\x12>\n\x07\x62uckets\x18\x03 \x03(\x0b\x32-.continual.rpc.management.v1.Histogram.Bucket\x1a\x45\n\x06\x42ucket\x12\x11\n\tlow_value\x18\x01 \x01(\x01\x12\x12\n\nhigh_value\x18\x02 \x01(\x01\x12\x14\n\x0csample_count\x18\x04 \x01(\x03\"\x84\x02\n\x12TimestampHistogram\x12\x0f\n\x07num_nan\x18\x01 \x01(\x04\x12\x15\n\rnum_undefined\x18\x02 \x01(\x04\x12G\n\x07\x62uckets\x18\x03 \x03(\x0b\x32\x36.continual.rpc.management.v1.TimestampHistogram.Bucket\x1a}\n\x06\x42ucket\x12-\n\tlow_value\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12.\n\nhigh_value\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x14\n\x0csample_count\x18\x04 \x01(\x03\"\x9d\x01\n\x17\x43\x61tegoricalDistribution\x12Q\n\ncategories\x18\x01 \x03(\x0b\x32=.continual.rpc.management.v1.CategoricalDistribution.Category\x1a/\n\x08\x43\x61tegory\x12\r\n\x05label\x18\x04 \x01(\t\x12\x14\n\x0csample_count\x18\x05 \x01(\x03\"w\n\x0bStringStats\x12\x13\n\x0b\x63olumn_name\x18\x01 \x01(\t\x12\x17\n\x0f\x64istinct_values\x18\x02 \x01(\x03\x12\x12\n\nnull_count\x18\x03 \x01(\x03\x12\x12\n\nmin_length\x18\x04 \x01(\x03\x12\x12\n\nmax_length\x18\x05 \x01(\x03\"\xac\x01\n\x10\x43\x61tegoricalStats\x12\x13\n\x0b\x63olumn_name\x18\x01 \x01(\t\x12\x17\n\x0f\x64istinct_values\x18\x02 \x01(\x03\x12\x12\n\nnull_count\x18\x03 \x01(\x03\x12V\n\x18\x63\x61tegorical_distribution\x18\x04 \x01(\x0b\x32\x34.continual.rpc.management.v1.CategoricalDistribution\"\xf2\x01\n\x0eTimestampStats\x12\x13\n\x0b\x63olumn_name\x18\x01 \x01(\t\x12\x17\n\x0f\x64istinct_values\x18\x02 \x01(\x03\x12\x12\n\nnull_count\x18\x03 \x01(\x03\x12\'\n\x03min\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\'\n\x03max\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12L\n\x13timestamp_histogram\x18\x06 \x01(\x0b\x32/.continual.rpc.management.v1.TimestampHistogram\"\xac\x02\n\x0cNumericStats\x12\x13\n\x0b\x63olumn_name\x18\x01 \x01(\t\x12\x17\n\x0f\x64istinct_values\x18\x02 \x01(\x03\x12\x0b\n\x03min\x18\x03 \x01(\x01\x12\x0b\n\x03max\x18\x04 \x01(\x01\x12\n\n\x02sd\x18\x05 \x01(\x01\x12\x0c\n\x04mean\x18\x06 \x01(\x01\x12\r\n\x05\x63ount\x18\x07 \x01(\x03\x12\x14\n\x0cpercentile25\x18\x08 \x01(\x01\x12\x14\n\x0cpercentile50\x18\t \x01(\x01\x12\x14\n\x0cpercentile75\x18\n \x01(\x01\x12\x12\n\nnull_count\x18\x0b \x01(\x03\x12\x0f\n\x07num_nan\x18\x0c \x01(\x03\x12\x44\n\x14num_values_histogram\x18\r \x01(\x0b\x32&.continual.rpc.management.v1.Histogram\"\xa7\x01\n\x0e\x42illingAccount\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05\x65mail\x18\x02 \x01(\t\x12\x0c\n\x04memo\x18\x03 \x01(\t\x12\x34\n\x0bupdate_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x34\n\x0b\x63reate_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\"\xdb\x01\n\x0cSubscription\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x17\n\x05price\x18\x02 \x01(\tB\x08\xc2\x45\x05Price\x12(\n\x0epayment_method\x18\x03 \x01(\tB\x10\xc2\x45\rPaymentMethod\x12\x34\n\x0bupdate_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x34\n\x0b\x63reate_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.TimestampB\x03\xb8\x45\x01\x12\x0e\n\x06\x61\x63tive\x18\x06 \x01(\x08\"\xb8\x01\n\x12ProjectUsageRecord\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07\x63redits\x18\x02 \x01(\x03\x12Q\n\ttypeCount\x18\x03 \x03(\x0b\x32>.continual.rpc.management.v1.ProjectUsageRecord.TypeCountEntry\x1a\x30\n\x0eTypeCountEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x03:\x02\x38\x01\"\x99\x01\n\x11\x43reditUsageRecord\x12*\n\x06period\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0f\n\x07\x63redits\x18\x02 \x01(\x03\x12G\n\x0eproject_usages\x18\x03 \x03(\x0b\x32/.continual.rpc.management.v1.ProjectUsageRecord\"\xc8\x01\n\x0fInvoiceLineItem\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0e\n\x06\x61mount\x18\x02 \x01(\x03\x12\x10\n\x08quantity\x18\x03 \x01(\x03\x12\x10\n\x08\x63urrency\x18\x04 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x05 \x01(\t\x12\x30\n\x0cperiod_start\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12.\n\nperiod_end\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\x9a\x02\n\x07Invoice\x12\x0c\n\x04name\x18\x01 \x01(\t\x12(\n\x04\x64\x61te\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0e\n\x06\x61mount\x18\x04 \x01(\x03\x12\x30\n\x0cperiod_start\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12.\n\nperiod_end\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0e\n\x06status\x18\x07 \x01(\t\x12\x13\n\x0binvoice_url\x18\x08 \x01(\t\x12@\n\nline_items\x18\t \x03(\x0b\x32,.continual.rpc.management.v1.InvoiceLineItem\"Q\n\x05Price\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05title\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x16\n\x0e\x64irect_invoice\x18\x04 \x01(\x08\"\xa2\x02\n\rPaymentMethod\x12\x0c\n\x04name\x18\x01 \x01(\t\x12R\n\x0f\x62illing_details\x18\x02 \x01(\x0b\x32\x39.continual.rpc.management.v1.PaymentMethod.BillingDetails\x12=\n\x04\x63\x61rd\x18\x03 \x01(\x0b\x32/.continual.rpc.management.v1.PaymentMethod.Card\x1a%\n\x0e\x42illingDetails\x12\x13\n\x0bpostal_code\x18\x01 \x01(\t\x1aI\n\x04\x43\x61rd\x12\r\n\x05\x62rand\x18\x01 \x01(\t\x12\r\n\x05last4\x18\x02 \x01(\t\x12\x10\n\x08\x65xp_year\x18\x03 \x01(\x03\x12\x11\n\texp_month\x18\x04 \x01(\x03\"\xd5\x01\n\x08\x41rtifact\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0b\n\x03run\x18\x02 \x01(\t\x12\x10\n\x08\x65xternal\x18\x03 \x01(\x08\x12\x0c\n\x04path\x18\x04 \x01(\t\x12\x0b\n\x03url\x18\x05 \x01(\t\x12\x11\n\tmime_type\x18\x07 \x01(\t\x12/\n\x0b\x63reate_time\x18\t \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12/\n\x0bupdate_time\x18\n \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0c\n\x04size\x18\x0b \x01(\x04\"\xa4\x01\n\x07\x43ronJob\x12\x14\n\x0corganization\x18\x01 \x01(\t\x12\r\n\x05model\x18\x02 \x01(\t\x12\x10\n\x08job_type\x18\x03 \x01(\t\x12\x10\n\x08schedule\x18\x04 \x01(\t\x12\x1b\n\x13organization_status\x18\x05 \x01(\t\x12\x0e\n\x06status\x18\x06 \x01(\t\x12\x13\n\x0bworkflow_id\x18\x07 \x01(\t\x12\x0e\n\x06run_id\x18\x08 \x01(\t\"i\n\x17UserIntegrationProvider\x12\x10\n\x08provider\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\r\n\x05title\x18\x03 \x01(\t\x12\x1f\n\x17integration_install_url\x18\x04 \x01(\t\"\x84\x04\n\x0fUserIntegration\x12\x0c\n\x04name\x18\x01 \x01(\t\x12/\n\x0b\x63reate_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12/\n\x0bupdate_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x10\n\x08provider\x18\x04 \x01(\t\x12\x15\n\rprovider_name\x18\x05 \x01(\t\x12\x0e\n\x06\x63onfig\x18\x06 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x07 \x01(\t\x12\x34\n\x10\x64\x61ta_update_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x14\n\x0c\x61\x63\x63ount_name\x18\t \x01(\t\x12\x19\n\x11\x61\x63\x63ount_full_name\x18\n \x01(\t\x12\x1d\n\x15\x61\x63\x63ount_primary_email\x18\x0b \x01(\t\x12\x1a\n\x12\x61\x63\x63ount_avatar_url\x18\x0c \x01(\t\x12\x13\n\x0b\x61\x63\x63ount_url\x18\r \x01(\t\x12\x1a\n\x12\x61\x63\x63ount_config_url\x18\x0e \x01(\t\x12\x33\n\x0f\x64isconnect_time\x18\x0f \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x19\n\x11\x64isconnect_reason\x18\x10 \x01(\t\x12\x17\n\x0f\x64isconnect_data\x18\x11 \x01(\t\"\x91\x01\n\x0bGitProvider\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04type\x18\x02 \x01(\t\x12\x10\n\x08\x61pp_name\x18\x03 \x01(\t\x12\x17\n\x0f\x61pp_external_id\x18\x04 \x01(\t\x12\x19\n\x11installation_name\x18\x05 \x01(\t\x12 \n\x18installation_external_id\x18\x06 \x01(\t\"\xdd\x01\n\x14GitProviderNamespace\x12\x0c\n\x04name\x18\x01 \x01(\t\x12:\n\x08provider\x18\x02 \x01(\x0b\x32(.continual.rpc.management.v1.GitProvider\x12\x13\n\x0b\x65xternal_id\x18\x03 \x01(\t\x12\x0c\n\x04type\x18\x04 \x01(\t\x12\x0c\n\x04slug\x18\x05 \x01(\t\x12\r\n\x05title\x18\x06 \x01(\t\x12\x12\n\navatar_url\x18\x07 \x01(\t\x12\x12\n\nconfig_url\x18\x08 \x01(\t\x12\x13\n\x0b\x61\x63\x63\x65ss_type\x18\t \x01(\t\"\xa9\x02\n\x15GitProviderRepository\x12\x0c\n\x04name\x18\x01 \x01(\t\x12M\n\x12provider_namespace\x18\x02 \x01(\x0b\x32\x31.continual.rpc.management.v1.GitProviderNamespace\x12\x13\n\x0b\x65xternal_id\x18\x03 \x01(\t\x12\x0c\n\x04slug\x18\x04 \x01(\t\x12\r\n\x05title\x18\x05 \x01(\t\x12\x1b\n\x13\x64\x65\x66\x61ult_branch_name\x18\x06 \x01(\t\x12\x13\n\x0b\x61\x63\x63\x65ss_type\x18\x07 \x01(\t\x12\x0b\n\x03url\x18\x08 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\t \x01(\t\x12\x34\n\x10\x64\x61ta_update_time\x18\n \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\xa7\x04\n\x15ProjectGitIntegration\x12\x0c\n\x04name\x18\x01 \x01(\t\x12/\n\x0b\x63reate_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12/\n\x0bupdate_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0f\n\x07project\x18\x04 \x01(\t\x12>\n\x0cgit_provider\x18\x05 \x01(\x0b\x32(.continual.rpc.management.v1.GitProvider\x12Q\n\x16git_provider_namespace\x18\x06 \x01(\x0b\x32\x31.continual.rpc.management.v1.GitProviderNamespace\x12S\n\x17git_provider_repository\x18\x07 \x01(\x0b\x32\x32.continual.rpc.management.v1.GitProviderRepository\x12\x17\n\x0f\x63reated_by_user\x18\x08 \x01(\t\x12#\n\x1b\x63reated_by_user_integration\x18\t \x01(\t\x12\x33\n\x0f\x64isconnect_time\x18\n \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x19\n\x11\x64isconnect_reason\x18\x0b \x01(\t\x12\x17\n\x0f\x64isconnect_data\x18\x0c \x01(\t\"\xe0\x02\n\x06Metric\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x14\n\x0c\x64isplay_name\x18\x02 \x01(\t\x12\x0b\n\x03run\x18\x03 \x01(\t\x12/\n\x0b\x63reate_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12/\n\x0bupdate_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12@\n\tdirection\x18\x06 \x01(\x0e\x32-.continual.rpc.management.v1.Metric.Direction\x12\x38\n\x06values\x18\x07 \x03(\x0b\x32(.continual.rpc.management.v1.MetricValue\x12\x12\n\ncreated_by\x18\x08 \x01(\t\"3\n\tDirection\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\n\n\x06HIGHER\x10\x01\x12\t\n\x05LOWER\x10\x02\"\xd7\x01\n\x0bMetricValue\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0b\n\x03run\x18\x02 \x01(\t\x12\x0c\n\x04step\x18\x03 \x01(\x04\x12\r\n\x05value\x18\x04 \x01(\x01\x12-\n\ttimestamp\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\r\n\x05group\x18\x06 \x01(\t\x12/\n\x0b\x63reate_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x12\n\ncreated_by\x18\x08 \x01(\t\x12\r\n\x05label\x18\t \x01(\t\"\x95\x01\n\x08Metadata\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0b\n\x03run\x18\x02 \x01(\t\x12/\n\x0b\x63reate_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12/\n\x0bupdate_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0c\n\x04\x64\x61ta\x18\x05 \x01(\t\"\x8b\t\n\nGitContext\x12\x46\n\nrepository\x18\x01 \x01(\x0b\x32\x32.continual.rpc.management.v1.GitContext.Repository\x12>\n\x06\x62ranch\x18\x02 \x01(\x0b\x32..continual.rpc.management.v1.GitContext.Branch\x12=\n\x04tags\x18\x03 \x01(\x0b\x32/.continual.rpc.management.v1.GitContext.GitTags\x12\x38\n\x03ref\x18\x04 \x01(\x0b\x32+.continual.rpc.management.v1.GitContext.Ref\x12>\n\x06\x63ommit\x18\x05 \x01(\x0b\x32..continual.rpc.management.v1.GitContext.Commit\x12:\n\x04head\x18\x06 \x01(\x0b\x32,.continual.rpc.management.v1.GitContext.Head\x12\x10\n\x08is_dirty\x18\x07 \x01(\x08\x1av\n\nRepository\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x12\n\nname_local\x18\x02 \x01(\t\x12\x1a\n\x12name_remote_origin\x18\x03 \x01(\t\x12\x19\n\x11remote_origin_url\x18\x04 \x01(\t\x12\x0f\n\x07is_bare\x18\x05 \x01(\x08\x1a\x9b\x02\n\x06\x43ommit\x12\x0b\n\x03sha\x18\x01 \x01(\t\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x35\n\x11\x61uthored_datetime\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12>\n\x06\x61uthor\x18\x04 \x01(\x0b\x32..continual.rpc.management.v1.GitContext.Author\x12\x36\n\x12\x63ommitted_datetime\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x44\n\tcommitter\x18\x06 \x01(\x0b\x32\x31.continual.rpc.management.v1.GitContext.Committer\x1a\x16\n\x06\x42ranch\x12\x0c\n\x04name\x18\x01 \x01(\t\x1a-\n\x06GitTag\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x15\n\rname_describe\x18\x02 \x01(\t\x1a\x8d\x01\n\x07GitTags\x12?\n\x07\x63urrent\x18\x01 \x01(\x0b\x32..continual.rpc.management.v1.GitContext.GitTag\x12\x41\n\treachable\x18\x02 \x01(\x0b\x32..continual.rpc.management.v1.GitContext.GitTag\x1a\x13\n\x03Ref\x12\x0c\n\x04path\x18\x01 \x01(\t\x1a%\n\x06\x41uthor\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05\x65mail\x18\x02 \x01(\t\x1a(\n\tCommitter\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05\x65mail\x18\x02 \x01(\t\x1a\x1b\n\x04Head\x12\x13\n\x0bis_detached\x18\x01 \x01(\x08\"\x9f\x01\n\x12GitProviderContext\x12\x10\n\x08provider\x18\x01 \x01(\t\x12N\n\nrepository\x18\x02 \x01(\x0b\x32:.continual.rpc.management.v1.GitProviderContext.Repository\x1a\'\n\nRepository\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0b\n\x03url\x18\x02 \x01(\t\"F\n\x11\x43icdRunnerContext\x12\x10\n\x08provider\x18\x01 \x01(\t\x12\x0e\n\x06run_id\x18\x02 \x01(\t\x12\x0f\n\x07run_url\x18\x03 \x01(\t\"\x91\x03\n\x11SdlcReviewContext\x12\x10\n\x08provider\x18\x01 \x01(\t\x12\n\n\x02id\x18\x02 \x01(\t\x12\x0b\n\x03url\x18\x03 \x01(\t\x12\r\n\x05state\x18\x04 \x01(\t\x12\r\n\x05title\x18\x05 \x01(\t\x12\x41\n\x04user\x18\x06 \x01(\x0b\x32\x33.continual.rpc.management.v1.SdlcReviewContext.User\x12@\n\x04\x62\x61se\x18\x07 \x01(\x0b\x32\x32.continual.rpc.management.v1.SdlcReviewContext.Ref\x12@\n\x04head\x18\x08 \x01(\x0b\x32\x32.continual.rpc.management.v1.SdlcReviewContext.Ref\x12\x0c\n\x04\x64\x61ta\x18\t \x01(\t\x1a:\n\x04User\x12\x11\n\tuser_name\x18\x01 \x01(\t\x12\x0b\n\x03url\x18\x02 \x01(\t\x12\x12\n\navatar_url\x18\x03 \x01(\t\x1a\"\n\x03Ref\x12\x0e\n\x06\x62ranch\x18\x01 \x01(\t\x12\x0b\n\x03sha\x18\x02 \x01(\t\"\xae\x07\n\nRunSummary\x12\x41\n\x0emodel_versions\x18\x01 \x03(\x0b\x32).continual.rpc.management.v1.ModelVersion\x12K\n\x11\x62\x61tch_predictions\x18\x02 \x03(\x0b\x32\x30.continual.rpc.management.v1.RunSummary.Resource\x12\x44\n\npromotions\x18\x03 \x03(\x0b\x32\x30.continual.rpc.management.v1.RunSummary.Resource\x12\x45\n\x0b\x64\x65ployments\x18\x04 \x03(\x0b\x32\x30.continual.rpc.management.v1.RunSummary.Resource\x12<\n\x0bgit_context\x18\x05 \x01(\x0b\x32\'.continual.rpc.management.v1.GitContext\x12\x39\n\tprincipal\x18\x06 \x01(\x0b\x32&.continual.rpc.management.v1.Principal\x12/\n\x0b\x63reate_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12=\n\x11\x61ssociated_models\x18\x08 \x03(\x0b\x32\".continual.rpc.management.v1.Model\x12\x41\n\x13\x61ssociated_datasets\x18\t \x03(\x0b\x32$.continual.rpc.management.v1.Dataset\x12\x45\n\x10\x64\x61taset_versions\x18\n \x03(\x0b\x32+.continual.rpc.management.v1.DatasetVersion\x12M\n\x14git_provider_context\x18\x0b \x01(\x0b\x32/.continual.rpc.management.v1.GitProviderContext\x12K\n\x13\x63icd_runner_context\x18\x0c \x01(\x0b\x32..continual.rpc.management.v1.CicdRunnerContext\x12K\n\x13sdlc_review_context\x18\r \x01(\x0b\x32..continual.rpc.management.v1.SdlcReviewContext\x1a\'\n\x08Resource\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05state\x18\x02 \x01(\t\"\xb6\x03\n\x03Run\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12\x0e\n\x06\x61uthor\x18\x03 \x01(\t\x12/\n\x0b\x63reate_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12/\n\x0bupdate_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x32\n\x0elast_heartbeat\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x1a\n\x12heartbeat_interval\x18\x07 \x01(\x04\x12\x35\n\x05state\x18\x08 \x01(\x0e\x32&.continual.rpc.management.v1.Run.State\x12\x15\n\rerror_message\x18\t \x01(\t\x12\x31\n\rcomplete_time\x18\n \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"I\n\x05State\x12\n\n\x06\x41\x43TIVE\x10\x00\x12\x0c\n\x08INACTIVE\x10\x01\x12\r\n\tCOMPLETED\x10\x02\x12\x0b\n\x07STOPPED\x10\x03\x12\n\n\x06\x46\x41ILED\x10\x04\"\x91\x01\n\x03Tag\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0b\n\x03key\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\t\x12/\n\x0b\x63reate_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12/\n\x0bupdate_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\xec\x01\n\x18NotificationPolicyConfig\x12J\n\x05\x65mail\x18\x01 \x01(\x0b\x32;.continual.rpc.management.v1.NotificationPolicyConfig.Email\x12J\n\x05slack\x18\x02 \x01(\x0b\x32;.continual.rpc.management.v1.NotificationPolicyConfig.Slack\x1a\x1e\n\x05\x45mail\x12\x15\n\remail_address\x18\x01 \x01(\t\x1a\x18\n\x05Slack\x12\x0f\n\x07\x63hannel\x18\x01 \x01(\t\"\xa9\x03\n\x12NotificationPolicy\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12\x45\n\x06\x63onfig\x18\x03 \x01(\x0b\x32\x35.continual.rpc.management.v1.NotificationPolicyConfig\x12\x42\n\x04type\x18\x04 \x01(\x0e\x32\x34.continual.rpc.management.v1.NotificationPolicy.Type\x12\x10\n\x08severity\x18\x05 \x01(\t\x12\x10\n\x08resource\x18\x06 \x01(\t\x12\x15\n\rresource_type\x18\x07 \x01(\t\x12\x12\n\nevent_type\x18\x08 \x01(\t\x12\x16\n\x0erequire_notify\x18\t \x01(\x08\x12/\n\x0b\x63reate_time\x18\n \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12/\n\x0bupdate_time\x18\x0b \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\x1c\n\x04Type\x12\t\n\x05\x45MAIL\x10\x00\x12\t\n\x05SLACK\x10\x01\"\xc0\x01\n\x0cNotification\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x42\n\x04type\x18\x02 \x01(\x0e\x32\x34.continual.rpc.management.v1.NotificationPolicy.Type\x12\r\n\x05\x65vent\x18\x03 \x01(\t\x12\x0e\n\x06target\x18\x04 \x01(\t\x12\x0e\n\x06result\x18\x05 \x01(\t\x12/\n\x0b\x63reate_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\xa4\x01\n\x0fRunCheckSummary\x12\x10\n\x08run_name\x18\x01 \x01(\t\x12:\n\x07outcome\x18\x02 \x01(\x0e\x32).continual.rpc.management.v1.CheckOutcome\x12\x43\n\x0fresource_checks\x18\x03 \x03(\x0b\x32*.continual.rpc.management.v1.ResourceCheck\"\xbf\x02\n\rResourceCheck\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08run_name\x18\x02 \x01(\t\x12\x12\n\ngroup_name\x18\x03 \x01(\t\x12:\n\x07outcome\x18\x05 \x01(\x0e\x32).continual.rpc.management.v1.CheckOutcome\x12\x10\n\x08\x64uration\x18\x06 \x01(\x02\x12\x0c\n\x04\x64\x61ta\x18\x07 \x01(\t\x12\x0e\n\x06\x65rrors\x18\x08 \x03(\t\x12\x10\n\x08warnings\x18\t \x03(\t\x12\r\n\x05infos\x18\n \x03(\t\x12\x0f\n\x07summary\x18\x0b \x01(\t\x12\x15\n\rartifact_name\x18\x0c \x01(\t\x12/\n\x0b\x63reate_time\x18\r \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x14\n\x0c\x64isplay_name\x18\x0e \x01(\t*\xd5\x01\n\tFieldType\x12\x0b\n\x07INVALID\x10\x00\x12\n\n\x06NUMBER\x10\x01\x12\x0b\n\x07\x42OOLEAN\x10\x02\x12\x0f\n\x0b\x43\x41TEGORICAL\x10\x03\x12\x08\n\x04TEXT\x10\x04\x12\r\n\tTIMESTAMP\x10\x05\x12\x08\n\x04TIME\x10\x06\x12\x08\n\x04\x44\x41TE\x10\x07\x12\x0c\n\x08\x44\x41TETIME\x10\x08\x12\x0f\n\x0bNUMBER_LIST\x10\t\x12\x10\n\x0c\x42OOLEAN_LIST\x10\n\x12\r\n\tTEXT_LIST\x10\x0b\x12\t\n\x05INDEX\x10\x0c\x12\x0e\n\nTIME_INDEX\x10\r\x12\t\n\x05SPLIT\x10\x0f*?\n\x18\x42\x61tchPredictionStoreType\x12\x0f\n\x0bUNSPECIFIED\x10\x00\x12\x07\n\x03\x43SV\x10\x01\x12\t\n\x05STORE\x10\x02*3\n\x0c\x43heckOutcome\x12\n\n\x06PASSED\x10\x00\x12\n\n\x06\x46\x41ILED\x10\x01\x12\x0b\n\x07SKIPPED\x10\x02\x42LZJgithub.com/continual-ai/continual/continual/rpc/management/v1;managementv1b\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,google_dot_protobuf_dot_duration__pb2.DESCRIPTOR,continual_dot_rpc_dot_graphql_dot_options__pb2.DESCRIPTOR,])

_FIELDTYPE = _descriptor.EnumDescriptor(
  name='FieldType',
  full_name='continual.rpc.management.v1.FieldType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='INVALID', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NUMBER', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BOOLEAN', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CATEGORICAL', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TEXT', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TIMESTAMP', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TIME', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DATE', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DATETIME', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NUMBER_LIST', index=9, number=9,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='BOOLEAN_LIST', index=10, number=10,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TEXT_LIST', index=11, number=11,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INDEX', index=12, number=12,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TIME_INDEX', index=13, number=13,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SPLIT', index=14, number=15,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=24349,
  serialized_end=24562,
)
_sym_db.RegisterEnumDescriptor(_FIELDTYPE)

FieldType = enum_type_wrapper.EnumTypeWrapper(_FIELDTYPE)
_BATCHPREDICTIONSTORETYPE = _descriptor.EnumDescriptor(
  name='BatchPredictionStoreType',
  full_name='continual.rpc.management.v1.BatchPredictionStoreType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CSV', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='STORE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=24564,
  serialized_end=24627,
)
_sym_db.RegisterEnumDescriptor(_BATCHPREDICTIONSTORETYPE)

BatchPredictionStoreType = enum_type_wrapper.EnumTypeWrapper(_BATCHPREDICTIONSTORETYPE)
_CHECKOUTCOME = _descriptor.EnumDescriptor(
  name='CheckOutcome',
  full_name='continual.rpc.management.v1.CheckOutcome',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='PASSED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FAILED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SKIPPED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=24629,
  serialized_end=24680,
)
_sym_db.RegisterEnumDescriptor(_CHECKOUTCOME)

CheckOutcome = enum_type_wrapper.EnumTypeWrapper(_CHECKOUTCOME)
INVALID = 0
NUMBER = 1
BOOLEAN = 2
CATEGORICAL = 3
TEXT = 4
TIMESTAMP = 5
TIME = 6
DATE = 7
DATETIME = 8
NUMBER_LIST = 9
BOOLEAN_LIST = 10
TEXT_LIST = 11
INDEX = 12
TIME_INDEX = 13
SPLIT = 15
UNSPECIFIED = 0
CSV = 1
STORE = 2
PASSED = 0
FAILED = 1
SKIPPED = 2


_PRINCIPAL_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='continual.rpc.management.v1.Principal.Type',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='USER', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='API_KEY', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=649,
  serialized_end=695,
)
_sym_db.RegisterEnumDescriptor(_PRINCIPAL_TYPE)

_SESSION_STATE = _descriptor.EnumDescriptor(
  name='State',
  full_name='continual.rpc.management.v1.Session.State',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ACTIVE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='EXPIRED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1004,
  serialized_end=1053,
)
_sym_db.RegisterEnumDescriptor(_SESSION_STATE)

_ORGANIZATION_STATUS = _descriptor.EnumDescriptor(
  name='Status',
  full_name='continual.rpc.management.v1.Organization.Status',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='EXPIRED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ACTIVE_TRIAL', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ACTIVE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='DISABLED', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2040,
  serialized_end=2105,
)
_sym_db.RegisterEnumDescriptor(_ORGANIZATION_STATUS)

_PROJECT_CREATESOURCETYPE = _descriptor.EnumDescriptor(
  name='CreateSourceType',
  full_name='continual.rpc.management.v1.Project.CreateSourceType',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='BLANK', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='REPO_EXISTING', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='REPO_TEMPLATE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=2811,
  serialized_end=2878,
)
_sym_db.RegisterEnumDescriptor(_PROJECT_CREATESOURCETYPE)

_APIKEY_STATE = _descriptor.EnumDescriptor(
  name='State',
  full_name='continual.rpc.management.v1.ApiKey.State',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ACTIVE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INACTIVE', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=6794,
  serialized_end=6844,
)
_sym_db.RegisterEnumDescriptor(_APIKEY_STATE)

_PROMOTION_REASON = _descriptor.EnumDescriptor(
  name='Reason',
  full_name='continual.rpc.management.v1.Promotion.Reason',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='REASON_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UPLIFT', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='NO_PROMOTED_MODEL_VERSION_EXISTS', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='LATEST_POLICY', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='UNDERPERFORMED', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='MANUAL', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FORCED', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INCOMPATIBLE_SCHEMA', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=7768,
  serialized_end=7938,
)
_sym_db.RegisterEnumDescriptor(_PROMOTION_REASON)

_PROMOTION_STATE = _descriptor.EnumDescriptor(
  name='State',
  full_name='continual.rpc.management.v1.Promotion.State',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='STATE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SUCCEEDED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FAILED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=7940,
  serialized_end=7997,
)
_sym_db.RegisterEnumDescriptor(_PROMOTION_STATE)

_MODELVERSION_STATE = _descriptor.EnumDescriptor(
  name='State',
  full_name='continual.rpc.management.v1.ModelVersion.State',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='STATE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CREATED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PENDING', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PREPROCESSING', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TRAINING', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='POSTPROCESSING', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CANCELLING', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SUCCEEDED', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FAILED', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CANCELLED', index=9, number=9,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PROFILING', index=10, number=10,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=8637,
  serialized_end=8819,
)
_sym_db.RegisterEnumDescriptor(_MODELVERSION_STATE)

_EXPERIMENT_STATE = _descriptor.EnumDescriptor(
  name='State',
  full_name='continual.rpc.management.v1.Experiment.State',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='STATE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PENDING', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PREPROCESSING', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='TRAINING', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='POSTPROCESSING', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CANCELLING', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SUCCEEDED', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FAILED', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CANCELLED', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=10753,
  serialized_end=10907,
)
_sym_db.RegisterEnumDescriptor(_EXPERIMENT_STATE)

_BATCHPREDICTION_STATE = _descriptor.EnumDescriptor(
  name='State',
  full_name='continual.rpc.management.v1.BatchPrediction.State',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='STATE_UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CREATED', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PENDING', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PREPROCESSING', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='PROCESSING', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='POSTPROCESSING', index=5, number=5,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CANCELLING', index=6, number=6,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SUCCEEDED', index=7, number=7,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FAILED', index=8, number=8,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='CANCELLED', index=9, number=9,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=11894,
  serialized_end=12063,
)
_sym_db.RegisterEnumDescriptor(_BATCHPREDICTION_STATE)

_METRIC_DIRECTION = _descriptor.EnumDescriptor(
  name='Direction',
  full_name='continual.rpc.management.v1.Metric.Direction',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNSPECIFIED', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='HIGHER', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='LOWER', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=19236,
  serialized_end=19287,
)
_sym_db.RegisterEnumDescriptor(_METRIC_DIRECTION)

_RUN_STATE = _descriptor.EnumDescriptor(
  name='State',
  full_name='continual.rpc.management.v1.Run.State',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='ACTIVE', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='INACTIVE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='COMPLETED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='STOPPED', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FAILED', index=4, number=4,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=22774,
  serialized_end=22847,
)
_sym_db.RegisterEnumDescriptor(_RUN_STATE)

_NOTIFICATIONPOLICY_TYPE = _descriptor.EnumDescriptor(
  name='Type',
  full_name='continual.rpc.management.v1.NotificationPolicy.Type',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='EMAIL', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SLACK', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=23634,
  serialized_end=23662,
)
_sym_db.RegisterEnumDescriptor(_NOTIFICATIONPOLICY_TYPE)


_USER = _descriptor.Descriptor(
  name='User',
  full_name='continual.rpc.management.v1.User',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.User.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='email', full_name='continual.rpc.management.v1.User.email', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='email_verified', full_name='continual.rpc.management.v1.User.email_verified', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='full_name', full_name='continual.rpc.management.v1.User.full_name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bio', full_name='continual.rpc.management.v1.User.bio', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='location', full_name='continual.rpc.management.v1.User.location', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='password', full_name='continual.rpc.management.v1.User.password', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\000', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='service_account', full_name='continual.rpc.management.v1.User.service_account', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.User.update_time', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.User.create_time', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='disabled', full_name='continual.rpc.management.v1.User.disabled', index=10,
      number=11, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='trial_available', full_name='continual.rpc.management.v1.User.trial_available', index=11,
      number=12, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='first_name', full_name='continual.rpc.management.v1.User.first_name', index=12,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_name', full_name='continual.rpc.management.v1.User.last_name', index=13,
      number=14, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=186,
  serialized_end=538,
)


_PRINCIPAL = _descriptor.Descriptor(
  name='Principal',
  full_name='continual.rpc.management.v1.Principal',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Principal.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='continual.rpc.management.v1.Principal.type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='display_name', full_name='continual.rpc.management.v1.Principal.display_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _PRINCIPAL_TYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=541,
  serialized_end=695,
)


_SESSION = _descriptor.Descriptor(
  name='Session',
  full_name='continual.rpc.management.v1.Session',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Session.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='subject', full_name='continual.rpc.management.v1.Session.subject', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_access_ip', full_name='continual.rpc.management.v1.Session.last_access_ip', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_access_time', full_name='continual.rpc.management.v1.Session.last_access_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='login_ip', full_name='continual.rpc.management.v1.Session.login_ip', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='login_time', full_name='continual.rpc.management.v1.Session.login_time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='login_user_agent', full_name='continual.rpc.management.v1.Session.login_user_agent', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='continual.rpc.management.v1.Session.state', index=7,
      number=8, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _SESSION_STATE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=698,
  serialized_end=1053,
)


_ROLE = _descriptor.Descriptor(
  name='Role',
  full_name='continual.rpc.management.v1.Role',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Role.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='title', full_name='continual.rpc.management.v1.Role.title', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='continual.rpc.management.v1.Role.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource_type', full_name='continual.rpc.management.v1.Role.resource_type', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='permissions', full_name='continual.rpc.management.v1.Role.permissions', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1055,
  serialized_end=1155,
)


_ACCESSPOLICY = _descriptor.Descriptor(
  name='AccessPolicy',
  full_name='continual.rpc.management.v1.AccessPolicy',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.AccessPolicy.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='subject', full_name='continual.rpc.management.v1.AccessPolicy.subject', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\004User', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource', full_name='continual.rpc.management.v1.AccessPolicy.resource', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='role', full_name='continual.rpc.management.v1.AccessPolicy.role', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\004Role', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.AccessPolicy.create_time', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1158,
  serialized_end=1302,
)


_INVITE = _descriptor.Descriptor(
  name='Invite',
  full_name='continual.rpc.management.v1.Invite',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Invite.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sender', full_name='continual.rpc.management.v1.Invite.sender', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\004User\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='email', full_name='continual.rpc.management.v1.Invite.email', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='role', full_name='continual.rpc.management.v1.Invite.role', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\004Role', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Invite.create_time', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='expire_time', full_name='continual.rpc.management.v1.Invite.expire_time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1305,
  serialized_end=1501,
)


_ORGANIZATION = _descriptor.Descriptor(
  name='Organization',
  full_name='continual.rpc.management.v1.Organization',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Organization.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='display_name', full_name='continual.rpc.management.v1.Organization.display_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.Organization.update_time', index=2,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Organization.create_time', index=3,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status', full_name='continual.rpc.management.v1.Organization.status', index=4,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='trial_end_date', full_name='continual.rpc.management.v1.Organization.trial_end_date', index=5,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='trial_credits', full_name='continual.rpc.management.v1.Organization.trial_credits', index=6,
      number=8, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='trial_credits_used', full_name='continual.rpc.management.v1.Organization.trial_credits_used', index=7,
      number=9, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='show_plan', full_name='continual.rpc.management.v1.Organization.show_plan', index=8,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sso_enabled', full_name='continual.rpc.management.v1.Organization.sso_enabled', index=9,
      number=11, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='requires_sso', full_name='continual.rpc.management.v1.Organization.requires_sso', index=10,
      number=12, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sso_domains', full_name='continual.rpc.management.v1.Organization.sso_domains', index=11,
      number=13, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='allow_external_users', full_name='continual.rpc.management.v1.Organization.allow_external_users', index=12,
      number=14, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sso_configured', full_name='continual.rpc.management.v1.Organization.sso_configured', index=13,
      number=15, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='directory_sync_configured', full_name='continual.rpc.management.v1.Organization.directory_sync_configured', index=14,
      number=16, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _ORGANIZATION_STATUS,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1504,
  serialized_end=2105,
)


_ONBOARDING = _descriptor.Descriptor(
  name='Onboarding',
  full_name='continual.rpc.management.v1.Onboarding',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Onboarding.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Onboarding.create_time', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.Onboarding.update_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='project_created', full_name='continual.rpc.management.v1.Onboarding.project_created', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_created', full_name='continual.rpc.management.v1.Onboarding.model_created', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_set_created', full_name='continual.rpc.management.v1.Onboarding.feature_set_created', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='billing_setup', full_name='continual.rpc.management.v1.Onboarding.billing_setup', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dismissed', full_name='continual.rpc.management.v1.Onboarding.dismissed', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2108,
  serialized_end=2361,
)


_PROJECT = _descriptor.Descriptor(
  name='Project',
  full_name='continual.rpc.management.v1.Project',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Project.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='display_name', full_name='continual.rpc.management.v1.Project.display_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='organization', full_name='continual.rpc.management.v1.Project.organization', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\014Organization', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='summary', full_name='continual.rpc.management.v1.Project.summary', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.Project.update_time', index=4,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Project.create_time', index=5,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='default_environment', full_name='continual.rpc.management.v1.Project.default_environment', index=6,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_source_type', full_name='continual.rpc.management.v1.Project.create_source_type', index=7,
      number=9, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_source', full_name='continual.rpc.management.v1.Project.create_source', index=8,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='git_integration', full_name='continual.rpc.management.v1.Project.git_integration', index=9,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\025ProjectGitIntegration', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _PROJECT_CREATESOURCETYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2364,
  serialized_end=2878,
)


_PROJECTSUMMARY = _descriptor.Descriptor(
  name='ProjectSummary',
  full_name='continual.rpc.management.v1.ProjectSummary',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='feature_set_count', full_name='continual.rpc.management.v1.ProjectSummary.feature_set_count', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_set_row_count', full_name='continual.rpc.management.v1.ProjectSummary.feature_set_row_count', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_set_bytes', full_name='continual.rpc.management.v1.ProjectSummary.feature_set_bytes', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_count', full_name='continual.rpc.management.v1.ProjectSummary.feature_count', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='connection_count', full_name='continual.rpc.management.v1.ProjectSummary.connection_count', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_count', full_name='continual.rpc.management.v1.ProjectSummary.model_count', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_version_count', full_name='continual.rpc.management.v1.ProjectSummary.model_version_count', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='experiment_count', full_name='continual.rpc.management.v1.ProjectSummary.experiment_count', index=7,
      number=8, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='prediction_count', full_name='continual.rpc.management.v1.ProjectSummary.prediction_count', index=8,
      number=9, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_set_health', full_name='continual.rpc.management.v1.ProjectSummary.feature_set_health', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_health', full_name='continual.rpc.management.v1.ProjectSummary.model_health', index=10,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_deployment_time', full_name='continual.rpc.management.v1.ProjectSummary.last_deployment_time', index=11,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_training_time', full_name='continual.rpc.management.v1.ProjectSummary.last_training_time', index=12,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run_count', full_name='continual.rpc.management.v1.ProjectSummary.run_count', index=13,
      number=14, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dataset_count', full_name='continual.rpc.management.v1.ProjectSummary.dataset_count', index=14,
      number=15, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dataset_version_count', full_name='continual.rpc.management.v1.ProjectSummary.dataset_version_count', index=15,
      number=16, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2881,
  serialized_end=3456,
)


_DATASTORE_BIGQUERY = _descriptor.Descriptor(
  name='BigQuery',
  full_name='continual.rpc.management.v1.DataStore.BigQuery',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='auth_file', full_name='continual.rpc.management.v1.DataStore.BigQuery.auth_file', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='auth_file_name', full_name='continual.rpc.management.v1.DataStore.BigQuery.auth_file_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dataset', full_name='continual.rpc.management.v1.DataStore.BigQuery.dataset', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='location', full_name='continual.rpc.management.v1.DataStore.BigQuery.location', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3866,
  serialized_end=3954,
)

_DATASTORE_SNOWFLAKE_PARAMSENTRY = _descriptor.Descriptor(
  name='ParamsEntry',
  full_name='continual.rpc.management.v1.DataStore.Snowflake.ParamsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='continual.rpc.management.v1.DataStore.Snowflake.ParamsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.DataStore.Snowflake.ParamsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4182,
  serialized_end=4227,
)

_DATASTORE_SNOWFLAKE = _descriptor.Descriptor(
  name='Snowflake',
  full_name='continual.rpc.management.v1.DataStore.Snowflake',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host', full_name='continual.rpc.management.v1.DataStore.Snowflake.host', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='port', full_name='continual.rpc.management.v1.DataStore.Snowflake.port', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='username', full_name='continual.rpc.management.v1.DataStore.Snowflake.username', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='password', full_name='continual.rpc.management.v1.DataStore.Snowflake.password', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='database', full_name='continual.rpc.management.v1.DataStore.Snowflake.database', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='warehouse', full_name='continual.rpc.management.v1.DataStore.Snowflake.warehouse', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='role', full_name='continual.rpc.management.v1.DataStore.Snowflake.role', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='params', full_name='continual.rpc.management.v1.DataStore.Snowflake.params', index=7,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='db_schema', full_name='continual.rpc.management.v1.DataStore.Snowflake.db_schema', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_DATASTORE_SNOWFLAKE_PARAMSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3957,
  serialized_end=4227,
)

_DATASTORE_POSTGRES_PARAMSENTRY = _descriptor.Descriptor(
  name='ParamsEntry',
  full_name='continual.rpc.management.v1.DataStore.Postgres.ParamsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='continual.rpc.management.v1.DataStore.Postgres.ParamsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.DataStore.Postgres.ParamsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4182,
  serialized_end=4227,
)

_DATASTORE_POSTGRES = _descriptor.Descriptor(
  name='Postgres',
  full_name='continual.rpc.management.v1.DataStore.Postgres',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host', full_name='continual.rpc.management.v1.DataStore.Postgres.host', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='port', full_name='continual.rpc.management.v1.DataStore.Postgres.port', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='username', full_name='continual.rpc.management.v1.DataStore.Postgres.username', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='password', full_name='continual.rpc.management.v1.DataStore.Postgres.password', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='database', full_name='continual.rpc.management.v1.DataStore.Postgres.database', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='params', full_name='continual.rpc.management.v1.DataStore.Postgres.params', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='db_schema', full_name='continual.rpc.management.v1.DataStore.Postgres.db_schema', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_DATASTORE_POSTGRES_PARAMSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4230,
  serialized_end=4465,
)

_DATASTORE_REDSHIFT_PARAMSENTRY = _descriptor.Descriptor(
  name='ParamsEntry',
  full_name='continual.rpc.management.v1.DataStore.Redshift.ParamsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='continual.rpc.management.v1.DataStore.Redshift.ParamsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.DataStore.Redshift.ParamsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4182,
  serialized_end=4227,
)

_DATASTORE_REDSHIFT = _descriptor.Descriptor(
  name='Redshift',
  full_name='continual.rpc.management.v1.DataStore.Redshift',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host', full_name='continual.rpc.management.v1.DataStore.Redshift.host', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='port', full_name='continual.rpc.management.v1.DataStore.Redshift.port', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='username', full_name='continual.rpc.management.v1.DataStore.Redshift.username', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='password', full_name='continual.rpc.management.v1.DataStore.Redshift.password', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='database', full_name='continual.rpc.management.v1.DataStore.Redshift.database', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='params', full_name='continual.rpc.management.v1.DataStore.Redshift.params', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='db_schema', full_name='continual.rpc.management.v1.DataStore.Redshift.db_schema', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_DATASTORE_REDSHIFT_PARAMSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4468,
  serialized_end=4703,
)

_DATASTORE_DATABRICKS_PARAMSENTRY = _descriptor.Descriptor(
  name='ParamsEntry',
  full_name='continual.rpc.management.v1.DataStore.Databricks.ParamsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='continual.rpc.management.v1.DataStore.Databricks.ParamsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.DataStore.Databricks.ParamsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4182,
  serialized_end=4227,
)

_DATASTORE_DATABRICKS = _descriptor.Descriptor(
  name='Databricks',
  full_name='continual.rpc.management.v1.DataStore.Databricks',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='host', full_name='continual.rpc.management.v1.DataStore.Databricks.host', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='port', full_name='continual.rpc.management.v1.DataStore.Databricks.port', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='token', full_name='continual.rpc.management.v1.DataStore.Databricks.token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='http_path', full_name='continual.rpc.management.v1.DataStore.Databricks.http_path', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='database', full_name='continual.rpc.management.v1.DataStore.Databricks.database', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='params', full_name='continual.rpc.management.v1.DataStore.Databricks.params', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='db_schema', full_name='continual.rpc.management.v1.DataStore.Databricks.db_schema', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_DATASTORE_DATABRICKS_PARAMSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4706,
  serialized_end=4943,
)

_DATASTORE = _descriptor.Descriptor(
  name='DataStore',
  full_name='continual.rpc.management.v1.DataStore',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='continual.rpc.management.v1.DataStore.type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='big_query', full_name='continual.rpc.management.v1.DataStore.big_query', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='snowflake', full_name='continual.rpc.management.v1.DataStore.snowflake', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='postgres', full_name='continual.rpc.management.v1.DataStore.postgres', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='redshift', full_name='continual.rpc.management.v1.DataStore.redshift', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='databricks', full_name='continual.rpc.management.v1.DataStore.databricks', index=5,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_demo_data_store', full_name='continual.rpc.management.v1.DataStore.is_demo_data_store', index=6,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_DATASTORE_BIGQUERY, _DATASTORE_SNOWFLAKE, _DATASTORE_POSTGRES, _DATASTORE_REDSHIFT, _DATASTORE_DATABRICKS, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='config', full_name='continual.rpc.management.v1.DataStore.config',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=3459,
  serialized_end=4953,
)


_HEALTHCOUNTS = _descriptor.Descriptor(
  name='HealthCounts',
  full_name='continual.rpc.management.v1.HealthCounts',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='healthy_count', full_name='continual.rpc.management.v1.HealthCounts.healthy_count', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='unhealthy_count', full_name='continual.rpc.management.v1.HealthCounts.unhealthy_count', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='critical_count', full_name='continual.rpc.management.v1.HealthCounts.critical_count', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4955,
  serialized_end=5041,
)


_ENVIRONMENT = _descriptor.Descriptor(
  name='Environment',
  full_name='continual.rpc.management.v1.Environment',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Environment.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='summary', full_name='continual.rpc.management.v1.Environment.summary', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data_store', full_name='continual.rpc.management.v1.Environment.data_store', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.Environment.update_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Environment.create_time', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='scheduling_enabled', full_name='continual.rpc.management.v1.Environment.scheduling_enabled', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='source', full_name='continual.rpc.management.v1.Environment.source', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='env_type', full_name='continual.rpc.management.v1.Environment.env_type', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_protected', full_name='continual.rpc.management.v1.Environment.is_protected', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_ephemeral', full_name='continual.rpc.management.v1.Environment.is_ephemeral', index=9,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5044,
  serialized_end=5412,
)


_FEATURESET = _descriptor.Descriptor(
  name='FeatureSet',
  full_name='continual.rpc.management.v1.FeatureSet',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.FeatureSet.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schema_text', full_name='continual.rpc.management.v1.FeatureSet.schema_text', index=1,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.FeatureSet.create_time', index=2,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.FeatureSet.update_time', index=3,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='table', full_name='continual.rpc.management.v1.FeatureSet.table', index=4,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5415,
  serialized_end=5580,
)


_EVENT_DATAENTRY = _descriptor.Descriptor(
  name='DataEntry',
  full_name='continual.rpc.management.v1.Event.DataEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='continual.rpc.management.v1.Event.DataEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.Event.DataEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5874,
  serialized_end=5917,
)

_EVENT = _descriptor.Descriptor(
  name='Event',
  full_name='continual.rpc.management.v1.Event',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Event.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='subject', full_name='continual.rpc.management.v1.Event.subject', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\tPrincipal', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='event', full_name='continual.rpc.management.v1.Event.event', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource', full_name='continual.rpc.management.v1.Event.resource', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='severity', full_name='continual.rpc.management.v1.Event.severity', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message', full_name='continual.rpc.management.v1.Event.message', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data', full_name='continual.rpc.management.v1.Event.data', index=6,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Event.create_time', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='notify', full_name='continual.rpc.management.v1.Event.notify', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run', full_name='continual.rpc.management.v1.Event.run', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='before', full_name='continual.rpc.management.v1.Event.before', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='after', full_name='continual.rpc.management.v1.Event.after', index=11,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_EVENT_DATAENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5583,
  serialized_end=5917,
)


_SERVINGMETRIC = _descriptor.Descriptor(
  name='ServingMetric',
  full_name='continual.rpc.management.v1.ServingMetric',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model', full_name='continual.rpc.management.v1.ServingMetric.model', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ServingMetric.parent', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tag', full_name='continual.rpc.management.v1.ServingMetric.tag', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='request_start_time', full_name='continual.rpc.management.v1.ServingMetric.request_start_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_start_time', full_name='continual.rpc.management.v1.ServingMetric.model_start_time', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='request_duration', full_name='continual.rpc.management.v1.ServingMetric.request_duration', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='route_duration', full_name='continual.rpc.management.v1.ServingMetric.route_duration', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_duration', full_name='continual.rpc.management.v1.ServingMetric.feature_duration', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='predict_duration', full_name='continual.rpc.management.v1.ServingMetric.predict_duration', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_duration', full_name='continual.rpc.management.v1.ServingMetric.model_duration', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_version', full_name='continual.rpc.management.v1.ServingMetric.model_version', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='result_code', full_name='continual.rpc.management.v1.ServingMetric.result_code', index=11,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='result_data', full_name='continual.rpc.management.v1.ServingMetric.result_data', index=12,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5920,
  serialized_end=6415,
)


_APIKEY = _descriptor.Descriptor(
  name='ApiKey',
  full_name='continual.rpc.management.v1.ApiKey',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.ApiKey.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='key', full_name='continual.rpc.management.v1.ApiKey.key', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='continual.rpc.management.v1.ApiKey.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='created_by', full_name='continual.rpc.management.v1.ApiKey.created_by', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\004User', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.ApiKey.create_time', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_access_ip', full_name='continual.rpc.management.v1.ApiKey.last_access_ip', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_access_time', full_name='continual.rpc.management.v1.ApiKey.last_access_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='continual.rpc.management.v1.ApiKey.state', index=7,
      number=8, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='revoked_time', full_name='continual.rpc.management.v1.ApiKey.revoked_time', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='revoked_by', full_name='continual.rpc.management.v1.ApiKey.revoked_by', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\004User', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _APIKEY_STATE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6418,
  serialized_end=6844,
)


_MODEL = _descriptor.Descriptor(
  name='Model',
  full_name='continual.rpc.management.v1.Model',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Model.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='continual.rpc.management.v1.Model.description', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='continual.rpc.management.v1.Model.state', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='author', full_name='continual.rpc.management.v1.Model.author', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='display_name', full_name='continual.rpc.management.v1.Model.display_name', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='current_version', full_name='continual.rpc.management.v1.Model.current_version', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\014ModelVersion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Model.create_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.Model.update_time', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest_model_version', full_name='continual.rpc.management.v1.Model.latest_model_version', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\014ModelVersion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest_batch_prediction', full_name='continual.rpc.management.v1.Model.latest_batch_prediction', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\017BatchPrediction', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run', full_name='continual.rpc.management.v1.Model.run', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6847,
  serialized_end=7205,
)


_PROMOTION = _descriptor.Descriptor(
  name='Promotion',
  full_name='continual.rpc.management.v1.Promotion',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Promotion.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Promotion.create_time', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='promoted_time', full_name='continual.rpc.management.v1.Promotion.promoted_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='demoted_time', full_name='continual.rpc.management.v1.Promotion.demoted_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_version', full_name='continual.rpc.management.v1.Promotion.model_version', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\014ModelVersion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='base_model_version', full_name='continual.rpc.management.v1.Promotion.base_model_version', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\014ModelVersion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='continual.rpc.management.v1.Promotion.state', index=6,
      number=7, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='improvement_metric', full_name='continual.rpc.management.v1.Promotion.improvement_metric', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='improvement_metric_value', full_name='continual.rpc.management.v1.Promotion.improvement_metric_value', index=8,
      number=9, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='base_improvement_metric_value', full_name='continual.rpc.management.v1.Promotion.base_improvement_metric_value', index=9,
      number=10, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='improvement_metric_diff', full_name='continual.rpc.management.v1.Promotion.improvement_metric_diff', index=10,
      number=11, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='reason', full_name='continual.rpc.management.v1.Promotion.reason', index=11,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message', full_name='continual.rpc.management.v1.Promotion.message', index=12,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='account', full_name='continual.rpc.management.v1.Promotion.account', index=13,
      number=14, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\tPrincipal', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run_name', full_name='continual.rpc.management.v1.Promotion.run_name', index=14,
      number=15, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _PROMOTION_REASON,
    _PROMOTION_STATE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7208,
  serialized_end=7997,
)


_MODELVERSION = _descriptor.Descriptor(
  name='ModelVersion',
  full_name='continual.rpc.management.v1.ModelVersion',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.ModelVersion.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run_name', full_name='continual.rpc.management.v1.ModelVersion.run_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='experiment_name', full_name='continual.rpc.management.v1.ModelVersion.experiment_name', index=2,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\nExperiment', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='continual.rpc.management.v1.ModelVersion.state', index=3,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='error_message', full_name='continual.rpc.management.v1.ModelVersion.error_message', index=4,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='stack_trace', full_name='continual.rpc.management.v1.ModelVersion.stack_trace', index=5,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.ModelVersion.create_time', index=6,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.ModelVersion.update_time', index=7,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ModelVersion.start_time', index=8,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ModelVersion.end_time', index=9,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='training_row_count', full_name='continual.rpc.management.v1.ModelVersion.training_row_count', index=10,
      number=12, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='validation_row_count', full_name='continual.rpc.management.v1.ModelVersion.validation_row_count', index=11,
      number=13, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='test_row_count', full_name='continual.rpc.management.v1.ModelVersion.test_row_count', index=12,
      number=14, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='performance_metric', full_name='continual.rpc.management.v1.ModelVersion.performance_metric', index=13,
      number=15, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='performance_metric_val', full_name='continual.rpc.management.v1.ModelVersion.performance_metric_val', index=14,
      number=16, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='promotion', full_name='continual.rpc.management.v1.ModelVersion.promotion', index=15,
      number=17, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\tPromotion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='promoted', full_name='continual.rpc.management.v1.ModelVersion.promoted', index=16,
      number=18, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _MODELVERSION_STATE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8000,
  serialized_end=8819,
)


_CONFUSIONMATRIX_ROW = _descriptor.Descriptor(
  name='Row',
  full_name='continual.rpc.management.v1.ConfusionMatrix.Row',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='example_count', full_name='continual.rpc.management.v1.ConfusionMatrix.Row.example_count', index=0,
      number=1, type=5, cpp_type=1, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8926,
  serialized_end=8954,
)

_CONFUSIONMATRIX = _descriptor.Descriptor(
  name='ConfusionMatrix',
  full_name='continual.rpc.management.v1.ConfusionMatrix',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='display_name', full_name='continual.rpc.management.v1.ConfusionMatrix.display_name', index=0,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='row', full_name='continual.rpc.management.v1.ConfusionMatrix.row', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_CONFUSIONMATRIX_ROW, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8822,
  serialized_end=8954,
)


_BOOLEANEVALUATIONMETRICS = _descriptor.Descriptor(
  name='BooleanEvaluationMetrics',
  full_name='continual.rpc.management.v1.BooleanEvaluationMetrics',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='log_loss', full_name='continual.rpc.management.v1.BooleanEvaluationMetrics.log_loss', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='accuracy', full_name='continual.rpc.management.v1.BooleanEvaluationMetrics.accuracy', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='roc_auc', full_name='continual.rpc.management.v1.BooleanEvaluationMetrics.roc_auc', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='error', full_name='continual.rpc.management.v1.BooleanEvaluationMetrics.error', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='confusion_matrix', full_name='continual.rpc.management.v1.BooleanEvaluationMetrics.confusion_matrix', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='precision', full_name='continual.rpc.management.v1.BooleanEvaluationMetrics.precision', index=5,
      number=6, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='recall', full_name='continual.rpc.management.v1.BooleanEvaluationMetrics.recall', index=6,
      number=7, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='f1', full_name='continual.rpc.management.v1.BooleanEvaluationMetrics.f1', index=7,
      number=8, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8957,
  serialized_end=9170,
)


_TEXTEVALUATIONMETRICS = _descriptor.Descriptor(
  name='TextEvaluationMetrics',
  full_name='continual.rpc.management.v1.TextEvaluationMetrics',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='log_loss', full_name='continual.rpc.management.v1.TextEvaluationMetrics.log_loss', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='f1', full_name='continual.rpc.management.v1.TextEvaluationMetrics.f1', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='perplexity', full_name='continual.rpc.management.v1.TextEvaluationMetrics.perplexity', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='edit_distance', full_name='continual.rpc.management.v1.TextEvaluationMetrics.edit_distance', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='token_accuracy', full_name='continual.rpc.management.v1.TextEvaluationMetrics.token_accuracy', index=4,
      number=5, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='bleu', full_name='continual.rpc.management.v1.TextEvaluationMetrics.bleu', index=5,
      number=6, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9173,
  serialized_end=9307,
)


_CATEGORICALEVALUATIONMETRICS = _descriptor.Descriptor(
  name='CategoricalEvaluationMetrics',
  full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='log_loss', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.log_loss', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='accuracy', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.accuracy', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='roc_auc', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.roc_auc', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='multiclass_error', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.multiclass_error', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='confusion_matrix', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.confusion_matrix', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='precision', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.precision', index=5,
      number=6, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='precision_micro', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.precision_micro', index=6,
      number=7, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='precision_macro', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.precision_macro', index=7,
      number=8, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='precision_weighted', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.precision_weighted', index=8,
      number=9, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='recall', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.recall', index=9,
      number=10, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='recall_micro', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.recall_micro', index=10,
      number=11, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='recall_macro', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.recall_macro', index=11,
      number=12, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='recall_weighted', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.recall_weighted', index=12,
      number=13, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='f1', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.f1', index=13,
      number=14, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='f1_micro', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.f1_micro', index=14,
      number=15, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='f1_macro', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.f1_macro', index=15,
      number=16, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='f1_weighted', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.f1_weighted', index=16,
      number=17, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='balanced_accuracy', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.balanced_accuracy', index=17,
      number=18, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='average_precision', full_name='continual.rpc.management.v1.CategoricalEvaluationMetrics.average_precision', index=18,
      number=19, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9310,
  serialized_end=9796,
)


_NUMBEREVALUATIONMETRICS = _descriptor.Descriptor(
  name='NumberEvaluationMetrics',
  full_name='continual.rpc.management.v1.NumberEvaluationMetrics',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='rmse', full_name='continual.rpc.management.v1.NumberEvaluationMetrics.rmse', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mae', full_name='continual.rpc.management.v1.NumberEvaluationMetrics.mae', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='r2', full_name='continual.rpc.management.v1.NumberEvaluationMetrics.r2', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='precision', full_name='continual.rpc.management.v1.NumberEvaluationMetrics.precision', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='recall', full_name='continual.rpc.management.v1.NumberEvaluationMetrics.recall', index=4,
      number=5, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mape', full_name='continual.rpc.management.v1.NumberEvaluationMetrics.mape', index=5,
      number=6, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='smape', full_name='continual.rpc.management.v1.NumberEvaluationMetrics.smape', index=6,
      number=7, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mean_squared_error', full_name='continual.rpc.management.v1.NumberEvaluationMetrics.mean_squared_error', index=7,
      number=8, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='median_absolute_error', full_name='continual.rpc.management.v1.NumberEvaluationMetrics.median_absolute_error', index=8,
      number=9, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9799,
  serialized_end=9986,
)


_EVALUATIONMETRICS = _descriptor.Descriptor(
  name='EvaluationMetrics',
  full_name='continual.rpc.management.v1.EvaluationMetrics',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='boolean', full_name='continual.rpc.management.v1.EvaluationMetrics.boolean', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='number', full_name='continual.rpc.management.v1.EvaluationMetrics.number', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='text', full_name='continual.rpc.management.v1.EvaluationMetrics.text', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='categorical', full_name='continual.rpc.management.v1.EvaluationMetrics.categorical', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='metrics', full_name='continual.rpc.management.v1.EvaluationMetrics.metrics',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=9989,
  serialized_end=10315,
)


_EXPERIMENT = _descriptor.Descriptor(
  name='Experiment',
  full_name='continual.rpc.management.v1.Experiment',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Experiment.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run_name', full_name='continual.rpc.management.v1.Experiment.run_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='continual.rpc.management.v1.Experiment.state', index=2,
      number=3, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Experiment.create_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.Experiment.update_time', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.Experiment.start_time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.Experiment.end_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='performance_metric_name', full_name='continual.rpc.management.v1.Experiment.performance_metric_name', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='performance_metric_val', full_name='continual.rpc.management.v1.Experiment.performance_metric_val', index=8,
      number=9, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='training_config', full_name='continual.rpc.management.v1.Experiment.training_config', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='error_message', full_name='continual.rpc.management.v1.Experiment.error_message', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='stack_trace', full_name='continual.rpc.management.v1.Experiment.stack_trace', index=11,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _EXPERIMENT_STATE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10318,
  serialized_end=10907,
)


_TEXTSEQUENCE = _descriptor.Descriptor(
  name='TextSequence',
  full_name='continual.rpc.management.v1.TextSequence',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.TextSequence.value', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10909,
  serialized_end=10938,
)


_NUMBERSEQUENCE = _descriptor.Descriptor(
  name='NumberSequence',
  full_name='continual.rpc.management.v1.NumberSequence',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.NumberSequence.value', index=0,
      number=1, type=1, cpp_type=5, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10940,
  serialized_end=10971,
)


_SIGNATUREFEATURE = _descriptor.Descriptor(
  name='SignatureFeature',
  full_name='continual.rpc.management.v1.SignatureFeature',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='qualified_name', full_name='continual.rpc.management.v1.SignatureFeature.qualified_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='continual.rpc.management.v1.SignatureFeature.type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10973,
  serialized_end=11069,
)


_QUERYFEATURE = _descriptor.Descriptor(
  name='QueryFeature',
  full_name='continual.rpc.management.v1.QueryFeature',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.QueryFeature.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='qualified_name', full_name='continual.rpc.management.v1.QueryFeature.qualified_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='relation', full_name='continual.rpc.management.v1.QueryFeature.relation', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.QueryFeature.value', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='continual.rpc.management.v1.QueryFeature.type', index=4,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dtype', full_name='continual.rpc.management.v1.QueryFeature.dtype', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model', full_name='continual.rpc.management.v1.QueryFeature.model', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='index', full_name='continual.rpc.management.v1.QueryFeature.index', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='time_index', full_name='continual.rpc.management.v1.QueryFeature.time_index', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='exclude', full_name='continual.rpc.management.v1.QueryFeature.exclude', index=9,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='split', full_name='continual.rpc.management.v1.QueryFeature.split', index=10,
      number=11, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11072,
  serialized_end=11344,
)


_QUERY = _descriptor.Descriptor(
  name='Query',
  full_name='continual.rpc.management.v1.Query',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Query.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='entity_id', full_name='continual.rpc.management.v1.Query.entity_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='table_name', full_name='continual.rpc.management.v1.Query.table_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fields', full_name='continual.rpc.management.v1.Query.fields', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='related', full_name='continual.rpc.management.v1.Query.related', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11347,
  serialized_end=11519,
)


_BATCHPREDICTION = _descriptor.Descriptor(
  name='BatchPrediction',
  full_name='continual.rpc.management.v1.BatchPrediction',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.BatchPrediction.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='continual.rpc.management.v1.BatchPrediction.state', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model_version', full_name='continual.rpc.management.v1.BatchPrediction.model_version', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\014ModelVersion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.BatchPrediction.create_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.BatchPrediction.start_time', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.BatchPrediction.end_time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='prediction_count', full_name='continual.rpc.management.v1.BatchPrediction.prediction_count', index=6,
      number=7, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='error_message', full_name='continual.rpc.management.v1.BatchPrediction.error_message', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='stack_trace', full_name='continual.rpc.management.v1.BatchPrediction.stack_trace', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run_name', full_name='continual.rpc.management.v1.BatchPrediction.run_name', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _BATCHPREDICTION_STATE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11522,
  serialized_end=12063,
)


_SIGNEDURL = _descriptor.Descriptor(
  name='SignedURL',
  full_name='continual.rpc.management.v1.SignedURL',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='signed_url', full_name='continual.rpc.management.v1.SignedURL.signed_url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12065,
  serialized_end=12096,
)


_ENTITY = _descriptor.Descriptor(
  name='Entity',
  full_name='continual.rpc.management.v1.Entity',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Entity.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='models', full_name='continual.rpc.management.v1.Entity.models', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_sets', full_name='continual.rpc.management.v1.Entity.feature_sets', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12098,
  serialized_end=12158,
)


_DATASET = _descriptor.Descriptor(
  name='Dataset',
  full_name='continual.rpc.management.v1.Dataset',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Dataset.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='continual.rpc.management.v1.Dataset.description', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='continual.rpc.management.v1.Dataset.state', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='author', full_name='continual.rpc.management.v1.Dataset.author', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='display_name', full_name='continual.rpc.management.v1.Dataset.display_name', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='current_version', full_name='continual.rpc.management.v1.Dataset.current_version', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\016DatasetVersion', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Dataset.create_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.Dataset.update_time', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run', full_name='continual.rpc.management.v1.Dataset.run', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12161,
  serialized_end=12423,
)


_DATASETVERSION = _descriptor.Descriptor(
  name='DatasetVersion',
  full_name='continual.rpc.management.v1.DatasetVersion',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DatasetVersion.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run', full_name='continual.rpc.management.v1.DatasetVersion.run', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.DatasetVersion.create_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.DatasetVersion.update_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12426,
  serialized_end=12582,
)


_DATASETVERSIONASSIGNMENT = _descriptor.Descriptor(
  name='DatasetVersionAssignment',
  full_name='continual.rpc.management.v1.DatasetVersionAssignment',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DatasetVersionAssignment.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource_name', full_name='continual.rpc.management.v1.DatasetVersionAssignment.resource_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.DatasetVersionAssignment.create_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12584,
  serialized_end=12696,
)


_DATAPROFILE = _descriptor.Descriptor(
  name='DataProfile',
  full_name='continual.rpc.management.v1.DataProfile',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DataProfile.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dataset_stats', full_name='continual.rpc.management.v1.DataProfile.dataset_stats', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.DataProfile.create_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12699,
  serialized_end=12846,
)


_DATASETSTATSENTRY = _descriptor.Descriptor(
  name='DatasetStatsEntry',
  full_name='continual.rpc.management.v1.DatasetStatsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='entry_name', full_name='continual.rpc.management.v1.DatasetStatsEntry.entry_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='num_examples', full_name='continual.rpc.management.v1.DatasetStatsEntry.num_examples', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='correlation_matrix', full_name='continual.rpc.management.v1.DatasetStatsEntry.correlation_matrix', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='string_stats', full_name='continual.rpc.management.v1.DatasetStatsEntry.string_stats', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='numeric_stats', full_name='continual.rpc.management.v1.DatasetStatsEntry.numeric_stats', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp_stats', full_name='continual.rpc.management.v1.DatasetStatsEntry.timestamp_stats', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='categorical_stats', full_name='continual.rpc.management.v1.DatasetStatsEntry.categorical_stats', index=6,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='duplicate_rows', full_name='continual.rpc.management.v1.DatasetStatsEntry.duplicate_rows', index=7,
      number=8, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='duplicate_indexes', full_name='continual.rpc.management.v1.DatasetStatsEntry.duplicate_indexes', index=8,
      number=9, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dataset_ranges', full_name='continual.rpc.management.v1.DatasetStatsEntry.dataset_ranges', index=9,
      number=10, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12849,
  serialized_end=13378,
)


_DATASETRANGE = _descriptor.Descriptor(
  name='DatasetRange',
  full_name='continual.rpc.management.v1.DatasetRange',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DatasetRange.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.DatasetRange.start_time', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.DatasetRange.end_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13380,
  serialized_end=13502,
)


_CORRELATIONMATRIX_ROW = _descriptor.Descriptor(
  name='Row',
  full_name='continual.rpc.management.v1.CorrelationMatrix.Row',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='correlations', full_name='continual.rpc.management.v1.CorrelationMatrix.Row.correlations', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13615,
  serialized_end=13642,
)

_CORRELATIONMATRIX = _descriptor.Descriptor(
  name='CorrelationMatrix',
  full_name='continual.rpc.management.v1.CorrelationMatrix',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='display_names', full_name='continual.rpc.management.v1.CorrelationMatrix.display_names', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='rows', full_name='continual.rpc.management.v1.CorrelationMatrix.rows', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_CORRELATIONMATRIX_ROW, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13505,
  serialized_end=13642,
)


_HISTOGRAM_BUCKET = _descriptor.Descriptor(
  name='Bucket',
  full_name='continual.rpc.management.v1.Histogram.Bucket',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='low_value', full_name='continual.rpc.management.v1.Histogram.Bucket.low_value', index=0,
      number=1, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='high_value', full_name='continual.rpc.management.v1.Histogram.Bucket.high_value', index=1,
      number=2, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sample_count', full_name='continual.rpc.management.v1.Histogram.Bucket.sample_count', index=2,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13762,
  serialized_end=13831,
)

_HISTOGRAM = _descriptor.Descriptor(
  name='Histogram',
  full_name='continual.rpc.management.v1.Histogram',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='num_nan', full_name='continual.rpc.management.v1.Histogram.num_nan', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='num_undefined', full_name='continual.rpc.management.v1.Histogram.num_undefined', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='buckets', full_name='continual.rpc.management.v1.Histogram.buckets', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_HISTOGRAM_BUCKET, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13645,
  serialized_end=13831,
)


_TIMESTAMPHISTOGRAM_BUCKET = _descriptor.Descriptor(
  name='Bucket',
  full_name='continual.rpc.management.v1.TimestampHistogram.Bucket',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='low_value', full_name='continual.rpc.management.v1.TimestampHistogram.Bucket.low_value', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='high_value', full_name='continual.rpc.management.v1.TimestampHistogram.Bucket.high_value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sample_count', full_name='continual.rpc.management.v1.TimestampHistogram.Bucket.sample_count', index=2,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13969,
  serialized_end=14094,
)

_TIMESTAMPHISTOGRAM = _descriptor.Descriptor(
  name='TimestampHistogram',
  full_name='continual.rpc.management.v1.TimestampHistogram',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='num_nan', full_name='continual.rpc.management.v1.TimestampHistogram.num_nan', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='num_undefined', full_name='continual.rpc.management.v1.TimestampHistogram.num_undefined', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='buckets', full_name='continual.rpc.management.v1.TimestampHistogram.buckets', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_TIMESTAMPHISTOGRAM_BUCKET, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13834,
  serialized_end=14094,
)


_CATEGORICALDISTRIBUTION_CATEGORY = _descriptor.Descriptor(
  name='Category',
  full_name='continual.rpc.management.v1.CategoricalDistribution.Category',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='label', full_name='continual.rpc.management.v1.CategoricalDistribution.Category.label', index=0,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sample_count', full_name='continual.rpc.management.v1.CategoricalDistribution.Category.sample_count', index=1,
      number=5, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14207,
  serialized_end=14254,
)

_CATEGORICALDISTRIBUTION = _descriptor.Descriptor(
  name='CategoricalDistribution',
  full_name='continual.rpc.management.v1.CategoricalDistribution',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='categories', full_name='continual.rpc.management.v1.CategoricalDistribution.categories', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_CATEGORICALDISTRIBUTION_CATEGORY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14097,
  serialized_end=14254,
)


_STRINGSTATS = _descriptor.Descriptor(
  name='StringStats',
  full_name='continual.rpc.management.v1.StringStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='column_name', full_name='continual.rpc.management.v1.StringStats.column_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='distinct_values', full_name='continual.rpc.management.v1.StringStats.distinct_values', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='null_count', full_name='continual.rpc.management.v1.StringStats.null_count', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='min_length', full_name='continual.rpc.management.v1.StringStats.min_length', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max_length', full_name='continual.rpc.management.v1.StringStats.max_length', index=4,
      number=5, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14256,
  serialized_end=14375,
)


_CATEGORICALSTATS = _descriptor.Descriptor(
  name='CategoricalStats',
  full_name='continual.rpc.management.v1.CategoricalStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='column_name', full_name='continual.rpc.management.v1.CategoricalStats.column_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='distinct_values', full_name='continual.rpc.management.v1.CategoricalStats.distinct_values', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='null_count', full_name='continual.rpc.management.v1.CategoricalStats.null_count', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='categorical_distribution', full_name='continual.rpc.management.v1.CategoricalStats.categorical_distribution', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14378,
  serialized_end=14550,
)


_TIMESTAMPSTATS = _descriptor.Descriptor(
  name='TimestampStats',
  full_name='continual.rpc.management.v1.TimestampStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='column_name', full_name='continual.rpc.management.v1.TimestampStats.column_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='distinct_values', full_name='continual.rpc.management.v1.TimestampStats.distinct_values', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='null_count', full_name='continual.rpc.management.v1.TimestampStats.null_count', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='min', full_name='continual.rpc.management.v1.TimestampStats.min', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max', full_name='continual.rpc.management.v1.TimestampStats.max', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp_histogram', full_name='continual.rpc.management.v1.TimestampStats.timestamp_histogram', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14553,
  serialized_end=14795,
)


_NUMERICSTATS = _descriptor.Descriptor(
  name='NumericStats',
  full_name='continual.rpc.management.v1.NumericStats',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='column_name', full_name='continual.rpc.management.v1.NumericStats.column_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='distinct_values', full_name='continual.rpc.management.v1.NumericStats.distinct_values', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='min', full_name='continual.rpc.management.v1.NumericStats.min', index=2,
      number=3, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max', full_name='continual.rpc.management.v1.NumericStats.max', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sd', full_name='continual.rpc.management.v1.NumericStats.sd', index=4,
      number=5, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mean', full_name='continual.rpc.management.v1.NumericStats.mean', index=5,
      number=6, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='count', full_name='continual.rpc.management.v1.NumericStats.count', index=6,
      number=7, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='percentile25', full_name='continual.rpc.management.v1.NumericStats.percentile25', index=7,
      number=8, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='percentile50', full_name='continual.rpc.management.v1.NumericStats.percentile50', index=8,
      number=9, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='percentile75', full_name='continual.rpc.management.v1.NumericStats.percentile75', index=9,
      number=10, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='null_count', full_name='continual.rpc.management.v1.NumericStats.null_count', index=10,
      number=11, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='num_nan', full_name='continual.rpc.management.v1.NumericStats.num_nan', index=11,
      number=12, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='num_values_histogram', full_name='continual.rpc.management.v1.NumericStats.num_values_histogram', index=12,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14798,
  serialized_end=15098,
)


_BILLINGACCOUNT = _descriptor.Descriptor(
  name='BillingAccount',
  full_name='continual.rpc.management.v1.BillingAccount',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.BillingAccount.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='email', full_name='continual.rpc.management.v1.BillingAccount.email', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='memo', full_name='continual.rpc.management.v1.BillingAccount.memo', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.BillingAccount.update_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.BillingAccount.create_time', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15101,
  serialized_end=15268,
)


_SUBSCRIPTION = _descriptor.Descriptor(
  name='Subscription',
  full_name='continual.rpc.management.v1.Subscription',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Subscription.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='price', full_name='continual.rpc.management.v1.Subscription.price', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\005Price', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='payment_method', full_name='continual.rpc.management.v1.Subscription.payment_method', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\302E\rPaymentMethod', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.Subscription.update_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Subscription.create_time', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=b'\270E\001', file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='active', full_name='continual.rpc.management.v1.Subscription.active', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15271,
  serialized_end=15490,
)


_PROJECTUSAGERECORD_TYPECOUNTENTRY = _descriptor.Descriptor(
  name='TypeCountEntry',
  full_name='continual.rpc.management.v1.ProjectUsageRecord.TypeCountEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='continual.rpc.management.v1.ProjectUsageRecord.TypeCountEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.ProjectUsageRecord.TypeCountEntry.value', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15629,
  serialized_end=15677,
)

_PROJECTUSAGERECORD = _descriptor.Descriptor(
  name='ProjectUsageRecord',
  full_name='continual.rpc.management.v1.ProjectUsageRecord',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.ProjectUsageRecord.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='credits', full_name='continual.rpc.management.v1.ProjectUsageRecord.credits', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='typeCount', full_name='continual.rpc.management.v1.ProjectUsageRecord.typeCount', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_PROJECTUSAGERECORD_TYPECOUNTENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15493,
  serialized_end=15677,
)


_CREDITUSAGERECORD = _descriptor.Descriptor(
  name='CreditUsageRecord',
  full_name='continual.rpc.management.v1.CreditUsageRecord',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='period', full_name='continual.rpc.management.v1.CreditUsageRecord.period', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='credits', full_name='continual.rpc.management.v1.CreditUsageRecord.credits', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='project_usages', full_name='continual.rpc.management.v1.CreditUsageRecord.project_usages', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15680,
  serialized_end=15833,
)


_INVOICELINEITEM = _descriptor.Descriptor(
  name='InvoiceLineItem',
  full_name='continual.rpc.management.v1.InvoiceLineItem',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='continual.rpc.management.v1.InvoiceLineItem.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='amount', full_name='continual.rpc.management.v1.InvoiceLineItem.amount', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='quantity', full_name='continual.rpc.management.v1.InvoiceLineItem.quantity', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='currency', full_name='continual.rpc.management.v1.InvoiceLineItem.currency', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='continual.rpc.management.v1.InvoiceLineItem.description', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='period_start', full_name='continual.rpc.management.v1.InvoiceLineItem.period_start', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='period_end', full_name='continual.rpc.management.v1.InvoiceLineItem.period_end', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15836,
  serialized_end=16036,
)


_INVOICE = _descriptor.Descriptor(
  name='Invoice',
  full_name='continual.rpc.management.v1.Invoice',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Invoice.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='date', full_name='continual.rpc.management.v1.Invoice.date', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='amount', full_name='continual.rpc.management.v1.Invoice.amount', index=2,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='period_start', full_name='continual.rpc.management.v1.Invoice.period_start', index=3,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='period_end', full_name='continual.rpc.management.v1.Invoice.period_end', index=4,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status', full_name='continual.rpc.management.v1.Invoice.status', index=5,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='invoice_url', full_name='continual.rpc.management.v1.Invoice.invoice_url', index=6,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='line_items', full_name='continual.rpc.management.v1.Invoice.line_items', index=7,
      number=9, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16039,
  serialized_end=16321,
)


_PRICE = _descriptor.Descriptor(
  name='Price',
  full_name='continual.rpc.management.v1.Price',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Price.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='title', full_name='continual.rpc.management.v1.Price.title', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='continual.rpc.management.v1.Price.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='direct_invoice', full_name='continual.rpc.management.v1.Price.direct_invoice', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16323,
  serialized_end=16404,
)


_PAYMENTMETHOD_BILLINGDETAILS = _descriptor.Descriptor(
  name='BillingDetails',
  full_name='continual.rpc.management.v1.PaymentMethod.BillingDetails',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='postal_code', full_name='continual.rpc.management.v1.PaymentMethod.BillingDetails.postal_code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16585,
  serialized_end=16622,
)

_PAYMENTMETHOD_CARD = _descriptor.Descriptor(
  name='Card',
  full_name='continual.rpc.management.v1.PaymentMethod.Card',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='brand', full_name='continual.rpc.management.v1.PaymentMethod.Card.brand', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last4', full_name='continual.rpc.management.v1.PaymentMethod.Card.last4', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='exp_year', full_name='continual.rpc.management.v1.PaymentMethod.Card.exp_year', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='exp_month', full_name='continual.rpc.management.v1.PaymentMethod.Card.exp_month', index=3,
      number=4, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16624,
  serialized_end=16697,
)

_PAYMENTMETHOD = _descriptor.Descriptor(
  name='PaymentMethod',
  full_name='continual.rpc.management.v1.PaymentMethod',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.PaymentMethod.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='billing_details', full_name='continual.rpc.management.v1.PaymentMethod.billing_details', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='card', full_name='continual.rpc.management.v1.PaymentMethod.card', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_PAYMENTMETHOD_BILLINGDETAILS, _PAYMENTMETHOD_CARD, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16407,
  serialized_end=16697,
)


_ARTIFACT = _descriptor.Descriptor(
  name='Artifact',
  full_name='continual.rpc.management.v1.Artifact',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Artifact.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run', full_name='continual.rpc.management.v1.Artifact.run', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='external', full_name='continual.rpc.management.v1.Artifact.external', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='path', full_name='continual.rpc.management.v1.Artifact.path', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='url', full_name='continual.rpc.management.v1.Artifact.url', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mime_type', full_name='continual.rpc.management.v1.Artifact.mime_type', index=5,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Artifact.create_time', index=6,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.Artifact.update_time', index=7,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='size', full_name='continual.rpc.management.v1.Artifact.size', index=8,
      number=11, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16700,
  serialized_end=16913,
)


_CRONJOB = _descriptor.Descriptor(
  name='CronJob',
  full_name='continual.rpc.management.v1.CronJob',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='organization', full_name='continual.rpc.management.v1.CronJob.organization', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='model', full_name='continual.rpc.management.v1.CronJob.model', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='job_type', full_name='continual.rpc.management.v1.CronJob.job_type', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schedule', full_name='continual.rpc.management.v1.CronJob.schedule', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='organization_status', full_name='continual.rpc.management.v1.CronJob.organization_status', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status', full_name='continual.rpc.management.v1.CronJob.status', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='workflow_id', full_name='continual.rpc.management.v1.CronJob.workflow_id', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run_id', full_name='continual.rpc.management.v1.CronJob.run_id', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16916,
  serialized_end=17080,
)


_USERINTEGRATIONPROVIDER = _descriptor.Descriptor(
  name='UserIntegrationProvider',
  full_name='continual.rpc.management.v1.UserIntegrationProvider',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='provider', full_name='continual.rpc.management.v1.UserIntegrationProvider.provider', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.UserIntegrationProvider.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='title', full_name='continual.rpc.management.v1.UserIntegrationProvider.title', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='integration_install_url', full_name='continual.rpc.management.v1.UserIntegrationProvider.integration_install_url', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17082,
  serialized_end=17187,
)


_USERINTEGRATION = _descriptor.Descriptor(
  name='UserIntegration',
  full_name='continual.rpc.management.v1.UserIntegration',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.UserIntegration.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.UserIntegration.create_time', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.UserIntegration.update_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='provider', full_name='continual.rpc.management.v1.UserIntegration.provider', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='provider_name', full_name='continual.rpc.management.v1.UserIntegration.provider_name', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='config', full_name='continual.rpc.management.v1.UserIntegration.config', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data', full_name='continual.rpc.management.v1.UserIntegration.data', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data_update_time', full_name='continual.rpc.management.v1.UserIntegration.data_update_time', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='account_name', full_name='continual.rpc.management.v1.UserIntegration.account_name', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='account_full_name', full_name='continual.rpc.management.v1.UserIntegration.account_full_name', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='account_primary_email', full_name='continual.rpc.management.v1.UserIntegration.account_primary_email', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='account_avatar_url', full_name='continual.rpc.management.v1.UserIntegration.account_avatar_url', index=11,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='account_url', full_name='continual.rpc.management.v1.UserIntegration.account_url', index=12,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='account_config_url', full_name='continual.rpc.management.v1.UserIntegration.account_config_url', index=13,
      number=14, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='disconnect_time', full_name='continual.rpc.management.v1.UserIntegration.disconnect_time', index=14,
      number=15, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='disconnect_reason', full_name='continual.rpc.management.v1.UserIntegration.disconnect_reason', index=15,
      number=16, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='disconnect_data', full_name='continual.rpc.management.v1.UserIntegration.disconnect_data', index=16,
      number=17, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17190,
  serialized_end=17706,
)


_GITPROVIDER = _descriptor.Descriptor(
  name='GitProvider',
  full_name='continual.rpc.management.v1.GitProvider',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GitProvider.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='continual.rpc.management.v1.GitProvider.type', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='app_name', full_name='continual.rpc.management.v1.GitProvider.app_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='app_external_id', full_name='continual.rpc.management.v1.GitProvider.app_external_id', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='installation_name', full_name='continual.rpc.management.v1.GitProvider.installation_name', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='installation_external_id', full_name='continual.rpc.management.v1.GitProvider.installation_external_id', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17709,
  serialized_end=17854,
)


_GITPROVIDERNAMESPACE = _descriptor.Descriptor(
  name='GitProviderNamespace',
  full_name='continual.rpc.management.v1.GitProviderNamespace',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GitProviderNamespace.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='provider', full_name='continual.rpc.management.v1.GitProviderNamespace.provider', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='external_id', full_name='continual.rpc.management.v1.GitProviderNamespace.external_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='continual.rpc.management.v1.GitProviderNamespace.type', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='slug', full_name='continual.rpc.management.v1.GitProviderNamespace.slug', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='title', full_name='continual.rpc.management.v1.GitProviderNamespace.title', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='avatar_url', full_name='continual.rpc.management.v1.GitProviderNamespace.avatar_url', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='config_url', full_name='continual.rpc.management.v1.GitProviderNamespace.config_url', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='access_type', full_name='continual.rpc.management.v1.GitProviderNamespace.access_type', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17857,
  serialized_end=18078,
)


_GITPROVIDERREPOSITORY = _descriptor.Descriptor(
  name='GitProviderRepository',
  full_name='continual.rpc.management.v1.GitProviderRepository',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GitProviderRepository.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='provider_namespace', full_name='continual.rpc.management.v1.GitProviderRepository.provider_namespace', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='external_id', full_name='continual.rpc.management.v1.GitProviderRepository.external_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='slug', full_name='continual.rpc.management.v1.GitProviderRepository.slug', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='title', full_name='continual.rpc.management.v1.GitProviderRepository.title', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='default_branch_name', full_name='continual.rpc.management.v1.GitProviderRepository.default_branch_name', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='access_type', full_name='continual.rpc.management.v1.GitProviderRepository.access_type', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='url', full_name='continual.rpc.management.v1.GitProviderRepository.url', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data', full_name='continual.rpc.management.v1.GitProviderRepository.data', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data_update_time', full_name='continual.rpc.management.v1.GitProviderRepository.data_update_time', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18081,
  serialized_end=18378,
)


_PROJECTGITINTEGRATION = _descriptor.Descriptor(
  name='ProjectGitIntegration',
  full_name='continual.rpc.management.v1.ProjectGitIntegration',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.ProjectGitIntegration.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.ProjectGitIntegration.create_time', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.ProjectGitIntegration.update_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='project', full_name='continual.rpc.management.v1.ProjectGitIntegration.project', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='git_provider', full_name='continual.rpc.management.v1.ProjectGitIntegration.git_provider', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='git_provider_namespace', full_name='continual.rpc.management.v1.ProjectGitIntegration.git_provider_namespace', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='git_provider_repository', full_name='continual.rpc.management.v1.ProjectGitIntegration.git_provider_repository', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='created_by_user', full_name='continual.rpc.management.v1.ProjectGitIntegration.created_by_user', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='created_by_user_integration', full_name='continual.rpc.management.v1.ProjectGitIntegration.created_by_user_integration', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='disconnect_time', full_name='continual.rpc.management.v1.ProjectGitIntegration.disconnect_time', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='disconnect_reason', full_name='continual.rpc.management.v1.ProjectGitIntegration.disconnect_reason', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='disconnect_data', full_name='continual.rpc.management.v1.ProjectGitIntegration.disconnect_data', index=11,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18381,
  serialized_end=18932,
)


_METRIC = _descriptor.Descriptor(
  name='Metric',
  full_name='continual.rpc.management.v1.Metric',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Metric.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='display_name', full_name='continual.rpc.management.v1.Metric.display_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run', full_name='continual.rpc.management.v1.Metric.run', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Metric.create_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.Metric.update_time', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='direction', full_name='continual.rpc.management.v1.Metric.direction', index=5,
      number=6, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='values', full_name='continual.rpc.management.v1.Metric.values', index=6,
      number=7, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='created_by', full_name='continual.rpc.management.v1.Metric.created_by', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _METRIC_DIRECTION,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18935,
  serialized_end=19287,
)


_METRICVALUE = _descriptor.Descriptor(
  name='MetricValue',
  full_name='continual.rpc.management.v1.MetricValue',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.MetricValue.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run', full_name='continual.rpc.management.v1.MetricValue.run', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='step', full_name='continual.rpc.management.v1.MetricValue.step', index=2,
      number=3, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.MetricValue.value', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='continual.rpc.management.v1.MetricValue.timestamp', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group', full_name='continual.rpc.management.v1.MetricValue.group', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.MetricValue.create_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='created_by', full_name='continual.rpc.management.v1.MetricValue.created_by', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='label', full_name='continual.rpc.management.v1.MetricValue.label', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19290,
  serialized_end=19505,
)


_METADATA = _descriptor.Descriptor(
  name='Metadata',
  full_name='continual.rpc.management.v1.Metadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Metadata.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run', full_name='continual.rpc.management.v1.Metadata.run', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Metadata.create_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.Metadata.update_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data', full_name='continual.rpc.management.v1.Metadata.data', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19508,
  serialized_end=19657,
)


_GITCONTEXT_REPOSITORY = _descriptor.Descriptor(
  name='Repository',
  full_name='continual.rpc.management.v1.GitContext.Repository',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GitContext.Repository.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name_local', full_name='continual.rpc.management.v1.GitContext.Repository.name_local', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name_remote_origin', full_name='continual.rpc.management.v1.GitContext.Repository.name_remote_origin', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='remote_origin_url', full_name='continual.rpc.management.v1.GitContext.Repository.remote_origin_url', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_bare', full_name='continual.rpc.management.v1.GitContext.Repository.is_bare', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20073,
  serialized_end=20191,
)

_GITCONTEXT_COMMIT = _descriptor.Descriptor(
  name='Commit',
  full_name='continual.rpc.management.v1.GitContext.Commit',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='sha', full_name='continual.rpc.management.v1.GitContext.Commit.sha', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message', full_name='continual.rpc.management.v1.GitContext.Commit.message', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='authored_datetime', full_name='continual.rpc.management.v1.GitContext.Commit.authored_datetime', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='author', full_name='continual.rpc.management.v1.GitContext.Commit.author', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='committed_datetime', full_name='continual.rpc.management.v1.GitContext.Commit.committed_datetime', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='committer', full_name='continual.rpc.management.v1.GitContext.Commit.committer', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20194,
  serialized_end=20477,
)

_GITCONTEXT_BRANCH = _descriptor.Descriptor(
  name='Branch',
  full_name='continual.rpc.management.v1.GitContext.Branch',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GitContext.Branch.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20479,
  serialized_end=20501,
)

_GITCONTEXT_GITTAG = _descriptor.Descriptor(
  name='GitTag',
  full_name='continual.rpc.management.v1.GitContext.GitTag',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GitContext.GitTag.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='name_describe', full_name='continual.rpc.management.v1.GitContext.GitTag.name_describe', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20503,
  serialized_end=20548,
)

_GITCONTEXT_GITTAGS = _descriptor.Descriptor(
  name='GitTags',
  full_name='continual.rpc.management.v1.GitContext.GitTags',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='current', full_name='continual.rpc.management.v1.GitContext.GitTags.current', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='reachable', full_name='continual.rpc.management.v1.GitContext.GitTags.reachable', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20551,
  serialized_end=20692,
)

_GITCONTEXT_REF = _descriptor.Descriptor(
  name='Ref',
  full_name='continual.rpc.management.v1.GitContext.Ref',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='path', full_name='continual.rpc.management.v1.GitContext.Ref.path', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20694,
  serialized_end=20713,
)

_GITCONTEXT_AUTHOR = _descriptor.Descriptor(
  name='Author',
  full_name='continual.rpc.management.v1.GitContext.Author',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GitContext.Author.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='email', full_name='continual.rpc.management.v1.GitContext.Author.email', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20715,
  serialized_end=20752,
)

_GITCONTEXT_COMMITTER = _descriptor.Descriptor(
  name='Committer',
  full_name='continual.rpc.management.v1.GitContext.Committer',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GitContext.Committer.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='email', full_name='continual.rpc.management.v1.GitContext.Committer.email', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20754,
  serialized_end=20794,
)

_GITCONTEXT_HEAD = _descriptor.Descriptor(
  name='Head',
  full_name='continual.rpc.management.v1.GitContext.Head',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='is_detached', full_name='continual.rpc.management.v1.GitContext.Head.is_detached', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20796,
  serialized_end=20823,
)

_GITCONTEXT = _descriptor.Descriptor(
  name='GitContext',
  full_name='continual.rpc.management.v1.GitContext',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='repository', full_name='continual.rpc.management.v1.GitContext.repository', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='branch', full_name='continual.rpc.management.v1.GitContext.branch', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='tags', full_name='continual.rpc.management.v1.GitContext.tags', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ref', full_name='continual.rpc.management.v1.GitContext.ref', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='commit', full_name='continual.rpc.management.v1.GitContext.commit', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='head', full_name='continual.rpc.management.v1.GitContext.head', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='is_dirty', full_name='continual.rpc.management.v1.GitContext.is_dirty', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_GITCONTEXT_REPOSITORY, _GITCONTEXT_COMMIT, _GITCONTEXT_BRANCH, _GITCONTEXT_GITTAG, _GITCONTEXT_GITTAGS, _GITCONTEXT_REF, _GITCONTEXT_AUTHOR, _GITCONTEXT_COMMITTER, _GITCONTEXT_HEAD, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19660,
  serialized_end=20823,
)


_GITPROVIDERCONTEXT_REPOSITORY = _descriptor.Descriptor(
  name='Repository',
  full_name='continual.rpc.management.v1.GitProviderContext.Repository',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GitProviderContext.Repository.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='url', full_name='continual.rpc.management.v1.GitProviderContext.Repository.url', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20946,
  serialized_end=20985,
)

_GITPROVIDERCONTEXT = _descriptor.Descriptor(
  name='GitProviderContext',
  full_name='continual.rpc.management.v1.GitProviderContext',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='provider', full_name='continual.rpc.management.v1.GitProviderContext.provider', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='repository', full_name='continual.rpc.management.v1.GitProviderContext.repository', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_GITPROVIDERCONTEXT_REPOSITORY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20826,
  serialized_end=20985,
)


_CICDRUNNERCONTEXT = _descriptor.Descriptor(
  name='CicdRunnerContext',
  full_name='continual.rpc.management.v1.CicdRunnerContext',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='provider', full_name='continual.rpc.management.v1.CicdRunnerContext.provider', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run_id', full_name='continual.rpc.management.v1.CicdRunnerContext.run_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run_url', full_name='continual.rpc.management.v1.CicdRunnerContext.run_url', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20987,
  serialized_end=21057,
)


_SDLCREVIEWCONTEXT_USER = _descriptor.Descriptor(
  name='User',
  full_name='continual.rpc.management.v1.SdlcReviewContext.User',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_name', full_name='continual.rpc.management.v1.SdlcReviewContext.User.user_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='url', full_name='continual.rpc.management.v1.SdlcReviewContext.User.url', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='avatar_url', full_name='continual.rpc.management.v1.SdlcReviewContext.User.avatar_url', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21367,
  serialized_end=21425,
)

_SDLCREVIEWCONTEXT_REF = _descriptor.Descriptor(
  name='Ref',
  full_name='continual.rpc.management.v1.SdlcReviewContext.Ref',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='branch', full_name='continual.rpc.management.v1.SdlcReviewContext.Ref.branch', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sha', full_name='continual.rpc.management.v1.SdlcReviewContext.Ref.sha', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21427,
  serialized_end=21461,
)

_SDLCREVIEWCONTEXT = _descriptor.Descriptor(
  name='SdlcReviewContext',
  full_name='continual.rpc.management.v1.SdlcReviewContext',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='provider', full_name='continual.rpc.management.v1.SdlcReviewContext.provider', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='id', full_name='continual.rpc.management.v1.SdlcReviewContext.id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='url', full_name='continual.rpc.management.v1.SdlcReviewContext.url', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='continual.rpc.management.v1.SdlcReviewContext.state', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='title', full_name='continual.rpc.management.v1.SdlcReviewContext.title', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user', full_name='continual.rpc.management.v1.SdlcReviewContext.user', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='base', full_name='continual.rpc.management.v1.SdlcReviewContext.base', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='head', full_name='continual.rpc.management.v1.SdlcReviewContext.head', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data', full_name='continual.rpc.management.v1.SdlcReviewContext.data', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_SDLCREVIEWCONTEXT_USER, _SDLCREVIEWCONTEXT_REF, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21060,
  serialized_end=21461,
)


_RUNSUMMARY_RESOURCE = _descriptor.Descriptor(
  name='Resource',
  full_name='continual.rpc.management.v1.RunSummary.Resource',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.RunSummary.Resource.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='continual.rpc.management.v1.RunSummary.Resource.state', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22367,
  serialized_end=22406,
)

_RUNSUMMARY = _descriptor.Descriptor(
  name='RunSummary',
  full_name='continual.rpc.management.v1.RunSummary',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model_versions', full_name='continual.rpc.management.v1.RunSummary.model_versions', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='batch_predictions', full_name='continual.rpc.management.v1.RunSummary.batch_predictions', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='promotions', full_name='continual.rpc.management.v1.RunSummary.promotions', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='deployments', full_name='continual.rpc.management.v1.RunSummary.deployments', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='git_context', full_name='continual.rpc.management.v1.RunSummary.git_context', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='principal', full_name='continual.rpc.management.v1.RunSummary.principal', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.RunSummary.create_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='associated_models', full_name='continual.rpc.management.v1.RunSummary.associated_models', index=7,
      number=8, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='associated_datasets', full_name='continual.rpc.management.v1.RunSummary.associated_datasets', index=8,
      number=9, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dataset_versions', full_name='continual.rpc.management.v1.RunSummary.dataset_versions', index=9,
      number=10, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='git_provider_context', full_name='continual.rpc.management.v1.RunSummary.git_provider_context', index=10,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='cicd_runner_context', full_name='continual.rpc.management.v1.RunSummary.cicd_runner_context', index=11,
      number=12, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='sdlc_review_context', full_name='continual.rpc.management.v1.RunSummary.sdlc_review_context', index=12,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_RUNSUMMARY_RESOURCE, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21464,
  serialized_end=22406,
)


_RUN = _descriptor.Descriptor(
  name='Run',
  full_name='continual.rpc.management.v1.Run',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Run.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='continual.rpc.management.v1.Run.description', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='author', full_name='continual.rpc.management.v1.Run.author', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Run.create_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.Run.update_time', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_heartbeat', full_name='continual.rpc.management.v1.Run.last_heartbeat', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='heartbeat_interval', full_name='continual.rpc.management.v1.Run.heartbeat_interval', index=6,
      number=7, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='state', full_name='continual.rpc.management.v1.Run.state', index=7,
      number=8, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='error_message', full_name='continual.rpc.management.v1.Run.error_message', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='complete_time', full_name='continual.rpc.management.v1.Run.complete_time', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _RUN_STATE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22409,
  serialized_end=22847,
)


_TAG = _descriptor.Descriptor(
  name='Tag',
  full_name='continual.rpc.management.v1.Tag',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Tag.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='key', full_name='continual.rpc.management.v1.Tag.key', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.Tag.value', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Tag.create_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.Tag.update_time', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22850,
  serialized_end=22995,
)


_NOTIFICATIONPOLICYCONFIG_EMAIL = _descriptor.Descriptor(
  name='Email',
  full_name='continual.rpc.management.v1.NotificationPolicyConfig.Email',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='email_address', full_name='continual.rpc.management.v1.NotificationPolicyConfig.Email.email_address', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23178,
  serialized_end=23208,
)

_NOTIFICATIONPOLICYCONFIG_SLACK = _descriptor.Descriptor(
  name='Slack',
  full_name='continual.rpc.management.v1.NotificationPolicyConfig.Slack',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='channel', full_name='continual.rpc.management.v1.NotificationPolicyConfig.Slack.channel', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23210,
  serialized_end=23234,
)

_NOTIFICATIONPOLICYCONFIG = _descriptor.Descriptor(
  name='NotificationPolicyConfig',
  full_name='continual.rpc.management.v1.NotificationPolicyConfig',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='email', full_name='continual.rpc.management.v1.NotificationPolicyConfig.email', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='slack', full_name='continual.rpc.management.v1.NotificationPolicyConfig.slack', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_NOTIFICATIONPOLICYCONFIG_EMAIL, _NOTIFICATIONPOLICYCONFIG_SLACK, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22998,
  serialized_end=23234,
)


_NOTIFICATIONPOLICY = _descriptor.Descriptor(
  name='NotificationPolicy',
  full_name='continual.rpc.management.v1.NotificationPolicy',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.NotificationPolicy.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='continual.rpc.management.v1.NotificationPolicy.description', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='config', full_name='continual.rpc.management.v1.NotificationPolicy.config', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='continual.rpc.management.v1.NotificationPolicy.type', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='severity', full_name='continual.rpc.management.v1.NotificationPolicy.severity', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource', full_name='continual.rpc.management.v1.NotificationPolicy.resource', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource_type', full_name='continual.rpc.management.v1.NotificationPolicy.resource_type', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='event_type', full_name='continual.rpc.management.v1.NotificationPolicy.event_type', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='require_notify', full_name='continual.rpc.management.v1.NotificationPolicy.require_notify', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.NotificationPolicy.create_time', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_time', full_name='continual.rpc.management.v1.NotificationPolicy.update_time', index=10,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _NOTIFICATIONPOLICY_TYPE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23237,
  serialized_end=23662,
)


_NOTIFICATION = _descriptor.Descriptor(
  name='Notification',
  full_name='continual.rpc.management.v1.Notification',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.Notification.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='type', full_name='continual.rpc.management.v1.Notification.type', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='event', full_name='continual.rpc.management.v1.Notification.event', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='target', full_name='continual.rpc.management.v1.Notification.target', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='result', full_name='continual.rpc.management.v1.Notification.result', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.Notification.create_time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23665,
  serialized_end=23857,
)


_RUNCHECKSUMMARY = _descriptor.Descriptor(
  name='RunCheckSummary',
  full_name='continual.rpc.management.v1.RunCheckSummary',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='run_name', full_name='continual.rpc.management.v1.RunCheckSummary.run_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='outcome', full_name='continual.rpc.management.v1.RunCheckSummary.outcome', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource_checks', full_name='continual.rpc.management.v1.RunCheckSummary.resource_checks', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=23860,
  serialized_end=24024,
)


_RESOURCECHECK = _descriptor.Descriptor(
  name='ResourceCheck',
  full_name='continual.rpc.management.v1.ResourceCheck',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.ResourceCheck.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run_name', full_name='continual.rpc.management.v1.ResourceCheck.run_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='group_name', full_name='continual.rpc.management.v1.ResourceCheck.group_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='outcome', full_name='continual.rpc.management.v1.ResourceCheck.outcome', index=3,
      number=5, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='duration', full_name='continual.rpc.management.v1.ResourceCheck.duration', index=4,
      number=6, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data', full_name='continual.rpc.management.v1.ResourceCheck.data', index=5,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='errors', full_name='continual.rpc.management.v1.ResourceCheck.errors', index=6,
      number=8, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='warnings', full_name='continual.rpc.management.v1.ResourceCheck.warnings', index=7,
      number=9, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='infos', full_name='continual.rpc.management.v1.ResourceCheck.infos', index=8,
      number=10, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='summary', full_name='continual.rpc.management.v1.ResourceCheck.summary', index=9,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='artifact_name', full_name='continual.rpc.management.v1.ResourceCheck.artifact_name', index=10,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='continual.rpc.management.v1.ResourceCheck.create_time', index=11,
      number=13, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='display_name', full_name='continual.rpc.management.v1.ResourceCheck.display_name', index=12,
      number=14, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=24027,
  serialized_end=24346,
)

_USER.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_USER.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PRINCIPAL.fields_by_name['type'].enum_type = _PRINCIPAL_TYPE
_PRINCIPAL_TYPE.containing_type = _PRINCIPAL
_SESSION.fields_by_name['last_access_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_SESSION.fields_by_name['login_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_SESSION.fields_by_name['state'].enum_type = _SESSION_STATE
_SESSION_STATE.containing_type = _SESSION
_ACCESSPOLICY.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_INVITE.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_INVITE.fields_by_name['expire_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_ORGANIZATION.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_ORGANIZATION.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_ORGANIZATION.fields_by_name['status'].enum_type = _ORGANIZATION_STATUS
_ORGANIZATION.fields_by_name['trial_end_date'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_ORGANIZATION_STATUS.containing_type = _ORGANIZATION
_ONBOARDING.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_ONBOARDING.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PROJECT.fields_by_name['summary'].message_type = _PROJECTSUMMARY
_PROJECT.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PROJECT.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PROJECT.fields_by_name['create_source_type'].enum_type = _PROJECT_CREATESOURCETYPE
_PROJECT_CREATESOURCETYPE.containing_type = _PROJECT
_PROJECTSUMMARY.fields_by_name['feature_set_health'].message_type = _HEALTHCOUNTS
_PROJECTSUMMARY.fields_by_name['model_health'].message_type = _HEALTHCOUNTS
_PROJECTSUMMARY.fields_by_name['last_deployment_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PROJECTSUMMARY.fields_by_name['last_training_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_DATASTORE_BIGQUERY.containing_type = _DATASTORE
_DATASTORE_SNOWFLAKE_PARAMSENTRY.containing_type = _DATASTORE_SNOWFLAKE
_DATASTORE_SNOWFLAKE.fields_by_name['params'].message_type = _DATASTORE_SNOWFLAKE_PARAMSENTRY
_DATASTORE_SNOWFLAKE.containing_type = _DATASTORE
_DATASTORE_POSTGRES_PARAMSENTRY.containing_type = _DATASTORE_POSTGRES
_DATASTORE_POSTGRES.fields_by_name['params'].message_type = _DATASTORE_POSTGRES_PARAMSENTRY
_DATASTORE_POSTGRES.containing_type = _DATASTORE
_DATASTORE_REDSHIFT_PARAMSENTRY.containing_type = _DATASTORE_REDSHIFT
_DATASTORE_REDSHIFT.fields_by_name['params'].message_type = _DATASTORE_REDSHIFT_PARAMSENTRY
_DATASTORE_REDSHIFT.containing_type = _DATASTORE
_DATASTORE_DATABRICKS_PARAMSENTRY.containing_type = _DATASTORE_DATABRICKS
_DATASTORE_DATABRICKS.fields_by_name['params'].message_type = _DATASTORE_DATABRICKS_PARAMSENTRY
_DATASTORE_DATABRICKS.containing_type = _DATASTORE
_DATASTORE.fields_by_name['big_query'].message_type = _DATASTORE_BIGQUERY
_DATASTORE.fields_by_name['snowflake'].message_type = _DATASTORE_SNOWFLAKE
_DATASTORE.fields_by_name['postgres'].message_type = _DATASTORE_POSTGRES
_DATASTORE.fields_by_name['redshift'].message_type = _DATASTORE_REDSHIFT
_DATASTORE.fields_by_name['databricks'].message_type = _DATASTORE_DATABRICKS
_DATASTORE.oneofs_by_name['config'].fields.append(
  _DATASTORE.fields_by_name['big_query'])
_DATASTORE.fields_by_name['big_query'].containing_oneof = _DATASTORE.oneofs_by_name['config']
_DATASTORE.oneofs_by_name['config'].fields.append(
  _DATASTORE.fields_by_name['snowflake'])
_DATASTORE.fields_by_name['snowflake'].containing_oneof = _DATASTORE.oneofs_by_name['config']
_DATASTORE.oneofs_by_name['config'].fields.append(
  _DATASTORE.fields_by_name['postgres'])
_DATASTORE.fields_by_name['postgres'].containing_oneof = _DATASTORE.oneofs_by_name['config']
_DATASTORE.oneofs_by_name['config'].fields.append(
  _DATASTORE.fields_by_name['redshift'])
_DATASTORE.fields_by_name['redshift'].containing_oneof = _DATASTORE.oneofs_by_name['config']
_DATASTORE.oneofs_by_name['config'].fields.append(
  _DATASTORE.fields_by_name['databricks'])
_DATASTORE.fields_by_name['databricks'].containing_oneof = _DATASTORE.oneofs_by_name['config']
_ENVIRONMENT.fields_by_name['summary'].message_type = _PROJECTSUMMARY
_ENVIRONMENT.fields_by_name['data_store'].message_type = _DATASTORE
_ENVIRONMENT.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_ENVIRONMENT.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_FEATURESET.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_FEATURESET.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_EVENT_DATAENTRY.containing_type = _EVENT
_EVENT.fields_by_name['data'].message_type = _EVENT_DATAENTRY
_EVENT.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_SERVINGMETRIC.fields_by_name['request_start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_SERVINGMETRIC.fields_by_name['model_start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_SERVINGMETRIC.fields_by_name['request_duration'].message_type = google_dot_protobuf_dot_duration__pb2._DURATION
_SERVINGMETRIC.fields_by_name['route_duration'].message_type = google_dot_protobuf_dot_duration__pb2._DURATION
_SERVINGMETRIC.fields_by_name['feature_duration'].message_type = google_dot_protobuf_dot_duration__pb2._DURATION
_SERVINGMETRIC.fields_by_name['predict_duration'].message_type = google_dot_protobuf_dot_duration__pb2._DURATION
_SERVINGMETRIC.fields_by_name['model_duration'].message_type = google_dot_protobuf_dot_duration__pb2._DURATION
_APIKEY.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_APIKEY.fields_by_name['last_access_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_APIKEY.fields_by_name['state'].enum_type = _APIKEY_STATE
_APIKEY.fields_by_name['revoked_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_APIKEY_STATE.containing_type = _APIKEY
_MODEL.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_MODEL.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PROMOTION.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PROMOTION.fields_by_name['promoted_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PROMOTION.fields_by_name['demoted_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PROMOTION.fields_by_name['state'].enum_type = _PROMOTION_STATE
_PROMOTION_REASON.containing_type = _PROMOTION
_PROMOTION_STATE.containing_type = _PROMOTION
_MODELVERSION.fields_by_name['state'].enum_type = _MODELVERSION_STATE
_MODELVERSION.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_MODELVERSION.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_MODELVERSION.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_MODELVERSION.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_MODELVERSION_STATE.containing_type = _MODELVERSION
_CONFUSIONMATRIX_ROW.containing_type = _CONFUSIONMATRIX
_CONFUSIONMATRIX.fields_by_name['row'].message_type = _CONFUSIONMATRIX_ROW
_BOOLEANEVALUATIONMETRICS.fields_by_name['confusion_matrix'].message_type = _CONFUSIONMATRIX
_CATEGORICALEVALUATIONMETRICS.fields_by_name['confusion_matrix'].message_type = _CONFUSIONMATRIX
_EVALUATIONMETRICS.fields_by_name['boolean'].message_type = _BOOLEANEVALUATIONMETRICS
_EVALUATIONMETRICS.fields_by_name['number'].message_type = _NUMBEREVALUATIONMETRICS
_EVALUATIONMETRICS.fields_by_name['text'].message_type = _TEXTEVALUATIONMETRICS
_EVALUATIONMETRICS.fields_by_name['categorical'].message_type = _CATEGORICALEVALUATIONMETRICS
_EVALUATIONMETRICS.oneofs_by_name['metrics'].fields.append(
  _EVALUATIONMETRICS.fields_by_name['boolean'])
_EVALUATIONMETRICS.fields_by_name['boolean'].containing_oneof = _EVALUATIONMETRICS.oneofs_by_name['metrics']
_EVALUATIONMETRICS.oneofs_by_name['metrics'].fields.append(
  _EVALUATIONMETRICS.fields_by_name['number'])
_EVALUATIONMETRICS.fields_by_name['number'].containing_oneof = _EVALUATIONMETRICS.oneofs_by_name['metrics']
_EVALUATIONMETRICS.oneofs_by_name['metrics'].fields.append(
  _EVALUATIONMETRICS.fields_by_name['text'])
_EVALUATIONMETRICS.fields_by_name['text'].containing_oneof = _EVALUATIONMETRICS.oneofs_by_name['metrics']
_EVALUATIONMETRICS.oneofs_by_name['metrics'].fields.append(
  _EVALUATIONMETRICS.fields_by_name['categorical'])
_EVALUATIONMETRICS.fields_by_name['categorical'].containing_oneof = _EVALUATIONMETRICS.oneofs_by_name['metrics']
_EXPERIMENT.fields_by_name['state'].enum_type = _EXPERIMENT_STATE
_EXPERIMENT.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_EXPERIMENT.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_EXPERIMENT.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_EXPERIMENT.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_EXPERIMENT_STATE.containing_type = _EXPERIMENT
_SIGNATUREFEATURE.fields_by_name['type'].enum_type = _FIELDTYPE
_QUERYFEATURE.fields_by_name['relation'].message_type = _QUERY
_QUERYFEATURE.fields_by_name['type'].enum_type = _FIELDTYPE
_QUERY.fields_by_name['fields'].message_type = _QUERYFEATURE
_QUERY.fields_by_name['related'].message_type = _QUERY
_BATCHPREDICTION.fields_by_name['state'].enum_type = _BATCHPREDICTION_STATE
_BATCHPREDICTION.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_BATCHPREDICTION.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_BATCHPREDICTION.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_BATCHPREDICTION_STATE.containing_type = _BATCHPREDICTION
_DATASET.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_DATASET.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_DATASETVERSION.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_DATASETVERSION.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_DATASETVERSIONASSIGNMENT.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_DATAPROFILE.fields_by_name['dataset_stats'].message_type = _DATASETSTATSENTRY
_DATAPROFILE.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_DATASETSTATSENTRY.fields_by_name['correlation_matrix'].message_type = _CORRELATIONMATRIX
_DATASETSTATSENTRY.fields_by_name['string_stats'].message_type = _STRINGSTATS
_DATASETSTATSENTRY.fields_by_name['numeric_stats'].message_type = _NUMERICSTATS
_DATASETSTATSENTRY.fields_by_name['timestamp_stats'].message_type = _TIMESTAMPSTATS
_DATASETSTATSENTRY.fields_by_name['categorical_stats'].message_type = _CATEGORICALSTATS
_DATASETSTATSENTRY.fields_by_name['dataset_ranges'].message_type = _DATASETRANGE
_DATASETRANGE.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_DATASETRANGE.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_CORRELATIONMATRIX_ROW.containing_type = _CORRELATIONMATRIX
_CORRELATIONMATRIX.fields_by_name['rows'].message_type = _CORRELATIONMATRIX_ROW
_HISTOGRAM_BUCKET.containing_type = _HISTOGRAM
_HISTOGRAM.fields_by_name['buckets'].message_type = _HISTOGRAM_BUCKET
_TIMESTAMPHISTOGRAM_BUCKET.fields_by_name['low_value'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_TIMESTAMPHISTOGRAM_BUCKET.fields_by_name['high_value'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_TIMESTAMPHISTOGRAM_BUCKET.containing_type = _TIMESTAMPHISTOGRAM
_TIMESTAMPHISTOGRAM.fields_by_name['buckets'].message_type = _TIMESTAMPHISTOGRAM_BUCKET
_CATEGORICALDISTRIBUTION_CATEGORY.containing_type = _CATEGORICALDISTRIBUTION
_CATEGORICALDISTRIBUTION.fields_by_name['categories'].message_type = _CATEGORICALDISTRIBUTION_CATEGORY
_CATEGORICALSTATS.fields_by_name['categorical_distribution'].message_type = _CATEGORICALDISTRIBUTION
_TIMESTAMPSTATS.fields_by_name['min'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_TIMESTAMPSTATS.fields_by_name['max'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_TIMESTAMPSTATS.fields_by_name['timestamp_histogram'].message_type = _TIMESTAMPHISTOGRAM
_NUMERICSTATS.fields_by_name['num_values_histogram'].message_type = _HISTOGRAM
_BILLINGACCOUNT.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_BILLINGACCOUNT.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_SUBSCRIPTION.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_SUBSCRIPTION.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PROJECTUSAGERECORD_TYPECOUNTENTRY.containing_type = _PROJECTUSAGERECORD
_PROJECTUSAGERECORD.fields_by_name['typeCount'].message_type = _PROJECTUSAGERECORD_TYPECOUNTENTRY
_CREDITUSAGERECORD.fields_by_name['period'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_CREDITUSAGERECORD.fields_by_name['project_usages'].message_type = _PROJECTUSAGERECORD
_INVOICELINEITEM.fields_by_name['period_start'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_INVOICELINEITEM.fields_by_name['period_end'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_INVOICE.fields_by_name['date'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_INVOICE.fields_by_name['period_start'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_INVOICE.fields_by_name['period_end'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_INVOICE.fields_by_name['line_items'].message_type = _INVOICELINEITEM
_PAYMENTMETHOD_BILLINGDETAILS.containing_type = _PAYMENTMETHOD
_PAYMENTMETHOD_CARD.containing_type = _PAYMENTMETHOD
_PAYMENTMETHOD.fields_by_name['billing_details'].message_type = _PAYMENTMETHOD_BILLINGDETAILS
_PAYMENTMETHOD.fields_by_name['card'].message_type = _PAYMENTMETHOD_CARD
_ARTIFACT.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_ARTIFACT.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_USERINTEGRATION.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_USERINTEGRATION.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_USERINTEGRATION.fields_by_name['data_update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_USERINTEGRATION.fields_by_name['disconnect_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_GITPROVIDERNAMESPACE.fields_by_name['provider'].message_type = _GITPROVIDER
_GITPROVIDERREPOSITORY.fields_by_name['provider_namespace'].message_type = _GITPROVIDERNAMESPACE
_GITPROVIDERREPOSITORY.fields_by_name['data_update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PROJECTGITINTEGRATION.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PROJECTGITINTEGRATION.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_PROJECTGITINTEGRATION.fields_by_name['git_provider'].message_type = _GITPROVIDER
_PROJECTGITINTEGRATION.fields_by_name['git_provider_namespace'].message_type = _GITPROVIDERNAMESPACE
_PROJECTGITINTEGRATION.fields_by_name['git_provider_repository'].message_type = _GITPROVIDERREPOSITORY
_PROJECTGITINTEGRATION.fields_by_name['disconnect_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_METRIC.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_METRIC.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_METRIC.fields_by_name['direction'].enum_type = _METRIC_DIRECTION
_METRIC.fields_by_name['values'].message_type = _METRICVALUE
_METRIC_DIRECTION.containing_type = _METRIC
_METRICVALUE.fields_by_name['timestamp'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_METRICVALUE.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_METADATA.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_METADATA.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_GITCONTEXT_REPOSITORY.containing_type = _GITCONTEXT
_GITCONTEXT_COMMIT.fields_by_name['authored_datetime'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_GITCONTEXT_COMMIT.fields_by_name['author'].message_type = _GITCONTEXT_AUTHOR
_GITCONTEXT_COMMIT.fields_by_name['committed_datetime'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_GITCONTEXT_COMMIT.fields_by_name['committer'].message_type = _GITCONTEXT_COMMITTER
_GITCONTEXT_COMMIT.containing_type = _GITCONTEXT
_GITCONTEXT_BRANCH.containing_type = _GITCONTEXT
_GITCONTEXT_GITTAG.containing_type = _GITCONTEXT
_GITCONTEXT_GITTAGS.fields_by_name['current'].message_type = _GITCONTEXT_GITTAG
_GITCONTEXT_GITTAGS.fields_by_name['reachable'].message_type = _GITCONTEXT_GITTAG
_GITCONTEXT_GITTAGS.containing_type = _GITCONTEXT
_GITCONTEXT_REF.containing_type = _GITCONTEXT
_GITCONTEXT_AUTHOR.containing_type = _GITCONTEXT
_GITCONTEXT_COMMITTER.containing_type = _GITCONTEXT
_GITCONTEXT_HEAD.containing_type = _GITCONTEXT
_GITCONTEXT.fields_by_name['repository'].message_type = _GITCONTEXT_REPOSITORY
_GITCONTEXT.fields_by_name['branch'].message_type = _GITCONTEXT_BRANCH
_GITCONTEXT.fields_by_name['tags'].message_type = _GITCONTEXT_GITTAGS
_GITCONTEXT.fields_by_name['ref'].message_type = _GITCONTEXT_REF
_GITCONTEXT.fields_by_name['commit'].message_type = _GITCONTEXT_COMMIT
_GITCONTEXT.fields_by_name['head'].message_type = _GITCONTEXT_HEAD
_GITPROVIDERCONTEXT_REPOSITORY.containing_type = _GITPROVIDERCONTEXT
_GITPROVIDERCONTEXT.fields_by_name['repository'].message_type = _GITPROVIDERCONTEXT_REPOSITORY
_SDLCREVIEWCONTEXT_USER.containing_type = _SDLCREVIEWCONTEXT
_SDLCREVIEWCONTEXT_REF.containing_type = _SDLCREVIEWCONTEXT
_SDLCREVIEWCONTEXT.fields_by_name['user'].message_type = _SDLCREVIEWCONTEXT_USER
_SDLCREVIEWCONTEXT.fields_by_name['base'].message_type = _SDLCREVIEWCONTEXT_REF
_SDLCREVIEWCONTEXT.fields_by_name['head'].message_type = _SDLCREVIEWCONTEXT_REF
_RUNSUMMARY_RESOURCE.containing_type = _RUNSUMMARY
_RUNSUMMARY.fields_by_name['model_versions'].message_type = _MODELVERSION
_RUNSUMMARY.fields_by_name['batch_predictions'].message_type = _RUNSUMMARY_RESOURCE
_RUNSUMMARY.fields_by_name['promotions'].message_type = _RUNSUMMARY_RESOURCE
_RUNSUMMARY.fields_by_name['deployments'].message_type = _RUNSUMMARY_RESOURCE
_RUNSUMMARY.fields_by_name['git_context'].message_type = _GITCONTEXT
_RUNSUMMARY.fields_by_name['principal'].message_type = _PRINCIPAL
_RUNSUMMARY.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_RUNSUMMARY.fields_by_name['associated_models'].message_type = _MODEL
_RUNSUMMARY.fields_by_name['associated_datasets'].message_type = _DATASET
_RUNSUMMARY.fields_by_name['dataset_versions'].message_type = _DATASETVERSION
_RUNSUMMARY.fields_by_name['git_provider_context'].message_type = _GITPROVIDERCONTEXT
_RUNSUMMARY.fields_by_name['cicd_runner_context'].message_type = _CICDRUNNERCONTEXT
_RUNSUMMARY.fields_by_name['sdlc_review_context'].message_type = _SDLCREVIEWCONTEXT
_RUN.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_RUN.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_RUN.fields_by_name['last_heartbeat'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_RUN.fields_by_name['state'].enum_type = _RUN_STATE
_RUN.fields_by_name['complete_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_RUN_STATE.containing_type = _RUN
_TAG.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_TAG.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_NOTIFICATIONPOLICYCONFIG_EMAIL.containing_type = _NOTIFICATIONPOLICYCONFIG
_NOTIFICATIONPOLICYCONFIG_SLACK.containing_type = _NOTIFICATIONPOLICYCONFIG
_NOTIFICATIONPOLICYCONFIG.fields_by_name['email'].message_type = _NOTIFICATIONPOLICYCONFIG_EMAIL
_NOTIFICATIONPOLICYCONFIG.fields_by_name['slack'].message_type = _NOTIFICATIONPOLICYCONFIG_SLACK
_NOTIFICATIONPOLICY.fields_by_name['config'].message_type = _NOTIFICATIONPOLICYCONFIG
_NOTIFICATIONPOLICY.fields_by_name['type'].enum_type = _NOTIFICATIONPOLICY_TYPE
_NOTIFICATIONPOLICY.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_NOTIFICATIONPOLICY.fields_by_name['update_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_NOTIFICATIONPOLICY_TYPE.containing_type = _NOTIFICATIONPOLICY
_NOTIFICATION.fields_by_name['type'].enum_type = _NOTIFICATIONPOLICY_TYPE
_NOTIFICATION.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_RUNCHECKSUMMARY.fields_by_name['outcome'].enum_type = _CHECKOUTCOME
_RUNCHECKSUMMARY.fields_by_name['resource_checks'].message_type = _RESOURCECHECK
_RESOURCECHECK.fields_by_name['outcome'].enum_type = _CHECKOUTCOME
_RESOURCECHECK.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
DESCRIPTOR.message_types_by_name['User'] = _USER
DESCRIPTOR.message_types_by_name['Principal'] = _PRINCIPAL
DESCRIPTOR.message_types_by_name['Session'] = _SESSION
DESCRIPTOR.message_types_by_name['Role'] = _ROLE
DESCRIPTOR.message_types_by_name['AccessPolicy'] = _ACCESSPOLICY
DESCRIPTOR.message_types_by_name['Invite'] = _INVITE
DESCRIPTOR.message_types_by_name['Organization'] = _ORGANIZATION
DESCRIPTOR.message_types_by_name['Onboarding'] = _ONBOARDING
DESCRIPTOR.message_types_by_name['Project'] = _PROJECT
DESCRIPTOR.message_types_by_name['ProjectSummary'] = _PROJECTSUMMARY
DESCRIPTOR.message_types_by_name['DataStore'] = _DATASTORE
DESCRIPTOR.message_types_by_name['HealthCounts'] = _HEALTHCOUNTS
DESCRIPTOR.message_types_by_name['Environment'] = _ENVIRONMENT
DESCRIPTOR.message_types_by_name['FeatureSet'] = _FEATURESET
DESCRIPTOR.message_types_by_name['Event'] = _EVENT
DESCRIPTOR.message_types_by_name['ServingMetric'] = _SERVINGMETRIC
DESCRIPTOR.message_types_by_name['ApiKey'] = _APIKEY
DESCRIPTOR.message_types_by_name['Model'] = _MODEL
DESCRIPTOR.message_types_by_name['Promotion'] = _PROMOTION
DESCRIPTOR.message_types_by_name['ModelVersion'] = _MODELVERSION
DESCRIPTOR.message_types_by_name['ConfusionMatrix'] = _CONFUSIONMATRIX
DESCRIPTOR.message_types_by_name['BooleanEvaluationMetrics'] = _BOOLEANEVALUATIONMETRICS
DESCRIPTOR.message_types_by_name['TextEvaluationMetrics'] = _TEXTEVALUATIONMETRICS
DESCRIPTOR.message_types_by_name['CategoricalEvaluationMetrics'] = _CATEGORICALEVALUATIONMETRICS
DESCRIPTOR.message_types_by_name['NumberEvaluationMetrics'] = _NUMBEREVALUATIONMETRICS
DESCRIPTOR.message_types_by_name['EvaluationMetrics'] = _EVALUATIONMETRICS
DESCRIPTOR.message_types_by_name['Experiment'] = _EXPERIMENT
DESCRIPTOR.message_types_by_name['TextSequence'] = _TEXTSEQUENCE
DESCRIPTOR.message_types_by_name['NumberSequence'] = _NUMBERSEQUENCE
DESCRIPTOR.message_types_by_name['SignatureFeature'] = _SIGNATUREFEATURE
DESCRIPTOR.message_types_by_name['QueryFeature'] = _QUERYFEATURE
DESCRIPTOR.message_types_by_name['Query'] = _QUERY
DESCRIPTOR.message_types_by_name['BatchPrediction'] = _BATCHPREDICTION
DESCRIPTOR.message_types_by_name['SignedURL'] = _SIGNEDURL
DESCRIPTOR.message_types_by_name['Entity'] = _ENTITY
DESCRIPTOR.message_types_by_name['Dataset'] = _DATASET
DESCRIPTOR.message_types_by_name['DatasetVersion'] = _DATASETVERSION
DESCRIPTOR.message_types_by_name['DatasetVersionAssignment'] = _DATASETVERSIONASSIGNMENT
DESCRIPTOR.message_types_by_name['DataProfile'] = _DATAPROFILE
DESCRIPTOR.message_types_by_name['DatasetStatsEntry'] = _DATASETSTATSENTRY
DESCRIPTOR.message_types_by_name['DatasetRange'] = _DATASETRANGE
DESCRIPTOR.message_types_by_name['CorrelationMatrix'] = _CORRELATIONMATRIX
DESCRIPTOR.message_types_by_name['Histogram'] = _HISTOGRAM
DESCRIPTOR.message_types_by_name['TimestampHistogram'] = _TIMESTAMPHISTOGRAM
DESCRIPTOR.message_types_by_name['CategoricalDistribution'] = _CATEGORICALDISTRIBUTION
DESCRIPTOR.message_types_by_name['StringStats'] = _STRINGSTATS
DESCRIPTOR.message_types_by_name['CategoricalStats'] = _CATEGORICALSTATS
DESCRIPTOR.message_types_by_name['TimestampStats'] = _TIMESTAMPSTATS
DESCRIPTOR.message_types_by_name['NumericStats'] = _NUMERICSTATS
DESCRIPTOR.message_types_by_name['BillingAccount'] = _BILLINGACCOUNT
DESCRIPTOR.message_types_by_name['Subscription'] = _SUBSCRIPTION
DESCRIPTOR.message_types_by_name['ProjectUsageRecord'] = _PROJECTUSAGERECORD
DESCRIPTOR.message_types_by_name['CreditUsageRecord'] = _CREDITUSAGERECORD
DESCRIPTOR.message_types_by_name['InvoiceLineItem'] = _INVOICELINEITEM
DESCRIPTOR.message_types_by_name['Invoice'] = _INVOICE
DESCRIPTOR.message_types_by_name['Price'] = _PRICE
DESCRIPTOR.message_types_by_name['PaymentMethod'] = _PAYMENTMETHOD
DESCRIPTOR.message_types_by_name['Artifact'] = _ARTIFACT
DESCRIPTOR.message_types_by_name['CronJob'] = _CRONJOB
DESCRIPTOR.message_types_by_name['UserIntegrationProvider'] = _USERINTEGRATIONPROVIDER
DESCRIPTOR.message_types_by_name['UserIntegration'] = _USERINTEGRATION
DESCRIPTOR.message_types_by_name['GitProvider'] = _GITPROVIDER
DESCRIPTOR.message_types_by_name['GitProviderNamespace'] = _GITPROVIDERNAMESPACE
DESCRIPTOR.message_types_by_name['GitProviderRepository'] = _GITPROVIDERREPOSITORY
DESCRIPTOR.message_types_by_name['ProjectGitIntegration'] = _PROJECTGITINTEGRATION
DESCRIPTOR.message_types_by_name['Metric'] = _METRIC
DESCRIPTOR.message_types_by_name['MetricValue'] = _METRICVALUE
DESCRIPTOR.message_types_by_name['Metadata'] = _METADATA
DESCRIPTOR.message_types_by_name['GitContext'] = _GITCONTEXT
DESCRIPTOR.message_types_by_name['GitProviderContext'] = _GITPROVIDERCONTEXT
DESCRIPTOR.message_types_by_name['CicdRunnerContext'] = _CICDRUNNERCONTEXT
DESCRIPTOR.message_types_by_name['SdlcReviewContext'] = _SDLCREVIEWCONTEXT
DESCRIPTOR.message_types_by_name['RunSummary'] = _RUNSUMMARY
DESCRIPTOR.message_types_by_name['Run'] = _RUN
DESCRIPTOR.message_types_by_name['Tag'] = _TAG
DESCRIPTOR.message_types_by_name['NotificationPolicyConfig'] = _NOTIFICATIONPOLICYCONFIG
DESCRIPTOR.message_types_by_name['NotificationPolicy'] = _NOTIFICATIONPOLICY
DESCRIPTOR.message_types_by_name['Notification'] = _NOTIFICATION
DESCRIPTOR.message_types_by_name['RunCheckSummary'] = _RUNCHECKSUMMARY
DESCRIPTOR.message_types_by_name['ResourceCheck'] = _RESOURCECHECK
DESCRIPTOR.enum_types_by_name['FieldType'] = _FIELDTYPE
DESCRIPTOR.enum_types_by_name['BatchPredictionStoreType'] = _BATCHPREDICTIONSTORETYPE
DESCRIPTOR.enum_types_by_name['CheckOutcome'] = _CHECKOUTCOME
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

User = _reflection.GeneratedProtocolMessageType('User', (_message.Message,), {
  'DESCRIPTOR' : _USER,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.User)
  })
_sym_db.RegisterMessage(User)

Principal = _reflection.GeneratedProtocolMessageType('Principal', (_message.Message,), {
  'DESCRIPTOR' : _PRINCIPAL,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Principal)
  })
_sym_db.RegisterMessage(Principal)

Session = _reflection.GeneratedProtocolMessageType('Session', (_message.Message,), {
  'DESCRIPTOR' : _SESSION,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Session)
  })
_sym_db.RegisterMessage(Session)

Role = _reflection.GeneratedProtocolMessageType('Role', (_message.Message,), {
  'DESCRIPTOR' : _ROLE,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Role)
  })
_sym_db.RegisterMessage(Role)

AccessPolicy = _reflection.GeneratedProtocolMessageType('AccessPolicy', (_message.Message,), {
  'DESCRIPTOR' : _ACCESSPOLICY,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.AccessPolicy)
  })
_sym_db.RegisterMessage(AccessPolicy)

Invite = _reflection.GeneratedProtocolMessageType('Invite', (_message.Message,), {
  'DESCRIPTOR' : _INVITE,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Invite)
  })
_sym_db.RegisterMessage(Invite)

Organization = _reflection.GeneratedProtocolMessageType('Organization', (_message.Message,), {
  'DESCRIPTOR' : _ORGANIZATION,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Organization)
  })
_sym_db.RegisterMessage(Organization)

Onboarding = _reflection.GeneratedProtocolMessageType('Onboarding', (_message.Message,), {
  'DESCRIPTOR' : _ONBOARDING,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Onboarding)
  })
_sym_db.RegisterMessage(Onboarding)

Project = _reflection.GeneratedProtocolMessageType('Project', (_message.Message,), {
  'DESCRIPTOR' : _PROJECT,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Project)
  })
_sym_db.RegisterMessage(Project)

ProjectSummary = _reflection.GeneratedProtocolMessageType('ProjectSummary', (_message.Message,), {
  'DESCRIPTOR' : _PROJECTSUMMARY,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ProjectSummary)
  })
_sym_db.RegisterMessage(ProjectSummary)

DataStore = _reflection.GeneratedProtocolMessageType('DataStore', (_message.Message,), {

  'BigQuery' : _reflection.GeneratedProtocolMessageType('BigQuery', (_message.Message,), {
    'DESCRIPTOR' : _DATASTORE_BIGQUERY,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DataStore.BigQuery)
    })
  ,

  'Snowflake' : _reflection.GeneratedProtocolMessageType('Snowflake', (_message.Message,), {

    'ParamsEntry' : _reflection.GeneratedProtocolMessageType('ParamsEntry', (_message.Message,), {
      'DESCRIPTOR' : _DATASTORE_SNOWFLAKE_PARAMSENTRY,
      '__module__' : 'continual.rpc.management.v1.management_types_pb2'
      # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DataStore.Snowflake.ParamsEntry)
      })
    ,
    'DESCRIPTOR' : _DATASTORE_SNOWFLAKE,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DataStore.Snowflake)
    })
  ,

  'Postgres' : _reflection.GeneratedProtocolMessageType('Postgres', (_message.Message,), {

    'ParamsEntry' : _reflection.GeneratedProtocolMessageType('ParamsEntry', (_message.Message,), {
      'DESCRIPTOR' : _DATASTORE_POSTGRES_PARAMSENTRY,
      '__module__' : 'continual.rpc.management.v1.management_types_pb2'
      # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DataStore.Postgres.ParamsEntry)
      })
    ,
    'DESCRIPTOR' : _DATASTORE_POSTGRES,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DataStore.Postgres)
    })
  ,

  'Redshift' : _reflection.GeneratedProtocolMessageType('Redshift', (_message.Message,), {

    'ParamsEntry' : _reflection.GeneratedProtocolMessageType('ParamsEntry', (_message.Message,), {
      'DESCRIPTOR' : _DATASTORE_REDSHIFT_PARAMSENTRY,
      '__module__' : 'continual.rpc.management.v1.management_types_pb2'
      # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DataStore.Redshift.ParamsEntry)
      })
    ,
    'DESCRIPTOR' : _DATASTORE_REDSHIFT,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DataStore.Redshift)
    })
  ,

  'Databricks' : _reflection.GeneratedProtocolMessageType('Databricks', (_message.Message,), {

    'ParamsEntry' : _reflection.GeneratedProtocolMessageType('ParamsEntry', (_message.Message,), {
      'DESCRIPTOR' : _DATASTORE_DATABRICKS_PARAMSENTRY,
      '__module__' : 'continual.rpc.management.v1.management_types_pb2'
      # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DataStore.Databricks.ParamsEntry)
      })
    ,
    'DESCRIPTOR' : _DATASTORE_DATABRICKS,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DataStore.Databricks)
    })
  ,
  'DESCRIPTOR' : _DATASTORE,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DataStore)
  })
_sym_db.RegisterMessage(DataStore)
_sym_db.RegisterMessage(DataStore.BigQuery)
_sym_db.RegisterMessage(DataStore.Snowflake)
_sym_db.RegisterMessage(DataStore.Snowflake.ParamsEntry)
_sym_db.RegisterMessage(DataStore.Postgres)
_sym_db.RegisterMessage(DataStore.Postgres.ParamsEntry)
_sym_db.RegisterMessage(DataStore.Redshift)
_sym_db.RegisterMessage(DataStore.Redshift.ParamsEntry)
_sym_db.RegisterMessage(DataStore.Databricks)
_sym_db.RegisterMessage(DataStore.Databricks.ParamsEntry)

HealthCounts = _reflection.GeneratedProtocolMessageType('HealthCounts', (_message.Message,), {
  'DESCRIPTOR' : _HEALTHCOUNTS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.HealthCounts)
  })
_sym_db.RegisterMessage(HealthCounts)

Environment = _reflection.GeneratedProtocolMessageType('Environment', (_message.Message,), {
  'DESCRIPTOR' : _ENVIRONMENT,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Environment)
  })
_sym_db.RegisterMessage(Environment)

FeatureSet = _reflection.GeneratedProtocolMessageType('FeatureSet', (_message.Message,), {
  'DESCRIPTOR' : _FEATURESET,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.FeatureSet)
  })
_sym_db.RegisterMessage(FeatureSet)

Event = _reflection.GeneratedProtocolMessageType('Event', (_message.Message,), {

  'DataEntry' : _reflection.GeneratedProtocolMessageType('DataEntry', (_message.Message,), {
    'DESCRIPTOR' : _EVENT_DATAENTRY,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Event.DataEntry)
    })
  ,
  'DESCRIPTOR' : _EVENT,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Event)
  })
_sym_db.RegisterMessage(Event)
_sym_db.RegisterMessage(Event.DataEntry)

ServingMetric = _reflection.GeneratedProtocolMessageType('ServingMetric', (_message.Message,), {
  'DESCRIPTOR' : _SERVINGMETRIC,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ServingMetric)
  })
_sym_db.RegisterMessage(ServingMetric)

ApiKey = _reflection.GeneratedProtocolMessageType('ApiKey', (_message.Message,), {
  'DESCRIPTOR' : _APIKEY,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ApiKey)
  })
_sym_db.RegisterMessage(ApiKey)

Model = _reflection.GeneratedProtocolMessageType('Model', (_message.Message,), {
  'DESCRIPTOR' : _MODEL,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Model)
  })
_sym_db.RegisterMessage(Model)

Promotion = _reflection.GeneratedProtocolMessageType('Promotion', (_message.Message,), {
  'DESCRIPTOR' : _PROMOTION,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Promotion)
  })
_sym_db.RegisterMessage(Promotion)

ModelVersion = _reflection.GeneratedProtocolMessageType('ModelVersion', (_message.Message,), {
  'DESCRIPTOR' : _MODELVERSION,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ModelVersion)
  })
_sym_db.RegisterMessage(ModelVersion)

ConfusionMatrix = _reflection.GeneratedProtocolMessageType('ConfusionMatrix', (_message.Message,), {

  'Row' : _reflection.GeneratedProtocolMessageType('Row', (_message.Message,), {
    'DESCRIPTOR' : _CONFUSIONMATRIX_ROW,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ConfusionMatrix.Row)
    })
  ,
  'DESCRIPTOR' : _CONFUSIONMATRIX,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ConfusionMatrix)
  })
_sym_db.RegisterMessage(ConfusionMatrix)
_sym_db.RegisterMessage(ConfusionMatrix.Row)

BooleanEvaluationMetrics = _reflection.GeneratedProtocolMessageType('BooleanEvaluationMetrics', (_message.Message,), {
  'DESCRIPTOR' : _BOOLEANEVALUATIONMETRICS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BooleanEvaluationMetrics)
  })
_sym_db.RegisterMessage(BooleanEvaluationMetrics)

TextEvaluationMetrics = _reflection.GeneratedProtocolMessageType('TextEvaluationMetrics', (_message.Message,), {
  'DESCRIPTOR' : _TEXTEVALUATIONMETRICS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.TextEvaluationMetrics)
  })
_sym_db.RegisterMessage(TextEvaluationMetrics)

CategoricalEvaluationMetrics = _reflection.GeneratedProtocolMessageType('CategoricalEvaluationMetrics', (_message.Message,), {
  'DESCRIPTOR' : _CATEGORICALEVALUATIONMETRICS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CategoricalEvaluationMetrics)
  })
_sym_db.RegisterMessage(CategoricalEvaluationMetrics)

NumberEvaluationMetrics = _reflection.GeneratedProtocolMessageType('NumberEvaluationMetrics', (_message.Message,), {
  'DESCRIPTOR' : _NUMBEREVALUATIONMETRICS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.NumberEvaluationMetrics)
  })
_sym_db.RegisterMessage(NumberEvaluationMetrics)

EvaluationMetrics = _reflection.GeneratedProtocolMessageType('EvaluationMetrics', (_message.Message,), {
  'DESCRIPTOR' : _EVALUATIONMETRICS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.EvaluationMetrics)
  })
_sym_db.RegisterMessage(EvaluationMetrics)

Experiment = _reflection.GeneratedProtocolMessageType('Experiment', (_message.Message,), {
  'DESCRIPTOR' : _EXPERIMENT,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Experiment)
  })
_sym_db.RegisterMessage(Experiment)

TextSequence = _reflection.GeneratedProtocolMessageType('TextSequence', (_message.Message,), {
  'DESCRIPTOR' : _TEXTSEQUENCE,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.TextSequence)
  })
_sym_db.RegisterMessage(TextSequence)

NumberSequence = _reflection.GeneratedProtocolMessageType('NumberSequence', (_message.Message,), {
  'DESCRIPTOR' : _NUMBERSEQUENCE,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.NumberSequence)
  })
_sym_db.RegisterMessage(NumberSequence)

SignatureFeature = _reflection.GeneratedProtocolMessageType('SignatureFeature', (_message.Message,), {
  'DESCRIPTOR' : _SIGNATUREFEATURE,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.SignatureFeature)
  })
_sym_db.RegisterMessage(SignatureFeature)

QueryFeature = _reflection.GeneratedProtocolMessageType('QueryFeature', (_message.Message,), {
  'DESCRIPTOR' : _QUERYFEATURE,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.QueryFeature)
  })
_sym_db.RegisterMessage(QueryFeature)

Query = _reflection.GeneratedProtocolMessageType('Query', (_message.Message,), {
  'DESCRIPTOR' : _QUERY,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Query)
  })
_sym_db.RegisterMessage(Query)

BatchPrediction = _reflection.GeneratedProtocolMessageType('BatchPrediction', (_message.Message,), {
  'DESCRIPTOR' : _BATCHPREDICTION,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchPrediction)
  })
_sym_db.RegisterMessage(BatchPrediction)

SignedURL = _reflection.GeneratedProtocolMessageType('SignedURL', (_message.Message,), {
  'DESCRIPTOR' : _SIGNEDURL,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.SignedURL)
  })
_sym_db.RegisterMessage(SignedURL)

Entity = _reflection.GeneratedProtocolMessageType('Entity', (_message.Message,), {
  'DESCRIPTOR' : _ENTITY,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Entity)
  })
_sym_db.RegisterMessage(Entity)

Dataset = _reflection.GeneratedProtocolMessageType('Dataset', (_message.Message,), {
  'DESCRIPTOR' : _DATASET,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Dataset)
  })
_sym_db.RegisterMessage(Dataset)

DatasetVersion = _reflection.GeneratedProtocolMessageType('DatasetVersion', (_message.Message,), {
  'DESCRIPTOR' : _DATASETVERSION,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DatasetVersion)
  })
_sym_db.RegisterMessage(DatasetVersion)

DatasetVersionAssignment = _reflection.GeneratedProtocolMessageType('DatasetVersionAssignment', (_message.Message,), {
  'DESCRIPTOR' : _DATASETVERSIONASSIGNMENT,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DatasetVersionAssignment)
  })
_sym_db.RegisterMessage(DatasetVersionAssignment)

DataProfile = _reflection.GeneratedProtocolMessageType('DataProfile', (_message.Message,), {
  'DESCRIPTOR' : _DATAPROFILE,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DataProfile)
  })
_sym_db.RegisterMessage(DataProfile)

DatasetStatsEntry = _reflection.GeneratedProtocolMessageType('DatasetStatsEntry', (_message.Message,), {
  'DESCRIPTOR' : _DATASETSTATSENTRY,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DatasetStatsEntry)
  })
_sym_db.RegisterMessage(DatasetStatsEntry)

DatasetRange = _reflection.GeneratedProtocolMessageType('DatasetRange', (_message.Message,), {
  'DESCRIPTOR' : _DATASETRANGE,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DatasetRange)
  })
_sym_db.RegisterMessage(DatasetRange)

CorrelationMatrix = _reflection.GeneratedProtocolMessageType('CorrelationMatrix', (_message.Message,), {

  'Row' : _reflection.GeneratedProtocolMessageType('Row', (_message.Message,), {
    'DESCRIPTOR' : _CORRELATIONMATRIX_ROW,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CorrelationMatrix.Row)
    })
  ,
  'DESCRIPTOR' : _CORRELATIONMATRIX,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CorrelationMatrix)
  })
_sym_db.RegisterMessage(CorrelationMatrix)
_sym_db.RegisterMessage(CorrelationMatrix.Row)

Histogram = _reflection.GeneratedProtocolMessageType('Histogram', (_message.Message,), {

  'Bucket' : _reflection.GeneratedProtocolMessageType('Bucket', (_message.Message,), {
    'DESCRIPTOR' : _HISTOGRAM_BUCKET,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Histogram.Bucket)
    })
  ,
  'DESCRIPTOR' : _HISTOGRAM,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Histogram)
  })
_sym_db.RegisterMessage(Histogram)
_sym_db.RegisterMessage(Histogram.Bucket)

TimestampHistogram = _reflection.GeneratedProtocolMessageType('TimestampHistogram', (_message.Message,), {

  'Bucket' : _reflection.GeneratedProtocolMessageType('Bucket', (_message.Message,), {
    'DESCRIPTOR' : _TIMESTAMPHISTOGRAM_BUCKET,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.TimestampHistogram.Bucket)
    })
  ,
  'DESCRIPTOR' : _TIMESTAMPHISTOGRAM,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.TimestampHistogram)
  })
_sym_db.RegisterMessage(TimestampHistogram)
_sym_db.RegisterMessage(TimestampHistogram.Bucket)

CategoricalDistribution = _reflection.GeneratedProtocolMessageType('CategoricalDistribution', (_message.Message,), {

  'Category' : _reflection.GeneratedProtocolMessageType('Category', (_message.Message,), {
    'DESCRIPTOR' : _CATEGORICALDISTRIBUTION_CATEGORY,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CategoricalDistribution.Category)
    })
  ,
  'DESCRIPTOR' : _CATEGORICALDISTRIBUTION,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CategoricalDistribution)
  })
_sym_db.RegisterMessage(CategoricalDistribution)
_sym_db.RegisterMessage(CategoricalDistribution.Category)

StringStats = _reflection.GeneratedProtocolMessageType('StringStats', (_message.Message,), {
  'DESCRIPTOR' : _STRINGSTATS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.StringStats)
  })
_sym_db.RegisterMessage(StringStats)

CategoricalStats = _reflection.GeneratedProtocolMessageType('CategoricalStats', (_message.Message,), {
  'DESCRIPTOR' : _CATEGORICALSTATS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CategoricalStats)
  })
_sym_db.RegisterMessage(CategoricalStats)

TimestampStats = _reflection.GeneratedProtocolMessageType('TimestampStats', (_message.Message,), {
  'DESCRIPTOR' : _TIMESTAMPSTATS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.TimestampStats)
  })
_sym_db.RegisterMessage(TimestampStats)

NumericStats = _reflection.GeneratedProtocolMessageType('NumericStats', (_message.Message,), {
  'DESCRIPTOR' : _NUMERICSTATS,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.NumericStats)
  })
_sym_db.RegisterMessage(NumericStats)

BillingAccount = _reflection.GeneratedProtocolMessageType('BillingAccount', (_message.Message,), {
  'DESCRIPTOR' : _BILLINGACCOUNT,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BillingAccount)
  })
_sym_db.RegisterMessage(BillingAccount)

Subscription = _reflection.GeneratedProtocolMessageType('Subscription', (_message.Message,), {
  'DESCRIPTOR' : _SUBSCRIPTION,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Subscription)
  })
_sym_db.RegisterMessage(Subscription)

ProjectUsageRecord = _reflection.GeneratedProtocolMessageType('ProjectUsageRecord', (_message.Message,), {

  'TypeCountEntry' : _reflection.GeneratedProtocolMessageType('TypeCountEntry', (_message.Message,), {
    'DESCRIPTOR' : _PROJECTUSAGERECORD_TYPECOUNTENTRY,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ProjectUsageRecord.TypeCountEntry)
    })
  ,
  'DESCRIPTOR' : _PROJECTUSAGERECORD,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ProjectUsageRecord)
  })
_sym_db.RegisterMessage(ProjectUsageRecord)
_sym_db.RegisterMessage(ProjectUsageRecord.TypeCountEntry)

CreditUsageRecord = _reflection.GeneratedProtocolMessageType('CreditUsageRecord', (_message.Message,), {
  'DESCRIPTOR' : _CREDITUSAGERECORD,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreditUsageRecord)
  })
_sym_db.RegisterMessage(CreditUsageRecord)

InvoiceLineItem = _reflection.GeneratedProtocolMessageType('InvoiceLineItem', (_message.Message,), {
  'DESCRIPTOR' : _INVOICELINEITEM,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.InvoiceLineItem)
  })
_sym_db.RegisterMessage(InvoiceLineItem)

Invoice = _reflection.GeneratedProtocolMessageType('Invoice', (_message.Message,), {
  'DESCRIPTOR' : _INVOICE,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Invoice)
  })
_sym_db.RegisterMessage(Invoice)

Price = _reflection.GeneratedProtocolMessageType('Price', (_message.Message,), {
  'DESCRIPTOR' : _PRICE,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Price)
  })
_sym_db.RegisterMessage(Price)

PaymentMethod = _reflection.GeneratedProtocolMessageType('PaymentMethod', (_message.Message,), {

  'BillingDetails' : _reflection.GeneratedProtocolMessageType('BillingDetails', (_message.Message,), {
    'DESCRIPTOR' : _PAYMENTMETHOD_BILLINGDETAILS,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.PaymentMethod.BillingDetails)
    })
  ,

  'Card' : _reflection.GeneratedProtocolMessageType('Card', (_message.Message,), {
    'DESCRIPTOR' : _PAYMENTMETHOD_CARD,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.PaymentMethod.Card)
    })
  ,
  'DESCRIPTOR' : _PAYMENTMETHOD,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.PaymentMethod)
  })
_sym_db.RegisterMessage(PaymentMethod)
_sym_db.RegisterMessage(PaymentMethod.BillingDetails)
_sym_db.RegisterMessage(PaymentMethod.Card)

Artifact = _reflection.GeneratedProtocolMessageType('Artifact', (_message.Message,), {
  'DESCRIPTOR' : _ARTIFACT,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Artifact)
  })
_sym_db.RegisterMessage(Artifact)

CronJob = _reflection.GeneratedProtocolMessageType('CronJob', (_message.Message,), {
  'DESCRIPTOR' : _CRONJOB,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CronJob)
  })
_sym_db.RegisterMessage(CronJob)

UserIntegrationProvider = _reflection.GeneratedProtocolMessageType('UserIntegrationProvider', (_message.Message,), {
  'DESCRIPTOR' : _USERINTEGRATIONPROVIDER,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UserIntegrationProvider)
  })
_sym_db.RegisterMessage(UserIntegrationProvider)

UserIntegration = _reflection.GeneratedProtocolMessageType('UserIntegration', (_message.Message,), {
  'DESCRIPTOR' : _USERINTEGRATION,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UserIntegration)
  })
_sym_db.RegisterMessage(UserIntegration)

GitProvider = _reflection.GeneratedProtocolMessageType('GitProvider', (_message.Message,), {
  'DESCRIPTOR' : _GITPROVIDER,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GitProvider)
  })
_sym_db.RegisterMessage(GitProvider)

GitProviderNamespace = _reflection.GeneratedProtocolMessageType('GitProviderNamespace', (_message.Message,), {
  'DESCRIPTOR' : _GITPROVIDERNAMESPACE,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GitProviderNamespace)
  })
_sym_db.RegisterMessage(GitProviderNamespace)

GitProviderRepository = _reflection.GeneratedProtocolMessageType('GitProviderRepository', (_message.Message,), {
  'DESCRIPTOR' : _GITPROVIDERREPOSITORY,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GitProviderRepository)
  })
_sym_db.RegisterMessage(GitProviderRepository)

ProjectGitIntegration = _reflection.GeneratedProtocolMessageType('ProjectGitIntegration', (_message.Message,), {
  'DESCRIPTOR' : _PROJECTGITINTEGRATION,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ProjectGitIntegration)
  })
_sym_db.RegisterMessage(ProjectGitIntegration)

Metric = _reflection.GeneratedProtocolMessageType('Metric', (_message.Message,), {
  'DESCRIPTOR' : _METRIC,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Metric)
  })
_sym_db.RegisterMessage(Metric)

MetricValue = _reflection.GeneratedProtocolMessageType('MetricValue', (_message.Message,), {
  'DESCRIPTOR' : _METRICVALUE,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.MetricValue)
  })
_sym_db.RegisterMessage(MetricValue)

Metadata = _reflection.GeneratedProtocolMessageType('Metadata', (_message.Message,), {
  'DESCRIPTOR' : _METADATA,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Metadata)
  })
_sym_db.RegisterMessage(Metadata)

GitContext = _reflection.GeneratedProtocolMessageType('GitContext', (_message.Message,), {

  'Repository' : _reflection.GeneratedProtocolMessageType('Repository', (_message.Message,), {
    'DESCRIPTOR' : _GITCONTEXT_REPOSITORY,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GitContext.Repository)
    })
  ,

  'Commit' : _reflection.GeneratedProtocolMessageType('Commit', (_message.Message,), {
    'DESCRIPTOR' : _GITCONTEXT_COMMIT,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GitContext.Commit)
    })
  ,

  'Branch' : _reflection.GeneratedProtocolMessageType('Branch', (_message.Message,), {
    'DESCRIPTOR' : _GITCONTEXT_BRANCH,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GitContext.Branch)
    })
  ,

  'GitTag' : _reflection.GeneratedProtocolMessageType('GitTag', (_message.Message,), {
    'DESCRIPTOR' : _GITCONTEXT_GITTAG,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GitContext.GitTag)
    })
  ,

  'GitTags' : _reflection.GeneratedProtocolMessageType('GitTags', (_message.Message,), {
    'DESCRIPTOR' : _GITCONTEXT_GITTAGS,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GitContext.GitTags)
    })
  ,

  'Ref' : _reflection.GeneratedProtocolMessageType('Ref', (_message.Message,), {
    'DESCRIPTOR' : _GITCONTEXT_REF,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GitContext.Ref)
    })
  ,

  'Author' : _reflection.GeneratedProtocolMessageType('Author', (_message.Message,), {
    'DESCRIPTOR' : _GITCONTEXT_AUTHOR,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GitContext.Author)
    })
  ,

  'Committer' : _reflection.GeneratedProtocolMessageType('Committer', (_message.Message,), {
    'DESCRIPTOR' : _GITCONTEXT_COMMITTER,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GitContext.Committer)
    })
  ,

  'Head' : _reflection.GeneratedProtocolMessageType('Head', (_message.Message,), {
    'DESCRIPTOR' : _GITCONTEXT_HEAD,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GitContext.Head)
    })
  ,
  'DESCRIPTOR' : _GITCONTEXT,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GitContext)
  })
_sym_db.RegisterMessage(GitContext)
_sym_db.RegisterMessage(GitContext.Repository)
_sym_db.RegisterMessage(GitContext.Commit)
_sym_db.RegisterMessage(GitContext.Branch)
_sym_db.RegisterMessage(GitContext.GitTag)
_sym_db.RegisterMessage(GitContext.GitTags)
_sym_db.RegisterMessage(GitContext.Ref)
_sym_db.RegisterMessage(GitContext.Author)
_sym_db.RegisterMessage(GitContext.Committer)
_sym_db.RegisterMessage(GitContext.Head)

GitProviderContext = _reflection.GeneratedProtocolMessageType('GitProviderContext', (_message.Message,), {

  'Repository' : _reflection.GeneratedProtocolMessageType('Repository', (_message.Message,), {
    'DESCRIPTOR' : _GITPROVIDERCONTEXT_REPOSITORY,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GitProviderContext.Repository)
    })
  ,
  'DESCRIPTOR' : _GITPROVIDERCONTEXT,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GitProviderContext)
  })
_sym_db.RegisterMessage(GitProviderContext)
_sym_db.RegisterMessage(GitProviderContext.Repository)

CicdRunnerContext = _reflection.GeneratedProtocolMessageType('CicdRunnerContext', (_message.Message,), {
  'DESCRIPTOR' : _CICDRUNNERCONTEXT,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CicdRunnerContext)
  })
_sym_db.RegisterMessage(CicdRunnerContext)

SdlcReviewContext = _reflection.GeneratedProtocolMessageType('SdlcReviewContext', (_message.Message,), {

  'User' : _reflection.GeneratedProtocolMessageType('User', (_message.Message,), {
    'DESCRIPTOR' : _SDLCREVIEWCONTEXT_USER,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.SdlcReviewContext.User)
    })
  ,

  'Ref' : _reflection.GeneratedProtocolMessageType('Ref', (_message.Message,), {
    'DESCRIPTOR' : _SDLCREVIEWCONTEXT_REF,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.SdlcReviewContext.Ref)
    })
  ,
  'DESCRIPTOR' : _SDLCREVIEWCONTEXT,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.SdlcReviewContext)
  })
_sym_db.RegisterMessage(SdlcReviewContext)
_sym_db.RegisterMessage(SdlcReviewContext.User)
_sym_db.RegisterMessage(SdlcReviewContext.Ref)

RunSummary = _reflection.GeneratedProtocolMessageType('RunSummary', (_message.Message,), {

  'Resource' : _reflection.GeneratedProtocolMessageType('Resource', (_message.Message,), {
    'DESCRIPTOR' : _RUNSUMMARY_RESOURCE,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.RunSummary.Resource)
    })
  ,
  'DESCRIPTOR' : _RUNSUMMARY,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.RunSummary)
  })
_sym_db.RegisterMessage(RunSummary)
_sym_db.RegisterMessage(RunSummary.Resource)

Run = _reflection.GeneratedProtocolMessageType('Run', (_message.Message,), {
  'DESCRIPTOR' : _RUN,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Run)
  })
_sym_db.RegisterMessage(Run)

Tag = _reflection.GeneratedProtocolMessageType('Tag', (_message.Message,), {
  'DESCRIPTOR' : _TAG,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Tag)
  })
_sym_db.RegisterMessage(Tag)

NotificationPolicyConfig = _reflection.GeneratedProtocolMessageType('NotificationPolicyConfig', (_message.Message,), {

  'Email' : _reflection.GeneratedProtocolMessageType('Email', (_message.Message,), {
    'DESCRIPTOR' : _NOTIFICATIONPOLICYCONFIG_EMAIL,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.NotificationPolicyConfig.Email)
    })
  ,

  'Slack' : _reflection.GeneratedProtocolMessageType('Slack', (_message.Message,), {
    'DESCRIPTOR' : _NOTIFICATIONPOLICYCONFIG_SLACK,
    '__module__' : 'continual.rpc.management.v1.management_types_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.NotificationPolicyConfig.Slack)
    })
  ,
  'DESCRIPTOR' : _NOTIFICATIONPOLICYCONFIG,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.NotificationPolicyConfig)
  })
_sym_db.RegisterMessage(NotificationPolicyConfig)
_sym_db.RegisterMessage(NotificationPolicyConfig.Email)
_sym_db.RegisterMessage(NotificationPolicyConfig.Slack)

NotificationPolicy = _reflection.GeneratedProtocolMessageType('NotificationPolicy', (_message.Message,), {
  'DESCRIPTOR' : _NOTIFICATIONPOLICY,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.NotificationPolicy)
  })
_sym_db.RegisterMessage(NotificationPolicy)

Notification = _reflection.GeneratedProtocolMessageType('Notification', (_message.Message,), {
  'DESCRIPTOR' : _NOTIFICATION,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.Notification)
  })
_sym_db.RegisterMessage(Notification)

RunCheckSummary = _reflection.GeneratedProtocolMessageType('RunCheckSummary', (_message.Message,), {
  'DESCRIPTOR' : _RUNCHECKSUMMARY,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.RunCheckSummary)
  })
_sym_db.RegisterMessage(RunCheckSummary)

ResourceCheck = _reflection.GeneratedProtocolMessageType('ResourceCheck', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCECHECK,
  '__module__' : 'continual.rpc.management.v1.management_types_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ResourceCheck)
  })
_sym_db.RegisterMessage(ResourceCheck)


DESCRIPTOR._options = None
_USER.fields_by_name['password']._options = None
_USER.fields_by_name['service_account']._options = None
_USER.fields_by_name['update_time']._options = None
_USER.fields_by_name['create_time']._options = None
_SESSION.fields_by_name['name']._options = None
_SESSION.fields_by_name['subject']._options = None
_SESSION.fields_by_name['last_access_ip']._options = None
_SESSION.fields_by_name['last_access_time']._options = None
_SESSION.fields_by_name['login_ip']._options = None
_SESSION.fields_by_name['login_time']._options = None
_SESSION.fields_by_name['login_user_agent']._options = None
_ACCESSPOLICY.fields_by_name['subject']._options = None
_ACCESSPOLICY.fields_by_name['role']._options = None
_INVITE.fields_by_name['sender']._options = None
_INVITE.fields_by_name['role']._options = None
_INVITE.fields_by_name['create_time']._options = None
_INVITE.fields_by_name['expire_time']._options = None
_ORGANIZATION.fields_by_name['update_time']._options = None
_ORGANIZATION.fields_by_name['create_time']._options = None
_ORGANIZATION.fields_by_name['status']._options = None
_ORGANIZATION.fields_by_name['trial_end_date']._options = None
_ORGANIZATION.fields_by_name['trial_credits']._options = None
_ORGANIZATION.fields_by_name['trial_credits_used']._options = None
_ORGANIZATION.fields_by_name['sso_enabled']._options = None
_ORGANIZATION.fields_by_name['sso_configured']._options = None
_ORGANIZATION.fields_by_name['directory_sync_configured']._options = None
_ONBOARDING.fields_by_name['create_time']._options = None
_ONBOARDING.fields_by_name['update_time']._options = None
_PROJECT.fields_by_name['organization']._options = None
_PROJECT.fields_by_name['summary']._options = None
_PROJECT.fields_by_name['update_time']._options = None
_PROJECT.fields_by_name['create_time']._options = None
_PROJECT.fields_by_name['git_integration']._options = None
_DATASTORE_SNOWFLAKE_PARAMSENTRY._options = None
_DATASTORE_POSTGRES_PARAMSENTRY._options = None
_DATASTORE_REDSHIFT_PARAMSENTRY._options = None
_DATASTORE_DATABRICKS_PARAMSENTRY._options = None
_ENVIRONMENT.fields_by_name['summary']._options = None
_ENVIRONMENT.fields_by_name['update_time']._options = None
_ENVIRONMENT.fields_by_name['create_time']._options = None
_FEATURESET.fields_by_name['table']._options = None
_EVENT_DATAENTRY._options = None
_EVENT.fields_by_name['subject']._options = None
_APIKEY.fields_by_name['key']._options = None
_APIKEY.fields_by_name['created_by']._options = None
_APIKEY.fields_by_name['create_time']._options = None
_APIKEY.fields_by_name['last_access_ip']._options = None
_APIKEY.fields_by_name['last_access_time']._options = None
_APIKEY.fields_by_name['revoked_time']._options = None
_APIKEY.fields_by_name['revoked_by']._options = None
_MODEL.fields_by_name['current_version']._options = None
_MODEL.fields_by_name['create_time']._options = None
_MODEL.fields_by_name['update_time']._options = None
_MODEL.fields_by_name['latest_model_version']._options = None
_MODEL.fields_by_name['latest_batch_prediction']._options = None
_PROMOTION.fields_by_name['create_time']._options = None
_PROMOTION.fields_by_name['promoted_time']._options = None
_PROMOTION.fields_by_name['demoted_time']._options = None
_PROMOTION.fields_by_name['model_version']._options = None
_PROMOTION.fields_by_name['base_model_version']._options = None
_PROMOTION.fields_by_name['state']._options = None
_PROMOTION.fields_by_name['account']._options = None
_MODELVERSION.fields_by_name['name']._options = None
_MODELVERSION.fields_by_name['experiment_name']._options = None
_MODELVERSION.fields_by_name['state']._options = None
_MODELVERSION.fields_by_name['error_message']._options = None
_MODELVERSION.fields_by_name['stack_trace']._options = None
_MODELVERSION.fields_by_name['create_time']._options = None
_MODELVERSION.fields_by_name['update_time']._options = None
_MODELVERSION.fields_by_name['training_row_count']._options = None
_MODELVERSION.fields_by_name['validation_row_count']._options = None
_MODELVERSION.fields_by_name['test_row_count']._options = None
_MODELVERSION.fields_by_name['performance_metric']._options = None
_MODELVERSION.fields_by_name['performance_metric_val']._options = None
_MODELVERSION.fields_by_name['promotion']._options = None
_BATCHPREDICTION.fields_by_name['model_version']._options = None
_DATASET.fields_by_name['current_version']._options = None
_DATASET.fields_by_name['create_time']._options = None
_DATASET.fields_by_name['update_time']._options = None
_DATASETVERSION.fields_by_name['name']._options = None
_DATASETVERSION.fields_by_name['create_time']._options = None
_DATASETVERSION.fields_by_name['update_time']._options = None
_BILLINGACCOUNT.fields_by_name['update_time']._options = None
_BILLINGACCOUNT.fields_by_name['create_time']._options = None
_SUBSCRIPTION.fields_by_name['price']._options = None
_SUBSCRIPTION.fields_by_name['payment_method']._options = None
_SUBSCRIPTION.fields_by_name['update_time']._options = None
_SUBSCRIPTION.fields_by_name['create_time']._options = None
_PROJECTUSAGERECORD_TYPECOUNTENTRY._options = None
# @@protoc_insertion_point(module_scope)
