# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: continual/rpc/management/v1/management.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.api import annotations_pb2 as google_dot_api_dot_annotations__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from google.protobuf import field_mask_pb2 as google_dot_protobuf_dot_field__mask__pb2
from continual.rpc.management.v1 import management_types_pb2 as continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2
from continual.rpc.graphql import options_pb2 as continual_dot_rpc_dot_graphql_dot_options__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='continual/rpc/management/v1/management.proto',
  package='continual.rpc.management.v1',
  syntax='proto3',
  serialized_options=b'ZJgithub.com/continual-ai/continual/continual/rpc/management/v1;managementv1',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n,continual/rpc/management/v1/management.proto\x12\x1b\x63ontinual.rpc.management.v1\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1cgoogle/api/annotations.proto\x1a\x1bgoogle/protobuf/empty.proto\x1a google/protobuf/field_mask.proto\x1a\x32\x63ontinual/rpc/management/v1/management_types.proto\x1a#continual/rpc/graphql/options.proto\"u\n\x11UpdateUserRequest\x12/\n\x04user\x18\x01 \x01(\x0b\x32!.continual.rpc.management.v1.User\x12/\n\x0bupdate_mask\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\"\x1e\n\x0eGetUserRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"%\n\x14\x42\x61tchGetUsersRequest\x12\r\n\x05names\x18\x01 \x03(\t\"I\n\x15\x42\x61tchGetUsersResponse\x12\x30\n\x05users\x18\x01 \x03(\x0b\x32!.continual.rpc.management.v1.User\"!\n\x11\x44\x65leteUserRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"k\n\x10ListUsersRequest\x12\x11\n\tpage_size\x18\x01 \x01(\x05\x12\x12\n\npage_token\x18\x02 \x01(\t\x12\x10\n\x08order_by\x18\x03 \x01(\t\x12\x0e\n\x06search\x18\x04 \x01(\t\x12\x0e\n\x06latest\x18\x05 \x01(\x08\"^\n\x11ListUsersResponse\x12\x30\n\x05users\x18\x01 \x03(\x0b\x32!.continual.rpc.management.v1.User\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"l\n\x0fRegisterRequest\x12\x11\n\tfull_name\x18\x01 \x01(\t\x12\r\n\x05\x65mail\x18\x02 \x01(\t\x12\x10\n\x08password\x18\x03 \x01(\t\x12\x12\n\nfirst_name\x18\x04 \x01(\t\x12\x11\n\tlast_name\x18\x05 \x01(\t\"W\n\x10RegisterResponse\x12/\n\x04user\x18\x01 \x01(\x0b\x32!.continual.rpc.management.v1.User\x12\x12\n\nauth_token\x18\x02 \x01(\t\"/\n\x0cLoginRequest\x12\r\n\x05\x65mail\x18\x01 \x01(\t\x12\x10\n\x08password\x18\x02 \x01(\t\"T\n\rLoginResponse\x12/\n\x04user\x18\x01 \x01(\x0b\x32!.continual.rpc.management.v1.User\x12\x12\n\nauth_token\x18\x02 \x01(\t\"6\n\x1e\x43heckOrganizationAccessRequest\x12\x14\n\x0corganization\x18\x01 \x01(\t\"d\n\x1f\x43heckOrganizationAccessResponse\x12\x15\n\rrequires_auth\x18\x01 \x01(\x08\x12\x14\n\x0corganization\x18\x02 \x01(\t\x12\x14\n\x0c\x64isplay_name\x18\x03 \x01(\t\",\n\x1cSendEmailVerificationRequest\x12\x0c\n\x04user\x18\x01 \x01(\t\"0\n\x12VerifyEmailRequest\x12\x0c\n\x04user\x18\x01 \x01(\t\x12\x0c\n\x04\x63ode\x18\x02 \x01(\t\"&\n\x13VerifyEmailResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\")\n\x18SendPasswordResetRequest\x12\r\n\x05\x65mail\x18\x01 \x01(\t\"\'\n\x17GetPasswordResetRequest\x12\x0c\n\x04\x63ode\x18\x01 \x01(\t\"L\n\x14ResetPasswordRequest\x12\x10\n\x08password\x18\x01 \x01(\t\x12\x0c\n\x04\x63ode\x18\x02 \x01(\t\x12\x14\n\x0cold_password\x18\x03 \x01(\t\"(\n\x15ResetPasswordResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\"?\n\x15\x43hangePasswordRequest\x12\x10\n\x08password\x18\x01 \x01(\t\x12\x14\n\x0cold_password\x18\x02 \x01(\t\"=\n\x16\x43hangePasswordResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x12\n\nauth_token\x18\x02 \x01(\t\"!\n\x11GetSessionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"L\n\x13ListSessionsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\"g\n\x14ListSessionsResponse\x12\x36\n\x08sessions\x18\x01 \x03(\x0b\x32$.continual.rpc.management.v1.Session\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"$\n\x14\x44\x65leteSessionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"*\n\x18\x44\x65leteAllSessionsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\"9\n\x10ListRolesRequest\x12\x11\n\tpage_size\x18\x01 \x01(\x05\x12\x12\n\npage_token\x18\x02 \x01(\t\"^\n\x11ListRolesResponse\x12\x30\n\x05roles\x18\x01 \x03(\x0b\x32!.continual.rpc.management.v1.Role\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"\x1e\n\x0eGetRoleRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"5\n\x14\x42\x61tchGetRolesRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\r\n\x05names\x18\x02 \x03(\t\"I\n\x15\x42\x61tchGetRolesResponse\x12\x30\n\x05roles\x18\x01 \x03(\x0b\x32!.continual.rpc.management.v1.Role\"l\n\x19ListAccessPoliciesRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\x12\x18\n\x10include_api_keys\x18\x04 \x01(\x08\"y\n\x1aListAccessPoliciesResponse\x12\x42\n\x0f\x61\x63\x63\x65ss_policies\x18\x01 \x03(\x0b\x32).continual.rpc.management.v1.AccessPolicy\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"m\n\x19\x43reateAccessPolicyRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12@\n\raccess_policy\x18\x02 \x01(\x0b\x32).continual.rpc.management.v1.AccessPolicy\"&\n\x16GetAccessPolicyRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x8e\x01\n\x19UpdateAccessPolicyRequest\x12@\n\raccess_policy\x18\x01 \x01(\x0b\x32).continual.rpc.management.v1.AccessPolicy\x12/\n\x0bupdate_mask\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\")\n\x19\x44\x65leteAccessPolicyRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"0\n\x1f\x42ulkDeleteAccessPoliciesRequest\x12\r\n\x05names\x18\x01 \x03(\t\"Z\n\x13\x43reateInviteRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x33\n\x06invite\x18\x02 \x01(\x0b\x32#.continual.rpc.management.v1.Invite\" \n\x10GetInviteRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"K\n\x12ListInvitesRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x12\n\npage_token\x18\x02 \x01(\t\x12\x11\n\tpage_size\x18\x03 \x01(\x05\"d\n\x13ListInvitesResponse\x12\x34\n\x07invites\x18\x01 \x03(\x0b\x32#.continual.rpc.management.v1.Invite\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"{\n\x13UpdateInviteRequest\x12\x33\n\x06invite\x18\x01 \x01(\x0b\x32#.continual.rpc.management.v1.Invite\x12/\n\x0bupdate_mask\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\"#\n\x13\x44\x65leteInviteRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"#\n\x13\x41\x63\x63\x65ptInviteRequest\x12\x0c\n\x04\x63ode\x18\x01 \x01(\t\"#\n\x13ResendInviteRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"?\n\x16TestPermissionsRequest\x12\x10\n\x08resource\x18\x01 \x01(\t\x12\x13\n\x0bpermissions\x18\x02 \x03(\t\".\n\x17TestPermissionsResponse\x12\x13\n\x0bpermissions\x18\x01 \x03(\t\"\\\n\x19\x43reateOrganizationRequest\x12?\n\x0corganization\x18\x01 \x01(\x0b\x32).continual.rpc.management.v1.Organization\"\x8d\x01\n\x19UpdateOrganizationRequest\x12?\n\x0corganization\x18\x01 \x01(\x0b\x32).continual.rpc.management.v1.Organization\x12/\n\x0bupdate_mask\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\"&\n\x16GetOrganizationRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"-\n\x1c\x42\x61tchGetOrganizationsRequest\x12\r\n\x05names\x18\x01 \x03(\t\"a\n\x1d\x42\x61tchGetOrganizationsResponse\x12@\n\rorganizations\x18\x01 \x03(\x0b\x32).continual.rpc.management.v1.Organization\")\n\x19\x44\x65leteOrganizationRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x88\x01\n\x18ListOrganizationsRequest\x12\x11\n\tpage_size\x18\x01 \x01(\x05\x12\x12\n\npage_token\x18\x02 \x01(\t\x12\x10\n\x08order_by\x18\x03 \x01(\t\x12\x0e\n\x06search\x18\x04 \x01(\t\x12\x0e\n\x06latest\x18\x05 \x01(\x08\x12\x13\n\x0binclude_sso\x18\x06 \x01(\x08\"v\n\x19ListOrganizationsResponse\x12@\n\rorganizations\x18\x01 \x03(\x0b\x32).continual.rpc.management.v1.Organization\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"$\n\x14GetOnboardingRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x87\x01\n\x17UpdateOnboardingRequest\x12;\n\nonboarding\x18\x01 \x01(\x0b\x32\'.continual.rpc.management.v1.Onboarding\x12/\n\x0bupdate_mask\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\"a\n\x14\x43reateProjectRequest\x12\x35\n\x07project\x18\x01 \x01(\x0b\x32$.continual.rpc.management.v1.Project\x12\x12\n\nproject_id\x18\x02 \x01(\t\"~\n\x14UpdateProjectRequest\x12\x35\n\x07project\x18\x01 \x01(\x0b\x32$.continual.rpc.management.v1.Project\x12/\n\x0bupdate_mask\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\"!\n\x11GetProjectRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"(\n\x17\x42\x61tchGetProjectsRequest\x12\r\n\x05names\x18\x01 \x03(\t\"R\n\x18\x42\x61tchGetProjectsResponse\x12\x36\n\x08projects\x18\x01 \x03(\x0b\x32$.continual.rpc.management.v1.Project\"$\n\x14\x44\x65leteProjectRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x85\x02\n\x13ListProjectsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\x12\x0f\n\x07\x66ilters\x18\x04 \x03(\t\x12\x16\n\x0eignore_filters\x18\x05 \x03(\t\x12.\n\nstart_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x10\n\x08order_by\x18\x08 \x01(\t\x12\x0e\n\x06search\x18\t \x01(\t\x12\x0e\n\x06latest\x18\n \x01(\x08\"g\n\x14ListProjectsResponse\x12\x36\n\x08projects\x18\x01 \x03(\x0b\x32$.continual.rpc.management.v1.Project\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\">\n\x0cReRunRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07message\x18\x06 \x01(\t\x12\x0f\n\x07trigger\x18\x07 \x01(\t\"$\n\x14GetModelStatsRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x7f\n\x0fGetUsageRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12.\n\nstart_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"[\n\x10GetUsageResponse\x12G\n\x0f\x63redits_per_day\x18\x01 \x03(\x0b\x32..continual.rpc.management.v1.CreditUsageRecord\"#\n\x11ListPricesRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\"H\n\x12ListPricesResponse\x12\x32\n\x06prices\x18\x01 \x03(\x0b\x32\".continual.rpc.management.v1.Price\"L\n\x13ListInvoicesRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\"g\n\x14ListInvoicesResponse\x12\x36\n\x08invoices\x18\x01 \x03(\x0b\x32$.continual.rpc.management.v1.Invoice\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"\x1f\n\x0fGetPriceRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\'\n\x17GetPaymentMethodRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\")\n\x19GetUpcomingInvoiceRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"&\n\x16GetSubscriptionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"(\n\x18GetBillingAccountRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x94\x01\n\x1bUpdateBillingAccountRequest\x12\x44\n\x0f\x62illing_account\x18\x01 \x01(\x0b\x32+.continual.rpc.management.v1.BillingAccount\x12/\n\x0bupdate_mask\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\"\x8d\x01\n\x19UpdateSubscriptionRequest\x12?\n\x0csubscription\x18\x01 \x01(\x0b\x32).continual.rpc.management.v1.Subscription\x12/\n\x0bupdate_mask\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\"V\n\x16\x43reatePromotionRequest\x12\x1a\n\x12model_version_name\x18\x01 \x01(\t\x12\x10\n\x08run_name\x18\x02 \x01(\t\x12\x0e\n\x06reason\x18\x03 \x01(\t\"\xad\x02\n\x15ListPromotionsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x12\n\npage_token\x18\x02 \x01(\t\x12\x11\n\tpage_size\x18\x03 \x01(\x05\x12\x0e\n\x06\x61\x63tive\x18\x04 \x01(\x08\x12\x0f\n\x07\x66ilters\x18\x05 \x03(\t\x12\x16\n\x0eignore_filters\x18\x06 \x03(\t\x12.\n\nstart_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x14\n\x0c\x61ll_projects\x18\t \x01(\x08\x12\x10\n\x08order_by\x18\n \x01(\t\x12\x0e\n\x06latest\x18\x0b \x01(\x08\x12\x0e\n\x06search\x18\x0c \x01(\t\"m\n\x16ListPromotionsResponse\x12:\n\npromotions\x18\x01 \x03(\x0b\x32&.continual.rpc.management.v1.Promotion\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"#\n\x13GetPromotionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\":\n\x19\x42\x61tchGetPromotionsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\r\n\x05names\x18\x02 \x03(\t\"X\n\x1a\x42\x61tchGetPromotionsResponse\x12:\n\npromotions\x18\x01 \x03(\x0b\x32&.continual.rpc.management.v1.Promotion\"W\n\x12\x43reateEventRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x31\n\x05\x65vent\x18\x02 \x01(\x0b\x32\".continual.rpc.management.v1.Event\"\xf7\x01\n\x11ListEventsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x0f\n\x07\x66ilters\x18\x02 \x03(\t\x12\x16\n\x0eignore_filters\x18\x03 \x03(\t\x12.\n\nstart_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x12\n\npage_token\x18\x06 \x01(\t\x12\x11\n\tpage_size\x18\x07 \x01(\x05\x12\x14\n\x0c\x61ll_projects\x18\x08 \x01(\x08\x12\x0e\n\x06search\x18\t \x01(\t\"a\n\x12ListEventsResponse\x12\x32\n\x06\x65vents\x18\x01 \x03(\x0b\x32\".continual.rpc.management.v1.Event\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"\x1f\n\x0fGetEventRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"6\n\x15\x42\x61tchGetEventsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\r\n\x05names\x18\x02 \x03(\t\"L\n\x16\x42\x61tchGetEventsResponse\x12\x32\n\x06\x65vents\x18\x01 \x03(\x0b\x32\".continual.rpc.management.v1.Event\"[\n\x12\x43reateModelRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12\x15\n\rget_if_exists\x18\x03 \x01(\x08\x12\x0b\n\x03run\x18\x04 \x01(\t\"\x1f\n\x0fGetModelRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"6\n\x15\x42\x61tchGetModelsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\r\n\x05names\x18\x02 \x03(\t\"L\n\x16\x42\x61tchGetModelsResponse\x12\x32\n\x06models\x18\x01 \x03(\x0b\x32\".continual.rpc.management.v1.Model\"\x99\x02\n\x11ListModelsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\x12\x0e\n\x06latest\x18\x05 \x01(\x08\x12\x0f\n\x07\x66ilters\x18\x06 \x03(\t\x12\x16\n\x0eignore_filters\x18\x07 \x03(\t\x12.\n\nstart_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\t \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x14\n\x0c\x61ll_projects\x18\n \x01(\x08\x12\x10\n\x08order_by\x18\x0b \x01(\t\x12\x0e\n\x06search\x18\x0c \x01(\t\"a\n\x12ListModelsResponse\x12\x32\n\x06models\x18\x01 \x03(\x0b\x32\".continual.rpc.management.v1.Model\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"4\n\x12\x44\x65leteModelRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08run_name\x18\x02 \x01(\t\";\n\x14ValidateModelRequest\x12\x13\n\x0b\x65nvironment\x18\x01 \x01(\t\x12\x0e\n\x06schema\x18\x02 \x01(\t\"O\n\x19\x43reateModelVersionRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x10\n\x08run_name\x18\x02 \x01(\t\x12\x10\n\x08\x64\x61tasets\x18\x03 \x03(\t\"&\n\x16GetModelVersionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"=\n\x1c\x42\x61tchGetModelVersionsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\r\n\x05names\x18\x02 \x03(\t\"b\n\x1d\x42\x61tchGetModelVersionsResponse\x12\x41\n\x0emodel_versions\x18\x01 \x03(\x0b\x32).continual.rpc.management.v1.ModelVersion\"\xa0\x02\n\x18ListModelVersionsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\x12\x0e\n\x06latest\x18\x04 \x01(\x08\x12\x0f\n\x07\x66ilters\x18\x05 \x03(\t\x12\x16\n\x0eignore_filters\x18\x06 \x03(\t\x12.\n\nstart_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x14\n\x0c\x61ll_projects\x18\t \x01(\x08\x12\x10\n\x08order_by\x18\n \x01(\t\x12\x0e\n\x06search\x18\x0b \x01(\t\"w\n\x19ListModelVersionsResponse\x12\x41\n\x0emodel_versions\x18\x01 \x03(\x0b\x32).continual.rpc.management.v1.ModelVersion\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\";\n\x17\x43reateExperimentRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x10\n\x08run_name\x18\x02 \x01(\t\"$\n\x14GetExperimentRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\";\n\x1a\x42\x61tchGetExperimentsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\r\n\x05names\x18\x02 \x03(\t\"[\n\x1b\x42\x61tchGetExperimentsResponse\x12<\n\x0b\x65xperiments\x18\x01 \x03(\x0b\x32\'.continual.rpc.management.v1.Experiment\"\x9e\x02\n\x16ListExperimentsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\x12\x0f\n\x07\x66ilters\x18\x04 \x03(\t\x12\x16\n\x0eignore_filters\x18\x05 \x03(\t\x12.\n\nstart_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x14\n\x0c\x61ll_projects\x18\x08 \x01(\x08\x12\x0e\n\x06latest\x18\t \x01(\x08\x12\x10\n\x08order_by\x18\n \x01(\t\x12\x0e\n\x06search\x18\x0b \x01(\t\"p\n\x17ListExperimentsResponse\x12<\n\x0b\x65xperiments\x18\x01 \x03(\x0b\x32\'.continual.rpc.management.v1.Experiment\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"d\n\x1c\x43reateBatchPredictionRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x10\n\x08run_name\x18\x02 \x01(\t\x12\"\n\x1atrained_model_version_name\x18\x03 \x01(\t\"\xa3\x02\n\x1bListBatchPredictionsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\x12\x0e\n\x06latest\x18\x04 \x01(\x08\x12\x0f\n\x07\x66ilters\x18\x05 \x03(\t\x12\x16\n\x0eignore_filters\x18\x06 \x03(\t\x12.\n\nstart_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x14\n\x0c\x61ll_projects\x18\t \x01(\x08\x12\x10\n\x08order_by\x18\n \x01(\t\x12\x0e\n\x06search\x18\x0b \x01(\t\"\x80\x01\n\x1cListBatchPredictionsResponse\x12G\n\x11\x62\x61tch_predictions\x18\x01 \x03(\x0b\x32,.continual.rpc.management.v1.BatchPrediction\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"&\n\x16\x42\x61tchPredictionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"@\n\x1f\x42\x61tchGetBatchPredictionsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\r\n\x05names\x18\x02 \x03(\t\"k\n BatchGetBatchPredictionsResponse\x12G\n\x11\x62\x61tch_predictions\x18\x01 \x03(\x0b\x32,.continual.rpc.management.v1.BatchPrediction\"e\n\x0fYamlDiffRequest\x12\r\n\x05right\x18\x01 \x01(\t\x12\x12\n\nright_text\x18\x02 \x01(\t\x12\x0c\n\x04left\x18\x03 \x01(\t\x12\x11\n\tleft_text\x18\x04 \x01(\t\x12\x0e\n\x06parent\x18\x05 \x01(\t\" \n\x10YamlDiffResponse\x12\x0c\n\x04text\x18\x01 \x01(\t\"i\n\x18\x43reateEnvironmentRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x16\n\x0e\x65nvironment_id\x18\x02 \x01(\t\x12\x0e\n\x06source\x18\x03 \x01(\t\x12\x15\n\rget_if_exists\x18\x05 \x01(\x08\"\x8a\x01\n\x18UpdateEnvironmentRequest\x12=\n\x0b\x65nvironment\x18\x01 \x01(\x0b\x32(.continual.rpc.management.v1.Environment\x12/\n\x0bupdate_mask\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\"%\n\x15GetEnvironmentRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\",\n\x1b\x42\x61tchGetEnvironmentsRequest\x12\r\n\x05names\x18\x01 \x03(\t\"^\n\x1c\x42\x61tchGetEnvironmentsResponse\x12>\n\x0c\x65nvironments\x18\x01 \x03(\x0b\x32(.continual.rpc.management.v1.Environment\":\n\x18\x44\x65leteEnvironmentRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08run_name\x18\x02 \x01(\t\"\x89\x02\n\x17ListEnvironmentsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\x12\x0f\n\x07\x66ilters\x18\x04 \x03(\t\x12\x16\n\x0eignore_filters\x18\x05 \x03(\t\x12.\n\nstart_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x10\n\x08order_by\x18\x08 \x01(\t\x12\x0e\n\x06search\x18\t \x01(\t\x12\x0e\n\x06latest\x18\n \x01(\x08\"s\n\x18ListEnvironmentsResponse\x12>\n\x0c\x65nvironments\x18\x01 \x03(\x0b\x32(.continual.rpc.management.v1.Environment\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"\x19\n\x17GetSystemMessageRequest\"+\n\x18GetSystemMessageResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\":\n\x15GetFeatureFlagRequest\x12\x10\n\x08resource\x18\x01 \x01(\t\x12\x0f\n\x07\x66\x65\x61ture\x18\x02 \x01(\t\"L\n\x16GetFeatureFlagResponse\x12\x10\n\x08resource\x18\x01 \x01(\t\x12\x0f\n\x07\x66\x65\x61ture\x18\x02 \x01(\t\x12\x0f\n\x07\x65nabled\x18\x03 \x01(\x08\"6\n\"GetOrganizationFeatureFlagsRequest\x12\x10\n\x08resource\x18\x01 \x01(\t\"\xc5\x01\n#GetOrganizationFeatureFlagsResponse\x12i\n\rfeature_flags\x18\x01 \x03(\x0b\x32R.continual.rpc.management.v1.GetOrganizationFeatureFlagsResponse.FeatureFlagsEntry\x1a\x33\n\x11\x46\x65\x61tureFlagsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x08:\x02\x38\x01\"\x8f\x01\n\x15\x43reateArtifactRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x37\n\x08\x61rtifact\x18\x02 \x01(\x0b\x32%.continual.rpc.management.v1.Artifact\x12\x13\n\x0b\x61rtifact_id\x18\x03 \x01(\t\x12\x18\n\x10update_if_exists\x18\x04 \x01(\x08\"\"\n\x12GetArtifactRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"9\n\x18\x42\x61tchGetArtifactsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\r\n\x05names\x18\x02 \x03(\t\"U\n\x19\x42\x61tchGetArtifactsResponse\x12\x38\n\tartifacts\x18\x01 \x03(\x0b\x32%.continual.rpc.management.v1.Artifact\"%\n\x15\x44\x65leteArtifactRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\xa8\x01\n\x14ListArtifactsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\x12\x0f\n\x07\x66ilters\x18\x04 \x03(\t\x12\x16\n\x0eignore_filters\x18\x05 \x03(\t\x12\x10\n\x08order_by\x18\x06 \x01(\t\x12\x0e\n\x06search\x18\x07 \x01(\t\x12\x0e\n\x06latest\x18\x08 \x01(\x08\"j\n\x15ListArtifactsResponse\x12\x38\n\tartifacts\x18\x01 \x03(\x0b\x32%.continual.rpc.management.v1.Artifact\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"\xa2\x01\n GenerateArtifactUploadURLRequest\x12\x13\n\x0b\x61rtifact_id\x18\x01 \x01(\t\x12\x0e\n\x06parent\x18\x02 \x01(\t\x12\x0c\n\x04path\x18\x03 \x01(\t\x12\x11\n\tmime_type\x18\x04 \x01(\t\x12\x11\n\tresumable\x18\x05 \x01(\x08\x12\x0b\n\x03run\x18\x06 \x01(\t\x12\x18\n\x10update_if_exists\x18\x07 \x01(\x08\"i\n!GenerateArtifactUploadURLResponse\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x37\n\x08\x61rtifact\x18\x02 \x01(\x0b\x32%.continual.rpc.management.v1.Artifact\"%\n#ListUserIntegrationProvidersRequest\"o\n$ListUserIntegrationProvidersResponse\x12G\n\tproviders\x18\x01 \x03(\x0b\x32\x34.continual.rpc.management.v1.UserIntegrationProvider\"\xbd\x01\n\x1bListUserIntegrationsRequest\x12\x11\n\tuser_name\x18\x01 \x01(\t\x12\x0f\n\x07\x66ilters\x18\x02 \x03(\t\x12\x16\n\x0eignore_filters\x18\x03 \x03(\t\x12\x12\n\npage_token\x18\x04 \x01(\t\x12\x11\n\tpage_size\x18\x05 \x01(\x05\x12\x10\n\x08order_by\x18\x06 \x01(\t\x12\x0e\n\x06search\x18\x07 \x01(\t\x12\x19\n\x11show_disconnected\x18\x08 \x01(\x08\"{\n\x1cListUserIntegrationsResponse\x12\x42\n\x0cintegrations\x18\x01 \x03(\x0b\x32,.continual.rpc.management.v1.UserIntegration\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"N\n DisconnectUserIntegrationRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06reason\x18\x02 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\t\"A\n ListGitProviderNamespacesRequest\x12\x1d\n\x15user_integration_name\x18\x01 \x01(\t\"\x89\x01\n!ListGitProviderNamespacesResponse\x12\x1d\n\x15user_integration_name\x18\x01 \x01(\t\x12\x45\n\nnamespaces\x18\x02 \x03(\x0b\x32\x31.continual.rpc.management.v1.GitProviderNamespace\"h\n\"ListGitProviderRepositoriesRequest\x12\x1d\n\x15user_integration_name\x18\x01 \x01(\t\x12#\n\x1bgit_provider_namespace_name\x18\x02 \x01(\t\"\xb3\x01\n#ListGitProviderRepositoriesResponse\x12\x1d\n\x15user_integration_name\x18\x01 \x01(\t\x12#\n\x1bgit_provider_namespace_name\x18\x02 \x01(\t\x12H\n\x0crepositories\x18\x03 \x03(\x0b\x32\x32.continual.rpc.management.v1.GitProviderRepository\"\xa2\x01\n$ConnectProjectToGitRepositoryRequest\x12\x14\n\x0cproject_name\x18\x01 \x01(\t\x12\x1d\n\x15user_integration_name\x18\x02 \x01(\t\x12$\n\x1cgit_provider_repository_name\x18\x03 \x01(\t\x12\x1f\n\x17\x64\x65\x66\x61ult_env_branch_name\x18\x04 \x01(\t\"|\n%ConnectProjectToGitRepositoryResponse\x12S\n\x17project_git_integration\x18\x01 \x01(\x0b\x32\x32.continual.rpc.management.v1.ProjectGitIntegration\"/\n\x1fGetProjectGitIntegrationRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"R\n\x1fSeedProjectGitRepositoryRequest\x12\x1f\n\x17project_git_integration\x18\x01 \x01(\t\x12\x0e\n\x06source\x18\x02 \x01(\t\"8\n GenerateSsoAdminPortalURLRequest\x12\x14\n\x0corganization\x18\x01 \x01(\t\"0\n!GenerateSsoAdminPortalURLResponse\x12\x0b\n\x03url\x18\x01 \x01(\t\":\n\"GenerateDSyncAdminPortalURLRequest\x12\x14\n\x0corganization\x18\x01 \x01(\t\"2\n#GenerateDSyncAdminPortalURLResponse\x12\x0b\n\x03url\x18\x01 \x01(\t\"%\n\x14\x45xchangeTokenRequest\x12\r\n\x05token\x18\x01 \x01(\t\"&\n\x15\x45xchangeTokenResponse\x12\r\n\x05token\x18\x01 \x01(\t\"#\n\x13GetPrincipalRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\" \n\x10GetApiKeyRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"7\n\x16\x42\x61tchGetApiKeysRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\r\n\x05names\x18\x02 \x03(\t\"P\n\x17\x42\x61tchGetApiKeysResponse\x12\x35\n\x08\x61pi_keys\x18\x01 \x03(\x0b\x32#.continual.rpc.management.v1.ApiKey\"\x96\x01\n\x12ListApiKeysRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\x12\x0f\n\x07\x66ilters\x18\x04 \x03(\t\x12\x16\n\x0eignore_filters\x18\x05 \x03(\t\x12\x10\n\x08order_by\x18\x06 \x01(\t\x12\x0e\n\x06search\x18\x07 \x01(\t\"e\n\x13ListApiKeysResponse\x12\x35\n\x08\x61pi_keys\x18\x01 \x03(\x0b\x32#.continual.rpc.management.v1.ApiKey\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"[\n\x13\x43reateApiKeyRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x34\n\x07\x61pi_key\x18\x02 \x01(\x0b\x32#.continual.rpc.management.v1.ApiKey\"|\n\x13UpdateApiKeyRequest\x12\x34\n\x07\x61pi_key\x18\x01 \x01(\x0b\x32#.continual.rpc.management.v1.ApiKey\x12/\n\x0bupdate_mask\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\"#\n\x13RevokeApiKeyRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\'\n\x17GetApiKeyProjectRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\x1d\n\rGetRunRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"$\n\x13\x42\x61tchGetRunsRequest\x12\r\n\x05names\x18\x01 \x03(\t\"F\n\x14\x42\x61tchGetRunsResponse\x12.\n\x04runs\x18\x01 \x03(\x0b\x32 .continual.rpc.management.v1.Run\"\x81\x02\n\x0fListRunsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\x12\x0f\n\x07\x66ilters\x18\x04 \x03(\t\x12\x16\n\x0eignore_filters\x18\x05 \x03(\t\x12.\n\nstart_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x10\n\x08order_by\x18\x08 \x01(\t\x12\x0e\n\x06search\x18\t \x01(\t\x12\x0e\n\x06latest\x18\n \x01(\x08\"[\n\x10ListRunsResponse\x12.\n\x04runs\x18\x01 \x03(\x0b\x32 .continual.rpc.management.v1.Run\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"z\n\x10\x43reateRunRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x0e\n\x06run_id\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x1a\n\x12heartbeat_interval\x18\x04 \x01(\x04\x12\x15\n\rget_if_exists\x18\x05 \x01(\x08\"W\n\x10UpdateRunRequest\x12-\n\x03run\x18\x01 \x01(\x0b\x32 .continual.rpc.management.v1.Run\x12\x14\n\x0cupdate_paths\x18\x02 \x03(\t\" \n\x10\x44\x65leteRunRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"$\n\x14GetRunSummaryRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\")\n\x19GetRunCheckSummaryRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"&\n\x16LogRunHeartbeatRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"4\n\x10GetMetricRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x12\n\nmax_values\x18\x03 \x01(\x05\"K\n\x16\x42\x61tchGetMetricsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\r\n\x05names\x18\x02 \x03(\t\x12\x12\n\nmax_values\x18\x03 \x01(\x05\"O\n\x17\x42\x61tchGetMetricsResponse\x12\x34\n\x07metrics\x18\x01 \x03(\x0b\x32#.continual.rpc.management.v1.Metric\"\x98\x02\n\x12ListMetricsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\x12\x0f\n\x07\x66ilters\x18\x04 \x03(\t\x12\x16\n\x0eignore_filters\x18\x05 \x03(\t\x12.\n\nstart_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x10\n\x08order_by\x18\x08 \x01(\t\x12\x0e\n\x06search\x18\t \x01(\t\x12\x0e\n\x06latest\x18\n \x01(\x08\x12\x12\n\nmax_values\x18\x0b \x01(\x05\"d\n\x13ListMetricsResponse\x12\x34\n\x07metrics\x18\x01 \x03(\x0b\x32#.continual.rpc.management.v1.Metric\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"\x87\x01\n\x13\x43reateMetricRequest\x12\x33\n\x06metric\x18\x01 \x01(\x0b\x32#.continual.rpc.management.v1.Metric\x12\x0e\n\x06parent\x18\x02 \x01(\t\x12\x11\n\tmetric_id\x18\x03 \x01(\t\x12\x18\n\x10update_if_exists\x18\x04 \x01(\x08\"[\n\x10LogMetricRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x37\n\x05value\x18\x02 \x01(\x0b\x32(.continual.rpc.management.v1.MetricValue\"{\n\x13UpdateMetricRequest\x12\x33\n\x06metric\x18\x01 \x01(\x0b\x32#.continual.rpc.management.v1.Metric\x12/\n\x0bupdate_mask\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\"\"\n\x12GetMetadataRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"(\n\x17\x42\x61tchGetMetadataRequest\x12\r\n\x05names\x18\x01 \x03(\t\"S\n\x18\x42\x61tchGetMetadataResponse\x12\x37\n\x08metadata\x18\x01 \x03(\x0b\x32%.continual.rpc.management.v1.Metadata\"\x85\x02\n\x13ListMetadataRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\x12\x0f\n\x07\x66ilters\x18\x04 \x03(\t\x12\x16\n\x0eignore_filters\x18\x05 \x03(\t\x12.\n\nstart_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x10\n\x08order_by\x18\x08 \x01(\t\x12\x0e\n\x06search\x18\t \x01(\t\x12\x0e\n\x06latest\x18\n \x01(\x08\"h\n\x14ListMetadataResponse\x12\x37\n\x08metadata\x18\x01 \x03(\x0b\x32%.continual.rpc.management.v1.Metadata\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"\x8f\x01\n\x15\x43reateMetadataRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x37\n\x08metadata\x18\x02 \x01(\x0b\x32%.continual.rpc.management.v1.Metadata\x12\x13\n\x0bmetadata_id\x18\x03 \x01(\t\x12\x18\n\x10update_if_exists\x18\x04 \x01(\x08\"P\n\x15UpdateMetadataRequest\x12\x37\n\x08metadata\x18\x01 \x01(\x0b\x32%.continual.rpc.management.v1.Metadata\"%\n\x15\x44\x65leteMetadataRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\",\n\x1cGetNotificationPolicyRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"4\n#BatchGetNotificationPoliciesRequest\x12\r\n\x05names\x18\x01 \x03(\t\"v\n$BatchGetNotificationPoliciesResponse\x12N\n\x15notification_policies\x18\x01 \x03(\x0b\x32/.continual.rpc.management.v1.NotificationPolicy\"\x81\x02\n\x1fListNotificationPoliciesRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\x12\x0f\n\x07\x66ilters\x18\x04 \x03(\t\x12\x16\n\x0eignore_filters\x18\x05 \x03(\t\x12.\n\nstart_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x10\n\x08order_by\x18\x08 \x01(\t\x12\x0e\n\x06search\x18\t \x01(\t\"\x8b\x01\n ListNotificationPoliciesResponse\x12N\n\x15notification_policies\x18\x01 \x03(\x0b\x32/.continual.rpc.management.v1.NotificationPolicy\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"\x7f\n\x1f\x43reateNotificationPolicyRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12L\n\x13notification_policy\x18\x02 \x01(\x0b\x32/.continual.rpc.management.v1.NotificationPolicy\"\xa0\x01\n\x1fUpdateNotificationPolicyRequest\x12L\n\x13notification_policy\x18\x01 \x01(\x0b\x32/.continual.rpc.management.v1.NotificationPolicy\x12/\n\x0bupdate_mask\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\"/\n\x1f\x44\x65leteNotificationPolicyRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\",\n\rGetTagRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x0b\n\x03key\x18\x02 \x01(\t\"$\n\x13\x42\x61tchGetTagsRequest\x12\r\n\x05names\x18\x01 \x03(\t\"F\n\x14\x42\x61tchGetTagsResponse\x12.\n\x04tags\x18\x01 \x03(\x0b\x32 .continual.rpc.management.v1.Tag\"\x81\x02\n\x0fListTagsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\x12\x0f\n\x07\x66ilters\x18\x04 \x03(\t\x12\x16\n\x0eignore_filters\x18\x05 \x03(\t\x12.\n\nstart_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x10\n\x08order_by\x18\x08 \x01(\t\x12\x0e\n\x06search\x18\t \x01(\t\x12\x0e\n\x06latest\x18\n \x01(\x08\"[\n\x10ListTagsResponse\x12.\n\x04tags\x18\x01 \x03(\x0b\x32 .continual.rpc.management.v1.Tag\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\">\n\x10\x43reateTagRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x0b\n\x03key\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\t\"r\n\x10UpdateTagRequest\x12-\n\x03tag\x18\x01 \x01(\x0b\x32 .continual.rpc.management.v1.Tag\x12/\n\x0bupdate_mask\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\" \n\x10\x44\x65leteTagRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"%\n\x15GetDataProfileRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"w\n\x18\x43reateDataProfileRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12>\n\x0c\x64\x61ta_profile\x18\x02 \x01(\x0b\x32(.continual.rpc.management.v1.DataProfile\x12\x0b\n\x03run\x18\x03 \x01(\t\"\xab\x01\n\x17ListDataProfilesRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\x12\x0f\n\x07\x66ilters\x18\x04 \x03(\t\x12\x16\n\x0eignore_filters\x18\x05 \x03(\t\x12\x10\n\x08order_by\x18\x06 \x01(\t\x12\x0e\n\x06search\x18\x07 \x01(\t\x12\x0e\n\x06latest\x18\x08 \x01(\x08\"t\n\x18ListDataProfilesResponse\x12?\n\rdata_profiles\x18\x01 \x03(\x0b\x32(.continual.rpc.management.v1.DataProfile\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"}\n\x18UpdateDataProfileRequest\x12>\n\x0c\x64\x61ta_profile\x18\x01 \x01(\x0b\x32(.continual.rpc.management.v1.DataProfile\x12\x14\n\x0cupdate_paths\x18\x02 \x03(\t\x12\x0b\n\x03run\x18\x03 \x01(\t\"\'\n\x17GetResourceCheckRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\xad\x01\n\x19ListResourceChecksRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\x12\x0f\n\x07\x66ilters\x18\x04 \x03(\t\x12\x16\n\x0eignore_filters\x18\x05 \x03(\t\x12\x10\n\x08order_by\x18\x06 \x01(\t\x12\x0e\n\x06search\x18\x07 \x01(\t\x12\x0e\n\x06latest\x18\x08 \x01(\x08\"z\n\x1aListResourceChecksResponse\x12\x43\n\x0fresource_checks\x18\x01 \x03(\x0b\x32*.continual.rpc.management.v1.ResourceCheck\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"p\n\x1a\x43reateResourceCheckRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x42\n\x0eresource_check\x18\x02 \x01(\x0b\x32*.continual.rpc.management.v1.ResourceCheck\"]\n\x14\x43reateDatasetRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12\x15\n\rget_if_exists\x18\x03 \x01(\x08\x12\x0b\n\x03run\x18\x04 \x01(\t\"!\n\x11GetDatasetRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"8\n\x17\x42\x61tchGetDatasetsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\r\n\x05names\x18\x02 \x03(\t\"R\n\x18\x42\x61tchGetDatasetsResponse\x12\x36\n\x08\x64\x61tasets\x18\x01 \x03(\x0b\x32$.continual.rpc.management.v1.Dataset\"\x9b\x02\n\x13ListDatasetsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\x12\x0e\n\x06latest\x18\x05 \x01(\x08\x12\x0f\n\x07\x66ilters\x18\x06 \x03(\t\x12\x16\n\x0eignore_filters\x18\x07 \x03(\t\x12.\n\nstart_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\t \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x14\n\x0c\x61ll_projects\x18\n \x01(\x08\x12\x10\n\x08order_by\x18\x0b \x01(\t\x12\x0e\n\x06search\x18\x0c \x01(\t\"g\n\x14ListDatasetsResponse\x12\x36\n\x08\x64\x61tasets\x18\x01 \x03(\x0b\x32$.continual.rpc.management.v1.Dataset\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"1\n\x14\x44\x65leteDatasetRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0b\n\x03run\x18\x02 \x01(\t\":\n\x1b\x43reateDatasetVersionRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x0b\n\x03run\x18\x02 \x01(\t\"(\n\x18GetDatasetVersionRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"?\n\x1e\x42\x61tchGetDatasetVersionsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\r\n\x05names\x18\x02 \x03(\t\"h\n\x1f\x42\x61tchGetDatasetVersionsResponse\x12\x45\n\x10\x64\x61taset_versions\x18\x01 \x03(\x0b\x32+.continual.rpc.management.v1.DatasetVersion\"\xa2\x02\n\x1aListDatasetVersionsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\x12\x0e\n\x06latest\x18\x04 \x01(\x08\x12\x0f\n\x07\x66ilters\x18\x05 \x03(\t\x12\x16\n\x0eignore_filters\x18\x06 \x03(\t\x12.\n\nstart_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x14\n\x0c\x61ll_projects\x18\t \x01(\x08\x12\x10\n\x08order_by\x18\n \x01(\t\x12\x0e\n\x06search\x18\x0b \x01(\t\"}\n\x1bListDatasetVersionsResponse\x12\x45\n\x10\x64\x61taset_versions\x18\x01 \x03(\x0b\x32+.continual.rpc.management.v1.DatasetVersion\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"\\\n%CreateDatasetVersionAssignmentRequest\x12\x1c\n\x14\x64\x61taset_version_name\x18\x01 \x01(\t\x12\x15\n\rresource_name\x18\x02 \x01(\t\"2\n\"GetDatasetVersionAssignmentRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"5\n%DeleteDatasetVersionAssignmentRequest\x12\x0c\n\x04name\x18\x01 \x01(\t\"\xac\x02\n$ListDatasetVersionAssignmentsRequest\x12\x0e\n\x06parent\x18\x01 \x01(\t\x12\x11\n\tpage_size\x18\x02 \x01(\x05\x12\x12\n\npage_token\x18\x03 \x01(\t\x12\x0e\n\x06latest\x18\x04 \x01(\x08\x12\x0f\n\x07\x66ilters\x18\x05 \x03(\t\x12\x16\n\x0eignore_filters\x18\x06 \x03(\t\x12.\n\nstart_time\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x14\n\x0c\x61ll_projects\x18\t \x01(\x08\x12\x10\n\x08order_by\x18\n \x01(\t\x12\x0e\n\x06search\x18\x0b \x01(\t\"\x8c\x01\n%ListDatasetVersionAssignmentsResponse\x12J\n\x0b\x61ssignments\x18\x01 \x03(\x0b\x32\x35.continual.rpc.management.v1.DatasetVersionAssignment\x12\x17\n\x0fnext_page_token\x18\x02 \x01(\t\"1\n\x17GetServerVersionRequest\x12\x16\n\x0e\x63lient_version\x18\x01 \x01(\t\"L\n\x18GetServerVersionResponse\x12\x16\n\x0eserver_version\x18\x01 \x01(\t\x12\x18\n\x10upgrade_required\x18\x02 \x01(\x08\x32\xda\xc6\x01\n\rManagementAPI\x12\x81\x01\n\x10GetServerVersion\x12\x34.continual.rpc.management.v1.GetServerVersionRequest\x1a\x35.continual.rpc.management.v1.GetServerVersionResponse\"\x00\x12u\n\x07GetUser\x12+.continual.rpc.management.v1.GetUserRequest\x1a!.continual.rpc.management.v1.User\"\x1a\x82\xd3\xe4\x93\x02\x14\x12\x12/v1/{name=users/*}\x12}\n\tListUsers\x12-.continual.rpc.management.v1.ListUsersRequest\x1a..continual.rpc.management.v1.ListUsersResponse\"\x11\x82\xd3\xe4\x93\x02\x0b\x12\t/v1/users\x12\x92\x01\n\rBatchGetUsers\x12\x31.continual.rpc.management.v1.BatchGetUsersRequest\x1a\x32.continual.rpc.management.v1.BatchGetUsersResponse\"\x1a\x82\xd3\xe4\x93\x02\x14\x12\x12/v1/users:batchGet\x12\x86\x01\n\nUpdateUser\x12..continual.rpc.management.v1.UpdateUserRequest\x1a!.continual.rpc.management.v1.User\"%\x82\xd3\xe4\x93\x02\x1f\x32\x17/v1/{user.name=users/*}:\x04user\x12p\n\nDeleteUser\x12..continual.rpc.management.v1.DeleteUserRequest\x1a\x16.google.protobuf.Empty\"\x1a\x82\xd3\xe4\x93\x02\x14*\x12/v1/{name=users/*}\x12Z\n\tGetViewer\x12\x16.google.protobuf.Empty\x1a!.continual.rpc.management.v1.User\"\x12\x82\xd3\xe4\x93\x02\x0c\x12\n/v1/viewer\x12\x83\x01\n\x08Register\x12,.continual.rpc.management.v1.RegisterRequest\x1a-.continual.rpc.management.v1.RegisterResponse\"\x1a\x82\xd3\xe4\x93\x02\x14\"\x12/v1/users:register\x12w\n\x05Login\x12).continual.rpc.management.v1.LoginRequest\x1a*.continual.rpc.management.v1.LoginResponse\"\x17\x82\xd3\xe4\x93\x02\x11\"\x0f/v1/users:login\x12R\n\x06Logout\x12\x16.google.protobuf.Empty\x1a\x16.google.protobuf.Empty\"\x18\x82\xd3\xe4\x93\x02\x12\"\x10/v1/users:logout\x12\xbf\x01\n\x17\x43heckOrganizationAccess\x12;.continual.rpc.management.v1.CheckOrganizationAccessRequest\x1a<.continual.rpc.management.v1.CheckOrganizationAccessResponse\")\x82\xd3\xe4\x93\x02#\"!/v1/users:checkOrganizationAccess\x12j\n\x0cGetPrincipal\x12\x30.continual.rpc.management.v1.GetPrincipalRequest\x1a&.continual.rpc.management.v1.Principal\"\x00\x12?\n\x0b\x43heckViewer\x12\x16.google.protobuf.Empty\x1a\x16.google.protobuf.Empty\"\x00\x12\x61\n\tGetApiKey\x12-.continual.rpc.management.v1.GetApiKeyRequest\x1a#.continual.rpc.management.v1.ApiKey\"\x00\x12~\n\x0f\x42\x61tchGetApiKeys\x12\x33.continual.rpc.management.v1.BatchGetApiKeysRequest\x1a\x34.continual.rpc.management.v1.BatchGetApiKeysResponse\"\x00\x12r\n\x0bListApiKeys\x12/.continual.rpc.management.v1.ListApiKeysRequest\x1a\x30.continual.rpc.management.v1.ListApiKeysResponse\"\x00\x12g\n\x0c\x43reateApiKey\x12\x30.continual.rpc.management.v1.CreateApiKeyRequest\x1a#.continual.rpc.management.v1.ApiKey\"\x00\x12g\n\x0cUpdateApiKey\x12\x30.continual.rpc.management.v1.UpdateApiKeyRequest\x1a#.continual.rpc.management.v1.ApiKey\"\x00\x12Z\n\x0cRevokeApiKey\x12\x30.continual.rpc.management.v1.RevokeApiKeyRequest\x1a\x16.google.protobuf.Empty\"\x00\x12p\n\x10GetApiKeyProject\x12\x34.continual.rpc.management.v1.GetApiKeyProjectRequest\x1a$.continual.rpc.management.v1.Project\"\x00\x12\x93\x01\n\x15SendEmailVerification\x12\x39.continual.rpc.management.v1.SendEmailVerificationRequest\x1a\x16.google.protobuf.Empty\"\'\x82\xd3\xe4\x93\x02!\"\x1f/v1/users:sendEmailVerification\x12\x8f\x01\n\x0bVerifyEmail\x12/.continual.rpc.management.v1.VerifyEmailRequest\x1a\x30.continual.rpc.management.v1.VerifyEmailResponse\"\x1d\x82\xd3\xe4\x93\x02\x17\"\x15/v1/users:verifyEmail\x12\x87\x01\n\x11SendPasswordReset\x12\x35.continual.rpc.management.v1.SendPasswordResetRequest\x1a\x16.google.protobuf.Empty\"#\x82\xd3\xe4\x93\x02\x1d\"\x1b/v1/users:sendPasswordReset\x12\x98\x01\n\rResetPassword\x12\x31.continual.rpc.management.v1.ResetPasswordRequest\x1a\x32.continual.rpc.management.v1.ResetPasswordResponse\" \x82\xd3\xe4\x93\x02\x1a\"\x18/v1/:users:resetPassword\x12\x9c\x01\n\x0e\x43hangePassword\x12\x32.continual.rpc.management.v1.ChangePasswordRequest\x1a\x33.continual.rpc.management.v1.ChangePasswordResponse\"!\x82\xd3\xe4\x93\x02\x1b\"\x19/v1/:users:changePassword\x12\x89\x01\n\nGetSession\x12..continual.rpc.management.v1.GetSessionRequest\x1a$.continual.rpc.management.v1.Session\"%\x82\xd3\xe4\x93\x02\x1f\x12\x1d/v1/{name=users/*/sessions/*}\x12\xac\x01\n\x0cListSessions\x12\x30.continual.rpc.management.v1.ListSessionsRequest\x1a\x31.continual.rpc.management.v1.ListSessionsResponse\"7\x82\xd3\xe4\x93\x02\x1f\x12\x1d/v1/{parent=users/*}/sessions\xca\x45\x08sessions\xd2\x45\x04User\x12\x81\x01\n\rDeleteSession\x12\x31.continual.rpc.management.v1.DeleteSessionRequest\x1a\x16.google.protobuf.Empty\"%\x82\xd3\xe4\x93\x02\x1f*\x1d/v1/{name=users/*/sessions/*}\x12\x89\x01\n\x11\x44\x65leteAllSessions\x12\x35.continual.rpc.management.v1.DeleteAllSessionsRequest\x1a\x16.google.protobuf.Empty\"%\x82\xd3\xe4\x93\x02\x1f*\x1d/v1/{parent=users/*}/sessions\x12\x95\x01\n\x0fGetOrganization\x12\x33.continual.rpc.management.v1.GetOrganizationRequest\x1a).continual.rpc.management.v1.Organization\"\"\x82\xd3\xe4\x93\x02\x1c\x12\x1a/v1/{name=organizations/*}\x12\xb2\x01\n\x15\x42\x61tchGetOrganizations\x12\x39.continual.rpc.management.v1.BatchGetOrganizationsRequest\x1a:.continual.rpc.management.v1.BatchGetOrganizationsResponse\"\"\x82\xd3\xe4\x93\x02\x1c\x12\x1a/v1/organizations:batchGet\x12\x9d\x01\n\x11ListOrganizations\x12\x35.continual.rpc.management.v1.ListOrganizationsRequest\x1a\x36.continual.rpc.management.v1.ListOrganizationsResponse\"\x19\x82\xd3\xe4\x93\x02\x13\x12\x11/v1/organizations\x12\xa0\x01\n\x12\x43reateOrganization\x12\x36.continual.rpc.management.v1.CreateOrganizationRequest\x1a).continual.rpc.management.v1.Organization\"\'\x82\xd3\xe4\x93\x02!\"\x11/v1/organizations:\x0corganization\x12\xb6\x01\n\x12UpdateOrganization\x12\x36.continual.rpc.management.v1.UpdateOrganizationRequest\x1a).continual.rpc.management.v1.Organization\"=\x82\xd3\xe4\x93\x02\x37\x32\'/v1/{organization.name=organizations/*}:\x0corganization\x12\x88\x01\n\x12\x44\x65leteOrganization\x12\x36.continual.rpc.management.v1.DeleteOrganizationRequest\x1a\x16.google.protobuf.Empty\"\"\x82\xd3\xe4\x93\x02\x1c*\x1a/v1/{name=organizations/*}\x12\xcf\x01\n\rGetOnboarding\x12\x31.continual.rpc.management.v1.GetOnboardingRequest\x1a\'.continual.rpc.management.v1.Onboarding\"b\x82\xd3\xe4\x93\x02\'\x12%/v1/{name=organizations/*/onboarding}\xd2\x45\x0cOrganization\xca\x45\nonboarding\xda\x45\x16name:name:$/onboarding\x12\xb7\x01\n\x10UpdateOnboarding\x12\x34.continual.rpc.management.v1.UpdateOnboardingRequest\x1a\'.continual.rpc.management.v1.Onboarding\"D\x82\xd3\xe4\x93\x02>20/v1/{onboarding.name=organizations/*/onboarding}:\nonboarding\x12}\n\tListRoles\x12-.continual.rpc.management.v1.ListRolesRequest\x1a..continual.rpc.management.v1.ListRolesResponse\"\x11\x82\xd3\xe4\x93\x02\x0b\x12\t/v1/roles\x12u\n\x07GetRole\x12+.continual.rpc.management.v1.GetRoleRequest\x1a!.continual.rpc.management.v1.Role\"\x1a\x82\xd3\xe4\x93\x02\x14\x12\x12/v1/{name=roles/*}\x12\x92\x01\n\rBatchGetRoles\x12\x31.continual.rpc.management.v1.BatchGetRolesRequest\x1a\x32.continual.rpc.management.v1.BatchGetRolesResponse\"\x1a\x82\xd3\xe4\x93\x02\x14\x12\x12/v1/roles:batchGet\x12\xae\x01\n\x12\x43reateAccessPolicy\x12\x36.continual.rpc.management.v1.CreateAccessPolicyRequest\x1a).continual.rpc.management.v1.AccessPolicy\"5\x82\xd3\xe4\x93\x02/\"\x1e/v1/{parent=**}/accessPolicies:\raccess_policy\x12\x99\x01\n\x0fGetAccessPolicy\x12\x33.continual.rpc.management.v1.GetAccessPolicyRequest\x1a).continual.rpc.management.v1.AccessPolicy\"&\x82\xd3\xe4\x93\x02 \x12\x1e/v1/{name=**/accessPolicies/*}\x12\x8a\x02\n\x12ListAccessPolicies\x12\x36.continual.rpc.management.v1.ListAccessPoliciesRequest\x1a\x37.continual.rpc.management.v1.ListAccessPoliciesResponse\"\x82\x01\x82\xd3\xe4\x93\x02 \x12\x1e/v1/{parent=**}/accessPolicies\xd2\x45:Organization,Project,Dataset,Model,ModelVersion,Experiment\xca\x45\x0e\x61\x63\x63\x65ssPolicies\xda\x45\x0bname:parent\x12r\n\x18\x42ulkDeleteAccessPolicies\x12<.continual.rpc.management.v1.BulkDeleteAccessPoliciesRequest\x1a\x16.google.protobuf.Empty\"\x00\x12\x8c\x01\n\x12\x44\x65leteAccessPolicy\x12\x36.continual.rpc.management.v1.DeleteAccessPolicyRequest\x1a\x16.google.protobuf.Empty\"&\x82\xd3\xe4\x93\x02 *\x1e/v1/{name=**/accessPolicies/*}\x12\xbc\x01\n\x12UpdateAccessPolicy\x12\x36.continual.rpc.management.v1.UpdateAccessPolicyRequest\x1a).continual.rpc.management.v1.AccessPolicy\"C\x82\xd3\xe4\x93\x02=2,/v1/{access_policy.name=**/accessPolicies/*}:\raccess_policy\x12\x86\x02\n\x0fTestPermissions\x12\x33.continual.rpc.management.v1.TestPermissionsRequest\x1a\x34.continual.rpc.management.v1.TestPermissionsResponse\"\x87\x01\x82\xd3\xe4\x93\x02&\"!/v1/{resource=**}:testPermissions:\x01*\xd2\x45:Organization,Project,Dataset,Model,ModelVersion,Experiment\xca\x45\x0bpermissions\xda\x45\rname:resource\x12\x86\x01\n\x0c\x43reateInvite\x12\x30.continual.rpc.management.v1.CreateInviteRequest\x1a#.continual.rpc.management.v1.Invite\"\x1f\x82\xd3\xe4\x93\x02\x19\"\x17/v1/{parent=**}/invites\x12\x80\x01\n\tGetInvite\x12-.continual.rpc.management.v1.GetInviteRequest\x1a#.continual.rpc.management.v1.Invite\"\x1f\x82\xd3\xe4\x93\x02\x19\x12\x17/v1/{name=**/invites/*}\x12\x91\x01\n\x0bListInvites\x12/.continual.rpc.management.v1.ListInvitesRequest\x1a\x30.continual.rpc.management.v1.ListInvitesResponse\"\x1f\x82\xd3\xe4\x93\x02\x19\x12\x17/v1/{parent=**}/invites\x12\x95\x01\n\x0cUpdateInvite\x12\x30.continual.rpc.management.v1.UpdateInviteRequest\x1a#.continual.rpc.management.v1.Invite\".\x82\xd3\xe4\x93\x02(2\x1e/v1/{invite.name=**/invites/*}:\x06invite\x12\x8d\x01\n\x0cResendInvite\x12\x30.continual.rpc.management.v1.ResendInviteRequest\x1a#.continual.rpc.management.v1.Invite\"&\x82\xd3\xe4\x93\x02 \"\x1e/v1/{name=**/invites/*}:resend\x12y\n\x0c\x44\x65leteInvite\x12\x30.continual.rpc.management.v1.DeleteInviteRequest\x1a\x16.google.protobuf.Empty\"\x1f\x82\xd3\xe4\x93\x02\x19*\x17/v1/{name=**/invites/*}\x12v\n\x0c\x41\x63\x63\x65ptInvite\x12\x30.continual.rpc.management.v1.AcceptInviteRequest\x1a\x16.google.protobuf.Empty\"\x1c\x82\xd3\xe4\x93\x02\x16\"\x14/v1/confirm/{code=*}\x12\x81\x01\n\nGetProject\x12..continual.rpc.management.v1.GetProjectRequest\x1a$.continual.rpc.management.v1.Project\"\x1d\x82\xd3\xe4\x93\x02\x17\x12\x15/v1/{name=projects/*}\x12\x9e\x01\n\x10\x42\x61tchGetProjects\x12\x34.continual.rpc.management.v1.BatchGetProjectsRequest\x1a\x35.continual.rpc.management.v1.BatchGetProjectsResponse\"\x1d\x82\xd3\xe4\x93\x02\x17\x12\x15/v1/projects:batchGet\x12\xda\x01\n\x0cListProjects\x12\x30.continual.rpc.management.v1.ListProjectsRequest\x1a\x31.continual.rpc.management.v1.ListProjectsResponse\"e\x82\xd3\xe4\x93\x02\x37\x12\x0c/v1/projectsZ\'\x12%/v1/{parent=organizations/*}/projects\xd2\x45\x0cOrganization\xca\x45\x08projects\xda\x45\x0bname:parent\x12\x87\x01\n\rCreateProject\x12\x31.continual.rpc.management.v1.CreateProjectRequest\x1a$.continual.rpc.management.v1.Project\"\x1d\x82\xd3\xe4\x93\x02\x17\"\x0c/v1/projects:\x07project\x12\x98\x01\n\rUpdateProject\x12\x31.continual.rpc.management.v1.UpdateProjectRequest\x1a$.continual.rpc.management.v1.Project\".\x82\xd3\xe4\x93\x02(2\x1d/v1/{project.name=projects/*}:\x07project\x12y\n\rDeleteProject\x12\x31.continual.rpc.management.v1.DeleteProjectRequest\x1a\x16.google.protobuf.Empty\"\x1d\x82\xd3\xe4\x93\x02\x17*\x15/v1/{name=projects/*}\x12p\n\x0f\x43reatePromotion\x12\x33.continual.rpc.management.v1.CreatePromotionRequest\x1a&.continual.rpc.management.v1.Promotion\"\x00\x12\x9c\x02\n\x0eListPromotions\x12\x32.continual.rpc.management.v1.ListPromotionsRequest\x1a\x33.continual.rpc.management.v1.ListPromotionsResponse\"\xa0\x01\x82\xd3\xe4\x93\x02g\x12+/v1/{parent=projects/*/models/*}/promotionsZ8\x12\x36/v1/{parent=projects/*/models/*/versions/*}/promotions\xca\x45\npromotions\xd2\x45\x12ModelVersion,Model\xe0\x45\x00\xda\x45\x0bname:parent\x12\x9d\x01\n\x0cGetPromotion\x12\x30.continual.rpc.management.v1.GetPromotionRequest\x1a&.continual.rpc.management.v1.Promotion\"3\x82\xd3\xe4\x93\x02-\x12+/v1/{name=projects/*/models/*/promotions/*}\x12\xc3\x01\n\x12\x42\x61tchGetPromotions\x12\x36.continual.rpc.management.v1.BatchGetPromotionsRequest\x1a\x37.continual.rpc.management.v1.BatchGetPromotionsResponse\"<\x82\xd3\xe4\x93\x02\x36\x12\x34/v1/{parent=projects/*/models/*}/promotions:batchGet\x12\x95\x01\n\nListEvents\x12..continual.rpc.management.v1.ListEventsRequest\x1a/.continual.rpc.management.v1.ListEventsResponse\"&\x82\xd3\xe4\x93\x02 \x12\x1e/v1/{parent=projects/*}:events\x12\x84\x01\n\x08GetEvent\x12,.continual.rpc.management.v1.GetEventRequest\x1a\".continual.rpc.management.v1.Event\"&\x82\xd3\xe4\x93\x02 \x12\x1e/v1/{name=projects/*/events/*}\x12\xaa\x01\n\x0e\x42\x61tchGetEvents\x12\x32.continual.rpc.management.v1.BatchGetEventsRequest\x1a\x33.continual.rpc.management.v1.BatchGetEventsResponse\"/\x82\xd3\xe4\x93\x02)\x12\'/v1/{parent=projects/*}/events:batchGet\x12X\n\x0b\x43reateEvent\x12/.continual.rpc.management.v1.CreateEventRequest\x1a\x16.google.protobuf.Empty\"\x00\x12\xd9\x01\n\x0fGetSubscription\x12\x33.continual.rpc.management.v1.GetSubscriptionRequest\x1a).continual.rpc.management.v1.Subscription\"f\x82\xd3\xe4\x93\x02\'\x12%/v1/{name=organizations/subscription}\xd2\x45\x0cOrganization\xca\x45\x0csubscription\xda\x45\x18name:name:$/subscription\x12\xe7\x01\n\x11GetBillingAccount\x12\x35.continual.rpc.management.v1.GetBillingAccountRequest\x1a+.continual.rpc.management.v1.BillingAccount\"n\x82\xd3\xe4\x93\x02+\x12)/v1/{name=organizations/*/billingAccount}\xd2\x45\x0cOrganization\xca\x45\x0e\x62illingAccount\xda\x45\x1aname:name:$/billingAccount\x12\xb8\x01\n\x08GetUsage\x12,.continual.rpc.management.v1.GetUsageRequest\x1a-.continual.rpc.management.v1.GetUsageResponse\"O\x82\xd3\xe4\x93\x02$\"\"/v1/{parent=organizations/*}:usage\xd2\x45\x0cOrganization\xca\x45\x05usage\xda\x45\x0bname:parent\x12\xca\x01\n\x0cListInvoices\x12\x30.continual.rpc.management.v1.ListInvoicesRequest\x1a\x31.continual.rpc.management.v1.ListInvoicesResponse\"U\x82\xd3\xe4\x93\x02\'\x12%/v1/{parent=organizations/*}/invoices\xd2\x45\x0cOrganization\xca\x45\x08invoices\xda\x45\x0bname:parent\x12\xc4\x01\n\x14UpdateBillingAccount\x12\x38.continual.rpc.management.v1.UpdateBillingAccountRequest\x1a+.continual.rpc.management.v1.BillingAccount\"E\x82\xd3\xe4\x93\x02?2,/v1/{billing_account.name=**/billingAccount}:\x0f\x62illing_account\x12\xb6\x01\n\x12UpdateSubscription\x12\x36.continual.rpc.management.v1.UpdateSubscriptionRequest\x1a).continual.rpc.management.v1.Subscription\"=\x82\xd3\xe4\x93\x02\x37\x32\'/v1/{subscription.name=**/subscription}:\x0csubscription\x12\xc0\x01\n\nListPrices\x12..continual.rpc.management.v1.ListPricesRequest\x1a/.continual.rpc.management.v1.ListPricesResponse\"Q\x82\xd3\xe4\x93\x02%\x12#/v1/{parent=organizations/*}/prices\xd2\x45\x0cOrganization\xca\x45\x06prices\xda\x45\x0bname:parent\x12\x89\x01\n\x08GetPrice\x12,.continual.rpc.management.v1.GetPriceRequest\x1a\".continual.rpc.management.v1.Price\"+\x82\xd3\xe4\x93\x02%\x12#/v1/{name=organizations/*/prices/*}\x12\xa9\x01\n\x10GetPaymentMethod\x12\x34.continual.rpc.management.v1.GetPaymentMethodRequest\x1a*.continual.rpc.management.v1.PaymentMethod\"3\x82\xd3\xe4\x93\x02-\x12+/v1/{name=organizations/*/paymentMethods/*}\x12\xe5\x01\n\x12GetUpcomingInvoice\x12\x36.continual.rpc.management.v1.GetUpcomingInvoiceRequest\x1a$.continual.rpc.management.v1.Invoice\"q\x82\xd3\xe4\x93\x02,\x12*/v1/{name=organizations/*/upcomingInvoice}\xd2\x45\x0cOrganization\xca\x45\x0fupcomingInvoice\xda\x45\x1bname:name:$/upcomingInvoice\x12\x8a\x01\n\x0b\x43reateModel\x12/.continual.rpc.management.v1.CreateModelRequest\x1a\".continual.rpc.management.v1.Model\"&\x82\xd3\xe4\x93\x02 \x12\x1e/v1/{name=projects/*/models/*}\x12\x84\x01\n\x08GetModel\x12,.continual.rpc.management.v1.GetModelRequest\x1a\".continual.rpc.management.v1.Model\"&\x82\xd3\xe4\x93\x02 \x12\x1e/v1/{name=projects/*/models/*}\x12\xaa\x01\n\x0e\x42\x61tchGetModels\x12\x32.continual.rpc.management.v1.BatchGetModelsRequest\x1a\x33.continual.rpc.management.v1.BatchGetModelsResponse\"/\x82\xd3\xe4\x93\x02)\x12\'/v1/{parent=projects/*}/models:batchGet\x12\xce\x01\n\nListModels\x12..continual.rpc.management.v1.ListModelsRequest\x1a/.continual.rpc.management.v1.ListModelsResponse\"_\x82\xd3\xe4\x93\x02 \x12\x1e/v1/{parent=projects/*}/models\xd2\x45\x13Project,Environment\xca\x45\x06models\xda\x45\x17name:parent;name:parent\x12~\n\x0b\x44\x65leteModel\x12/.continual.rpc.management.v1.DeleteModelRequest\x1a\x16.google.protobuf.Empty\"&\x82\xd3\xe4\x93\x02 *\x1e/v1/{name=projects/*/models/*}\x12y\n\x12\x43reateModelVersion\x12\x36.continual.rpc.management.v1.CreateModelVersionRequest\x1a).continual.rpc.management.v1.ModelVersion\"\x00\x12\xa4\x01\n\x0fGetModelVersion\x12\x33.continual.rpc.management.v1.GetModelVersionRequest\x1a).continual.rpc.management.v1.ModelVersion\"1\x82\xd3\xe4\x93\x02+\x12)/v1/{name=projects/*/models/*/versions/*}\x12\xca\x01\n\x15\x42\x61tchGetModelVersions\x12\x39.continual.rpc.management.v1.BatchGetModelVersionsRequest\x1a:.continual.rpc.management.v1.BatchGetModelVersionsResponse\":\x82\xd3\xe4\x93\x02\x34\x12\x32/v1/{parent=projects/*/models/*}/versions:batchGet\x12\xa1\x02\n\x11ListModelVersions\x12\x35.continual.rpc.management.v1.ListModelVersionsRequest\x1a\x36.continual.rpc.management.v1.ListModelVersionsResponse\"\x9c\x01\x82\xd3\xe4\x93\x02+\x12)/v1/{parent=projects/*/models/*}/versions\xca\x45\rmodelVersions\xd2\x45\x19Model,Project,Environment\xe0\x45\x00\xda\x45\x39name:parent;name:parent:$/models/-;name:parent:$/models/-\x12\x82\x01\n\x15\x43reateBatchPrediction\x12\x39.continual.rpc.management.v1.CreateBatchPredictionRequest\x1a,.continual.rpc.management.v1.BatchPrediction\"\x00\x12\xb2\x01\n\x12GetBatchPrediction\x12\x33.continual.rpc.management.v1.BatchPredictionRequest\x1a,.continual.rpc.management.v1.BatchPrediction\"9\x82\xd3\xe4\x93\x02\x33\x12\x31/v1/{name=projects/*/models/*/batchPredictions/*}\x12\xdb\x01\n\x18\x42\x61tchGetBatchPredictions\x12<.continual.rpc.management.v1.BatchGetBatchPredictionsRequest\x1a=.continual.rpc.management.v1.BatchGetBatchPredictionsResponse\"B\x82\xd3\xe4\x93\x02<\x12:/v1/{parent=projects/*/models/*}/batchPredictions:batchGet\x12\x9f\x02\n\x14ListBatchPredictions\x12\x38.continual.rpc.management.v1.ListBatchPredictionsRequest\x1a\x39.continual.rpc.management.v1.ListBatchPredictionsResponse\"\x91\x01\x82\xd3\xe4\x93\x02\x33\x12\x31/v1/{parent=projects/*/models/*}/batchPredictions\xca\x45\x10\x62\x61tchPredictions\xd2\x45\x19Project,Model,Environment\xe0\x45\x00\xda\x45#name:parent;name:parent;name:parent\x12s\n\x10\x43reateExperiment\x12\x34.continual.rpc.management.v1.CreateExperimentRequest\x1a\'.continual.rpc.management.v1.Experiment\"\x00\x12\xac\x01\n\rGetExperiment\x12\x31.continual.rpc.management.v1.GetExperimentRequest\x1a\'.continual.rpc.management.v1.Experiment\"?\x82\xd3\xe4\x93\x02\x39\x12\x37/v1/{name=projects/*/models/*/versions/*/experiments/*}\x12\xd2\x01\n\x13\x42\x61tchGetExperiments\x12\x37.continual.rpc.management.v1.BatchGetExperimentsRequest\x1a\x38.continual.rpc.management.v1.BatchGetExperimentsResponse\"H\x82\xd3\xe4\x93\x02\x42\x12@/v1/{parent=projects/*/models/*/versions/*}/experiments:batchGet\x12\xeb\x01\n\x0fListExperiments\x12\x33.continual.rpc.management.v1.ListExperimentsRequest\x1a\x34.continual.rpc.management.v1.ListExperimentsResponse\"m\x82\xd3\xe4\x93\x02\x39\x12\x37/v1/{parent=projects/*/models/*/versions/*}/experiments\xca\x45\x0b\x65xperiments\xd2\x45\x0cModelVersion\xe0\x45\x00\xda\x45\x0bname:parent\x12p\n\x0eGetEnvironment\x12\x32.continual.rpc.management.v1.GetEnvironmentRequest\x1a(.continual.rpc.management.v1.Environment\"\x00\x12\x8d\x01\n\x14\x42\x61tchGetEnvironments\x12\x38.continual.rpc.management.v1.BatchGetEnvironmentsRequest\x1a\x39.continual.rpc.management.v1.BatchGetEnvironmentsResponse\"\x00\x12\xa8\x01\n\x10ListEnvironments\x12\x34.continual.rpc.management.v1.ListEnvironmentsRequest\x1a\x35.continual.rpc.management.v1.ListEnvironmentsResponse\"\'\xd2\x45\x07Project\xca\x45\x0c\x65nvironments\xda\x45\x0bname:parent\x12v\n\x11\x43reateEnvironment\x12\x35.continual.rpc.management.v1.CreateEnvironmentRequest\x1a(.continual.rpc.management.v1.Environment\"\x00\x12v\n\x11UpdateEnvironment\x12\x35.continual.rpc.management.v1.UpdateEnvironmentRequest\x1a(.continual.rpc.management.v1.Environment\"\x00\x12\x64\n\x11\x44\x65leteEnvironment\x12\x35.continual.rpc.management.v1.DeleteEnvironmentRequest\x1a\x16.google.protobuf.Empty\"\x00\x12\x9a\x01\n\x10GetSystemMessage\x12\x34.continual.rpc.management.v1.GetSystemMessageRequest\x1a\x35.continual.rpc.management.v1.GetSystemMessageResponse\"\x19\x82\xd3\xe4\x93\x02\x13\"\x11/v1/systemMessage\x12{\n\x0eGetFeatureFlag\x12\x32.continual.rpc.management.v1.GetFeatureFlagRequest\x1a\x33.continual.rpc.management.v1.GetFeatureFlagResponse\"\x00\x12\xd0\x01\n\x1bGetOrganizationFeatureFlags\x12?.continual.rpc.management.v1.GetOrganizationFeatureFlagsRequest\x1a@.continual.rpc.management.v1.GetOrganizationFeatureFlagsResponse\".\xd2\x45\x0cOrganization\xca\x45\x0c\x66\x65\x61tureFlags\xda\x45\rname:resource\x12k\n\x0e\x43reateArtifact\x12\x32.continual.rpc.management.v1.CreateArtifactRequest\x1a%.continual.rpc.management.v1.Artifact\x12\x65\n\x0bGetArtifact\x12/.continual.rpc.management.v1.GetArtifactRequest\x1a%.continual.rpc.management.v1.Artifact\x12\x82\x01\n\x11\x42\x61tchGetArtifacts\x12\x35.continual.rpc.management.v1.BatchGetArtifactsRequest\x1a\x36.continual.rpc.management.v1.BatchGetArtifactsResponse\x12\\\n\x0e\x44\x65leteArtifact\x12\x32.continual.rpc.management.v1.DeleteArtifactRequest\x1a\x16.google.protobuf.Empty\x12\x9a\x01\n\x19GenerateArtifactUploadURL\x12=.continual.rpc.management.v1.GenerateArtifactUploadURLRequest\x1a>.continual.rpc.management.v1.GenerateArtifactUploadURLResponse\x12\xc0\x01\n\rListArtifacts\x12\x31.continual.rpc.management.v1.ListArtifactsRequest\x1a\x32.continual.rpc.management.v1.ListArtifactsResponse\"H\xca\x45\tartifacts\xd2\x45\x1cModelVersion,BatchPrediction\xe0\x45\x00\xda\x45\x17name:parent;name:parent\x12\xb1\x01\n\x1cListUserIntegrationProviders\x12@.continual.rpc.management.v1.ListUserIntegrationProvidersRequest\x1a\x41.continual.rpc.management.v1.ListUserIntegrationProvidersResponse\"\x0c\xca\x45\tproviders\x12\xa3\x01\n\x14ListUserIntegrations\x12\x38.continual.rpc.management.v1.ListUserIntegrationsRequest\x1a\x39.continual.rpc.management.v1.ListUserIntegrationsResponse\"\x16\xca\x45\x0cintegrations\xd2\x45\x04User\x12t\n\x19\x44isconnectUserIntegration\x12=.continual.rpc.management.v1.DisconnectUserIntegrationRequest\x1a\x16.google.protobuf.Empty\"\x00\x12\x9a\x01\n\x19GenerateSsoAdminPortalURL\x12=.continual.rpc.management.v1.GenerateSsoAdminPortalURLRequest\x1a>.continual.rpc.management.v1.GenerateSsoAdminPortalURLResponse\x12\xa0\x01\n\x1bGenerateDSyncAdminPortalURL\x12?.continual.rpc.management.v1.GenerateDSyncAdminPortalURLRequest\x1a@.continual.rpc.management.v1.GenerateDSyncAdminPortalURLResponse\x12v\n\rExchangeToken\x12\x31.continual.rpc.management.v1.ExchangeTokenRequest\x1a\x32.continual.rpc.management.v1.ExchangeTokenResponse\x12\xa9\x01\n\x19ListGitProviderNamespaces\x12=.continual.rpc.management.v1.ListGitProviderNamespacesRequest\x1a>.continual.rpc.management.v1.ListGitProviderNamespacesResponse\"\r\xca\x45\nnamespaces\x12\xb1\x01\n\x1bListGitProviderRepositories\x12?.continual.rpc.management.v1.ListGitProviderRepositoriesRequest\x1a@.continual.rpc.management.v1.ListGitProviderRepositoriesResponse\"\x0f\xca\x45\x0crepositories\x12\xa8\x01\n\x1d\x43onnectProjectToGitRepository\x12\x41.continual.rpc.management.v1.ConnectProjectToGitRepositoryRequest\x1a\x42.continual.rpc.management.v1.ConnectProjectToGitRepositoryResponse\"\x00\x12\x8e\x01\n\x18GetProjectGitIntegration\x12<.continual.rpc.management.v1.GetProjectGitIntegrationRequest\x1a\x32.continual.rpc.management.v1.ProjectGitIntegration\"\x00\x12r\n\x18SeedProjectGitRepository\x12<.continual.rpc.management.v1.SeedProjectGitRepositoryRequest\x1a\x16.google.protobuf.Empty\"\x00\x12X\n\x06GetRun\x12*.continual.rpc.management.v1.GetRunRequest\x1a .continual.rpc.management.v1.Run\"\x00\x12u\n\x0c\x42\x61tchGetRuns\x12\x30.continual.rpc.management.v1.BatchGetRunsRequest\x1a\x31.continual.rpc.management.v1.BatchGetRunsResponse\"\x00\x12\x8f\x01\n\x08ListRuns\x12,.continual.rpc.management.v1.ListRunsRequest\x1a-.continual.rpc.management.v1.ListRunsResponse\"&\xca\x45\x04runs\xd2\x45\x0b\x45nvironment\xe0\x45\x00\xda\x45\x0bname:parent\x12^\n\tCreateRun\x12-.continual.rpc.management.v1.CreateRunRequest\x1a .continual.rpc.management.v1.Run\"\x00\x12^\n\tUpdateRun\x12-.continual.rpc.management.v1.UpdateRunRequest\x1a .continual.rpc.management.v1.Run\"\x00\x12T\n\tDeleteRun\x12-.continual.rpc.management.v1.DeleteRunRequest\x1a\x16.google.protobuf.Empty\"\x00\x12\x89\x01\n\rGetRunSummary\x12\x31.continual.rpc.management.v1.GetRunSummaryRequest\x1a\'.continual.rpc.management.v1.RunSummary\"\x1c\xd2\x45\x03Run\xca\x45\x07summary\xda\x45\tname:name\x12\x9e\x01\n\x12GetRunCheckSummary\x12\x36.continual.rpc.management.v1.GetRunCheckSummaryRequest\x1a,.continual.rpc.management.v1.RunCheckSummary\"\"\xd2\x45\x03Run\xca\x45\rresultSummary\xda\x45\tname:name\x12\x64\n\x0fLogRunHeartbeat\x12\x33.continual.rpc.management.v1.LogRunHeartbeatRequest\x1a\x1a.google.protobuf.Timestamp\"\x00\x12\x61\n\tGetMetric\x12-.continual.rpc.management.v1.GetMetricRequest\x1a#.continual.rpc.management.v1.Metric\"\x00\x12~\n\x0f\x42\x61tchGetMetrics\x12\x33.continual.rpc.management.v1.BatchGetMetricsRequest\x1a\x34.continual.rpc.management.v1.BatchGetMetricsResponse\"\x00\x12\xa4\x01\n\x0bListMetrics\x12/.continual.rpc.management.v1.ListMetricsRequest\x1a\x30.continual.rpc.management.v1.ListMetricsResponse\"2\xd2\x45\x17ModelVersion,Experiment\xca\x45\x07metrics\xda\x45\x0bname:parent\x12g\n\x0c\x43reateMetric\x12\x30.continual.rpc.management.v1.CreateMetricRequest\x1a#.continual.rpc.management.v1.Metric\"\x00\x12T\n\tLogMetric\x12-.continual.rpc.management.v1.LogMetricRequest\x1a\x16.google.protobuf.Empty\"\x00\x12g\n\x0cUpdateMetric\x12\x30.continual.rpc.management.v1.UpdateMetricRequest\x1a#.continual.rpc.management.v1.Metric\"\x00\x12g\n\x0bGetMetadata\x12/.continual.rpc.management.v1.GetMetadataRequest\x1a%.continual.rpc.management.v1.Metadata\"\x00\x12\x81\x01\n\x10\x42\x61tchGetMetadata\x12\x34.continual.rpc.management.v1.BatchGetMetadataRequest\x1a\x35.continual.rpc.management.v1.BatchGetMetadataResponse\"\x00\x12u\n\x0cListMetadata\x12\x30.continual.rpc.management.v1.ListMetadataRequest\x1a\x31.continual.rpc.management.v1.ListMetadataResponse\"\x00\x12m\n\x0e\x43reateMetadata\x12\x32.continual.rpc.management.v1.CreateMetadataRequest\x1a%.continual.rpc.management.v1.Metadata\"\x00\x12m\n\x0eUpdateMetadata\x12\x32.continual.rpc.management.v1.UpdateMetadataRequest\x1a%.continual.rpc.management.v1.Metadata\"\x00\x12^\n\x0e\x44\x65leteMetadata\x12\x32.continual.rpc.management.v1.DeleteMetadataRequest\x1a\x16.google.protobuf.Empty\"\x00\x12\x85\x01\n\x15GetNotificationPolicy\x12\x39.continual.rpc.management.v1.GetNotificationPolicyRequest\x1a/.continual.rpc.management.v1.NotificationPolicy\"\x00\x12\xa5\x01\n\x1c\x42\x61tchGetNotificationPolicies\x12@.continual.rpc.management.v1.BatchGetNotificationPoliciesRequest\x1a\x41.continual.rpc.management.v1.BatchGetNotificationPoliciesResponse\"\x00\x12\x99\x01\n\x18ListNotificationPolicies\x12<.continual.rpc.management.v1.ListNotificationPoliciesRequest\x1a=.continual.rpc.management.v1.ListNotificationPoliciesResponse\"\x00\x12\x8b\x01\n\x18\x43reateNotificationPolicy\x12<.continual.rpc.management.v1.CreateNotificationPolicyRequest\x1a/.continual.rpc.management.v1.NotificationPolicy\"\x00\x12\x8b\x01\n\x18UpdateNotificationPolicy\x12<.continual.rpc.management.v1.UpdateNotificationPolicyRequest\x1a/.continual.rpc.management.v1.NotificationPolicy\"\x00\x12r\n\x18\x44\x65leteNotificationPolicy\x12<.continual.rpc.management.v1.DeleteNotificationPolicyRequest\x1a\x16.google.protobuf.Empty\"\x00\x12X\n\x06GetTag\x12*.continual.rpc.management.v1.GetTagRequest\x1a .continual.rpc.management.v1.Tag\"\x00\x12u\n\x0c\x42\x61tchGetTags\x12\x30.continual.rpc.management.v1.BatchGetTagsRequest\x1a\x31.continual.rpc.management.v1.BatchGetTagsResponse\"\x00\x12i\n\x08ListTags\x12,.continual.rpc.management.v1.ListTagsRequest\x1a-.continual.rpc.management.v1.ListTagsResponse\"\x00\x12^\n\tCreateTag\x12-.continual.rpc.management.v1.CreateTagRequest\x1a .continual.rpc.management.v1.Tag\"\x00\x12^\n\tUpdateTag\x12-.continual.rpc.management.v1.UpdateTagRequest\x1a .continual.rpc.management.v1.Tag\"\x00\x12T\n\tDeleteTag\x12-.continual.rpc.management.v1.DeleteTagRequest\x1a\x16.google.protobuf.Empty\"\x00\x12v\n\x10GetResourceCheck\x12\x34.continual.rpc.management.v1.GetResourceCheckRequest\x1a*.continual.rpc.management.v1.ResourceCheck\"\x00\x12\x87\x01\n\x12ListResourceChecks\x12\x36.continual.rpc.management.v1.ListResourceChecksRequest\x1a\x37.continual.rpc.management.v1.ListResourceChecksResponse\"\x00\x12|\n\x13\x43reateResourceCheck\x12\x37.continual.rpc.management.v1.CreateResourceCheckRequest\x1a*.continual.rpc.management.v1.ResourceCheck\"\x00\x12p\n\x0eGetDataProfile\x12\x32.continual.rpc.management.v1.GetDataProfileRequest\x1a(.continual.rpc.management.v1.DataProfile\"\x00\x12\x81\x01\n\x10ListDataProfiles\x12\x34.continual.rpc.management.v1.ListDataProfilesRequest\x1a\x35.continual.rpc.management.v1.ListDataProfilesResponse\"\x00\x12v\n\x11\x43reateDataProfile\x12\x35.continual.rpc.management.v1.CreateDataProfileRequest\x1a(.continual.rpc.management.v1.DataProfile\"\x00\x12v\n\x11UpdateDataProfile\x12\x35.continual.rpc.management.v1.UpdateDataProfileRequest\x1a(.continual.rpc.management.v1.DataProfile\"\x00\x12j\n\rCreateDataset\x12\x31.continual.rpc.management.v1.CreateDatasetRequest\x1a$.continual.rpc.management.v1.Dataset\"\x00\x12\x64\n\nGetDataset\x12..continual.rpc.management.v1.GetDatasetRequest\x1a$.continual.rpc.management.v1.Dataset\"\x00\x12\x81\x01\n\x10\x42\x61tchGetDatasets\x12\x34.continual.rpc.management.v1.BatchGetDatasetsRequest\x1a\x35.continual.rpc.management.v1.BatchGetDatasetsResponse\"\x00\x12\xb0\x01\n\x0cListDatasets\x12\x30.continual.rpc.management.v1.ListDatasetsRequest\x1a\x31.continual.rpc.management.v1.ListDatasetsResponse\";\xd2\x45\x13Project,Environment\xca\x45\x08\x64\x61tasets\xda\x45\x17name:parent;name:parent\x12\\\n\rDeleteDataset\x12\x31.continual.rpc.management.v1.DeleteDatasetRequest\x1a\x16.google.protobuf.Empty\"\x00\x12\x7f\n\x14\x43reateDatasetVersion\x12\x38.continual.rpc.management.v1.CreateDatasetVersionRequest\x1a+.continual.rpc.management.v1.DatasetVersion\"\x00\x12y\n\x11GetDatasetVersion\x12\x35.continual.rpc.management.v1.GetDatasetVersionRequest\x1a+.continual.rpc.management.v1.DatasetVersion\"\x00\x12\x96\x01\n\x17\x42\x61tchGetDatasetVersions\x12;.continual.rpc.management.v1.BatchGetDatasetVersionsRequest\x1a<.continual.rpc.management.v1.BatchGetDatasetVersionsResponse\"\x00\x12\xfd\x01\n\x13ListDatasetVersions\x12\x37.continual.rpc.management.v1.ListDatasetVersionsRequest\x1a\x38.continual.rpc.management.v1.ListDatasetVersionsResponse\"s\xca\x45\x0f\x64\x61tasetVersions\xd2\x45\x1b\x44\x61taset,Project,Environment\xe0\x45\x00\xda\x45=name:parent;name:parent:$/datasets/-;name:parent:$/datasets/-\x12\x9d\x01\n\x1e\x43reateDatasetVersionAssignment\x12\x42.continual.rpc.management.v1.CreateDatasetVersionAssignmentRequest\x1a\x35.continual.rpc.management.v1.DatasetVersionAssignment\"\x00\x12\x97\x01\n\x1bGetDatasetVersionAssignment\x12?.continual.rpc.management.v1.GetDatasetVersionAssignmentRequest\x1a\x35.continual.rpc.management.v1.DatasetVersionAssignment\"\x00\x12~\n\x1e\x44\x65leteDatasetVersionAssignment\x12\x42.continual.rpc.management.v1.DeleteDatasetVersionAssignmentRequest\x1a\x16.google.protobuf.Empty\"\x00\x12\xa8\x01\n\x1dListDatasetVersionAssignments\x12\x41.continual.rpc.management.v1.ListDatasetVersionAssignmentsRequest\x1a\x42.continual.rpc.management.v1.ListDatasetVersionAssignmentsResponse\"\x00\x42LZJgithub.com/continual-ai/continual/continual/rpc/management/v1;managementv1b\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,google_dot_api_dot_annotations__pb2.DESCRIPTOR,google_dot_protobuf_dot_empty__pb2.DESCRIPTOR,google_dot_protobuf_dot_field__mask__pb2.DESCRIPTOR,continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2.DESCRIPTOR,continual_dot_rpc_dot_graphql_dot_options__pb2.DESCRIPTOR,])




_UPDATEUSERREQUEST = _descriptor.Descriptor(
  name='UpdateUserRequest',
  full_name='continual.rpc.management.v1.UpdateUserRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user', full_name='continual.rpc.management.v1.UpdateUserRequest.user', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_mask', full_name='continual.rpc.management.v1.UpdateUserRequest.update_mask', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=292,
  serialized_end=409,
)


_GETUSERREQUEST = _descriptor.Descriptor(
  name='GetUserRequest',
  full_name='continual.rpc.management.v1.GetUserRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetUserRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=411,
  serialized_end=441,
)


_BATCHGETUSERSREQUEST = _descriptor.Descriptor(
  name='BatchGetUsersRequest',
  full_name='continual.rpc.management.v1.BatchGetUsersRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetUsersRequest.names', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=443,
  serialized_end=480,
)


_BATCHGETUSERSRESPONSE = _descriptor.Descriptor(
  name='BatchGetUsersResponse',
  full_name='continual.rpc.management.v1.BatchGetUsersResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='users', full_name='continual.rpc.management.v1.BatchGetUsersResponse.users', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=482,
  serialized_end=555,
)


_DELETEUSERREQUEST = _descriptor.Descriptor(
  name='DeleteUserRequest',
  full_name='continual.rpc.management.v1.DeleteUserRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DeleteUserRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=557,
  serialized_end=590,
)


_LISTUSERSREQUEST = _descriptor.Descriptor(
  name='ListUsersRequest',
  full_name='continual.rpc.management.v1.ListUsersRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListUsersRequest.page_size', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListUsersRequest.page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='order_by', full_name='continual.rpc.management.v1.ListUsersRequest.order_by', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='search', full_name='continual.rpc.management.v1.ListUsersRequest.search', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest', full_name='continual.rpc.management.v1.ListUsersRequest.latest', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=592,
  serialized_end=699,
)


_LISTUSERSRESPONSE = _descriptor.Descriptor(
  name='ListUsersResponse',
  full_name='continual.rpc.management.v1.ListUsersResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='users', full_name='continual.rpc.management.v1.ListUsersResponse.users', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListUsersResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=701,
  serialized_end=795,
)


_REGISTERREQUEST = _descriptor.Descriptor(
  name='RegisterRequest',
  full_name='continual.rpc.management.v1.RegisterRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='full_name', full_name='continual.rpc.management.v1.RegisterRequest.full_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='email', full_name='continual.rpc.management.v1.RegisterRequest.email', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='password', full_name='continual.rpc.management.v1.RegisterRequest.password', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='first_name', full_name='continual.rpc.management.v1.RegisterRequest.first_name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='last_name', full_name='continual.rpc.management.v1.RegisterRequest.last_name', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=797,
  serialized_end=905,
)


_REGISTERRESPONSE = _descriptor.Descriptor(
  name='RegisterResponse',
  full_name='continual.rpc.management.v1.RegisterResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user', full_name='continual.rpc.management.v1.RegisterResponse.user', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='auth_token', full_name='continual.rpc.management.v1.RegisterResponse.auth_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=907,
  serialized_end=994,
)


_LOGINREQUEST = _descriptor.Descriptor(
  name='LoginRequest',
  full_name='continual.rpc.management.v1.LoginRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='email', full_name='continual.rpc.management.v1.LoginRequest.email', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='password', full_name='continual.rpc.management.v1.LoginRequest.password', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=996,
  serialized_end=1043,
)


_LOGINRESPONSE = _descriptor.Descriptor(
  name='LoginResponse',
  full_name='continual.rpc.management.v1.LoginResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user', full_name='continual.rpc.management.v1.LoginResponse.user', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='auth_token', full_name='continual.rpc.management.v1.LoginResponse.auth_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1045,
  serialized_end=1129,
)


_CHECKORGANIZATIONACCESSREQUEST = _descriptor.Descriptor(
  name='CheckOrganizationAccessRequest',
  full_name='continual.rpc.management.v1.CheckOrganizationAccessRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='organization', full_name='continual.rpc.management.v1.CheckOrganizationAccessRequest.organization', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1131,
  serialized_end=1185,
)


_CHECKORGANIZATIONACCESSRESPONSE = _descriptor.Descriptor(
  name='CheckOrganizationAccessResponse',
  full_name='continual.rpc.management.v1.CheckOrganizationAccessResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='requires_auth', full_name='continual.rpc.management.v1.CheckOrganizationAccessResponse.requires_auth', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='organization', full_name='continual.rpc.management.v1.CheckOrganizationAccessResponse.organization', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='display_name', full_name='continual.rpc.management.v1.CheckOrganizationAccessResponse.display_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1187,
  serialized_end=1287,
)


_SENDEMAILVERIFICATIONREQUEST = _descriptor.Descriptor(
  name='SendEmailVerificationRequest',
  full_name='continual.rpc.management.v1.SendEmailVerificationRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user', full_name='continual.rpc.management.v1.SendEmailVerificationRequest.user', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1289,
  serialized_end=1333,
)


_VERIFYEMAILREQUEST = _descriptor.Descriptor(
  name='VerifyEmailRequest',
  full_name='continual.rpc.management.v1.VerifyEmailRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user', full_name='continual.rpc.management.v1.VerifyEmailRequest.user', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='code', full_name='continual.rpc.management.v1.VerifyEmailRequest.code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1335,
  serialized_end=1383,
)


_VERIFYEMAILRESPONSE = _descriptor.Descriptor(
  name='VerifyEmailResponse',
  full_name='continual.rpc.management.v1.VerifyEmailResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='success', full_name='continual.rpc.management.v1.VerifyEmailResponse.success', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1385,
  serialized_end=1423,
)


_SENDPASSWORDRESETREQUEST = _descriptor.Descriptor(
  name='SendPasswordResetRequest',
  full_name='continual.rpc.management.v1.SendPasswordResetRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='email', full_name='continual.rpc.management.v1.SendPasswordResetRequest.email', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1425,
  serialized_end=1466,
)


_GETPASSWORDRESETREQUEST = _descriptor.Descriptor(
  name='GetPasswordResetRequest',
  full_name='continual.rpc.management.v1.GetPasswordResetRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='continual.rpc.management.v1.GetPasswordResetRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1468,
  serialized_end=1507,
)


_RESETPASSWORDREQUEST = _descriptor.Descriptor(
  name='ResetPasswordRequest',
  full_name='continual.rpc.management.v1.ResetPasswordRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='password', full_name='continual.rpc.management.v1.ResetPasswordRequest.password', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='code', full_name='continual.rpc.management.v1.ResetPasswordRequest.code', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='old_password', full_name='continual.rpc.management.v1.ResetPasswordRequest.old_password', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1509,
  serialized_end=1585,
)


_RESETPASSWORDRESPONSE = _descriptor.Descriptor(
  name='ResetPasswordResponse',
  full_name='continual.rpc.management.v1.ResetPasswordResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='success', full_name='continual.rpc.management.v1.ResetPasswordResponse.success', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1587,
  serialized_end=1627,
)


_CHANGEPASSWORDREQUEST = _descriptor.Descriptor(
  name='ChangePasswordRequest',
  full_name='continual.rpc.management.v1.ChangePasswordRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='password', full_name='continual.rpc.management.v1.ChangePasswordRequest.password', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='old_password', full_name='continual.rpc.management.v1.ChangePasswordRequest.old_password', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1629,
  serialized_end=1692,
)


_CHANGEPASSWORDRESPONSE = _descriptor.Descriptor(
  name='ChangePasswordResponse',
  full_name='continual.rpc.management.v1.ChangePasswordResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='success', full_name='continual.rpc.management.v1.ChangePasswordResponse.success', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='auth_token', full_name='continual.rpc.management.v1.ChangePasswordResponse.auth_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1694,
  serialized_end=1755,
)


_GETSESSIONREQUEST = _descriptor.Descriptor(
  name='GetSessionRequest',
  full_name='continual.rpc.management.v1.GetSessionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetSessionRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1757,
  serialized_end=1790,
)


_LISTSESSIONSREQUEST = _descriptor.Descriptor(
  name='ListSessionsRequest',
  full_name='continual.rpc.management.v1.ListSessionsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListSessionsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListSessionsRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListSessionsRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1792,
  serialized_end=1868,
)


_LISTSESSIONSRESPONSE = _descriptor.Descriptor(
  name='ListSessionsResponse',
  full_name='continual.rpc.management.v1.ListSessionsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='sessions', full_name='continual.rpc.management.v1.ListSessionsResponse.sessions', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListSessionsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1870,
  serialized_end=1973,
)


_DELETESESSIONREQUEST = _descriptor.Descriptor(
  name='DeleteSessionRequest',
  full_name='continual.rpc.management.v1.DeleteSessionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DeleteSessionRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1975,
  serialized_end=2011,
)


_DELETEALLSESSIONSREQUEST = _descriptor.Descriptor(
  name='DeleteAllSessionsRequest',
  full_name='continual.rpc.management.v1.DeleteAllSessionsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.DeleteAllSessionsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2013,
  serialized_end=2055,
)


_LISTROLESREQUEST = _descriptor.Descriptor(
  name='ListRolesRequest',
  full_name='continual.rpc.management.v1.ListRolesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListRolesRequest.page_size', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListRolesRequest.page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2057,
  serialized_end=2114,
)


_LISTROLESRESPONSE = _descriptor.Descriptor(
  name='ListRolesResponse',
  full_name='continual.rpc.management.v1.ListRolesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='roles', full_name='continual.rpc.management.v1.ListRolesResponse.roles', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListRolesResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2116,
  serialized_end=2210,
)


_GETROLEREQUEST = _descriptor.Descriptor(
  name='GetRoleRequest',
  full_name='continual.rpc.management.v1.GetRoleRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetRoleRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2212,
  serialized_end=2242,
)


_BATCHGETROLESREQUEST = _descriptor.Descriptor(
  name='BatchGetRolesRequest',
  full_name='continual.rpc.management.v1.BatchGetRolesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.BatchGetRolesRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetRolesRequest.names', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2244,
  serialized_end=2297,
)


_BATCHGETROLESRESPONSE = _descriptor.Descriptor(
  name='BatchGetRolesResponse',
  full_name='continual.rpc.management.v1.BatchGetRolesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='roles', full_name='continual.rpc.management.v1.BatchGetRolesResponse.roles', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2299,
  serialized_end=2372,
)


_LISTACCESSPOLICIESREQUEST = _descriptor.Descriptor(
  name='ListAccessPoliciesRequest',
  full_name='continual.rpc.management.v1.ListAccessPoliciesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListAccessPoliciesRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListAccessPoliciesRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListAccessPoliciesRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='include_api_keys', full_name='continual.rpc.management.v1.ListAccessPoliciesRequest.include_api_keys', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2374,
  serialized_end=2482,
)


_LISTACCESSPOLICIESRESPONSE = _descriptor.Descriptor(
  name='ListAccessPoliciesResponse',
  full_name='continual.rpc.management.v1.ListAccessPoliciesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='access_policies', full_name='continual.rpc.management.v1.ListAccessPoliciesResponse.access_policies', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListAccessPoliciesResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2484,
  serialized_end=2605,
)


_CREATEACCESSPOLICYREQUEST = _descriptor.Descriptor(
  name='CreateAccessPolicyRequest',
  full_name='continual.rpc.management.v1.CreateAccessPolicyRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.CreateAccessPolicyRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='access_policy', full_name='continual.rpc.management.v1.CreateAccessPolicyRequest.access_policy', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2607,
  serialized_end=2716,
)


_GETACCESSPOLICYREQUEST = _descriptor.Descriptor(
  name='GetAccessPolicyRequest',
  full_name='continual.rpc.management.v1.GetAccessPolicyRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetAccessPolicyRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2718,
  serialized_end=2756,
)


_UPDATEACCESSPOLICYREQUEST = _descriptor.Descriptor(
  name='UpdateAccessPolicyRequest',
  full_name='continual.rpc.management.v1.UpdateAccessPolicyRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='access_policy', full_name='continual.rpc.management.v1.UpdateAccessPolicyRequest.access_policy', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_mask', full_name='continual.rpc.management.v1.UpdateAccessPolicyRequest.update_mask', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2759,
  serialized_end=2901,
)


_DELETEACCESSPOLICYREQUEST = _descriptor.Descriptor(
  name='DeleteAccessPolicyRequest',
  full_name='continual.rpc.management.v1.DeleteAccessPolicyRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DeleteAccessPolicyRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2903,
  serialized_end=2944,
)


_BULKDELETEACCESSPOLICIESREQUEST = _descriptor.Descriptor(
  name='BulkDeleteAccessPoliciesRequest',
  full_name='continual.rpc.management.v1.BulkDeleteAccessPoliciesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BulkDeleteAccessPoliciesRequest.names', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2946,
  serialized_end=2994,
)


_CREATEINVITEREQUEST = _descriptor.Descriptor(
  name='CreateInviteRequest',
  full_name='continual.rpc.management.v1.CreateInviteRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.CreateInviteRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='invite', full_name='continual.rpc.management.v1.CreateInviteRequest.invite', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2996,
  serialized_end=3086,
)


_GETINVITEREQUEST = _descriptor.Descriptor(
  name='GetInviteRequest',
  full_name='continual.rpc.management.v1.GetInviteRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetInviteRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3088,
  serialized_end=3120,
)


_LISTINVITESREQUEST = _descriptor.Descriptor(
  name='ListInvitesRequest',
  full_name='continual.rpc.management.v1.ListInvitesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListInvitesRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListInvitesRequest.page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListInvitesRequest.page_size', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3122,
  serialized_end=3197,
)


_LISTINVITESRESPONSE = _descriptor.Descriptor(
  name='ListInvitesResponse',
  full_name='continual.rpc.management.v1.ListInvitesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='invites', full_name='continual.rpc.management.v1.ListInvitesResponse.invites', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListInvitesResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3199,
  serialized_end=3299,
)


_UPDATEINVITEREQUEST = _descriptor.Descriptor(
  name='UpdateInviteRequest',
  full_name='continual.rpc.management.v1.UpdateInviteRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='invite', full_name='continual.rpc.management.v1.UpdateInviteRequest.invite', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_mask', full_name='continual.rpc.management.v1.UpdateInviteRequest.update_mask', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3301,
  serialized_end=3424,
)


_DELETEINVITEREQUEST = _descriptor.Descriptor(
  name='DeleteInviteRequest',
  full_name='continual.rpc.management.v1.DeleteInviteRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DeleteInviteRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3426,
  serialized_end=3461,
)


_ACCEPTINVITEREQUEST = _descriptor.Descriptor(
  name='AcceptInviteRequest',
  full_name='continual.rpc.management.v1.AcceptInviteRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='continual.rpc.management.v1.AcceptInviteRequest.code', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3463,
  serialized_end=3498,
)


_RESENDINVITEREQUEST = _descriptor.Descriptor(
  name='ResendInviteRequest',
  full_name='continual.rpc.management.v1.ResendInviteRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.ResendInviteRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3500,
  serialized_end=3535,
)


_TESTPERMISSIONSREQUEST = _descriptor.Descriptor(
  name='TestPermissionsRequest',
  full_name='continual.rpc.management.v1.TestPermissionsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='resource', full_name='continual.rpc.management.v1.TestPermissionsRequest.resource', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='permissions', full_name='continual.rpc.management.v1.TestPermissionsRequest.permissions', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3537,
  serialized_end=3600,
)


_TESTPERMISSIONSRESPONSE = _descriptor.Descriptor(
  name='TestPermissionsResponse',
  full_name='continual.rpc.management.v1.TestPermissionsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='permissions', full_name='continual.rpc.management.v1.TestPermissionsResponse.permissions', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3602,
  serialized_end=3648,
)


_CREATEORGANIZATIONREQUEST = _descriptor.Descriptor(
  name='CreateOrganizationRequest',
  full_name='continual.rpc.management.v1.CreateOrganizationRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='organization', full_name='continual.rpc.management.v1.CreateOrganizationRequest.organization', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3650,
  serialized_end=3742,
)


_UPDATEORGANIZATIONREQUEST = _descriptor.Descriptor(
  name='UpdateOrganizationRequest',
  full_name='continual.rpc.management.v1.UpdateOrganizationRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='organization', full_name='continual.rpc.management.v1.UpdateOrganizationRequest.organization', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_mask', full_name='continual.rpc.management.v1.UpdateOrganizationRequest.update_mask', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3745,
  serialized_end=3886,
)


_GETORGANIZATIONREQUEST = _descriptor.Descriptor(
  name='GetOrganizationRequest',
  full_name='continual.rpc.management.v1.GetOrganizationRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetOrganizationRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3888,
  serialized_end=3926,
)


_BATCHGETORGANIZATIONSREQUEST = _descriptor.Descriptor(
  name='BatchGetOrganizationsRequest',
  full_name='continual.rpc.management.v1.BatchGetOrganizationsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetOrganizationsRequest.names', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3928,
  serialized_end=3973,
)


_BATCHGETORGANIZATIONSRESPONSE = _descriptor.Descriptor(
  name='BatchGetOrganizationsResponse',
  full_name='continual.rpc.management.v1.BatchGetOrganizationsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='organizations', full_name='continual.rpc.management.v1.BatchGetOrganizationsResponse.organizations', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=3975,
  serialized_end=4072,
)


_DELETEORGANIZATIONREQUEST = _descriptor.Descriptor(
  name='DeleteOrganizationRequest',
  full_name='continual.rpc.management.v1.DeleteOrganizationRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DeleteOrganizationRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4074,
  serialized_end=4115,
)


_LISTORGANIZATIONSREQUEST = _descriptor.Descriptor(
  name='ListOrganizationsRequest',
  full_name='continual.rpc.management.v1.ListOrganizationsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListOrganizationsRequest.page_size', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListOrganizationsRequest.page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='order_by', full_name='continual.rpc.management.v1.ListOrganizationsRequest.order_by', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='search', full_name='continual.rpc.management.v1.ListOrganizationsRequest.search', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest', full_name='continual.rpc.management.v1.ListOrganizationsRequest.latest', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='include_sso', full_name='continual.rpc.management.v1.ListOrganizationsRequest.include_sso', index=5,
      number=6, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4118,
  serialized_end=4254,
)


_LISTORGANIZATIONSRESPONSE = _descriptor.Descriptor(
  name='ListOrganizationsResponse',
  full_name='continual.rpc.management.v1.ListOrganizationsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='organizations', full_name='continual.rpc.management.v1.ListOrganizationsResponse.organizations', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListOrganizationsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4256,
  serialized_end=4374,
)


_GETONBOARDINGREQUEST = _descriptor.Descriptor(
  name='GetOnboardingRequest',
  full_name='continual.rpc.management.v1.GetOnboardingRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetOnboardingRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4376,
  serialized_end=4412,
)


_UPDATEONBOARDINGREQUEST = _descriptor.Descriptor(
  name='UpdateOnboardingRequest',
  full_name='continual.rpc.management.v1.UpdateOnboardingRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='onboarding', full_name='continual.rpc.management.v1.UpdateOnboardingRequest.onboarding', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_mask', full_name='continual.rpc.management.v1.UpdateOnboardingRequest.update_mask', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4415,
  serialized_end=4550,
)


_CREATEPROJECTREQUEST = _descriptor.Descriptor(
  name='CreateProjectRequest',
  full_name='continual.rpc.management.v1.CreateProjectRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='project', full_name='continual.rpc.management.v1.CreateProjectRequest.project', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='project_id', full_name='continual.rpc.management.v1.CreateProjectRequest.project_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4552,
  serialized_end=4649,
)


_UPDATEPROJECTREQUEST = _descriptor.Descriptor(
  name='UpdateProjectRequest',
  full_name='continual.rpc.management.v1.UpdateProjectRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='project', full_name='continual.rpc.management.v1.UpdateProjectRequest.project', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_mask', full_name='continual.rpc.management.v1.UpdateProjectRequest.update_mask', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4651,
  serialized_end=4777,
)


_GETPROJECTREQUEST = _descriptor.Descriptor(
  name='GetProjectRequest',
  full_name='continual.rpc.management.v1.GetProjectRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetProjectRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4779,
  serialized_end=4812,
)


_BATCHGETPROJECTSREQUEST = _descriptor.Descriptor(
  name='BatchGetProjectsRequest',
  full_name='continual.rpc.management.v1.BatchGetProjectsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetProjectsRequest.names', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4814,
  serialized_end=4854,
)


_BATCHGETPROJECTSRESPONSE = _descriptor.Descriptor(
  name='BatchGetProjectsResponse',
  full_name='continual.rpc.management.v1.BatchGetProjectsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='projects', full_name='continual.rpc.management.v1.BatchGetProjectsResponse.projects', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4856,
  serialized_end=4938,
)


_DELETEPROJECTREQUEST = _descriptor.Descriptor(
  name='DeleteProjectRequest',
  full_name='continual.rpc.management.v1.DeleteProjectRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DeleteProjectRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4940,
  serialized_end=4976,
)


_LISTPROJECTSREQUEST = _descriptor.Descriptor(
  name='ListProjectsRequest',
  full_name='continual.rpc.management.v1.ListProjectsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListProjectsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListProjectsRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListProjectsRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListProjectsRequest.filters', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListProjectsRequest.ignore_filters', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListProjectsRequest.start_time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListProjectsRequest.end_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='order_by', full_name='continual.rpc.management.v1.ListProjectsRequest.order_by', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='search', full_name='continual.rpc.management.v1.ListProjectsRequest.search', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest', full_name='continual.rpc.management.v1.ListProjectsRequest.latest', index=9,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=4979,
  serialized_end=5240,
)


_LISTPROJECTSRESPONSE = _descriptor.Descriptor(
  name='ListProjectsResponse',
  full_name='continual.rpc.management.v1.ListProjectsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='projects', full_name='continual.rpc.management.v1.ListProjectsResponse.projects', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListProjectsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5242,
  serialized_end=5345,
)


_RERUNREQUEST = _descriptor.Descriptor(
  name='ReRunRequest',
  full_name='continual.rpc.management.v1.ReRunRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.ReRunRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='message', full_name='continual.rpc.management.v1.ReRunRequest.message', index=1,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='trigger', full_name='continual.rpc.management.v1.ReRunRequest.trigger', index=2,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5347,
  serialized_end=5409,
)


_GETMODELSTATSREQUEST = _descriptor.Descriptor(
  name='GetModelStatsRequest',
  full_name='continual.rpc.management.v1.GetModelStatsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetModelStatsRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5411,
  serialized_end=5447,
)


_GETUSAGEREQUEST = _descriptor.Descriptor(
  name='GetUsageRequest',
  full_name='continual.rpc.management.v1.GetUsageRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.GetUsageRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.GetUsageRequest.start_time', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.GetUsageRequest.end_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5449,
  serialized_end=5576,
)


_GETUSAGERESPONSE = _descriptor.Descriptor(
  name='GetUsageResponse',
  full_name='continual.rpc.management.v1.GetUsageResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='credits_per_day', full_name='continual.rpc.management.v1.GetUsageResponse.credits_per_day', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5578,
  serialized_end=5669,
)


_LISTPRICESREQUEST = _descriptor.Descriptor(
  name='ListPricesRequest',
  full_name='continual.rpc.management.v1.ListPricesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListPricesRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5671,
  serialized_end=5706,
)


_LISTPRICESRESPONSE = _descriptor.Descriptor(
  name='ListPricesResponse',
  full_name='continual.rpc.management.v1.ListPricesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='prices', full_name='continual.rpc.management.v1.ListPricesResponse.prices', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5708,
  serialized_end=5780,
)


_LISTINVOICESREQUEST = _descriptor.Descriptor(
  name='ListInvoicesRequest',
  full_name='continual.rpc.management.v1.ListInvoicesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListInvoicesRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListInvoicesRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListInvoicesRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5782,
  serialized_end=5858,
)


_LISTINVOICESRESPONSE = _descriptor.Descriptor(
  name='ListInvoicesResponse',
  full_name='continual.rpc.management.v1.ListInvoicesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='invoices', full_name='continual.rpc.management.v1.ListInvoicesResponse.invoices', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListInvoicesResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5860,
  serialized_end=5963,
)


_GETPRICEREQUEST = _descriptor.Descriptor(
  name='GetPriceRequest',
  full_name='continual.rpc.management.v1.GetPriceRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetPriceRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5965,
  serialized_end=5996,
)


_GETPAYMENTMETHODREQUEST = _descriptor.Descriptor(
  name='GetPaymentMethodRequest',
  full_name='continual.rpc.management.v1.GetPaymentMethodRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetPaymentMethodRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=5998,
  serialized_end=6037,
)


_GETUPCOMINGINVOICEREQUEST = _descriptor.Descriptor(
  name='GetUpcomingInvoiceRequest',
  full_name='continual.rpc.management.v1.GetUpcomingInvoiceRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetUpcomingInvoiceRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6039,
  serialized_end=6080,
)


_GETSUBSCRIPTIONREQUEST = _descriptor.Descriptor(
  name='GetSubscriptionRequest',
  full_name='continual.rpc.management.v1.GetSubscriptionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetSubscriptionRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6082,
  serialized_end=6120,
)


_GETBILLINGACCOUNTREQUEST = _descriptor.Descriptor(
  name='GetBillingAccountRequest',
  full_name='continual.rpc.management.v1.GetBillingAccountRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetBillingAccountRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6122,
  serialized_end=6162,
)


_UPDATEBILLINGACCOUNTREQUEST = _descriptor.Descriptor(
  name='UpdateBillingAccountRequest',
  full_name='continual.rpc.management.v1.UpdateBillingAccountRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='billing_account', full_name='continual.rpc.management.v1.UpdateBillingAccountRequest.billing_account', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_mask', full_name='continual.rpc.management.v1.UpdateBillingAccountRequest.update_mask', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6165,
  serialized_end=6313,
)


_UPDATESUBSCRIPTIONREQUEST = _descriptor.Descriptor(
  name='UpdateSubscriptionRequest',
  full_name='continual.rpc.management.v1.UpdateSubscriptionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='subscription', full_name='continual.rpc.management.v1.UpdateSubscriptionRequest.subscription', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_mask', full_name='continual.rpc.management.v1.UpdateSubscriptionRequest.update_mask', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6316,
  serialized_end=6457,
)


_CREATEPROMOTIONREQUEST = _descriptor.Descriptor(
  name='CreatePromotionRequest',
  full_name='continual.rpc.management.v1.CreatePromotionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model_version_name', full_name='continual.rpc.management.v1.CreatePromotionRequest.model_version_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run_name', full_name='continual.rpc.management.v1.CreatePromotionRequest.run_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='reason', full_name='continual.rpc.management.v1.CreatePromotionRequest.reason', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6459,
  serialized_end=6545,
)


_LISTPROMOTIONSREQUEST = _descriptor.Descriptor(
  name='ListPromotionsRequest',
  full_name='continual.rpc.management.v1.ListPromotionsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListPromotionsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListPromotionsRequest.page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListPromotionsRequest.page_size', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='active', full_name='continual.rpc.management.v1.ListPromotionsRequest.active', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListPromotionsRequest.filters', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListPromotionsRequest.ignore_filters', index=5,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListPromotionsRequest.start_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListPromotionsRequest.end_time', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='all_projects', full_name='continual.rpc.management.v1.ListPromotionsRequest.all_projects', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='order_by', full_name='continual.rpc.management.v1.ListPromotionsRequest.order_by', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest', full_name='continual.rpc.management.v1.ListPromotionsRequest.latest', index=10,
      number=11, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='search', full_name='continual.rpc.management.v1.ListPromotionsRequest.search', index=11,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6548,
  serialized_end=6849,
)


_LISTPROMOTIONSRESPONSE = _descriptor.Descriptor(
  name='ListPromotionsResponse',
  full_name='continual.rpc.management.v1.ListPromotionsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='promotions', full_name='continual.rpc.management.v1.ListPromotionsResponse.promotions', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListPromotionsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6851,
  serialized_end=6960,
)


_GETPROMOTIONREQUEST = _descriptor.Descriptor(
  name='GetPromotionRequest',
  full_name='continual.rpc.management.v1.GetPromotionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetPromotionRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6962,
  serialized_end=6997,
)


_BATCHGETPROMOTIONSREQUEST = _descriptor.Descriptor(
  name='BatchGetPromotionsRequest',
  full_name='continual.rpc.management.v1.BatchGetPromotionsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.BatchGetPromotionsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetPromotionsRequest.names', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=6999,
  serialized_end=7057,
)


_BATCHGETPROMOTIONSRESPONSE = _descriptor.Descriptor(
  name='BatchGetPromotionsResponse',
  full_name='continual.rpc.management.v1.BatchGetPromotionsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='promotions', full_name='continual.rpc.management.v1.BatchGetPromotionsResponse.promotions', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7059,
  serialized_end=7147,
)


_CREATEEVENTREQUEST = _descriptor.Descriptor(
  name='CreateEventRequest',
  full_name='continual.rpc.management.v1.CreateEventRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.CreateEventRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='event', full_name='continual.rpc.management.v1.CreateEventRequest.event', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7149,
  serialized_end=7236,
)


_LISTEVENTSREQUEST = _descriptor.Descriptor(
  name='ListEventsRequest',
  full_name='continual.rpc.management.v1.ListEventsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListEventsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListEventsRequest.filters', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListEventsRequest.ignore_filters', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListEventsRequest.start_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListEventsRequest.end_time', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListEventsRequest.page_token', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListEventsRequest.page_size', index=6,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='all_projects', full_name='continual.rpc.management.v1.ListEventsRequest.all_projects', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='search', full_name='continual.rpc.management.v1.ListEventsRequest.search', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7239,
  serialized_end=7486,
)


_LISTEVENTSRESPONSE = _descriptor.Descriptor(
  name='ListEventsResponse',
  full_name='continual.rpc.management.v1.ListEventsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='events', full_name='continual.rpc.management.v1.ListEventsResponse.events', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListEventsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7488,
  serialized_end=7585,
)


_GETEVENTREQUEST = _descriptor.Descriptor(
  name='GetEventRequest',
  full_name='continual.rpc.management.v1.GetEventRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetEventRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7587,
  serialized_end=7618,
)


_BATCHGETEVENTSREQUEST = _descriptor.Descriptor(
  name='BatchGetEventsRequest',
  full_name='continual.rpc.management.v1.BatchGetEventsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.BatchGetEventsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetEventsRequest.names', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7620,
  serialized_end=7674,
)


_BATCHGETEVENTSRESPONSE = _descriptor.Descriptor(
  name='BatchGetEventsResponse',
  full_name='continual.rpc.management.v1.BatchGetEventsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='events', full_name='continual.rpc.management.v1.BatchGetEventsResponse.events', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7676,
  serialized_end=7752,
)


_CREATEMODELREQUEST = _descriptor.Descriptor(
  name='CreateModelRequest',
  full_name='continual.rpc.management.v1.CreateModelRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.CreateModelRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='continual.rpc.management.v1.CreateModelRequest.description', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='get_if_exists', full_name='continual.rpc.management.v1.CreateModelRequest.get_if_exists', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run', full_name='continual.rpc.management.v1.CreateModelRequest.run', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7754,
  serialized_end=7845,
)


_GETMODELREQUEST = _descriptor.Descriptor(
  name='GetModelRequest',
  full_name='continual.rpc.management.v1.GetModelRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetModelRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7847,
  serialized_end=7878,
)


_BATCHGETMODELSREQUEST = _descriptor.Descriptor(
  name='BatchGetModelsRequest',
  full_name='continual.rpc.management.v1.BatchGetModelsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.BatchGetModelsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetModelsRequest.names', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7880,
  serialized_end=7934,
)


_BATCHGETMODELSRESPONSE = _descriptor.Descriptor(
  name='BatchGetModelsResponse',
  full_name='continual.rpc.management.v1.BatchGetModelsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='models', full_name='continual.rpc.management.v1.BatchGetModelsResponse.models', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=7936,
  serialized_end=8012,
)


_LISTMODELSREQUEST = _descriptor.Descriptor(
  name='ListModelsRequest',
  full_name='continual.rpc.management.v1.ListModelsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListModelsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListModelsRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListModelsRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest', full_name='continual.rpc.management.v1.ListModelsRequest.latest', index=3,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListModelsRequest.filters', index=4,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListModelsRequest.ignore_filters', index=5,
      number=7, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListModelsRequest.start_time', index=6,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListModelsRequest.end_time', index=7,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='all_projects', full_name='continual.rpc.management.v1.ListModelsRequest.all_projects', index=8,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='order_by', full_name='continual.rpc.management.v1.ListModelsRequest.order_by', index=9,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='search', full_name='continual.rpc.management.v1.ListModelsRequest.search', index=10,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8015,
  serialized_end=8296,
)


_LISTMODELSRESPONSE = _descriptor.Descriptor(
  name='ListModelsResponse',
  full_name='continual.rpc.management.v1.ListModelsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='models', full_name='continual.rpc.management.v1.ListModelsResponse.models', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListModelsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8298,
  serialized_end=8395,
)


_DELETEMODELREQUEST = _descriptor.Descriptor(
  name='DeleteModelRequest',
  full_name='continual.rpc.management.v1.DeleteModelRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DeleteModelRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run_name', full_name='continual.rpc.management.v1.DeleteModelRequest.run_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8397,
  serialized_end=8449,
)


_VALIDATEMODELREQUEST = _descriptor.Descriptor(
  name='ValidateModelRequest',
  full_name='continual.rpc.management.v1.ValidateModelRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='environment', full_name='continual.rpc.management.v1.ValidateModelRequest.environment', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='schema', full_name='continual.rpc.management.v1.ValidateModelRequest.schema', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8451,
  serialized_end=8510,
)


_CREATEMODELVERSIONREQUEST = _descriptor.Descriptor(
  name='CreateModelVersionRequest',
  full_name='continual.rpc.management.v1.CreateModelVersionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.CreateModelVersionRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run_name', full_name='continual.rpc.management.v1.CreateModelVersionRequest.run_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='datasets', full_name='continual.rpc.management.v1.CreateModelVersionRequest.datasets', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8512,
  serialized_end=8591,
)


_GETMODELVERSIONREQUEST = _descriptor.Descriptor(
  name='GetModelVersionRequest',
  full_name='continual.rpc.management.v1.GetModelVersionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetModelVersionRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8593,
  serialized_end=8631,
)


_BATCHGETMODELVERSIONSREQUEST = _descriptor.Descriptor(
  name='BatchGetModelVersionsRequest',
  full_name='continual.rpc.management.v1.BatchGetModelVersionsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.BatchGetModelVersionsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetModelVersionsRequest.names', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8633,
  serialized_end=8694,
)


_BATCHGETMODELVERSIONSRESPONSE = _descriptor.Descriptor(
  name='BatchGetModelVersionsResponse',
  full_name='continual.rpc.management.v1.BatchGetModelVersionsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model_versions', full_name='continual.rpc.management.v1.BatchGetModelVersionsResponse.model_versions', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8696,
  serialized_end=8794,
)


_LISTMODELVERSIONSREQUEST = _descriptor.Descriptor(
  name='ListModelVersionsRequest',
  full_name='continual.rpc.management.v1.ListModelVersionsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListModelVersionsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListModelVersionsRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListModelVersionsRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest', full_name='continual.rpc.management.v1.ListModelVersionsRequest.latest', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListModelVersionsRequest.filters', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListModelVersionsRequest.ignore_filters', index=5,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListModelVersionsRequest.start_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListModelVersionsRequest.end_time', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='all_projects', full_name='continual.rpc.management.v1.ListModelVersionsRequest.all_projects', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='order_by', full_name='continual.rpc.management.v1.ListModelVersionsRequest.order_by', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='search', full_name='continual.rpc.management.v1.ListModelVersionsRequest.search', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=8797,
  serialized_end=9085,
)


_LISTMODELVERSIONSRESPONSE = _descriptor.Descriptor(
  name='ListModelVersionsResponse',
  full_name='continual.rpc.management.v1.ListModelVersionsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='model_versions', full_name='continual.rpc.management.v1.ListModelVersionsResponse.model_versions', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListModelVersionsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9087,
  serialized_end=9206,
)


_CREATEEXPERIMENTREQUEST = _descriptor.Descriptor(
  name='CreateExperimentRequest',
  full_name='continual.rpc.management.v1.CreateExperimentRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.CreateExperimentRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run_name', full_name='continual.rpc.management.v1.CreateExperimentRequest.run_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9208,
  serialized_end=9267,
)


_GETEXPERIMENTREQUEST = _descriptor.Descriptor(
  name='GetExperimentRequest',
  full_name='continual.rpc.management.v1.GetExperimentRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetExperimentRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9269,
  serialized_end=9305,
)


_BATCHGETEXPERIMENTSREQUEST = _descriptor.Descriptor(
  name='BatchGetExperimentsRequest',
  full_name='continual.rpc.management.v1.BatchGetExperimentsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.BatchGetExperimentsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetExperimentsRequest.names', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9307,
  serialized_end=9366,
)


_BATCHGETEXPERIMENTSRESPONSE = _descriptor.Descriptor(
  name='BatchGetExperimentsResponse',
  full_name='continual.rpc.management.v1.BatchGetExperimentsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='experiments', full_name='continual.rpc.management.v1.BatchGetExperimentsResponse.experiments', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9368,
  serialized_end=9459,
)


_LISTEXPERIMENTSREQUEST = _descriptor.Descriptor(
  name='ListExperimentsRequest',
  full_name='continual.rpc.management.v1.ListExperimentsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListExperimentsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListExperimentsRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListExperimentsRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListExperimentsRequest.filters', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListExperimentsRequest.ignore_filters', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListExperimentsRequest.start_time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListExperimentsRequest.end_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='all_projects', full_name='continual.rpc.management.v1.ListExperimentsRequest.all_projects', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest', full_name='continual.rpc.management.v1.ListExperimentsRequest.latest', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='order_by', full_name='continual.rpc.management.v1.ListExperimentsRequest.order_by', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='search', full_name='continual.rpc.management.v1.ListExperimentsRequest.search', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9462,
  serialized_end=9748,
)


_LISTEXPERIMENTSRESPONSE = _descriptor.Descriptor(
  name='ListExperimentsResponse',
  full_name='continual.rpc.management.v1.ListExperimentsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='experiments', full_name='continual.rpc.management.v1.ListExperimentsResponse.experiments', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListExperimentsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9750,
  serialized_end=9862,
)


_CREATEBATCHPREDICTIONREQUEST = _descriptor.Descriptor(
  name='CreateBatchPredictionRequest',
  full_name='continual.rpc.management.v1.CreateBatchPredictionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.CreateBatchPredictionRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run_name', full_name='continual.rpc.management.v1.CreateBatchPredictionRequest.run_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='trained_model_version_name', full_name='continual.rpc.management.v1.CreateBatchPredictionRequest.trained_model_version_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9864,
  serialized_end=9964,
)


_LISTBATCHPREDICTIONSREQUEST = _descriptor.Descriptor(
  name='ListBatchPredictionsRequest',
  full_name='continual.rpc.management.v1.ListBatchPredictionsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListBatchPredictionsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListBatchPredictionsRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListBatchPredictionsRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest', full_name='continual.rpc.management.v1.ListBatchPredictionsRequest.latest', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListBatchPredictionsRequest.filters', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListBatchPredictionsRequest.ignore_filters', index=5,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListBatchPredictionsRequest.start_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListBatchPredictionsRequest.end_time', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='all_projects', full_name='continual.rpc.management.v1.ListBatchPredictionsRequest.all_projects', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='order_by', full_name='continual.rpc.management.v1.ListBatchPredictionsRequest.order_by', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='search', full_name='continual.rpc.management.v1.ListBatchPredictionsRequest.search', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=9967,
  serialized_end=10258,
)


_LISTBATCHPREDICTIONSRESPONSE = _descriptor.Descriptor(
  name='ListBatchPredictionsResponse',
  full_name='continual.rpc.management.v1.ListBatchPredictionsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='batch_predictions', full_name='continual.rpc.management.v1.ListBatchPredictionsResponse.batch_predictions', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListBatchPredictionsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10261,
  serialized_end=10389,
)


_BATCHPREDICTIONREQUEST = _descriptor.Descriptor(
  name='BatchPredictionRequest',
  full_name='continual.rpc.management.v1.BatchPredictionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.BatchPredictionRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10391,
  serialized_end=10429,
)


_BATCHGETBATCHPREDICTIONSREQUEST = _descriptor.Descriptor(
  name='BatchGetBatchPredictionsRequest',
  full_name='continual.rpc.management.v1.BatchGetBatchPredictionsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.BatchGetBatchPredictionsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetBatchPredictionsRequest.names', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10431,
  serialized_end=10495,
)


_BATCHGETBATCHPREDICTIONSRESPONSE = _descriptor.Descriptor(
  name='BatchGetBatchPredictionsResponse',
  full_name='continual.rpc.management.v1.BatchGetBatchPredictionsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='batch_predictions', full_name='continual.rpc.management.v1.BatchGetBatchPredictionsResponse.batch_predictions', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10497,
  serialized_end=10604,
)


_YAMLDIFFREQUEST = _descriptor.Descriptor(
  name='YamlDiffRequest',
  full_name='continual.rpc.management.v1.YamlDiffRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='right', full_name='continual.rpc.management.v1.YamlDiffRequest.right', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='right_text', full_name='continual.rpc.management.v1.YamlDiffRequest.right_text', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='left', full_name='continual.rpc.management.v1.YamlDiffRequest.left', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='left_text', full_name='continual.rpc.management.v1.YamlDiffRequest.left_text', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.YamlDiffRequest.parent', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10606,
  serialized_end=10707,
)


_YAMLDIFFRESPONSE = _descriptor.Descriptor(
  name='YamlDiffResponse',
  full_name='continual.rpc.management.v1.YamlDiffResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='text', full_name='continual.rpc.management.v1.YamlDiffResponse.text', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10709,
  serialized_end=10741,
)


_CREATEENVIRONMENTREQUEST = _descriptor.Descriptor(
  name='CreateEnvironmentRequest',
  full_name='continual.rpc.management.v1.CreateEnvironmentRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.CreateEnvironmentRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='environment_id', full_name='continual.rpc.management.v1.CreateEnvironmentRequest.environment_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='source', full_name='continual.rpc.management.v1.CreateEnvironmentRequest.source', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='get_if_exists', full_name='continual.rpc.management.v1.CreateEnvironmentRequest.get_if_exists', index=3,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10743,
  serialized_end=10848,
)


_UPDATEENVIRONMENTREQUEST = _descriptor.Descriptor(
  name='UpdateEnvironmentRequest',
  full_name='continual.rpc.management.v1.UpdateEnvironmentRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='environment', full_name='continual.rpc.management.v1.UpdateEnvironmentRequest.environment', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_mask', full_name='continual.rpc.management.v1.UpdateEnvironmentRequest.update_mask', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10851,
  serialized_end=10989,
)


_GETENVIRONMENTREQUEST = _descriptor.Descriptor(
  name='GetEnvironmentRequest',
  full_name='continual.rpc.management.v1.GetEnvironmentRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetEnvironmentRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=10991,
  serialized_end=11028,
)


_BATCHGETENVIRONMENTSREQUEST = _descriptor.Descriptor(
  name='BatchGetEnvironmentsRequest',
  full_name='continual.rpc.management.v1.BatchGetEnvironmentsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetEnvironmentsRequest.names', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11030,
  serialized_end=11074,
)


_BATCHGETENVIRONMENTSRESPONSE = _descriptor.Descriptor(
  name='BatchGetEnvironmentsResponse',
  full_name='continual.rpc.management.v1.BatchGetEnvironmentsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='environments', full_name='continual.rpc.management.v1.BatchGetEnvironmentsResponse.environments', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11076,
  serialized_end=11170,
)


_DELETEENVIRONMENTREQUEST = _descriptor.Descriptor(
  name='DeleteEnvironmentRequest',
  full_name='continual.rpc.management.v1.DeleteEnvironmentRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DeleteEnvironmentRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run_name', full_name='continual.rpc.management.v1.DeleteEnvironmentRequest.run_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11172,
  serialized_end=11230,
)


_LISTENVIRONMENTSREQUEST = _descriptor.Descriptor(
  name='ListEnvironmentsRequest',
  full_name='continual.rpc.management.v1.ListEnvironmentsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListEnvironmentsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListEnvironmentsRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListEnvironmentsRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListEnvironmentsRequest.filters', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListEnvironmentsRequest.ignore_filters', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListEnvironmentsRequest.start_time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListEnvironmentsRequest.end_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='order_by', full_name='continual.rpc.management.v1.ListEnvironmentsRequest.order_by', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='search', full_name='continual.rpc.management.v1.ListEnvironmentsRequest.search', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest', full_name='continual.rpc.management.v1.ListEnvironmentsRequest.latest', index=9,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11233,
  serialized_end=11498,
)


_LISTENVIRONMENTSRESPONSE = _descriptor.Descriptor(
  name='ListEnvironmentsResponse',
  full_name='continual.rpc.management.v1.ListEnvironmentsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='environments', full_name='continual.rpc.management.v1.ListEnvironmentsResponse.environments', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListEnvironmentsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11500,
  serialized_end=11615,
)


_GETSYSTEMMESSAGEREQUEST = _descriptor.Descriptor(
  name='GetSystemMessageRequest',
  full_name='continual.rpc.management.v1.GetSystemMessageRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11617,
  serialized_end=11642,
)


_GETSYSTEMMESSAGERESPONSE = _descriptor.Descriptor(
  name='GetSystemMessageResponse',
  full_name='continual.rpc.management.v1.GetSystemMessageResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='message', full_name='continual.rpc.management.v1.GetSystemMessageResponse.message', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11644,
  serialized_end=11687,
)


_GETFEATUREFLAGREQUEST = _descriptor.Descriptor(
  name='GetFeatureFlagRequest',
  full_name='continual.rpc.management.v1.GetFeatureFlagRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='resource', full_name='continual.rpc.management.v1.GetFeatureFlagRequest.resource', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature', full_name='continual.rpc.management.v1.GetFeatureFlagRequest.feature', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11689,
  serialized_end=11747,
)


_GETFEATUREFLAGRESPONSE = _descriptor.Descriptor(
  name='GetFeatureFlagResponse',
  full_name='continual.rpc.management.v1.GetFeatureFlagResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='resource', full_name='continual.rpc.management.v1.GetFeatureFlagResponse.resource', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature', full_name='continual.rpc.management.v1.GetFeatureFlagResponse.feature', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='enabled', full_name='continual.rpc.management.v1.GetFeatureFlagResponse.enabled', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11749,
  serialized_end=11825,
)


_GETORGANIZATIONFEATUREFLAGSREQUEST = _descriptor.Descriptor(
  name='GetOrganizationFeatureFlagsRequest',
  full_name='continual.rpc.management.v1.GetOrganizationFeatureFlagsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='resource', full_name='continual.rpc.management.v1.GetOrganizationFeatureFlagsRequest.resource', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11827,
  serialized_end=11881,
)


_GETORGANIZATIONFEATUREFLAGSRESPONSE_FEATUREFLAGSENTRY = _descriptor.Descriptor(
  name='FeatureFlagsEntry',
  full_name='continual.rpc.management.v1.GetOrganizationFeatureFlagsResponse.FeatureFlagsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='continual.rpc.management.v1.GetOrganizationFeatureFlagsResponse.FeatureFlagsEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.GetOrganizationFeatureFlagsResponse.FeatureFlagsEntry.value', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12030,
  serialized_end=12081,
)

_GETORGANIZATIONFEATUREFLAGSRESPONSE = _descriptor.Descriptor(
  name='GetOrganizationFeatureFlagsResponse',
  full_name='continual.rpc.management.v1.GetOrganizationFeatureFlagsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='feature_flags', full_name='continual.rpc.management.v1.GetOrganizationFeatureFlagsResponse.feature_flags', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_GETORGANIZATIONFEATUREFLAGSRESPONSE_FEATUREFLAGSENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=11884,
  serialized_end=12081,
)


_CREATEARTIFACTREQUEST = _descriptor.Descriptor(
  name='CreateArtifactRequest',
  full_name='continual.rpc.management.v1.CreateArtifactRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.CreateArtifactRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='artifact', full_name='continual.rpc.management.v1.CreateArtifactRequest.artifact', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='artifact_id', full_name='continual.rpc.management.v1.CreateArtifactRequest.artifact_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_if_exists', full_name='continual.rpc.management.v1.CreateArtifactRequest.update_if_exists', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12084,
  serialized_end=12227,
)


_GETARTIFACTREQUEST = _descriptor.Descriptor(
  name='GetArtifactRequest',
  full_name='continual.rpc.management.v1.GetArtifactRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetArtifactRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12229,
  serialized_end=12263,
)


_BATCHGETARTIFACTSREQUEST = _descriptor.Descriptor(
  name='BatchGetArtifactsRequest',
  full_name='continual.rpc.management.v1.BatchGetArtifactsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.BatchGetArtifactsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetArtifactsRequest.names', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12265,
  serialized_end=12322,
)


_BATCHGETARTIFACTSRESPONSE = _descriptor.Descriptor(
  name='BatchGetArtifactsResponse',
  full_name='continual.rpc.management.v1.BatchGetArtifactsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='artifacts', full_name='continual.rpc.management.v1.BatchGetArtifactsResponse.artifacts', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12324,
  serialized_end=12409,
)


_DELETEARTIFACTREQUEST = _descriptor.Descriptor(
  name='DeleteArtifactRequest',
  full_name='continual.rpc.management.v1.DeleteArtifactRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DeleteArtifactRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12411,
  serialized_end=12448,
)


_LISTARTIFACTSREQUEST = _descriptor.Descriptor(
  name='ListArtifactsRequest',
  full_name='continual.rpc.management.v1.ListArtifactsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListArtifactsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListArtifactsRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListArtifactsRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListArtifactsRequest.filters', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListArtifactsRequest.ignore_filters', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='order_by', full_name='continual.rpc.management.v1.ListArtifactsRequest.order_by', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='search', full_name='continual.rpc.management.v1.ListArtifactsRequest.search', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest', full_name='continual.rpc.management.v1.ListArtifactsRequest.latest', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12451,
  serialized_end=12619,
)


_LISTARTIFACTSRESPONSE = _descriptor.Descriptor(
  name='ListArtifactsResponse',
  full_name='continual.rpc.management.v1.ListArtifactsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='artifacts', full_name='continual.rpc.management.v1.ListArtifactsResponse.artifacts', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListArtifactsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12621,
  serialized_end=12727,
)


_GENERATEARTIFACTUPLOADURLREQUEST = _descriptor.Descriptor(
  name='GenerateArtifactUploadURLRequest',
  full_name='continual.rpc.management.v1.GenerateArtifactUploadURLRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='artifact_id', full_name='continual.rpc.management.v1.GenerateArtifactUploadURLRequest.artifact_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.GenerateArtifactUploadURLRequest.parent', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='path', full_name='continual.rpc.management.v1.GenerateArtifactUploadURLRequest.path', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='mime_type', full_name='continual.rpc.management.v1.GenerateArtifactUploadURLRequest.mime_type', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resumable', full_name='continual.rpc.management.v1.GenerateArtifactUploadURLRequest.resumable', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run', full_name='continual.rpc.management.v1.GenerateArtifactUploadURLRequest.run', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_if_exists', full_name='continual.rpc.management.v1.GenerateArtifactUploadURLRequest.update_if_exists', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12730,
  serialized_end=12892,
)


_GENERATEARTIFACTUPLOADURLRESPONSE = _descriptor.Descriptor(
  name='GenerateArtifactUploadURLResponse',
  full_name='continual.rpc.management.v1.GenerateArtifactUploadURLResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='url', full_name='continual.rpc.management.v1.GenerateArtifactUploadURLResponse.url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='artifact', full_name='continual.rpc.management.v1.GenerateArtifactUploadURLResponse.artifact', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=12894,
  serialized_end=12999,
)


_LISTUSERINTEGRATIONPROVIDERSREQUEST = _descriptor.Descriptor(
  name='ListUserIntegrationProvidersRequest',
  full_name='continual.rpc.management.v1.ListUserIntegrationProvidersRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13001,
  serialized_end=13038,
)


_LISTUSERINTEGRATIONPROVIDERSRESPONSE = _descriptor.Descriptor(
  name='ListUserIntegrationProvidersResponse',
  full_name='continual.rpc.management.v1.ListUserIntegrationProvidersResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='providers', full_name='continual.rpc.management.v1.ListUserIntegrationProvidersResponse.providers', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13040,
  serialized_end=13151,
)


_LISTUSERINTEGRATIONSREQUEST = _descriptor.Descriptor(
  name='ListUserIntegrationsRequest',
  full_name='continual.rpc.management.v1.ListUserIntegrationsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_name', full_name='continual.rpc.management.v1.ListUserIntegrationsRequest.user_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListUserIntegrationsRequest.filters', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListUserIntegrationsRequest.ignore_filters', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListUserIntegrationsRequest.page_token', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListUserIntegrationsRequest.page_size', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='order_by', full_name='continual.rpc.management.v1.ListUserIntegrationsRequest.order_by', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='search', full_name='continual.rpc.management.v1.ListUserIntegrationsRequest.search', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='show_disconnected', full_name='continual.rpc.management.v1.ListUserIntegrationsRequest.show_disconnected', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13154,
  serialized_end=13343,
)


_LISTUSERINTEGRATIONSRESPONSE = _descriptor.Descriptor(
  name='ListUserIntegrationsResponse',
  full_name='continual.rpc.management.v1.ListUserIntegrationsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='integrations', full_name='continual.rpc.management.v1.ListUserIntegrationsResponse.integrations', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListUserIntegrationsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13345,
  serialized_end=13468,
)


_DISCONNECTUSERINTEGRATIONREQUEST = _descriptor.Descriptor(
  name='DisconnectUserIntegrationRequest',
  full_name='continual.rpc.management.v1.DisconnectUserIntegrationRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DisconnectUserIntegrationRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='reason', full_name='continual.rpc.management.v1.DisconnectUserIntegrationRequest.reason', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data', full_name='continual.rpc.management.v1.DisconnectUserIntegrationRequest.data', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13470,
  serialized_end=13548,
)


_LISTGITPROVIDERNAMESPACESREQUEST = _descriptor.Descriptor(
  name='ListGitProviderNamespacesRequest',
  full_name='continual.rpc.management.v1.ListGitProviderNamespacesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_integration_name', full_name='continual.rpc.management.v1.ListGitProviderNamespacesRequest.user_integration_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13550,
  serialized_end=13615,
)


_LISTGITPROVIDERNAMESPACESRESPONSE = _descriptor.Descriptor(
  name='ListGitProviderNamespacesResponse',
  full_name='continual.rpc.management.v1.ListGitProviderNamespacesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_integration_name', full_name='continual.rpc.management.v1.ListGitProviderNamespacesResponse.user_integration_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='namespaces', full_name='continual.rpc.management.v1.ListGitProviderNamespacesResponse.namespaces', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13618,
  serialized_end=13755,
)


_LISTGITPROVIDERREPOSITORIESREQUEST = _descriptor.Descriptor(
  name='ListGitProviderRepositoriesRequest',
  full_name='continual.rpc.management.v1.ListGitProviderRepositoriesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_integration_name', full_name='continual.rpc.management.v1.ListGitProviderRepositoriesRequest.user_integration_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='git_provider_namespace_name', full_name='continual.rpc.management.v1.ListGitProviderRepositoriesRequest.git_provider_namespace_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13757,
  serialized_end=13861,
)


_LISTGITPROVIDERREPOSITORIESRESPONSE = _descriptor.Descriptor(
  name='ListGitProviderRepositoriesResponse',
  full_name='continual.rpc.management.v1.ListGitProviderRepositoriesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='user_integration_name', full_name='continual.rpc.management.v1.ListGitProviderRepositoriesResponse.user_integration_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='git_provider_namespace_name', full_name='continual.rpc.management.v1.ListGitProviderRepositoriesResponse.git_provider_namespace_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='repositories', full_name='continual.rpc.management.v1.ListGitProviderRepositoriesResponse.repositories', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13864,
  serialized_end=14043,
)


_CONNECTPROJECTTOGITREPOSITORYREQUEST = _descriptor.Descriptor(
  name='ConnectProjectToGitRepositoryRequest',
  full_name='continual.rpc.management.v1.ConnectProjectToGitRepositoryRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='project_name', full_name='continual.rpc.management.v1.ConnectProjectToGitRepositoryRequest.project_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='user_integration_name', full_name='continual.rpc.management.v1.ConnectProjectToGitRepositoryRequest.user_integration_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='git_provider_repository_name', full_name='continual.rpc.management.v1.ConnectProjectToGitRepositoryRequest.git_provider_repository_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='default_env_branch_name', full_name='continual.rpc.management.v1.ConnectProjectToGitRepositoryRequest.default_env_branch_name', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14046,
  serialized_end=14208,
)


_CONNECTPROJECTTOGITREPOSITORYRESPONSE = _descriptor.Descriptor(
  name='ConnectProjectToGitRepositoryResponse',
  full_name='continual.rpc.management.v1.ConnectProjectToGitRepositoryResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='project_git_integration', full_name='continual.rpc.management.v1.ConnectProjectToGitRepositoryResponse.project_git_integration', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14210,
  serialized_end=14334,
)


_GETPROJECTGITINTEGRATIONREQUEST = _descriptor.Descriptor(
  name='GetProjectGitIntegrationRequest',
  full_name='continual.rpc.management.v1.GetProjectGitIntegrationRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetProjectGitIntegrationRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14336,
  serialized_end=14383,
)


_SEEDPROJECTGITREPOSITORYREQUEST = _descriptor.Descriptor(
  name='SeedProjectGitRepositoryRequest',
  full_name='continual.rpc.management.v1.SeedProjectGitRepositoryRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='project_git_integration', full_name='continual.rpc.management.v1.SeedProjectGitRepositoryRequest.project_git_integration', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='source', full_name='continual.rpc.management.v1.SeedProjectGitRepositoryRequest.source', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14385,
  serialized_end=14467,
)


_GENERATESSOADMINPORTALURLREQUEST = _descriptor.Descriptor(
  name='GenerateSsoAdminPortalURLRequest',
  full_name='continual.rpc.management.v1.GenerateSsoAdminPortalURLRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='organization', full_name='continual.rpc.management.v1.GenerateSsoAdminPortalURLRequest.organization', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14469,
  serialized_end=14525,
)


_GENERATESSOADMINPORTALURLRESPONSE = _descriptor.Descriptor(
  name='GenerateSsoAdminPortalURLResponse',
  full_name='continual.rpc.management.v1.GenerateSsoAdminPortalURLResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='url', full_name='continual.rpc.management.v1.GenerateSsoAdminPortalURLResponse.url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14527,
  serialized_end=14575,
)


_GENERATEDSYNCADMINPORTALURLREQUEST = _descriptor.Descriptor(
  name='GenerateDSyncAdminPortalURLRequest',
  full_name='continual.rpc.management.v1.GenerateDSyncAdminPortalURLRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='organization', full_name='continual.rpc.management.v1.GenerateDSyncAdminPortalURLRequest.organization', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14577,
  serialized_end=14635,
)


_GENERATEDSYNCADMINPORTALURLRESPONSE = _descriptor.Descriptor(
  name='GenerateDSyncAdminPortalURLResponse',
  full_name='continual.rpc.management.v1.GenerateDSyncAdminPortalURLResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='url', full_name='continual.rpc.management.v1.GenerateDSyncAdminPortalURLResponse.url', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14637,
  serialized_end=14687,
)


_EXCHANGETOKENREQUEST = _descriptor.Descriptor(
  name='ExchangeTokenRequest',
  full_name='continual.rpc.management.v1.ExchangeTokenRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='token', full_name='continual.rpc.management.v1.ExchangeTokenRequest.token', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14689,
  serialized_end=14726,
)


_EXCHANGETOKENRESPONSE = _descriptor.Descriptor(
  name='ExchangeTokenResponse',
  full_name='continual.rpc.management.v1.ExchangeTokenResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='token', full_name='continual.rpc.management.v1.ExchangeTokenResponse.token', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14728,
  serialized_end=14766,
)


_GETPRINCIPALREQUEST = _descriptor.Descriptor(
  name='GetPrincipalRequest',
  full_name='continual.rpc.management.v1.GetPrincipalRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetPrincipalRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14768,
  serialized_end=14803,
)


_GETAPIKEYREQUEST = _descriptor.Descriptor(
  name='GetApiKeyRequest',
  full_name='continual.rpc.management.v1.GetApiKeyRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetApiKeyRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14805,
  serialized_end=14837,
)


_BATCHGETAPIKEYSREQUEST = _descriptor.Descriptor(
  name='BatchGetApiKeysRequest',
  full_name='continual.rpc.management.v1.BatchGetApiKeysRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.BatchGetApiKeysRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetApiKeysRequest.names', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14839,
  serialized_end=14894,
)


_BATCHGETAPIKEYSRESPONSE = _descriptor.Descriptor(
  name='BatchGetApiKeysResponse',
  full_name='continual.rpc.management.v1.BatchGetApiKeysResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='api_keys', full_name='continual.rpc.management.v1.BatchGetApiKeysResponse.api_keys', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14896,
  serialized_end=14976,
)


_LISTAPIKEYSREQUEST = _descriptor.Descriptor(
  name='ListApiKeysRequest',
  full_name='continual.rpc.management.v1.ListApiKeysRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListApiKeysRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListApiKeysRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListApiKeysRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListApiKeysRequest.filters', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListApiKeysRequest.ignore_filters', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='order_by', full_name='continual.rpc.management.v1.ListApiKeysRequest.order_by', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='search', full_name='continual.rpc.management.v1.ListApiKeysRequest.search', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=14979,
  serialized_end=15129,
)


_LISTAPIKEYSRESPONSE = _descriptor.Descriptor(
  name='ListApiKeysResponse',
  full_name='continual.rpc.management.v1.ListApiKeysResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='api_keys', full_name='continual.rpc.management.v1.ListApiKeysResponse.api_keys', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListApiKeysResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15131,
  serialized_end=15232,
)


_CREATEAPIKEYREQUEST = _descriptor.Descriptor(
  name='CreateApiKeyRequest',
  full_name='continual.rpc.management.v1.CreateApiKeyRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.CreateApiKeyRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='api_key', full_name='continual.rpc.management.v1.CreateApiKeyRequest.api_key', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15234,
  serialized_end=15325,
)


_UPDATEAPIKEYREQUEST = _descriptor.Descriptor(
  name='UpdateApiKeyRequest',
  full_name='continual.rpc.management.v1.UpdateApiKeyRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='api_key', full_name='continual.rpc.management.v1.UpdateApiKeyRequest.api_key', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_mask', full_name='continual.rpc.management.v1.UpdateApiKeyRequest.update_mask', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15327,
  serialized_end=15451,
)


_REVOKEAPIKEYREQUEST = _descriptor.Descriptor(
  name='RevokeApiKeyRequest',
  full_name='continual.rpc.management.v1.RevokeApiKeyRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.RevokeApiKeyRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15453,
  serialized_end=15488,
)


_GETAPIKEYPROJECTREQUEST = _descriptor.Descriptor(
  name='GetApiKeyProjectRequest',
  full_name='continual.rpc.management.v1.GetApiKeyProjectRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetApiKeyProjectRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15490,
  serialized_end=15529,
)


_GETRUNREQUEST = _descriptor.Descriptor(
  name='GetRunRequest',
  full_name='continual.rpc.management.v1.GetRunRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetRunRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15531,
  serialized_end=15560,
)


_BATCHGETRUNSREQUEST = _descriptor.Descriptor(
  name='BatchGetRunsRequest',
  full_name='continual.rpc.management.v1.BatchGetRunsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetRunsRequest.names', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15562,
  serialized_end=15598,
)


_BATCHGETRUNSRESPONSE = _descriptor.Descriptor(
  name='BatchGetRunsResponse',
  full_name='continual.rpc.management.v1.BatchGetRunsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='runs', full_name='continual.rpc.management.v1.BatchGetRunsResponse.runs', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15600,
  serialized_end=15670,
)


_LISTRUNSREQUEST = _descriptor.Descriptor(
  name='ListRunsRequest',
  full_name='continual.rpc.management.v1.ListRunsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListRunsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListRunsRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListRunsRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListRunsRequest.filters', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListRunsRequest.ignore_filters', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListRunsRequest.start_time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListRunsRequest.end_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='order_by', full_name='continual.rpc.management.v1.ListRunsRequest.order_by', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='search', full_name='continual.rpc.management.v1.ListRunsRequest.search', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest', full_name='continual.rpc.management.v1.ListRunsRequest.latest', index=9,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15673,
  serialized_end=15930,
)


_LISTRUNSRESPONSE = _descriptor.Descriptor(
  name='ListRunsResponse',
  full_name='continual.rpc.management.v1.ListRunsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='runs', full_name='continual.rpc.management.v1.ListRunsResponse.runs', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListRunsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15932,
  serialized_end=16023,
)


_CREATERUNREQUEST = _descriptor.Descriptor(
  name='CreateRunRequest',
  full_name='continual.rpc.management.v1.CreateRunRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.CreateRunRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run_id', full_name='continual.rpc.management.v1.CreateRunRequest.run_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='continual.rpc.management.v1.CreateRunRequest.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='heartbeat_interval', full_name='continual.rpc.management.v1.CreateRunRequest.heartbeat_interval', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='get_if_exists', full_name='continual.rpc.management.v1.CreateRunRequest.get_if_exists', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16025,
  serialized_end=16147,
)


_UPDATERUNREQUEST = _descriptor.Descriptor(
  name='UpdateRunRequest',
  full_name='continual.rpc.management.v1.UpdateRunRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='run', full_name='continual.rpc.management.v1.UpdateRunRequest.run', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_paths', full_name='continual.rpc.management.v1.UpdateRunRequest.update_paths', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16149,
  serialized_end=16236,
)


_DELETERUNREQUEST = _descriptor.Descriptor(
  name='DeleteRunRequest',
  full_name='continual.rpc.management.v1.DeleteRunRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DeleteRunRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16238,
  serialized_end=16270,
)


_GETRUNSUMMARYREQUEST = _descriptor.Descriptor(
  name='GetRunSummaryRequest',
  full_name='continual.rpc.management.v1.GetRunSummaryRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetRunSummaryRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16272,
  serialized_end=16308,
)


_GETRUNCHECKSUMMARYREQUEST = _descriptor.Descriptor(
  name='GetRunCheckSummaryRequest',
  full_name='continual.rpc.management.v1.GetRunCheckSummaryRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetRunCheckSummaryRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16310,
  serialized_end=16351,
)


_LOGRUNHEARTBEATREQUEST = _descriptor.Descriptor(
  name='LogRunHeartbeatRequest',
  full_name='continual.rpc.management.v1.LogRunHeartbeatRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.LogRunHeartbeatRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16353,
  serialized_end=16391,
)


_GETMETRICREQUEST = _descriptor.Descriptor(
  name='GetMetricRequest',
  full_name='continual.rpc.management.v1.GetMetricRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetMetricRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max_values', full_name='continual.rpc.management.v1.GetMetricRequest.max_values', index=1,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16393,
  serialized_end=16445,
)


_BATCHGETMETRICSREQUEST = _descriptor.Descriptor(
  name='BatchGetMetricsRequest',
  full_name='continual.rpc.management.v1.BatchGetMetricsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.BatchGetMetricsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetMetricsRequest.names', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max_values', full_name='continual.rpc.management.v1.BatchGetMetricsRequest.max_values', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16447,
  serialized_end=16522,
)


_BATCHGETMETRICSRESPONSE = _descriptor.Descriptor(
  name='BatchGetMetricsResponse',
  full_name='continual.rpc.management.v1.BatchGetMetricsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metrics', full_name='continual.rpc.management.v1.BatchGetMetricsResponse.metrics', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16524,
  serialized_end=16603,
)


_LISTMETRICSREQUEST = _descriptor.Descriptor(
  name='ListMetricsRequest',
  full_name='continual.rpc.management.v1.ListMetricsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListMetricsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListMetricsRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListMetricsRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListMetricsRequest.filters', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListMetricsRequest.ignore_filters', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListMetricsRequest.start_time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListMetricsRequest.end_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='order_by', full_name='continual.rpc.management.v1.ListMetricsRequest.order_by', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='search', full_name='continual.rpc.management.v1.ListMetricsRequest.search', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest', full_name='continual.rpc.management.v1.ListMetricsRequest.latest', index=9,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max_values', full_name='continual.rpc.management.v1.ListMetricsRequest.max_values', index=10,
      number=11, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16606,
  serialized_end=16886,
)


_LISTMETRICSRESPONSE = _descriptor.Descriptor(
  name='ListMetricsResponse',
  full_name='continual.rpc.management.v1.ListMetricsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metrics', full_name='continual.rpc.management.v1.ListMetricsResponse.metrics', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListMetricsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16888,
  serialized_end=16988,
)


_CREATEMETRICREQUEST = _descriptor.Descriptor(
  name='CreateMetricRequest',
  full_name='continual.rpc.management.v1.CreateMetricRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metric', full_name='continual.rpc.management.v1.CreateMetricRequest.metric', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.CreateMetricRequest.parent', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metric_id', full_name='continual.rpc.management.v1.CreateMetricRequest.metric_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_if_exists', full_name='continual.rpc.management.v1.CreateMetricRequest.update_if_exists', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=16991,
  serialized_end=17126,
)


_LOGMETRICREQUEST = _descriptor.Descriptor(
  name='LogMetricRequest',
  full_name='continual.rpc.management.v1.LogMetricRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.LogMetricRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.LogMetricRequest.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17128,
  serialized_end=17219,
)


_UPDATEMETRICREQUEST = _descriptor.Descriptor(
  name='UpdateMetricRequest',
  full_name='continual.rpc.management.v1.UpdateMetricRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metric', full_name='continual.rpc.management.v1.UpdateMetricRequest.metric', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_mask', full_name='continual.rpc.management.v1.UpdateMetricRequest.update_mask', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17221,
  serialized_end=17344,
)


_GETMETADATAREQUEST = _descriptor.Descriptor(
  name='GetMetadataRequest',
  full_name='continual.rpc.management.v1.GetMetadataRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetMetadataRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17346,
  serialized_end=17380,
)


_BATCHGETMETADATAREQUEST = _descriptor.Descriptor(
  name='BatchGetMetadataRequest',
  full_name='continual.rpc.management.v1.BatchGetMetadataRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetMetadataRequest.names', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17382,
  serialized_end=17422,
)


_BATCHGETMETADATARESPONSE = _descriptor.Descriptor(
  name='BatchGetMetadataResponse',
  full_name='continual.rpc.management.v1.BatchGetMetadataResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metadata', full_name='continual.rpc.management.v1.BatchGetMetadataResponse.metadata', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17424,
  serialized_end=17507,
)


_LISTMETADATAREQUEST = _descriptor.Descriptor(
  name='ListMetadataRequest',
  full_name='continual.rpc.management.v1.ListMetadataRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListMetadataRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListMetadataRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListMetadataRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListMetadataRequest.filters', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListMetadataRequest.ignore_filters', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListMetadataRequest.start_time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListMetadataRequest.end_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='order_by', full_name='continual.rpc.management.v1.ListMetadataRequest.order_by', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='search', full_name='continual.rpc.management.v1.ListMetadataRequest.search', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest', full_name='continual.rpc.management.v1.ListMetadataRequest.latest', index=9,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17510,
  serialized_end=17771,
)


_LISTMETADATARESPONSE = _descriptor.Descriptor(
  name='ListMetadataResponse',
  full_name='continual.rpc.management.v1.ListMetadataResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metadata', full_name='continual.rpc.management.v1.ListMetadataResponse.metadata', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListMetadataResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17773,
  serialized_end=17877,
)


_CREATEMETADATAREQUEST = _descriptor.Descriptor(
  name='CreateMetadataRequest',
  full_name='continual.rpc.management.v1.CreateMetadataRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.CreateMetadataRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metadata', full_name='continual.rpc.management.v1.CreateMetadataRequest.metadata', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='metadata_id', full_name='continual.rpc.management.v1.CreateMetadataRequest.metadata_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_if_exists', full_name='continual.rpc.management.v1.CreateMetadataRequest.update_if_exists', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17880,
  serialized_end=18023,
)


_UPDATEMETADATAREQUEST = _descriptor.Descriptor(
  name='UpdateMetadataRequest',
  full_name='continual.rpc.management.v1.UpdateMetadataRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='metadata', full_name='continual.rpc.management.v1.UpdateMetadataRequest.metadata', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18025,
  serialized_end=18105,
)


_DELETEMETADATAREQUEST = _descriptor.Descriptor(
  name='DeleteMetadataRequest',
  full_name='continual.rpc.management.v1.DeleteMetadataRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DeleteMetadataRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18107,
  serialized_end=18144,
)


_GETNOTIFICATIONPOLICYREQUEST = _descriptor.Descriptor(
  name='GetNotificationPolicyRequest',
  full_name='continual.rpc.management.v1.GetNotificationPolicyRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetNotificationPolicyRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18146,
  serialized_end=18190,
)


_BATCHGETNOTIFICATIONPOLICIESREQUEST = _descriptor.Descriptor(
  name='BatchGetNotificationPoliciesRequest',
  full_name='continual.rpc.management.v1.BatchGetNotificationPoliciesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetNotificationPoliciesRequest.names', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18192,
  serialized_end=18244,
)


_BATCHGETNOTIFICATIONPOLICIESRESPONSE = _descriptor.Descriptor(
  name='BatchGetNotificationPoliciesResponse',
  full_name='continual.rpc.management.v1.BatchGetNotificationPoliciesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='notification_policies', full_name='continual.rpc.management.v1.BatchGetNotificationPoliciesResponse.notification_policies', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18246,
  serialized_end=18364,
)


_LISTNOTIFICATIONPOLICIESREQUEST = _descriptor.Descriptor(
  name='ListNotificationPoliciesRequest',
  full_name='continual.rpc.management.v1.ListNotificationPoliciesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListNotificationPoliciesRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListNotificationPoliciesRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListNotificationPoliciesRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListNotificationPoliciesRequest.filters', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListNotificationPoliciesRequest.ignore_filters', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListNotificationPoliciesRequest.start_time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListNotificationPoliciesRequest.end_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='order_by', full_name='continual.rpc.management.v1.ListNotificationPoliciesRequest.order_by', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='search', full_name='continual.rpc.management.v1.ListNotificationPoliciesRequest.search', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18367,
  serialized_end=18624,
)


_LISTNOTIFICATIONPOLICIESRESPONSE = _descriptor.Descriptor(
  name='ListNotificationPoliciesResponse',
  full_name='continual.rpc.management.v1.ListNotificationPoliciesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='notification_policies', full_name='continual.rpc.management.v1.ListNotificationPoliciesResponse.notification_policies', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListNotificationPoliciesResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18627,
  serialized_end=18766,
)


_CREATENOTIFICATIONPOLICYREQUEST = _descriptor.Descriptor(
  name='CreateNotificationPolicyRequest',
  full_name='continual.rpc.management.v1.CreateNotificationPolicyRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.CreateNotificationPolicyRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='notification_policy', full_name='continual.rpc.management.v1.CreateNotificationPolicyRequest.notification_policy', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18768,
  serialized_end=18895,
)


_UPDATENOTIFICATIONPOLICYREQUEST = _descriptor.Descriptor(
  name='UpdateNotificationPolicyRequest',
  full_name='continual.rpc.management.v1.UpdateNotificationPolicyRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='notification_policy', full_name='continual.rpc.management.v1.UpdateNotificationPolicyRequest.notification_policy', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_mask', full_name='continual.rpc.management.v1.UpdateNotificationPolicyRequest.update_mask', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18898,
  serialized_end=19058,
)


_DELETENOTIFICATIONPOLICYREQUEST = _descriptor.Descriptor(
  name='DeleteNotificationPolicyRequest',
  full_name='continual.rpc.management.v1.DeleteNotificationPolicyRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DeleteNotificationPolicyRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19060,
  serialized_end=19107,
)


_GETTAGREQUEST = _descriptor.Descriptor(
  name='GetTagRequest',
  full_name='continual.rpc.management.v1.GetTagRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.GetTagRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='key', full_name='continual.rpc.management.v1.GetTagRequest.key', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19109,
  serialized_end=19153,
)


_BATCHGETTAGSREQUEST = _descriptor.Descriptor(
  name='BatchGetTagsRequest',
  full_name='continual.rpc.management.v1.BatchGetTagsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetTagsRequest.names', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19155,
  serialized_end=19191,
)


_BATCHGETTAGSRESPONSE = _descriptor.Descriptor(
  name='BatchGetTagsResponse',
  full_name='continual.rpc.management.v1.BatchGetTagsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='tags', full_name='continual.rpc.management.v1.BatchGetTagsResponse.tags', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19193,
  serialized_end=19263,
)


_LISTTAGSREQUEST = _descriptor.Descriptor(
  name='ListTagsRequest',
  full_name='continual.rpc.management.v1.ListTagsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListTagsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListTagsRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListTagsRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListTagsRequest.filters', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListTagsRequest.ignore_filters', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListTagsRequest.start_time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListTagsRequest.end_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='order_by', full_name='continual.rpc.management.v1.ListTagsRequest.order_by', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='search', full_name='continual.rpc.management.v1.ListTagsRequest.search', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest', full_name='continual.rpc.management.v1.ListTagsRequest.latest', index=9,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19266,
  serialized_end=19523,
)


_LISTTAGSRESPONSE = _descriptor.Descriptor(
  name='ListTagsResponse',
  full_name='continual.rpc.management.v1.ListTagsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='tags', full_name='continual.rpc.management.v1.ListTagsResponse.tags', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListTagsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19525,
  serialized_end=19616,
)


_CREATETAGREQUEST = _descriptor.Descriptor(
  name='CreateTagRequest',
  full_name='continual.rpc.management.v1.CreateTagRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.CreateTagRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='key', full_name='continual.rpc.management.v1.CreateTagRequest.key', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='continual.rpc.management.v1.CreateTagRequest.value', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19618,
  serialized_end=19680,
)


_UPDATETAGREQUEST = _descriptor.Descriptor(
  name='UpdateTagRequest',
  full_name='continual.rpc.management.v1.UpdateTagRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='tag', full_name='continual.rpc.management.v1.UpdateTagRequest.tag', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_mask', full_name='continual.rpc.management.v1.UpdateTagRequest.update_mask', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19682,
  serialized_end=19796,
)


_DELETETAGREQUEST = _descriptor.Descriptor(
  name='DeleteTagRequest',
  full_name='continual.rpc.management.v1.DeleteTagRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DeleteTagRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19798,
  serialized_end=19830,
)


_GETDATAPROFILEREQUEST = _descriptor.Descriptor(
  name='GetDataProfileRequest',
  full_name='continual.rpc.management.v1.GetDataProfileRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetDataProfileRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19832,
  serialized_end=19869,
)


_CREATEDATAPROFILEREQUEST = _descriptor.Descriptor(
  name='CreateDataProfileRequest',
  full_name='continual.rpc.management.v1.CreateDataProfileRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.CreateDataProfileRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data_profile', full_name='continual.rpc.management.v1.CreateDataProfileRequest.data_profile', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run', full_name='continual.rpc.management.v1.CreateDataProfileRequest.run', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19871,
  serialized_end=19990,
)


_LISTDATAPROFILESREQUEST = _descriptor.Descriptor(
  name='ListDataProfilesRequest',
  full_name='continual.rpc.management.v1.ListDataProfilesRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListDataProfilesRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListDataProfilesRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListDataProfilesRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListDataProfilesRequest.filters', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListDataProfilesRequest.ignore_filters', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='order_by', full_name='continual.rpc.management.v1.ListDataProfilesRequest.order_by', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='search', full_name='continual.rpc.management.v1.ListDataProfilesRequest.search', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest', full_name='continual.rpc.management.v1.ListDataProfilesRequest.latest', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=19993,
  serialized_end=20164,
)


_LISTDATAPROFILESRESPONSE = _descriptor.Descriptor(
  name='ListDataProfilesResponse',
  full_name='continual.rpc.management.v1.ListDataProfilesResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='data_profiles', full_name='continual.rpc.management.v1.ListDataProfilesResponse.data_profiles', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListDataProfilesResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20166,
  serialized_end=20282,
)


_UPDATEDATAPROFILEREQUEST = _descriptor.Descriptor(
  name='UpdateDataProfileRequest',
  full_name='continual.rpc.management.v1.UpdateDataProfileRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='data_profile', full_name='continual.rpc.management.v1.UpdateDataProfileRequest.data_profile', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='update_paths', full_name='continual.rpc.management.v1.UpdateDataProfileRequest.update_paths', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run', full_name='continual.rpc.management.v1.UpdateDataProfileRequest.run', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20284,
  serialized_end=20409,
)


_GETRESOURCECHECKREQUEST = _descriptor.Descriptor(
  name='GetResourceCheckRequest',
  full_name='continual.rpc.management.v1.GetResourceCheckRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetResourceCheckRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20411,
  serialized_end=20450,
)


_LISTRESOURCECHECKSREQUEST = _descriptor.Descriptor(
  name='ListResourceChecksRequest',
  full_name='continual.rpc.management.v1.ListResourceChecksRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListResourceChecksRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListResourceChecksRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListResourceChecksRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListResourceChecksRequest.filters', index=3,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListResourceChecksRequest.ignore_filters', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='order_by', full_name='continual.rpc.management.v1.ListResourceChecksRequest.order_by', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='search', full_name='continual.rpc.management.v1.ListResourceChecksRequest.search', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest', full_name='continual.rpc.management.v1.ListResourceChecksRequest.latest', index=7,
      number=8, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20453,
  serialized_end=20626,
)


_LISTRESOURCECHECKSRESPONSE = _descriptor.Descriptor(
  name='ListResourceChecksResponse',
  full_name='continual.rpc.management.v1.ListResourceChecksResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='resource_checks', full_name='continual.rpc.management.v1.ListResourceChecksResponse.resource_checks', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListResourceChecksResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20628,
  serialized_end=20750,
)


_CREATERESOURCECHECKREQUEST = _descriptor.Descriptor(
  name='CreateResourceCheckRequest',
  full_name='continual.rpc.management.v1.CreateResourceCheckRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.CreateResourceCheckRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource_check', full_name='continual.rpc.management.v1.CreateResourceCheckRequest.resource_check', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20752,
  serialized_end=20864,
)


_CREATEDATASETREQUEST = _descriptor.Descriptor(
  name='CreateDatasetRequest',
  full_name='continual.rpc.management.v1.CreateDatasetRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.CreateDatasetRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='continual.rpc.management.v1.CreateDatasetRequest.description', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='get_if_exists', full_name='continual.rpc.management.v1.CreateDatasetRequest.get_if_exists', index=2,
      number=3, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run', full_name='continual.rpc.management.v1.CreateDatasetRequest.run', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20866,
  serialized_end=20959,
)


_GETDATASETREQUEST = _descriptor.Descriptor(
  name='GetDatasetRequest',
  full_name='continual.rpc.management.v1.GetDatasetRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetDatasetRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20961,
  serialized_end=20994,
)


_BATCHGETDATASETSREQUEST = _descriptor.Descriptor(
  name='BatchGetDatasetsRequest',
  full_name='continual.rpc.management.v1.BatchGetDatasetsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.BatchGetDatasetsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetDatasetsRequest.names', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=20996,
  serialized_end=21052,
)


_BATCHGETDATASETSRESPONSE = _descriptor.Descriptor(
  name='BatchGetDatasetsResponse',
  full_name='continual.rpc.management.v1.BatchGetDatasetsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='datasets', full_name='continual.rpc.management.v1.BatchGetDatasetsResponse.datasets', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21054,
  serialized_end=21136,
)


_LISTDATASETSREQUEST = _descriptor.Descriptor(
  name='ListDatasetsRequest',
  full_name='continual.rpc.management.v1.ListDatasetsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListDatasetsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListDatasetsRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListDatasetsRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest', full_name='continual.rpc.management.v1.ListDatasetsRequest.latest', index=3,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListDatasetsRequest.filters', index=4,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListDatasetsRequest.ignore_filters', index=5,
      number=7, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListDatasetsRequest.start_time', index=6,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListDatasetsRequest.end_time', index=7,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='all_projects', full_name='continual.rpc.management.v1.ListDatasetsRequest.all_projects', index=8,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='order_by', full_name='continual.rpc.management.v1.ListDatasetsRequest.order_by', index=9,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='search', full_name='continual.rpc.management.v1.ListDatasetsRequest.search', index=10,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21139,
  serialized_end=21422,
)


_LISTDATASETSRESPONSE = _descriptor.Descriptor(
  name='ListDatasetsResponse',
  full_name='continual.rpc.management.v1.ListDatasetsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='datasets', full_name='continual.rpc.management.v1.ListDatasetsResponse.datasets', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListDatasetsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21424,
  serialized_end=21527,
)


_DELETEDATASETREQUEST = _descriptor.Descriptor(
  name='DeleteDatasetRequest',
  full_name='continual.rpc.management.v1.DeleteDatasetRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DeleteDatasetRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run', full_name='continual.rpc.management.v1.DeleteDatasetRequest.run', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21529,
  serialized_end=21578,
)


_CREATEDATASETVERSIONREQUEST = _descriptor.Descriptor(
  name='CreateDatasetVersionRequest',
  full_name='continual.rpc.management.v1.CreateDatasetVersionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.CreateDatasetVersionRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='run', full_name='continual.rpc.management.v1.CreateDatasetVersionRequest.run', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21580,
  serialized_end=21638,
)


_GETDATASETVERSIONREQUEST = _descriptor.Descriptor(
  name='GetDatasetVersionRequest',
  full_name='continual.rpc.management.v1.GetDatasetVersionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetDatasetVersionRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21640,
  serialized_end=21680,
)


_BATCHGETDATASETVERSIONSREQUEST = _descriptor.Descriptor(
  name='BatchGetDatasetVersionsRequest',
  full_name='continual.rpc.management.v1.BatchGetDatasetVersionsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.BatchGetDatasetVersionsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='names', full_name='continual.rpc.management.v1.BatchGetDatasetVersionsRequest.names', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21682,
  serialized_end=21745,
)


_BATCHGETDATASETVERSIONSRESPONSE = _descriptor.Descriptor(
  name='BatchGetDatasetVersionsResponse',
  full_name='continual.rpc.management.v1.BatchGetDatasetVersionsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='dataset_versions', full_name='continual.rpc.management.v1.BatchGetDatasetVersionsResponse.dataset_versions', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21747,
  serialized_end=21851,
)


_LISTDATASETVERSIONSREQUEST = _descriptor.Descriptor(
  name='ListDatasetVersionsRequest',
  full_name='continual.rpc.management.v1.ListDatasetVersionsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListDatasetVersionsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListDatasetVersionsRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListDatasetVersionsRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest', full_name='continual.rpc.management.v1.ListDatasetVersionsRequest.latest', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListDatasetVersionsRequest.filters', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListDatasetVersionsRequest.ignore_filters', index=5,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListDatasetVersionsRequest.start_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListDatasetVersionsRequest.end_time', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='all_projects', full_name='continual.rpc.management.v1.ListDatasetVersionsRequest.all_projects', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='order_by', full_name='continual.rpc.management.v1.ListDatasetVersionsRequest.order_by', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='search', full_name='continual.rpc.management.v1.ListDatasetVersionsRequest.search', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=21854,
  serialized_end=22144,
)


_LISTDATASETVERSIONSRESPONSE = _descriptor.Descriptor(
  name='ListDatasetVersionsResponse',
  full_name='continual.rpc.management.v1.ListDatasetVersionsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='dataset_versions', full_name='continual.rpc.management.v1.ListDatasetVersionsResponse.dataset_versions', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListDatasetVersionsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22146,
  serialized_end=22271,
)


_CREATEDATASETVERSIONASSIGNMENTREQUEST = _descriptor.Descriptor(
  name='CreateDatasetVersionAssignmentRequest',
  full_name='continual.rpc.management.v1.CreateDatasetVersionAssignmentRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='dataset_version_name', full_name='continual.rpc.management.v1.CreateDatasetVersionAssignmentRequest.dataset_version_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='resource_name', full_name='continual.rpc.management.v1.CreateDatasetVersionAssignmentRequest.resource_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22273,
  serialized_end=22365,
)


_GETDATASETVERSIONASSIGNMENTREQUEST = _descriptor.Descriptor(
  name='GetDatasetVersionAssignmentRequest',
  full_name='continual.rpc.management.v1.GetDatasetVersionAssignmentRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.GetDatasetVersionAssignmentRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22367,
  serialized_end=22417,
)


_DELETEDATASETVERSIONASSIGNMENTREQUEST = _descriptor.Descriptor(
  name='DeleteDatasetVersionAssignmentRequest',
  full_name='continual.rpc.management.v1.DeleteDatasetVersionAssignmentRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='continual.rpc.management.v1.DeleteDatasetVersionAssignmentRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22419,
  serialized_end=22472,
)


_LISTDATASETVERSIONASSIGNMENTSREQUEST = _descriptor.Descriptor(
  name='ListDatasetVersionAssignmentsRequest',
  full_name='continual.rpc.management.v1.ListDatasetVersionAssignmentsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='continual.rpc.management.v1.ListDatasetVersionAssignmentsRequest.parent', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_size', full_name='continual.rpc.management.v1.ListDatasetVersionAssignmentsRequest.page_size', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='page_token', full_name='continual.rpc.management.v1.ListDatasetVersionAssignmentsRequest.page_token', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='latest', full_name='continual.rpc.management.v1.ListDatasetVersionAssignmentsRequest.latest', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='filters', full_name='continual.rpc.management.v1.ListDatasetVersionAssignmentsRequest.filters', index=4,
      number=5, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ignore_filters', full_name='continual.rpc.management.v1.ListDatasetVersionAssignmentsRequest.ignore_filters', index=5,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='continual.rpc.management.v1.ListDatasetVersionAssignmentsRequest.start_time', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='continual.rpc.management.v1.ListDatasetVersionAssignmentsRequest.end_time', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='all_projects', full_name='continual.rpc.management.v1.ListDatasetVersionAssignmentsRequest.all_projects', index=8,
      number=9, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='order_by', full_name='continual.rpc.management.v1.ListDatasetVersionAssignmentsRequest.order_by', index=9,
      number=10, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='search', full_name='continual.rpc.management.v1.ListDatasetVersionAssignmentsRequest.search', index=10,
      number=11, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22475,
  serialized_end=22775,
)


_LISTDATASETVERSIONASSIGNMENTSRESPONSE = _descriptor.Descriptor(
  name='ListDatasetVersionAssignmentsResponse',
  full_name='continual.rpc.management.v1.ListDatasetVersionAssignmentsResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='assignments', full_name='continual.rpc.management.v1.ListDatasetVersionAssignmentsResponse.assignments', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='next_page_token', full_name='continual.rpc.management.v1.ListDatasetVersionAssignmentsResponse.next_page_token', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22778,
  serialized_end=22918,
)


_GETSERVERVERSIONREQUEST = _descriptor.Descriptor(
  name='GetServerVersionRequest',
  full_name='continual.rpc.management.v1.GetServerVersionRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='client_version', full_name='continual.rpc.management.v1.GetServerVersionRequest.client_version', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22920,
  serialized_end=22969,
)


_GETSERVERVERSIONRESPONSE = _descriptor.Descriptor(
  name='GetServerVersionResponse',
  full_name='continual.rpc.management.v1.GetServerVersionResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='server_version', full_name='continual.rpc.management.v1.GetServerVersionResponse.server_version', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='upgrade_required', full_name='continual.rpc.management.v1.GetServerVersionResponse.upgrade_required', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22971,
  serialized_end=23047,
)

_UPDATEUSERREQUEST.fields_by_name['user'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._USER
_UPDATEUSERREQUEST.fields_by_name['update_mask'].message_type = google_dot_protobuf_dot_field__mask__pb2._FIELDMASK
_BATCHGETUSERSRESPONSE.fields_by_name['users'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._USER
_LISTUSERSRESPONSE.fields_by_name['users'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._USER
_REGISTERRESPONSE.fields_by_name['user'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._USER
_LOGINRESPONSE.fields_by_name['user'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._USER
_LISTSESSIONSRESPONSE.fields_by_name['sessions'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._SESSION
_LISTROLESRESPONSE.fields_by_name['roles'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ROLE
_BATCHGETROLESRESPONSE.fields_by_name['roles'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ROLE
_LISTACCESSPOLICIESRESPONSE.fields_by_name['access_policies'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ACCESSPOLICY
_CREATEACCESSPOLICYREQUEST.fields_by_name['access_policy'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ACCESSPOLICY
_UPDATEACCESSPOLICYREQUEST.fields_by_name['access_policy'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ACCESSPOLICY
_UPDATEACCESSPOLICYREQUEST.fields_by_name['update_mask'].message_type = google_dot_protobuf_dot_field__mask__pb2._FIELDMASK
_CREATEINVITEREQUEST.fields_by_name['invite'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._INVITE
_LISTINVITESRESPONSE.fields_by_name['invites'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._INVITE
_UPDATEINVITEREQUEST.fields_by_name['invite'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._INVITE
_UPDATEINVITEREQUEST.fields_by_name['update_mask'].message_type = google_dot_protobuf_dot_field__mask__pb2._FIELDMASK
_CREATEORGANIZATIONREQUEST.fields_by_name['organization'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ORGANIZATION
_UPDATEORGANIZATIONREQUEST.fields_by_name['organization'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ORGANIZATION
_UPDATEORGANIZATIONREQUEST.fields_by_name['update_mask'].message_type = google_dot_protobuf_dot_field__mask__pb2._FIELDMASK
_BATCHGETORGANIZATIONSRESPONSE.fields_by_name['organizations'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ORGANIZATION
_LISTORGANIZATIONSRESPONSE.fields_by_name['organizations'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ORGANIZATION
_UPDATEONBOARDINGREQUEST.fields_by_name['onboarding'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ONBOARDING
_UPDATEONBOARDINGREQUEST.fields_by_name['update_mask'].message_type = google_dot_protobuf_dot_field__mask__pb2._FIELDMASK
_CREATEPROJECTREQUEST.fields_by_name['project'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PROJECT
_UPDATEPROJECTREQUEST.fields_by_name['project'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PROJECT
_UPDATEPROJECTREQUEST.fields_by_name['update_mask'].message_type = google_dot_protobuf_dot_field__mask__pb2._FIELDMASK
_BATCHGETPROJECTSRESPONSE.fields_by_name['projects'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PROJECT
_LISTPROJECTSREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTPROJECTSREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTPROJECTSRESPONSE.fields_by_name['projects'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PROJECT
_GETUSAGEREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_GETUSAGEREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_GETUSAGERESPONSE.fields_by_name['credits_per_day'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._CREDITUSAGERECORD
_LISTPRICESRESPONSE.fields_by_name['prices'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PRICE
_LISTINVOICESRESPONSE.fields_by_name['invoices'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._INVOICE
_UPDATEBILLINGACCOUNTREQUEST.fields_by_name['billing_account'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._BILLINGACCOUNT
_UPDATEBILLINGACCOUNTREQUEST.fields_by_name['update_mask'].message_type = google_dot_protobuf_dot_field__mask__pb2._FIELDMASK
_UPDATESUBSCRIPTIONREQUEST.fields_by_name['subscription'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._SUBSCRIPTION
_UPDATESUBSCRIPTIONREQUEST.fields_by_name['update_mask'].message_type = google_dot_protobuf_dot_field__mask__pb2._FIELDMASK
_LISTPROMOTIONSREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTPROMOTIONSREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTPROMOTIONSRESPONSE.fields_by_name['promotions'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PROMOTION
_BATCHGETPROMOTIONSRESPONSE.fields_by_name['promotions'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PROMOTION
_CREATEEVENTREQUEST.fields_by_name['event'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._EVENT
_LISTEVENTSREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTEVENTSREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTEVENTSRESPONSE.fields_by_name['events'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._EVENT
_BATCHGETEVENTSRESPONSE.fields_by_name['events'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._EVENT
_BATCHGETMODELSRESPONSE.fields_by_name['models'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._MODEL
_LISTMODELSREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTMODELSREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTMODELSRESPONSE.fields_by_name['models'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._MODEL
_BATCHGETMODELVERSIONSRESPONSE.fields_by_name['model_versions'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._MODELVERSION
_LISTMODELVERSIONSREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTMODELVERSIONSREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTMODELVERSIONSRESPONSE.fields_by_name['model_versions'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._MODELVERSION
_BATCHGETEXPERIMENTSRESPONSE.fields_by_name['experiments'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._EXPERIMENT
_LISTEXPERIMENTSREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTEXPERIMENTSREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTEXPERIMENTSRESPONSE.fields_by_name['experiments'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._EXPERIMENT
_LISTBATCHPREDICTIONSREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTBATCHPREDICTIONSREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTBATCHPREDICTIONSRESPONSE.fields_by_name['batch_predictions'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._BATCHPREDICTION
_BATCHGETBATCHPREDICTIONSRESPONSE.fields_by_name['batch_predictions'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._BATCHPREDICTION
_UPDATEENVIRONMENTREQUEST.fields_by_name['environment'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ENVIRONMENT
_UPDATEENVIRONMENTREQUEST.fields_by_name['update_mask'].message_type = google_dot_protobuf_dot_field__mask__pb2._FIELDMASK
_BATCHGETENVIRONMENTSRESPONSE.fields_by_name['environments'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ENVIRONMENT
_LISTENVIRONMENTSREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTENVIRONMENTSREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTENVIRONMENTSRESPONSE.fields_by_name['environments'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ENVIRONMENT
_GETORGANIZATIONFEATUREFLAGSRESPONSE_FEATUREFLAGSENTRY.containing_type = _GETORGANIZATIONFEATUREFLAGSRESPONSE
_GETORGANIZATIONFEATUREFLAGSRESPONSE.fields_by_name['feature_flags'].message_type = _GETORGANIZATIONFEATUREFLAGSRESPONSE_FEATUREFLAGSENTRY
_CREATEARTIFACTREQUEST.fields_by_name['artifact'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ARTIFACT
_BATCHGETARTIFACTSRESPONSE.fields_by_name['artifacts'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ARTIFACT
_LISTARTIFACTSRESPONSE.fields_by_name['artifacts'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ARTIFACT
_GENERATEARTIFACTUPLOADURLRESPONSE.fields_by_name['artifact'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ARTIFACT
_LISTUSERINTEGRATIONPROVIDERSRESPONSE.fields_by_name['providers'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._USERINTEGRATIONPROVIDER
_LISTUSERINTEGRATIONSRESPONSE.fields_by_name['integrations'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._USERINTEGRATION
_LISTGITPROVIDERNAMESPACESRESPONSE.fields_by_name['namespaces'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._GITPROVIDERNAMESPACE
_LISTGITPROVIDERREPOSITORIESRESPONSE.fields_by_name['repositories'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._GITPROVIDERREPOSITORY
_CONNECTPROJECTTOGITREPOSITORYRESPONSE.fields_by_name['project_git_integration'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PROJECTGITINTEGRATION
_BATCHGETAPIKEYSRESPONSE.fields_by_name['api_keys'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._APIKEY
_LISTAPIKEYSRESPONSE.fields_by_name['api_keys'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._APIKEY
_CREATEAPIKEYREQUEST.fields_by_name['api_key'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._APIKEY
_UPDATEAPIKEYREQUEST.fields_by_name['api_key'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._APIKEY
_UPDATEAPIKEYREQUEST.fields_by_name['update_mask'].message_type = google_dot_protobuf_dot_field__mask__pb2._FIELDMASK
_BATCHGETRUNSRESPONSE.fields_by_name['runs'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._RUN
_LISTRUNSREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTRUNSREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTRUNSRESPONSE.fields_by_name['runs'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._RUN
_UPDATERUNREQUEST.fields_by_name['run'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._RUN
_BATCHGETMETRICSRESPONSE.fields_by_name['metrics'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._METRIC
_LISTMETRICSREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTMETRICSREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTMETRICSRESPONSE.fields_by_name['metrics'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._METRIC
_CREATEMETRICREQUEST.fields_by_name['metric'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._METRIC
_LOGMETRICREQUEST.fields_by_name['value'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._METRICVALUE
_UPDATEMETRICREQUEST.fields_by_name['metric'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._METRIC
_UPDATEMETRICREQUEST.fields_by_name['update_mask'].message_type = google_dot_protobuf_dot_field__mask__pb2._FIELDMASK
_BATCHGETMETADATARESPONSE.fields_by_name['metadata'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._METADATA
_LISTMETADATAREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTMETADATAREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTMETADATARESPONSE.fields_by_name['metadata'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._METADATA
_CREATEMETADATAREQUEST.fields_by_name['metadata'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._METADATA
_UPDATEMETADATAREQUEST.fields_by_name['metadata'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._METADATA
_BATCHGETNOTIFICATIONPOLICIESRESPONSE.fields_by_name['notification_policies'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._NOTIFICATIONPOLICY
_LISTNOTIFICATIONPOLICIESREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTNOTIFICATIONPOLICIESREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTNOTIFICATIONPOLICIESRESPONSE.fields_by_name['notification_policies'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._NOTIFICATIONPOLICY
_CREATENOTIFICATIONPOLICYREQUEST.fields_by_name['notification_policy'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._NOTIFICATIONPOLICY
_UPDATENOTIFICATIONPOLICYREQUEST.fields_by_name['notification_policy'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._NOTIFICATIONPOLICY
_UPDATENOTIFICATIONPOLICYREQUEST.fields_by_name['update_mask'].message_type = google_dot_protobuf_dot_field__mask__pb2._FIELDMASK
_BATCHGETTAGSRESPONSE.fields_by_name['tags'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._TAG
_LISTTAGSREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTTAGSREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTTAGSRESPONSE.fields_by_name['tags'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._TAG
_UPDATETAGREQUEST.fields_by_name['tag'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._TAG
_UPDATETAGREQUEST.fields_by_name['update_mask'].message_type = google_dot_protobuf_dot_field__mask__pb2._FIELDMASK
_CREATEDATAPROFILEREQUEST.fields_by_name['data_profile'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATAPROFILE
_LISTDATAPROFILESRESPONSE.fields_by_name['data_profiles'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATAPROFILE
_UPDATEDATAPROFILEREQUEST.fields_by_name['data_profile'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATAPROFILE
_LISTRESOURCECHECKSRESPONSE.fields_by_name['resource_checks'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._RESOURCECHECK
_CREATERESOURCECHECKREQUEST.fields_by_name['resource_check'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._RESOURCECHECK
_BATCHGETDATASETSRESPONSE.fields_by_name['datasets'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATASET
_LISTDATASETSREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTDATASETSREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTDATASETSRESPONSE.fields_by_name['datasets'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATASET
_BATCHGETDATASETVERSIONSRESPONSE.fields_by_name['dataset_versions'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATASETVERSION
_LISTDATASETVERSIONSREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTDATASETVERSIONSREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTDATASETVERSIONSRESPONSE.fields_by_name['dataset_versions'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATASETVERSION
_LISTDATASETVERSIONASSIGNMENTSREQUEST.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTDATASETVERSIONASSIGNMENTSREQUEST.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_LISTDATASETVERSIONASSIGNMENTSRESPONSE.fields_by_name['assignments'].message_type = continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATASETVERSIONASSIGNMENT
DESCRIPTOR.message_types_by_name['UpdateUserRequest'] = _UPDATEUSERREQUEST
DESCRIPTOR.message_types_by_name['GetUserRequest'] = _GETUSERREQUEST
DESCRIPTOR.message_types_by_name['BatchGetUsersRequest'] = _BATCHGETUSERSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetUsersResponse'] = _BATCHGETUSERSRESPONSE
DESCRIPTOR.message_types_by_name['DeleteUserRequest'] = _DELETEUSERREQUEST
DESCRIPTOR.message_types_by_name['ListUsersRequest'] = _LISTUSERSREQUEST
DESCRIPTOR.message_types_by_name['ListUsersResponse'] = _LISTUSERSRESPONSE
DESCRIPTOR.message_types_by_name['RegisterRequest'] = _REGISTERREQUEST
DESCRIPTOR.message_types_by_name['RegisterResponse'] = _REGISTERRESPONSE
DESCRIPTOR.message_types_by_name['LoginRequest'] = _LOGINREQUEST
DESCRIPTOR.message_types_by_name['LoginResponse'] = _LOGINRESPONSE
DESCRIPTOR.message_types_by_name['CheckOrganizationAccessRequest'] = _CHECKORGANIZATIONACCESSREQUEST
DESCRIPTOR.message_types_by_name['CheckOrganizationAccessResponse'] = _CHECKORGANIZATIONACCESSRESPONSE
DESCRIPTOR.message_types_by_name['SendEmailVerificationRequest'] = _SENDEMAILVERIFICATIONREQUEST
DESCRIPTOR.message_types_by_name['VerifyEmailRequest'] = _VERIFYEMAILREQUEST
DESCRIPTOR.message_types_by_name['VerifyEmailResponse'] = _VERIFYEMAILRESPONSE
DESCRIPTOR.message_types_by_name['SendPasswordResetRequest'] = _SENDPASSWORDRESETREQUEST
DESCRIPTOR.message_types_by_name['GetPasswordResetRequest'] = _GETPASSWORDRESETREQUEST
DESCRIPTOR.message_types_by_name['ResetPasswordRequest'] = _RESETPASSWORDREQUEST
DESCRIPTOR.message_types_by_name['ResetPasswordResponse'] = _RESETPASSWORDRESPONSE
DESCRIPTOR.message_types_by_name['ChangePasswordRequest'] = _CHANGEPASSWORDREQUEST
DESCRIPTOR.message_types_by_name['ChangePasswordResponse'] = _CHANGEPASSWORDRESPONSE
DESCRIPTOR.message_types_by_name['GetSessionRequest'] = _GETSESSIONREQUEST
DESCRIPTOR.message_types_by_name['ListSessionsRequest'] = _LISTSESSIONSREQUEST
DESCRIPTOR.message_types_by_name['ListSessionsResponse'] = _LISTSESSIONSRESPONSE
DESCRIPTOR.message_types_by_name['DeleteSessionRequest'] = _DELETESESSIONREQUEST
DESCRIPTOR.message_types_by_name['DeleteAllSessionsRequest'] = _DELETEALLSESSIONSREQUEST
DESCRIPTOR.message_types_by_name['ListRolesRequest'] = _LISTROLESREQUEST
DESCRIPTOR.message_types_by_name['ListRolesResponse'] = _LISTROLESRESPONSE
DESCRIPTOR.message_types_by_name['GetRoleRequest'] = _GETROLEREQUEST
DESCRIPTOR.message_types_by_name['BatchGetRolesRequest'] = _BATCHGETROLESREQUEST
DESCRIPTOR.message_types_by_name['BatchGetRolesResponse'] = _BATCHGETROLESRESPONSE
DESCRIPTOR.message_types_by_name['ListAccessPoliciesRequest'] = _LISTACCESSPOLICIESREQUEST
DESCRIPTOR.message_types_by_name['ListAccessPoliciesResponse'] = _LISTACCESSPOLICIESRESPONSE
DESCRIPTOR.message_types_by_name['CreateAccessPolicyRequest'] = _CREATEACCESSPOLICYREQUEST
DESCRIPTOR.message_types_by_name['GetAccessPolicyRequest'] = _GETACCESSPOLICYREQUEST
DESCRIPTOR.message_types_by_name['UpdateAccessPolicyRequest'] = _UPDATEACCESSPOLICYREQUEST
DESCRIPTOR.message_types_by_name['DeleteAccessPolicyRequest'] = _DELETEACCESSPOLICYREQUEST
DESCRIPTOR.message_types_by_name['BulkDeleteAccessPoliciesRequest'] = _BULKDELETEACCESSPOLICIESREQUEST
DESCRIPTOR.message_types_by_name['CreateInviteRequest'] = _CREATEINVITEREQUEST
DESCRIPTOR.message_types_by_name['GetInviteRequest'] = _GETINVITEREQUEST
DESCRIPTOR.message_types_by_name['ListInvitesRequest'] = _LISTINVITESREQUEST
DESCRIPTOR.message_types_by_name['ListInvitesResponse'] = _LISTINVITESRESPONSE
DESCRIPTOR.message_types_by_name['UpdateInviteRequest'] = _UPDATEINVITEREQUEST
DESCRIPTOR.message_types_by_name['DeleteInviteRequest'] = _DELETEINVITEREQUEST
DESCRIPTOR.message_types_by_name['AcceptInviteRequest'] = _ACCEPTINVITEREQUEST
DESCRIPTOR.message_types_by_name['ResendInviteRequest'] = _RESENDINVITEREQUEST
DESCRIPTOR.message_types_by_name['TestPermissionsRequest'] = _TESTPERMISSIONSREQUEST
DESCRIPTOR.message_types_by_name['TestPermissionsResponse'] = _TESTPERMISSIONSRESPONSE
DESCRIPTOR.message_types_by_name['CreateOrganizationRequest'] = _CREATEORGANIZATIONREQUEST
DESCRIPTOR.message_types_by_name['UpdateOrganizationRequest'] = _UPDATEORGANIZATIONREQUEST
DESCRIPTOR.message_types_by_name['GetOrganizationRequest'] = _GETORGANIZATIONREQUEST
DESCRIPTOR.message_types_by_name['BatchGetOrganizationsRequest'] = _BATCHGETORGANIZATIONSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetOrganizationsResponse'] = _BATCHGETORGANIZATIONSRESPONSE
DESCRIPTOR.message_types_by_name['DeleteOrganizationRequest'] = _DELETEORGANIZATIONREQUEST
DESCRIPTOR.message_types_by_name['ListOrganizationsRequest'] = _LISTORGANIZATIONSREQUEST
DESCRIPTOR.message_types_by_name['ListOrganizationsResponse'] = _LISTORGANIZATIONSRESPONSE
DESCRIPTOR.message_types_by_name['GetOnboardingRequest'] = _GETONBOARDINGREQUEST
DESCRIPTOR.message_types_by_name['UpdateOnboardingRequest'] = _UPDATEONBOARDINGREQUEST
DESCRIPTOR.message_types_by_name['CreateProjectRequest'] = _CREATEPROJECTREQUEST
DESCRIPTOR.message_types_by_name['UpdateProjectRequest'] = _UPDATEPROJECTREQUEST
DESCRIPTOR.message_types_by_name['GetProjectRequest'] = _GETPROJECTREQUEST
DESCRIPTOR.message_types_by_name['BatchGetProjectsRequest'] = _BATCHGETPROJECTSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetProjectsResponse'] = _BATCHGETPROJECTSRESPONSE
DESCRIPTOR.message_types_by_name['DeleteProjectRequest'] = _DELETEPROJECTREQUEST
DESCRIPTOR.message_types_by_name['ListProjectsRequest'] = _LISTPROJECTSREQUEST
DESCRIPTOR.message_types_by_name['ListProjectsResponse'] = _LISTPROJECTSRESPONSE
DESCRIPTOR.message_types_by_name['ReRunRequest'] = _RERUNREQUEST
DESCRIPTOR.message_types_by_name['GetModelStatsRequest'] = _GETMODELSTATSREQUEST
DESCRIPTOR.message_types_by_name['GetUsageRequest'] = _GETUSAGEREQUEST
DESCRIPTOR.message_types_by_name['GetUsageResponse'] = _GETUSAGERESPONSE
DESCRIPTOR.message_types_by_name['ListPricesRequest'] = _LISTPRICESREQUEST
DESCRIPTOR.message_types_by_name['ListPricesResponse'] = _LISTPRICESRESPONSE
DESCRIPTOR.message_types_by_name['ListInvoicesRequest'] = _LISTINVOICESREQUEST
DESCRIPTOR.message_types_by_name['ListInvoicesResponse'] = _LISTINVOICESRESPONSE
DESCRIPTOR.message_types_by_name['GetPriceRequest'] = _GETPRICEREQUEST
DESCRIPTOR.message_types_by_name['GetPaymentMethodRequest'] = _GETPAYMENTMETHODREQUEST
DESCRIPTOR.message_types_by_name['GetUpcomingInvoiceRequest'] = _GETUPCOMINGINVOICEREQUEST
DESCRIPTOR.message_types_by_name['GetSubscriptionRequest'] = _GETSUBSCRIPTIONREQUEST
DESCRIPTOR.message_types_by_name['GetBillingAccountRequest'] = _GETBILLINGACCOUNTREQUEST
DESCRIPTOR.message_types_by_name['UpdateBillingAccountRequest'] = _UPDATEBILLINGACCOUNTREQUEST
DESCRIPTOR.message_types_by_name['UpdateSubscriptionRequest'] = _UPDATESUBSCRIPTIONREQUEST
DESCRIPTOR.message_types_by_name['CreatePromotionRequest'] = _CREATEPROMOTIONREQUEST
DESCRIPTOR.message_types_by_name['ListPromotionsRequest'] = _LISTPROMOTIONSREQUEST
DESCRIPTOR.message_types_by_name['ListPromotionsResponse'] = _LISTPROMOTIONSRESPONSE
DESCRIPTOR.message_types_by_name['GetPromotionRequest'] = _GETPROMOTIONREQUEST
DESCRIPTOR.message_types_by_name['BatchGetPromotionsRequest'] = _BATCHGETPROMOTIONSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetPromotionsResponse'] = _BATCHGETPROMOTIONSRESPONSE
DESCRIPTOR.message_types_by_name['CreateEventRequest'] = _CREATEEVENTREQUEST
DESCRIPTOR.message_types_by_name['ListEventsRequest'] = _LISTEVENTSREQUEST
DESCRIPTOR.message_types_by_name['ListEventsResponse'] = _LISTEVENTSRESPONSE
DESCRIPTOR.message_types_by_name['GetEventRequest'] = _GETEVENTREQUEST
DESCRIPTOR.message_types_by_name['BatchGetEventsRequest'] = _BATCHGETEVENTSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetEventsResponse'] = _BATCHGETEVENTSRESPONSE
DESCRIPTOR.message_types_by_name['CreateModelRequest'] = _CREATEMODELREQUEST
DESCRIPTOR.message_types_by_name['GetModelRequest'] = _GETMODELREQUEST
DESCRIPTOR.message_types_by_name['BatchGetModelsRequest'] = _BATCHGETMODELSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetModelsResponse'] = _BATCHGETMODELSRESPONSE
DESCRIPTOR.message_types_by_name['ListModelsRequest'] = _LISTMODELSREQUEST
DESCRIPTOR.message_types_by_name['ListModelsResponse'] = _LISTMODELSRESPONSE
DESCRIPTOR.message_types_by_name['DeleteModelRequest'] = _DELETEMODELREQUEST
DESCRIPTOR.message_types_by_name['ValidateModelRequest'] = _VALIDATEMODELREQUEST
DESCRIPTOR.message_types_by_name['CreateModelVersionRequest'] = _CREATEMODELVERSIONREQUEST
DESCRIPTOR.message_types_by_name['GetModelVersionRequest'] = _GETMODELVERSIONREQUEST
DESCRIPTOR.message_types_by_name['BatchGetModelVersionsRequest'] = _BATCHGETMODELVERSIONSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetModelVersionsResponse'] = _BATCHGETMODELVERSIONSRESPONSE
DESCRIPTOR.message_types_by_name['ListModelVersionsRequest'] = _LISTMODELVERSIONSREQUEST
DESCRIPTOR.message_types_by_name['ListModelVersionsResponse'] = _LISTMODELVERSIONSRESPONSE
DESCRIPTOR.message_types_by_name['CreateExperimentRequest'] = _CREATEEXPERIMENTREQUEST
DESCRIPTOR.message_types_by_name['GetExperimentRequest'] = _GETEXPERIMENTREQUEST
DESCRIPTOR.message_types_by_name['BatchGetExperimentsRequest'] = _BATCHGETEXPERIMENTSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetExperimentsResponse'] = _BATCHGETEXPERIMENTSRESPONSE
DESCRIPTOR.message_types_by_name['ListExperimentsRequest'] = _LISTEXPERIMENTSREQUEST
DESCRIPTOR.message_types_by_name['ListExperimentsResponse'] = _LISTEXPERIMENTSRESPONSE
DESCRIPTOR.message_types_by_name['CreateBatchPredictionRequest'] = _CREATEBATCHPREDICTIONREQUEST
DESCRIPTOR.message_types_by_name['ListBatchPredictionsRequest'] = _LISTBATCHPREDICTIONSREQUEST
DESCRIPTOR.message_types_by_name['ListBatchPredictionsResponse'] = _LISTBATCHPREDICTIONSRESPONSE
DESCRIPTOR.message_types_by_name['BatchPredictionRequest'] = _BATCHPREDICTIONREQUEST
DESCRIPTOR.message_types_by_name['BatchGetBatchPredictionsRequest'] = _BATCHGETBATCHPREDICTIONSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetBatchPredictionsResponse'] = _BATCHGETBATCHPREDICTIONSRESPONSE
DESCRIPTOR.message_types_by_name['YamlDiffRequest'] = _YAMLDIFFREQUEST
DESCRIPTOR.message_types_by_name['YamlDiffResponse'] = _YAMLDIFFRESPONSE
DESCRIPTOR.message_types_by_name['CreateEnvironmentRequest'] = _CREATEENVIRONMENTREQUEST
DESCRIPTOR.message_types_by_name['UpdateEnvironmentRequest'] = _UPDATEENVIRONMENTREQUEST
DESCRIPTOR.message_types_by_name['GetEnvironmentRequest'] = _GETENVIRONMENTREQUEST
DESCRIPTOR.message_types_by_name['BatchGetEnvironmentsRequest'] = _BATCHGETENVIRONMENTSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetEnvironmentsResponse'] = _BATCHGETENVIRONMENTSRESPONSE
DESCRIPTOR.message_types_by_name['DeleteEnvironmentRequest'] = _DELETEENVIRONMENTREQUEST
DESCRIPTOR.message_types_by_name['ListEnvironmentsRequest'] = _LISTENVIRONMENTSREQUEST
DESCRIPTOR.message_types_by_name['ListEnvironmentsResponse'] = _LISTENVIRONMENTSRESPONSE
DESCRIPTOR.message_types_by_name['GetSystemMessageRequest'] = _GETSYSTEMMESSAGEREQUEST
DESCRIPTOR.message_types_by_name['GetSystemMessageResponse'] = _GETSYSTEMMESSAGERESPONSE
DESCRIPTOR.message_types_by_name['GetFeatureFlagRequest'] = _GETFEATUREFLAGREQUEST
DESCRIPTOR.message_types_by_name['GetFeatureFlagResponse'] = _GETFEATUREFLAGRESPONSE
DESCRIPTOR.message_types_by_name['GetOrganizationFeatureFlagsRequest'] = _GETORGANIZATIONFEATUREFLAGSREQUEST
DESCRIPTOR.message_types_by_name['GetOrganizationFeatureFlagsResponse'] = _GETORGANIZATIONFEATUREFLAGSRESPONSE
DESCRIPTOR.message_types_by_name['CreateArtifactRequest'] = _CREATEARTIFACTREQUEST
DESCRIPTOR.message_types_by_name['GetArtifactRequest'] = _GETARTIFACTREQUEST
DESCRIPTOR.message_types_by_name['BatchGetArtifactsRequest'] = _BATCHGETARTIFACTSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetArtifactsResponse'] = _BATCHGETARTIFACTSRESPONSE
DESCRIPTOR.message_types_by_name['DeleteArtifactRequest'] = _DELETEARTIFACTREQUEST
DESCRIPTOR.message_types_by_name['ListArtifactsRequest'] = _LISTARTIFACTSREQUEST
DESCRIPTOR.message_types_by_name['ListArtifactsResponse'] = _LISTARTIFACTSRESPONSE
DESCRIPTOR.message_types_by_name['GenerateArtifactUploadURLRequest'] = _GENERATEARTIFACTUPLOADURLREQUEST
DESCRIPTOR.message_types_by_name['GenerateArtifactUploadURLResponse'] = _GENERATEARTIFACTUPLOADURLRESPONSE
DESCRIPTOR.message_types_by_name['ListUserIntegrationProvidersRequest'] = _LISTUSERINTEGRATIONPROVIDERSREQUEST
DESCRIPTOR.message_types_by_name['ListUserIntegrationProvidersResponse'] = _LISTUSERINTEGRATIONPROVIDERSRESPONSE
DESCRIPTOR.message_types_by_name['ListUserIntegrationsRequest'] = _LISTUSERINTEGRATIONSREQUEST
DESCRIPTOR.message_types_by_name['ListUserIntegrationsResponse'] = _LISTUSERINTEGRATIONSRESPONSE
DESCRIPTOR.message_types_by_name['DisconnectUserIntegrationRequest'] = _DISCONNECTUSERINTEGRATIONREQUEST
DESCRIPTOR.message_types_by_name['ListGitProviderNamespacesRequest'] = _LISTGITPROVIDERNAMESPACESREQUEST
DESCRIPTOR.message_types_by_name['ListGitProviderNamespacesResponse'] = _LISTGITPROVIDERNAMESPACESRESPONSE
DESCRIPTOR.message_types_by_name['ListGitProviderRepositoriesRequest'] = _LISTGITPROVIDERREPOSITORIESREQUEST
DESCRIPTOR.message_types_by_name['ListGitProviderRepositoriesResponse'] = _LISTGITPROVIDERREPOSITORIESRESPONSE
DESCRIPTOR.message_types_by_name['ConnectProjectToGitRepositoryRequest'] = _CONNECTPROJECTTOGITREPOSITORYREQUEST
DESCRIPTOR.message_types_by_name['ConnectProjectToGitRepositoryResponse'] = _CONNECTPROJECTTOGITREPOSITORYRESPONSE
DESCRIPTOR.message_types_by_name['GetProjectGitIntegrationRequest'] = _GETPROJECTGITINTEGRATIONREQUEST
DESCRIPTOR.message_types_by_name['SeedProjectGitRepositoryRequest'] = _SEEDPROJECTGITREPOSITORYREQUEST
DESCRIPTOR.message_types_by_name['GenerateSsoAdminPortalURLRequest'] = _GENERATESSOADMINPORTALURLREQUEST
DESCRIPTOR.message_types_by_name['GenerateSsoAdminPortalURLResponse'] = _GENERATESSOADMINPORTALURLRESPONSE
DESCRIPTOR.message_types_by_name['GenerateDSyncAdminPortalURLRequest'] = _GENERATEDSYNCADMINPORTALURLREQUEST
DESCRIPTOR.message_types_by_name['GenerateDSyncAdminPortalURLResponse'] = _GENERATEDSYNCADMINPORTALURLRESPONSE
DESCRIPTOR.message_types_by_name['ExchangeTokenRequest'] = _EXCHANGETOKENREQUEST
DESCRIPTOR.message_types_by_name['ExchangeTokenResponse'] = _EXCHANGETOKENRESPONSE
DESCRIPTOR.message_types_by_name['GetPrincipalRequest'] = _GETPRINCIPALREQUEST
DESCRIPTOR.message_types_by_name['GetApiKeyRequest'] = _GETAPIKEYREQUEST
DESCRIPTOR.message_types_by_name['BatchGetApiKeysRequest'] = _BATCHGETAPIKEYSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetApiKeysResponse'] = _BATCHGETAPIKEYSRESPONSE
DESCRIPTOR.message_types_by_name['ListApiKeysRequest'] = _LISTAPIKEYSREQUEST
DESCRIPTOR.message_types_by_name['ListApiKeysResponse'] = _LISTAPIKEYSRESPONSE
DESCRIPTOR.message_types_by_name['CreateApiKeyRequest'] = _CREATEAPIKEYREQUEST
DESCRIPTOR.message_types_by_name['UpdateApiKeyRequest'] = _UPDATEAPIKEYREQUEST
DESCRIPTOR.message_types_by_name['RevokeApiKeyRequest'] = _REVOKEAPIKEYREQUEST
DESCRIPTOR.message_types_by_name['GetApiKeyProjectRequest'] = _GETAPIKEYPROJECTREQUEST
DESCRIPTOR.message_types_by_name['GetRunRequest'] = _GETRUNREQUEST
DESCRIPTOR.message_types_by_name['BatchGetRunsRequest'] = _BATCHGETRUNSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetRunsResponse'] = _BATCHGETRUNSRESPONSE
DESCRIPTOR.message_types_by_name['ListRunsRequest'] = _LISTRUNSREQUEST
DESCRIPTOR.message_types_by_name['ListRunsResponse'] = _LISTRUNSRESPONSE
DESCRIPTOR.message_types_by_name['CreateRunRequest'] = _CREATERUNREQUEST
DESCRIPTOR.message_types_by_name['UpdateRunRequest'] = _UPDATERUNREQUEST
DESCRIPTOR.message_types_by_name['DeleteRunRequest'] = _DELETERUNREQUEST
DESCRIPTOR.message_types_by_name['GetRunSummaryRequest'] = _GETRUNSUMMARYREQUEST
DESCRIPTOR.message_types_by_name['GetRunCheckSummaryRequest'] = _GETRUNCHECKSUMMARYREQUEST
DESCRIPTOR.message_types_by_name['LogRunHeartbeatRequest'] = _LOGRUNHEARTBEATREQUEST
DESCRIPTOR.message_types_by_name['GetMetricRequest'] = _GETMETRICREQUEST
DESCRIPTOR.message_types_by_name['BatchGetMetricsRequest'] = _BATCHGETMETRICSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetMetricsResponse'] = _BATCHGETMETRICSRESPONSE
DESCRIPTOR.message_types_by_name['ListMetricsRequest'] = _LISTMETRICSREQUEST
DESCRIPTOR.message_types_by_name['ListMetricsResponse'] = _LISTMETRICSRESPONSE
DESCRIPTOR.message_types_by_name['CreateMetricRequest'] = _CREATEMETRICREQUEST
DESCRIPTOR.message_types_by_name['LogMetricRequest'] = _LOGMETRICREQUEST
DESCRIPTOR.message_types_by_name['UpdateMetricRequest'] = _UPDATEMETRICREQUEST
DESCRIPTOR.message_types_by_name['GetMetadataRequest'] = _GETMETADATAREQUEST
DESCRIPTOR.message_types_by_name['BatchGetMetadataRequest'] = _BATCHGETMETADATAREQUEST
DESCRIPTOR.message_types_by_name['BatchGetMetadataResponse'] = _BATCHGETMETADATARESPONSE
DESCRIPTOR.message_types_by_name['ListMetadataRequest'] = _LISTMETADATAREQUEST
DESCRIPTOR.message_types_by_name['ListMetadataResponse'] = _LISTMETADATARESPONSE
DESCRIPTOR.message_types_by_name['CreateMetadataRequest'] = _CREATEMETADATAREQUEST
DESCRIPTOR.message_types_by_name['UpdateMetadataRequest'] = _UPDATEMETADATAREQUEST
DESCRIPTOR.message_types_by_name['DeleteMetadataRequest'] = _DELETEMETADATAREQUEST
DESCRIPTOR.message_types_by_name['GetNotificationPolicyRequest'] = _GETNOTIFICATIONPOLICYREQUEST
DESCRIPTOR.message_types_by_name['BatchGetNotificationPoliciesRequest'] = _BATCHGETNOTIFICATIONPOLICIESREQUEST
DESCRIPTOR.message_types_by_name['BatchGetNotificationPoliciesResponse'] = _BATCHGETNOTIFICATIONPOLICIESRESPONSE
DESCRIPTOR.message_types_by_name['ListNotificationPoliciesRequest'] = _LISTNOTIFICATIONPOLICIESREQUEST
DESCRIPTOR.message_types_by_name['ListNotificationPoliciesResponse'] = _LISTNOTIFICATIONPOLICIESRESPONSE
DESCRIPTOR.message_types_by_name['CreateNotificationPolicyRequest'] = _CREATENOTIFICATIONPOLICYREQUEST
DESCRIPTOR.message_types_by_name['UpdateNotificationPolicyRequest'] = _UPDATENOTIFICATIONPOLICYREQUEST
DESCRIPTOR.message_types_by_name['DeleteNotificationPolicyRequest'] = _DELETENOTIFICATIONPOLICYREQUEST
DESCRIPTOR.message_types_by_name['GetTagRequest'] = _GETTAGREQUEST
DESCRIPTOR.message_types_by_name['BatchGetTagsRequest'] = _BATCHGETTAGSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetTagsResponse'] = _BATCHGETTAGSRESPONSE
DESCRIPTOR.message_types_by_name['ListTagsRequest'] = _LISTTAGSREQUEST
DESCRIPTOR.message_types_by_name['ListTagsResponse'] = _LISTTAGSRESPONSE
DESCRIPTOR.message_types_by_name['CreateTagRequest'] = _CREATETAGREQUEST
DESCRIPTOR.message_types_by_name['UpdateTagRequest'] = _UPDATETAGREQUEST
DESCRIPTOR.message_types_by_name['DeleteTagRequest'] = _DELETETAGREQUEST
DESCRIPTOR.message_types_by_name['GetDataProfileRequest'] = _GETDATAPROFILEREQUEST
DESCRIPTOR.message_types_by_name['CreateDataProfileRequest'] = _CREATEDATAPROFILEREQUEST
DESCRIPTOR.message_types_by_name['ListDataProfilesRequest'] = _LISTDATAPROFILESREQUEST
DESCRIPTOR.message_types_by_name['ListDataProfilesResponse'] = _LISTDATAPROFILESRESPONSE
DESCRIPTOR.message_types_by_name['UpdateDataProfileRequest'] = _UPDATEDATAPROFILEREQUEST
DESCRIPTOR.message_types_by_name['GetResourceCheckRequest'] = _GETRESOURCECHECKREQUEST
DESCRIPTOR.message_types_by_name['ListResourceChecksRequest'] = _LISTRESOURCECHECKSREQUEST
DESCRIPTOR.message_types_by_name['ListResourceChecksResponse'] = _LISTRESOURCECHECKSRESPONSE
DESCRIPTOR.message_types_by_name['CreateResourceCheckRequest'] = _CREATERESOURCECHECKREQUEST
DESCRIPTOR.message_types_by_name['CreateDatasetRequest'] = _CREATEDATASETREQUEST
DESCRIPTOR.message_types_by_name['GetDatasetRequest'] = _GETDATASETREQUEST
DESCRIPTOR.message_types_by_name['BatchGetDatasetsRequest'] = _BATCHGETDATASETSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetDatasetsResponse'] = _BATCHGETDATASETSRESPONSE
DESCRIPTOR.message_types_by_name['ListDatasetsRequest'] = _LISTDATASETSREQUEST
DESCRIPTOR.message_types_by_name['ListDatasetsResponse'] = _LISTDATASETSRESPONSE
DESCRIPTOR.message_types_by_name['DeleteDatasetRequest'] = _DELETEDATASETREQUEST
DESCRIPTOR.message_types_by_name['CreateDatasetVersionRequest'] = _CREATEDATASETVERSIONREQUEST
DESCRIPTOR.message_types_by_name['GetDatasetVersionRequest'] = _GETDATASETVERSIONREQUEST
DESCRIPTOR.message_types_by_name['BatchGetDatasetVersionsRequest'] = _BATCHGETDATASETVERSIONSREQUEST
DESCRIPTOR.message_types_by_name['BatchGetDatasetVersionsResponse'] = _BATCHGETDATASETVERSIONSRESPONSE
DESCRIPTOR.message_types_by_name['ListDatasetVersionsRequest'] = _LISTDATASETVERSIONSREQUEST
DESCRIPTOR.message_types_by_name['ListDatasetVersionsResponse'] = _LISTDATASETVERSIONSRESPONSE
DESCRIPTOR.message_types_by_name['CreateDatasetVersionAssignmentRequest'] = _CREATEDATASETVERSIONASSIGNMENTREQUEST
DESCRIPTOR.message_types_by_name['GetDatasetVersionAssignmentRequest'] = _GETDATASETVERSIONASSIGNMENTREQUEST
DESCRIPTOR.message_types_by_name['DeleteDatasetVersionAssignmentRequest'] = _DELETEDATASETVERSIONASSIGNMENTREQUEST
DESCRIPTOR.message_types_by_name['ListDatasetVersionAssignmentsRequest'] = _LISTDATASETVERSIONASSIGNMENTSREQUEST
DESCRIPTOR.message_types_by_name['ListDatasetVersionAssignmentsResponse'] = _LISTDATASETVERSIONASSIGNMENTSRESPONSE
DESCRIPTOR.message_types_by_name['GetServerVersionRequest'] = _GETSERVERVERSIONREQUEST
DESCRIPTOR.message_types_by_name['GetServerVersionResponse'] = _GETSERVERVERSIONRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

UpdateUserRequest = _reflection.GeneratedProtocolMessageType('UpdateUserRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEUSERREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UpdateUserRequest)
  })
_sym_db.RegisterMessage(UpdateUserRequest)

GetUserRequest = _reflection.GeneratedProtocolMessageType('GetUserRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETUSERREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetUserRequest)
  })
_sym_db.RegisterMessage(GetUserRequest)

BatchGetUsersRequest = _reflection.GeneratedProtocolMessageType('BatchGetUsersRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETUSERSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetUsersRequest)
  })
_sym_db.RegisterMessage(BatchGetUsersRequest)

BatchGetUsersResponse = _reflection.GeneratedProtocolMessageType('BatchGetUsersResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETUSERSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetUsersResponse)
  })
_sym_db.RegisterMessage(BatchGetUsersResponse)

DeleteUserRequest = _reflection.GeneratedProtocolMessageType('DeleteUserRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEUSERREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteUserRequest)
  })
_sym_db.RegisterMessage(DeleteUserRequest)

ListUsersRequest = _reflection.GeneratedProtocolMessageType('ListUsersRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTUSERSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListUsersRequest)
  })
_sym_db.RegisterMessage(ListUsersRequest)

ListUsersResponse = _reflection.GeneratedProtocolMessageType('ListUsersResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTUSERSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListUsersResponse)
  })
_sym_db.RegisterMessage(ListUsersResponse)

RegisterRequest = _reflection.GeneratedProtocolMessageType('RegisterRequest', (_message.Message,), {
  'DESCRIPTOR' : _REGISTERREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.RegisterRequest)
  })
_sym_db.RegisterMessage(RegisterRequest)

RegisterResponse = _reflection.GeneratedProtocolMessageType('RegisterResponse', (_message.Message,), {
  'DESCRIPTOR' : _REGISTERRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.RegisterResponse)
  })
_sym_db.RegisterMessage(RegisterResponse)

LoginRequest = _reflection.GeneratedProtocolMessageType('LoginRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOGINREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.LoginRequest)
  })
_sym_db.RegisterMessage(LoginRequest)

LoginResponse = _reflection.GeneratedProtocolMessageType('LoginResponse', (_message.Message,), {
  'DESCRIPTOR' : _LOGINRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.LoginResponse)
  })
_sym_db.RegisterMessage(LoginResponse)

CheckOrganizationAccessRequest = _reflection.GeneratedProtocolMessageType('CheckOrganizationAccessRequest', (_message.Message,), {
  'DESCRIPTOR' : _CHECKORGANIZATIONACCESSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CheckOrganizationAccessRequest)
  })
_sym_db.RegisterMessage(CheckOrganizationAccessRequest)

CheckOrganizationAccessResponse = _reflection.GeneratedProtocolMessageType('CheckOrganizationAccessResponse', (_message.Message,), {
  'DESCRIPTOR' : _CHECKORGANIZATIONACCESSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CheckOrganizationAccessResponse)
  })
_sym_db.RegisterMessage(CheckOrganizationAccessResponse)

SendEmailVerificationRequest = _reflection.GeneratedProtocolMessageType('SendEmailVerificationRequest', (_message.Message,), {
  'DESCRIPTOR' : _SENDEMAILVERIFICATIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.SendEmailVerificationRequest)
  })
_sym_db.RegisterMessage(SendEmailVerificationRequest)

VerifyEmailRequest = _reflection.GeneratedProtocolMessageType('VerifyEmailRequest', (_message.Message,), {
  'DESCRIPTOR' : _VERIFYEMAILREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.VerifyEmailRequest)
  })
_sym_db.RegisterMessage(VerifyEmailRequest)

VerifyEmailResponse = _reflection.GeneratedProtocolMessageType('VerifyEmailResponse', (_message.Message,), {
  'DESCRIPTOR' : _VERIFYEMAILRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.VerifyEmailResponse)
  })
_sym_db.RegisterMessage(VerifyEmailResponse)

SendPasswordResetRequest = _reflection.GeneratedProtocolMessageType('SendPasswordResetRequest', (_message.Message,), {
  'DESCRIPTOR' : _SENDPASSWORDRESETREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.SendPasswordResetRequest)
  })
_sym_db.RegisterMessage(SendPasswordResetRequest)

GetPasswordResetRequest = _reflection.GeneratedProtocolMessageType('GetPasswordResetRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETPASSWORDRESETREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetPasswordResetRequest)
  })
_sym_db.RegisterMessage(GetPasswordResetRequest)

ResetPasswordRequest = _reflection.GeneratedProtocolMessageType('ResetPasswordRequest', (_message.Message,), {
  'DESCRIPTOR' : _RESETPASSWORDREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ResetPasswordRequest)
  })
_sym_db.RegisterMessage(ResetPasswordRequest)

ResetPasswordResponse = _reflection.GeneratedProtocolMessageType('ResetPasswordResponse', (_message.Message,), {
  'DESCRIPTOR' : _RESETPASSWORDRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ResetPasswordResponse)
  })
_sym_db.RegisterMessage(ResetPasswordResponse)

ChangePasswordRequest = _reflection.GeneratedProtocolMessageType('ChangePasswordRequest', (_message.Message,), {
  'DESCRIPTOR' : _CHANGEPASSWORDREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ChangePasswordRequest)
  })
_sym_db.RegisterMessage(ChangePasswordRequest)

ChangePasswordResponse = _reflection.GeneratedProtocolMessageType('ChangePasswordResponse', (_message.Message,), {
  'DESCRIPTOR' : _CHANGEPASSWORDRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ChangePasswordResponse)
  })
_sym_db.RegisterMessage(ChangePasswordResponse)

GetSessionRequest = _reflection.GeneratedProtocolMessageType('GetSessionRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETSESSIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetSessionRequest)
  })
_sym_db.RegisterMessage(GetSessionRequest)

ListSessionsRequest = _reflection.GeneratedProtocolMessageType('ListSessionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTSESSIONSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListSessionsRequest)
  })
_sym_db.RegisterMessage(ListSessionsRequest)

ListSessionsResponse = _reflection.GeneratedProtocolMessageType('ListSessionsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTSESSIONSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListSessionsResponse)
  })
_sym_db.RegisterMessage(ListSessionsResponse)

DeleteSessionRequest = _reflection.GeneratedProtocolMessageType('DeleteSessionRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETESESSIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteSessionRequest)
  })
_sym_db.RegisterMessage(DeleteSessionRequest)

DeleteAllSessionsRequest = _reflection.GeneratedProtocolMessageType('DeleteAllSessionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEALLSESSIONSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteAllSessionsRequest)
  })
_sym_db.RegisterMessage(DeleteAllSessionsRequest)

ListRolesRequest = _reflection.GeneratedProtocolMessageType('ListRolesRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTROLESREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListRolesRequest)
  })
_sym_db.RegisterMessage(ListRolesRequest)

ListRolesResponse = _reflection.GeneratedProtocolMessageType('ListRolesResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTROLESRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListRolesResponse)
  })
_sym_db.RegisterMessage(ListRolesResponse)

GetRoleRequest = _reflection.GeneratedProtocolMessageType('GetRoleRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETROLEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetRoleRequest)
  })
_sym_db.RegisterMessage(GetRoleRequest)

BatchGetRolesRequest = _reflection.GeneratedProtocolMessageType('BatchGetRolesRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETROLESREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetRolesRequest)
  })
_sym_db.RegisterMessage(BatchGetRolesRequest)

BatchGetRolesResponse = _reflection.GeneratedProtocolMessageType('BatchGetRolesResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETROLESRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetRolesResponse)
  })
_sym_db.RegisterMessage(BatchGetRolesResponse)

ListAccessPoliciesRequest = _reflection.GeneratedProtocolMessageType('ListAccessPoliciesRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTACCESSPOLICIESREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListAccessPoliciesRequest)
  })
_sym_db.RegisterMessage(ListAccessPoliciesRequest)

ListAccessPoliciesResponse = _reflection.GeneratedProtocolMessageType('ListAccessPoliciesResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTACCESSPOLICIESRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListAccessPoliciesResponse)
  })
_sym_db.RegisterMessage(ListAccessPoliciesResponse)

CreateAccessPolicyRequest = _reflection.GeneratedProtocolMessageType('CreateAccessPolicyRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEACCESSPOLICYREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateAccessPolicyRequest)
  })
_sym_db.RegisterMessage(CreateAccessPolicyRequest)

GetAccessPolicyRequest = _reflection.GeneratedProtocolMessageType('GetAccessPolicyRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETACCESSPOLICYREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetAccessPolicyRequest)
  })
_sym_db.RegisterMessage(GetAccessPolicyRequest)

UpdateAccessPolicyRequest = _reflection.GeneratedProtocolMessageType('UpdateAccessPolicyRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEACCESSPOLICYREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UpdateAccessPolicyRequest)
  })
_sym_db.RegisterMessage(UpdateAccessPolicyRequest)

DeleteAccessPolicyRequest = _reflection.GeneratedProtocolMessageType('DeleteAccessPolicyRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEACCESSPOLICYREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteAccessPolicyRequest)
  })
_sym_db.RegisterMessage(DeleteAccessPolicyRequest)

BulkDeleteAccessPoliciesRequest = _reflection.GeneratedProtocolMessageType('BulkDeleteAccessPoliciesRequest', (_message.Message,), {
  'DESCRIPTOR' : _BULKDELETEACCESSPOLICIESREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BulkDeleteAccessPoliciesRequest)
  })
_sym_db.RegisterMessage(BulkDeleteAccessPoliciesRequest)

CreateInviteRequest = _reflection.GeneratedProtocolMessageType('CreateInviteRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEINVITEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateInviteRequest)
  })
_sym_db.RegisterMessage(CreateInviteRequest)

GetInviteRequest = _reflection.GeneratedProtocolMessageType('GetInviteRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETINVITEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetInviteRequest)
  })
_sym_db.RegisterMessage(GetInviteRequest)

ListInvitesRequest = _reflection.GeneratedProtocolMessageType('ListInvitesRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTINVITESREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListInvitesRequest)
  })
_sym_db.RegisterMessage(ListInvitesRequest)

ListInvitesResponse = _reflection.GeneratedProtocolMessageType('ListInvitesResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTINVITESRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListInvitesResponse)
  })
_sym_db.RegisterMessage(ListInvitesResponse)

UpdateInviteRequest = _reflection.GeneratedProtocolMessageType('UpdateInviteRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEINVITEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UpdateInviteRequest)
  })
_sym_db.RegisterMessage(UpdateInviteRequest)

DeleteInviteRequest = _reflection.GeneratedProtocolMessageType('DeleteInviteRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEINVITEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteInviteRequest)
  })
_sym_db.RegisterMessage(DeleteInviteRequest)

AcceptInviteRequest = _reflection.GeneratedProtocolMessageType('AcceptInviteRequest', (_message.Message,), {
  'DESCRIPTOR' : _ACCEPTINVITEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.AcceptInviteRequest)
  })
_sym_db.RegisterMessage(AcceptInviteRequest)

ResendInviteRequest = _reflection.GeneratedProtocolMessageType('ResendInviteRequest', (_message.Message,), {
  'DESCRIPTOR' : _RESENDINVITEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ResendInviteRequest)
  })
_sym_db.RegisterMessage(ResendInviteRequest)

TestPermissionsRequest = _reflection.GeneratedProtocolMessageType('TestPermissionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _TESTPERMISSIONSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.TestPermissionsRequest)
  })
_sym_db.RegisterMessage(TestPermissionsRequest)

TestPermissionsResponse = _reflection.GeneratedProtocolMessageType('TestPermissionsResponse', (_message.Message,), {
  'DESCRIPTOR' : _TESTPERMISSIONSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.TestPermissionsResponse)
  })
_sym_db.RegisterMessage(TestPermissionsResponse)

CreateOrganizationRequest = _reflection.GeneratedProtocolMessageType('CreateOrganizationRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEORGANIZATIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateOrganizationRequest)
  })
_sym_db.RegisterMessage(CreateOrganizationRequest)

UpdateOrganizationRequest = _reflection.GeneratedProtocolMessageType('UpdateOrganizationRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEORGANIZATIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UpdateOrganizationRequest)
  })
_sym_db.RegisterMessage(UpdateOrganizationRequest)

GetOrganizationRequest = _reflection.GeneratedProtocolMessageType('GetOrganizationRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETORGANIZATIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetOrganizationRequest)
  })
_sym_db.RegisterMessage(GetOrganizationRequest)

BatchGetOrganizationsRequest = _reflection.GeneratedProtocolMessageType('BatchGetOrganizationsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETORGANIZATIONSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetOrganizationsRequest)
  })
_sym_db.RegisterMessage(BatchGetOrganizationsRequest)

BatchGetOrganizationsResponse = _reflection.GeneratedProtocolMessageType('BatchGetOrganizationsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETORGANIZATIONSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetOrganizationsResponse)
  })
_sym_db.RegisterMessage(BatchGetOrganizationsResponse)

DeleteOrganizationRequest = _reflection.GeneratedProtocolMessageType('DeleteOrganizationRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEORGANIZATIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteOrganizationRequest)
  })
_sym_db.RegisterMessage(DeleteOrganizationRequest)

ListOrganizationsRequest = _reflection.GeneratedProtocolMessageType('ListOrganizationsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTORGANIZATIONSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListOrganizationsRequest)
  })
_sym_db.RegisterMessage(ListOrganizationsRequest)

ListOrganizationsResponse = _reflection.GeneratedProtocolMessageType('ListOrganizationsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTORGANIZATIONSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListOrganizationsResponse)
  })
_sym_db.RegisterMessage(ListOrganizationsResponse)

GetOnboardingRequest = _reflection.GeneratedProtocolMessageType('GetOnboardingRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETONBOARDINGREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetOnboardingRequest)
  })
_sym_db.RegisterMessage(GetOnboardingRequest)

UpdateOnboardingRequest = _reflection.GeneratedProtocolMessageType('UpdateOnboardingRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEONBOARDINGREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UpdateOnboardingRequest)
  })
_sym_db.RegisterMessage(UpdateOnboardingRequest)

CreateProjectRequest = _reflection.GeneratedProtocolMessageType('CreateProjectRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEPROJECTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateProjectRequest)
  })
_sym_db.RegisterMessage(CreateProjectRequest)

UpdateProjectRequest = _reflection.GeneratedProtocolMessageType('UpdateProjectRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEPROJECTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UpdateProjectRequest)
  })
_sym_db.RegisterMessage(UpdateProjectRequest)

GetProjectRequest = _reflection.GeneratedProtocolMessageType('GetProjectRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETPROJECTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetProjectRequest)
  })
_sym_db.RegisterMessage(GetProjectRequest)

BatchGetProjectsRequest = _reflection.GeneratedProtocolMessageType('BatchGetProjectsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETPROJECTSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetProjectsRequest)
  })
_sym_db.RegisterMessage(BatchGetProjectsRequest)

BatchGetProjectsResponse = _reflection.GeneratedProtocolMessageType('BatchGetProjectsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETPROJECTSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetProjectsResponse)
  })
_sym_db.RegisterMessage(BatchGetProjectsResponse)

DeleteProjectRequest = _reflection.GeneratedProtocolMessageType('DeleteProjectRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEPROJECTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteProjectRequest)
  })
_sym_db.RegisterMessage(DeleteProjectRequest)

ListProjectsRequest = _reflection.GeneratedProtocolMessageType('ListProjectsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTPROJECTSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListProjectsRequest)
  })
_sym_db.RegisterMessage(ListProjectsRequest)

ListProjectsResponse = _reflection.GeneratedProtocolMessageType('ListProjectsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTPROJECTSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListProjectsResponse)
  })
_sym_db.RegisterMessage(ListProjectsResponse)

ReRunRequest = _reflection.GeneratedProtocolMessageType('ReRunRequest', (_message.Message,), {
  'DESCRIPTOR' : _RERUNREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ReRunRequest)
  })
_sym_db.RegisterMessage(ReRunRequest)

GetModelStatsRequest = _reflection.GeneratedProtocolMessageType('GetModelStatsRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETMODELSTATSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetModelStatsRequest)
  })
_sym_db.RegisterMessage(GetModelStatsRequest)

GetUsageRequest = _reflection.GeneratedProtocolMessageType('GetUsageRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETUSAGEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetUsageRequest)
  })
_sym_db.RegisterMessage(GetUsageRequest)

GetUsageResponse = _reflection.GeneratedProtocolMessageType('GetUsageResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETUSAGERESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetUsageResponse)
  })
_sym_db.RegisterMessage(GetUsageResponse)

ListPricesRequest = _reflection.GeneratedProtocolMessageType('ListPricesRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTPRICESREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListPricesRequest)
  })
_sym_db.RegisterMessage(ListPricesRequest)

ListPricesResponse = _reflection.GeneratedProtocolMessageType('ListPricesResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTPRICESRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListPricesResponse)
  })
_sym_db.RegisterMessage(ListPricesResponse)

ListInvoicesRequest = _reflection.GeneratedProtocolMessageType('ListInvoicesRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTINVOICESREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListInvoicesRequest)
  })
_sym_db.RegisterMessage(ListInvoicesRequest)

ListInvoicesResponse = _reflection.GeneratedProtocolMessageType('ListInvoicesResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTINVOICESRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListInvoicesResponse)
  })
_sym_db.RegisterMessage(ListInvoicesResponse)

GetPriceRequest = _reflection.GeneratedProtocolMessageType('GetPriceRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETPRICEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetPriceRequest)
  })
_sym_db.RegisterMessage(GetPriceRequest)

GetPaymentMethodRequest = _reflection.GeneratedProtocolMessageType('GetPaymentMethodRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETPAYMENTMETHODREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetPaymentMethodRequest)
  })
_sym_db.RegisterMessage(GetPaymentMethodRequest)

GetUpcomingInvoiceRequest = _reflection.GeneratedProtocolMessageType('GetUpcomingInvoiceRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETUPCOMINGINVOICEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetUpcomingInvoiceRequest)
  })
_sym_db.RegisterMessage(GetUpcomingInvoiceRequest)

GetSubscriptionRequest = _reflection.GeneratedProtocolMessageType('GetSubscriptionRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETSUBSCRIPTIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetSubscriptionRequest)
  })
_sym_db.RegisterMessage(GetSubscriptionRequest)

GetBillingAccountRequest = _reflection.GeneratedProtocolMessageType('GetBillingAccountRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETBILLINGACCOUNTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetBillingAccountRequest)
  })
_sym_db.RegisterMessage(GetBillingAccountRequest)

UpdateBillingAccountRequest = _reflection.GeneratedProtocolMessageType('UpdateBillingAccountRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEBILLINGACCOUNTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UpdateBillingAccountRequest)
  })
_sym_db.RegisterMessage(UpdateBillingAccountRequest)

UpdateSubscriptionRequest = _reflection.GeneratedProtocolMessageType('UpdateSubscriptionRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATESUBSCRIPTIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UpdateSubscriptionRequest)
  })
_sym_db.RegisterMessage(UpdateSubscriptionRequest)

CreatePromotionRequest = _reflection.GeneratedProtocolMessageType('CreatePromotionRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEPROMOTIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreatePromotionRequest)
  })
_sym_db.RegisterMessage(CreatePromotionRequest)

ListPromotionsRequest = _reflection.GeneratedProtocolMessageType('ListPromotionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTPROMOTIONSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListPromotionsRequest)
  })
_sym_db.RegisterMessage(ListPromotionsRequest)

ListPromotionsResponse = _reflection.GeneratedProtocolMessageType('ListPromotionsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTPROMOTIONSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListPromotionsResponse)
  })
_sym_db.RegisterMessage(ListPromotionsResponse)

GetPromotionRequest = _reflection.GeneratedProtocolMessageType('GetPromotionRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETPROMOTIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetPromotionRequest)
  })
_sym_db.RegisterMessage(GetPromotionRequest)

BatchGetPromotionsRequest = _reflection.GeneratedProtocolMessageType('BatchGetPromotionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETPROMOTIONSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetPromotionsRequest)
  })
_sym_db.RegisterMessage(BatchGetPromotionsRequest)

BatchGetPromotionsResponse = _reflection.GeneratedProtocolMessageType('BatchGetPromotionsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETPROMOTIONSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetPromotionsResponse)
  })
_sym_db.RegisterMessage(BatchGetPromotionsResponse)

CreateEventRequest = _reflection.GeneratedProtocolMessageType('CreateEventRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEEVENTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateEventRequest)
  })
_sym_db.RegisterMessage(CreateEventRequest)

ListEventsRequest = _reflection.GeneratedProtocolMessageType('ListEventsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTEVENTSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListEventsRequest)
  })
_sym_db.RegisterMessage(ListEventsRequest)

ListEventsResponse = _reflection.GeneratedProtocolMessageType('ListEventsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTEVENTSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListEventsResponse)
  })
_sym_db.RegisterMessage(ListEventsResponse)

GetEventRequest = _reflection.GeneratedProtocolMessageType('GetEventRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETEVENTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetEventRequest)
  })
_sym_db.RegisterMessage(GetEventRequest)

BatchGetEventsRequest = _reflection.GeneratedProtocolMessageType('BatchGetEventsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETEVENTSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetEventsRequest)
  })
_sym_db.RegisterMessage(BatchGetEventsRequest)

BatchGetEventsResponse = _reflection.GeneratedProtocolMessageType('BatchGetEventsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETEVENTSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetEventsResponse)
  })
_sym_db.RegisterMessage(BatchGetEventsResponse)

CreateModelRequest = _reflection.GeneratedProtocolMessageType('CreateModelRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEMODELREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateModelRequest)
  })
_sym_db.RegisterMessage(CreateModelRequest)

GetModelRequest = _reflection.GeneratedProtocolMessageType('GetModelRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETMODELREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetModelRequest)
  })
_sym_db.RegisterMessage(GetModelRequest)

BatchGetModelsRequest = _reflection.GeneratedProtocolMessageType('BatchGetModelsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETMODELSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetModelsRequest)
  })
_sym_db.RegisterMessage(BatchGetModelsRequest)

BatchGetModelsResponse = _reflection.GeneratedProtocolMessageType('BatchGetModelsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETMODELSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetModelsResponse)
  })
_sym_db.RegisterMessage(BatchGetModelsResponse)

ListModelsRequest = _reflection.GeneratedProtocolMessageType('ListModelsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTMODELSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListModelsRequest)
  })
_sym_db.RegisterMessage(ListModelsRequest)

ListModelsResponse = _reflection.GeneratedProtocolMessageType('ListModelsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTMODELSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListModelsResponse)
  })
_sym_db.RegisterMessage(ListModelsResponse)

DeleteModelRequest = _reflection.GeneratedProtocolMessageType('DeleteModelRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEMODELREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteModelRequest)
  })
_sym_db.RegisterMessage(DeleteModelRequest)

ValidateModelRequest = _reflection.GeneratedProtocolMessageType('ValidateModelRequest', (_message.Message,), {
  'DESCRIPTOR' : _VALIDATEMODELREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ValidateModelRequest)
  })
_sym_db.RegisterMessage(ValidateModelRequest)

CreateModelVersionRequest = _reflection.GeneratedProtocolMessageType('CreateModelVersionRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEMODELVERSIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateModelVersionRequest)
  })
_sym_db.RegisterMessage(CreateModelVersionRequest)

GetModelVersionRequest = _reflection.GeneratedProtocolMessageType('GetModelVersionRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETMODELVERSIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetModelVersionRequest)
  })
_sym_db.RegisterMessage(GetModelVersionRequest)

BatchGetModelVersionsRequest = _reflection.GeneratedProtocolMessageType('BatchGetModelVersionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETMODELVERSIONSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetModelVersionsRequest)
  })
_sym_db.RegisterMessage(BatchGetModelVersionsRequest)

BatchGetModelVersionsResponse = _reflection.GeneratedProtocolMessageType('BatchGetModelVersionsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETMODELVERSIONSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetModelVersionsResponse)
  })
_sym_db.RegisterMessage(BatchGetModelVersionsResponse)

ListModelVersionsRequest = _reflection.GeneratedProtocolMessageType('ListModelVersionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTMODELVERSIONSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListModelVersionsRequest)
  })
_sym_db.RegisterMessage(ListModelVersionsRequest)

ListModelVersionsResponse = _reflection.GeneratedProtocolMessageType('ListModelVersionsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTMODELVERSIONSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListModelVersionsResponse)
  })
_sym_db.RegisterMessage(ListModelVersionsResponse)

CreateExperimentRequest = _reflection.GeneratedProtocolMessageType('CreateExperimentRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEEXPERIMENTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateExperimentRequest)
  })
_sym_db.RegisterMessage(CreateExperimentRequest)

GetExperimentRequest = _reflection.GeneratedProtocolMessageType('GetExperimentRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETEXPERIMENTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetExperimentRequest)
  })
_sym_db.RegisterMessage(GetExperimentRequest)

BatchGetExperimentsRequest = _reflection.GeneratedProtocolMessageType('BatchGetExperimentsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETEXPERIMENTSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetExperimentsRequest)
  })
_sym_db.RegisterMessage(BatchGetExperimentsRequest)

BatchGetExperimentsResponse = _reflection.GeneratedProtocolMessageType('BatchGetExperimentsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETEXPERIMENTSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetExperimentsResponse)
  })
_sym_db.RegisterMessage(BatchGetExperimentsResponse)

ListExperimentsRequest = _reflection.GeneratedProtocolMessageType('ListExperimentsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTEXPERIMENTSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListExperimentsRequest)
  })
_sym_db.RegisterMessage(ListExperimentsRequest)

ListExperimentsResponse = _reflection.GeneratedProtocolMessageType('ListExperimentsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTEXPERIMENTSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListExperimentsResponse)
  })
_sym_db.RegisterMessage(ListExperimentsResponse)

CreateBatchPredictionRequest = _reflection.GeneratedProtocolMessageType('CreateBatchPredictionRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEBATCHPREDICTIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateBatchPredictionRequest)
  })
_sym_db.RegisterMessage(CreateBatchPredictionRequest)

ListBatchPredictionsRequest = _reflection.GeneratedProtocolMessageType('ListBatchPredictionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTBATCHPREDICTIONSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListBatchPredictionsRequest)
  })
_sym_db.RegisterMessage(ListBatchPredictionsRequest)

ListBatchPredictionsResponse = _reflection.GeneratedProtocolMessageType('ListBatchPredictionsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTBATCHPREDICTIONSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListBatchPredictionsResponse)
  })
_sym_db.RegisterMessage(ListBatchPredictionsResponse)

BatchPredictionRequest = _reflection.GeneratedProtocolMessageType('BatchPredictionRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHPREDICTIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchPredictionRequest)
  })
_sym_db.RegisterMessage(BatchPredictionRequest)

BatchGetBatchPredictionsRequest = _reflection.GeneratedProtocolMessageType('BatchGetBatchPredictionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETBATCHPREDICTIONSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetBatchPredictionsRequest)
  })
_sym_db.RegisterMessage(BatchGetBatchPredictionsRequest)

BatchGetBatchPredictionsResponse = _reflection.GeneratedProtocolMessageType('BatchGetBatchPredictionsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETBATCHPREDICTIONSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetBatchPredictionsResponse)
  })
_sym_db.RegisterMessage(BatchGetBatchPredictionsResponse)

YamlDiffRequest = _reflection.GeneratedProtocolMessageType('YamlDiffRequest', (_message.Message,), {
  'DESCRIPTOR' : _YAMLDIFFREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.YamlDiffRequest)
  })
_sym_db.RegisterMessage(YamlDiffRequest)

YamlDiffResponse = _reflection.GeneratedProtocolMessageType('YamlDiffResponse', (_message.Message,), {
  'DESCRIPTOR' : _YAMLDIFFRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.YamlDiffResponse)
  })
_sym_db.RegisterMessage(YamlDiffResponse)

CreateEnvironmentRequest = _reflection.GeneratedProtocolMessageType('CreateEnvironmentRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEENVIRONMENTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateEnvironmentRequest)
  })
_sym_db.RegisterMessage(CreateEnvironmentRequest)

UpdateEnvironmentRequest = _reflection.GeneratedProtocolMessageType('UpdateEnvironmentRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEENVIRONMENTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UpdateEnvironmentRequest)
  })
_sym_db.RegisterMessage(UpdateEnvironmentRequest)

GetEnvironmentRequest = _reflection.GeneratedProtocolMessageType('GetEnvironmentRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETENVIRONMENTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetEnvironmentRequest)
  })
_sym_db.RegisterMessage(GetEnvironmentRequest)

BatchGetEnvironmentsRequest = _reflection.GeneratedProtocolMessageType('BatchGetEnvironmentsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETENVIRONMENTSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetEnvironmentsRequest)
  })
_sym_db.RegisterMessage(BatchGetEnvironmentsRequest)

BatchGetEnvironmentsResponse = _reflection.GeneratedProtocolMessageType('BatchGetEnvironmentsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETENVIRONMENTSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetEnvironmentsResponse)
  })
_sym_db.RegisterMessage(BatchGetEnvironmentsResponse)

DeleteEnvironmentRequest = _reflection.GeneratedProtocolMessageType('DeleteEnvironmentRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEENVIRONMENTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteEnvironmentRequest)
  })
_sym_db.RegisterMessage(DeleteEnvironmentRequest)

ListEnvironmentsRequest = _reflection.GeneratedProtocolMessageType('ListEnvironmentsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTENVIRONMENTSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListEnvironmentsRequest)
  })
_sym_db.RegisterMessage(ListEnvironmentsRequest)

ListEnvironmentsResponse = _reflection.GeneratedProtocolMessageType('ListEnvironmentsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTENVIRONMENTSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListEnvironmentsResponse)
  })
_sym_db.RegisterMessage(ListEnvironmentsResponse)

GetSystemMessageRequest = _reflection.GeneratedProtocolMessageType('GetSystemMessageRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETSYSTEMMESSAGEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetSystemMessageRequest)
  })
_sym_db.RegisterMessage(GetSystemMessageRequest)

GetSystemMessageResponse = _reflection.GeneratedProtocolMessageType('GetSystemMessageResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETSYSTEMMESSAGERESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetSystemMessageResponse)
  })
_sym_db.RegisterMessage(GetSystemMessageResponse)

GetFeatureFlagRequest = _reflection.GeneratedProtocolMessageType('GetFeatureFlagRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETFEATUREFLAGREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetFeatureFlagRequest)
  })
_sym_db.RegisterMessage(GetFeatureFlagRequest)

GetFeatureFlagResponse = _reflection.GeneratedProtocolMessageType('GetFeatureFlagResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETFEATUREFLAGRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetFeatureFlagResponse)
  })
_sym_db.RegisterMessage(GetFeatureFlagResponse)

GetOrganizationFeatureFlagsRequest = _reflection.GeneratedProtocolMessageType('GetOrganizationFeatureFlagsRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETORGANIZATIONFEATUREFLAGSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetOrganizationFeatureFlagsRequest)
  })
_sym_db.RegisterMessage(GetOrganizationFeatureFlagsRequest)

GetOrganizationFeatureFlagsResponse = _reflection.GeneratedProtocolMessageType('GetOrganizationFeatureFlagsResponse', (_message.Message,), {

  'FeatureFlagsEntry' : _reflection.GeneratedProtocolMessageType('FeatureFlagsEntry', (_message.Message,), {
    'DESCRIPTOR' : _GETORGANIZATIONFEATUREFLAGSRESPONSE_FEATUREFLAGSENTRY,
    '__module__' : 'continual.rpc.management.v1.management_pb2'
    # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetOrganizationFeatureFlagsResponse.FeatureFlagsEntry)
    })
  ,
  'DESCRIPTOR' : _GETORGANIZATIONFEATUREFLAGSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetOrganizationFeatureFlagsResponse)
  })
_sym_db.RegisterMessage(GetOrganizationFeatureFlagsResponse)
_sym_db.RegisterMessage(GetOrganizationFeatureFlagsResponse.FeatureFlagsEntry)

CreateArtifactRequest = _reflection.GeneratedProtocolMessageType('CreateArtifactRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEARTIFACTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateArtifactRequest)
  })
_sym_db.RegisterMessage(CreateArtifactRequest)

GetArtifactRequest = _reflection.GeneratedProtocolMessageType('GetArtifactRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETARTIFACTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetArtifactRequest)
  })
_sym_db.RegisterMessage(GetArtifactRequest)

BatchGetArtifactsRequest = _reflection.GeneratedProtocolMessageType('BatchGetArtifactsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETARTIFACTSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetArtifactsRequest)
  })
_sym_db.RegisterMessage(BatchGetArtifactsRequest)

BatchGetArtifactsResponse = _reflection.GeneratedProtocolMessageType('BatchGetArtifactsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETARTIFACTSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetArtifactsResponse)
  })
_sym_db.RegisterMessage(BatchGetArtifactsResponse)

DeleteArtifactRequest = _reflection.GeneratedProtocolMessageType('DeleteArtifactRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEARTIFACTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteArtifactRequest)
  })
_sym_db.RegisterMessage(DeleteArtifactRequest)

ListArtifactsRequest = _reflection.GeneratedProtocolMessageType('ListArtifactsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTARTIFACTSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListArtifactsRequest)
  })
_sym_db.RegisterMessage(ListArtifactsRequest)

ListArtifactsResponse = _reflection.GeneratedProtocolMessageType('ListArtifactsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTARTIFACTSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListArtifactsResponse)
  })
_sym_db.RegisterMessage(ListArtifactsResponse)

GenerateArtifactUploadURLRequest = _reflection.GeneratedProtocolMessageType('GenerateArtifactUploadURLRequest', (_message.Message,), {
  'DESCRIPTOR' : _GENERATEARTIFACTUPLOADURLREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GenerateArtifactUploadURLRequest)
  })
_sym_db.RegisterMessage(GenerateArtifactUploadURLRequest)

GenerateArtifactUploadURLResponse = _reflection.GeneratedProtocolMessageType('GenerateArtifactUploadURLResponse', (_message.Message,), {
  'DESCRIPTOR' : _GENERATEARTIFACTUPLOADURLRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GenerateArtifactUploadURLResponse)
  })
_sym_db.RegisterMessage(GenerateArtifactUploadURLResponse)

ListUserIntegrationProvidersRequest = _reflection.GeneratedProtocolMessageType('ListUserIntegrationProvidersRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTUSERINTEGRATIONPROVIDERSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListUserIntegrationProvidersRequest)
  })
_sym_db.RegisterMessage(ListUserIntegrationProvidersRequest)

ListUserIntegrationProvidersResponse = _reflection.GeneratedProtocolMessageType('ListUserIntegrationProvidersResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTUSERINTEGRATIONPROVIDERSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListUserIntegrationProvidersResponse)
  })
_sym_db.RegisterMessage(ListUserIntegrationProvidersResponse)

ListUserIntegrationsRequest = _reflection.GeneratedProtocolMessageType('ListUserIntegrationsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTUSERINTEGRATIONSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListUserIntegrationsRequest)
  })
_sym_db.RegisterMessage(ListUserIntegrationsRequest)

ListUserIntegrationsResponse = _reflection.GeneratedProtocolMessageType('ListUserIntegrationsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTUSERINTEGRATIONSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListUserIntegrationsResponse)
  })
_sym_db.RegisterMessage(ListUserIntegrationsResponse)

DisconnectUserIntegrationRequest = _reflection.GeneratedProtocolMessageType('DisconnectUserIntegrationRequest', (_message.Message,), {
  'DESCRIPTOR' : _DISCONNECTUSERINTEGRATIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DisconnectUserIntegrationRequest)
  })
_sym_db.RegisterMessage(DisconnectUserIntegrationRequest)

ListGitProviderNamespacesRequest = _reflection.GeneratedProtocolMessageType('ListGitProviderNamespacesRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTGITPROVIDERNAMESPACESREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListGitProviderNamespacesRequest)
  })
_sym_db.RegisterMessage(ListGitProviderNamespacesRequest)

ListGitProviderNamespacesResponse = _reflection.GeneratedProtocolMessageType('ListGitProviderNamespacesResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTGITPROVIDERNAMESPACESRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListGitProviderNamespacesResponse)
  })
_sym_db.RegisterMessage(ListGitProviderNamespacesResponse)

ListGitProviderRepositoriesRequest = _reflection.GeneratedProtocolMessageType('ListGitProviderRepositoriesRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTGITPROVIDERREPOSITORIESREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListGitProviderRepositoriesRequest)
  })
_sym_db.RegisterMessage(ListGitProviderRepositoriesRequest)

ListGitProviderRepositoriesResponse = _reflection.GeneratedProtocolMessageType('ListGitProviderRepositoriesResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTGITPROVIDERREPOSITORIESRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListGitProviderRepositoriesResponse)
  })
_sym_db.RegisterMessage(ListGitProviderRepositoriesResponse)

ConnectProjectToGitRepositoryRequest = _reflection.GeneratedProtocolMessageType('ConnectProjectToGitRepositoryRequest', (_message.Message,), {
  'DESCRIPTOR' : _CONNECTPROJECTTOGITREPOSITORYREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ConnectProjectToGitRepositoryRequest)
  })
_sym_db.RegisterMessage(ConnectProjectToGitRepositoryRequest)

ConnectProjectToGitRepositoryResponse = _reflection.GeneratedProtocolMessageType('ConnectProjectToGitRepositoryResponse', (_message.Message,), {
  'DESCRIPTOR' : _CONNECTPROJECTTOGITREPOSITORYRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ConnectProjectToGitRepositoryResponse)
  })
_sym_db.RegisterMessage(ConnectProjectToGitRepositoryResponse)

GetProjectGitIntegrationRequest = _reflection.GeneratedProtocolMessageType('GetProjectGitIntegrationRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETPROJECTGITINTEGRATIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetProjectGitIntegrationRequest)
  })
_sym_db.RegisterMessage(GetProjectGitIntegrationRequest)

SeedProjectGitRepositoryRequest = _reflection.GeneratedProtocolMessageType('SeedProjectGitRepositoryRequest', (_message.Message,), {
  'DESCRIPTOR' : _SEEDPROJECTGITREPOSITORYREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.SeedProjectGitRepositoryRequest)
  })
_sym_db.RegisterMessage(SeedProjectGitRepositoryRequest)

GenerateSsoAdminPortalURLRequest = _reflection.GeneratedProtocolMessageType('GenerateSsoAdminPortalURLRequest', (_message.Message,), {
  'DESCRIPTOR' : _GENERATESSOADMINPORTALURLREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GenerateSsoAdminPortalURLRequest)
  })
_sym_db.RegisterMessage(GenerateSsoAdminPortalURLRequest)

GenerateSsoAdminPortalURLResponse = _reflection.GeneratedProtocolMessageType('GenerateSsoAdminPortalURLResponse', (_message.Message,), {
  'DESCRIPTOR' : _GENERATESSOADMINPORTALURLRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GenerateSsoAdminPortalURLResponse)
  })
_sym_db.RegisterMessage(GenerateSsoAdminPortalURLResponse)

GenerateDSyncAdminPortalURLRequest = _reflection.GeneratedProtocolMessageType('GenerateDSyncAdminPortalURLRequest', (_message.Message,), {
  'DESCRIPTOR' : _GENERATEDSYNCADMINPORTALURLREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GenerateDSyncAdminPortalURLRequest)
  })
_sym_db.RegisterMessage(GenerateDSyncAdminPortalURLRequest)

GenerateDSyncAdminPortalURLResponse = _reflection.GeneratedProtocolMessageType('GenerateDSyncAdminPortalURLResponse', (_message.Message,), {
  'DESCRIPTOR' : _GENERATEDSYNCADMINPORTALURLRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GenerateDSyncAdminPortalURLResponse)
  })
_sym_db.RegisterMessage(GenerateDSyncAdminPortalURLResponse)

ExchangeTokenRequest = _reflection.GeneratedProtocolMessageType('ExchangeTokenRequest', (_message.Message,), {
  'DESCRIPTOR' : _EXCHANGETOKENREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ExchangeTokenRequest)
  })
_sym_db.RegisterMessage(ExchangeTokenRequest)

ExchangeTokenResponse = _reflection.GeneratedProtocolMessageType('ExchangeTokenResponse', (_message.Message,), {
  'DESCRIPTOR' : _EXCHANGETOKENRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ExchangeTokenResponse)
  })
_sym_db.RegisterMessage(ExchangeTokenResponse)

GetPrincipalRequest = _reflection.GeneratedProtocolMessageType('GetPrincipalRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETPRINCIPALREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetPrincipalRequest)
  })
_sym_db.RegisterMessage(GetPrincipalRequest)

GetApiKeyRequest = _reflection.GeneratedProtocolMessageType('GetApiKeyRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETAPIKEYREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetApiKeyRequest)
  })
_sym_db.RegisterMessage(GetApiKeyRequest)

BatchGetApiKeysRequest = _reflection.GeneratedProtocolMessageType('BatchGetApiKeysRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETAPIKEYSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetApiKeysRequest)
  })
_sym_db.RegisterMessage(BatchGetApiKeysRequest)

BatchGetApiKeysResponse = _reflection.GeneratedProtocolMessageType('BatchGetApiKeysResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETAPIKEYSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetApiKeysResponse)
  })
_sym_db.RegisterMessage(BatchGetApiKeysResponse)

ListApiKeysRequest = _reflection.GeneratedProtocolMessageType('ListApiKeysRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTAPIKEYSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListApiKeysRequest)
  })
_sym_db.RegisterMessage(ListApiKeysRequest)

ListApiKeysResponse = _reflection.GeneratedProtocolMessageType('ListApiKeysResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTAPIKEYSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListApiKeysResponse)
  })
_sym_db.RegisterMessage(ListApiKeysResponse)

CreateApiKeyRequest = _reflection.GeneratedProtocolMessageType('CreateApiKeyRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEAPIKEYREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateApiKeyRequest)
  })
_sym_db.RegisterMessage(CreateApiKeyRequest)

UpdateApiKeyRequest = _reflection.GeneratedProtocolMessageType('UpdateApiKeyRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEAPIKEYREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UpdateApiKeyRequest)
  })
_sym_db.RegisterMessage(UpdateApiKeyRequest)

RevokeApiKeyRequest = _reflection.GeneratedProtocolMessageType('RevokeApiKeyRequest', (_message.Message,), {
  'DESCRIPTOR' : _REVOKEAPIKEYREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.RevokeApiKeyRequest)
  })
_sym_db.RegisterMessage(RevokeApiKeyRequest)

GetApiKeyProjectRequest = _reflection.GeneratedProtocolMessageType('GetApiKeyProjectRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETAPIKEYPROJECTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetApiKeyProjectRequest)
  })
_sym_db.RegisterMessage(GetApiKeyProjectRequest)

GetRunRequest = _reflection.GeneratedProtocolMessageType('GetRunRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETRUNREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetRunRequest)
  })
_sym_db.RegisterMessage(GetRunRequest)

BatchGetRunsRequest = _reflection.GeneratedProtocolMessageType('BatchGetRunsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETRUNSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetRunsRequest)
  })
_sym_db.RegisterMessage(BatchGetRunsRequest)

BatchGetRunsResponse = _reflection.GeneratedProtocolMessageType('BatchGetRunsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETRUNSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetRunsResponse)
  })
_sym_db.RegisterMessage(BatchGetRunsResponse)

ListRunsRequest = _reflection.GeneratedProtocolMessageType('ListRunsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTRUNSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListRunsRequest)
  })
_sym_db.RegisterMessage(ListRunsRequest)

ListRunsResponse = _reflection.GeneratedProtocolMessageType('ListRunsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTRUNSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListRunsResponse)
  })
_sym_db.RegisterMessage(ListRunsResponse)

CreateRunRequest = _reflection.GeneratedProtocolMessageType('CreateRunRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATERUNREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateRunRequest)
  })
_sym_db.RegisterMessage(CreateRunRequest)

UpdateRunRequest = _reflection.GeneratedProtocolMessageType('UpdateRunRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATERUNREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UpdateRunRequest)
  })
_sym_db.RegisterMessage(UpdateRunRequest)

DeleteRunRequest = _reflection.GeneratedProtocolMessageType('DeleteRunRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETERUNREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteRunRequest)
  })
_sym_db.RegisterMessage(DeleteRunRequest)

GetRunSummaryRequest = _reflection.GeneratedProtocolMessageType('GetRunSummaryRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETRUNSUMMARYREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetRunSummaryRequest)
  })
_sym_db.RegisterMessage(GetRunSummaryRequest)

GetRunCheckSummaryRequest = _reflection.GeneratedProtocolMessageType('GetRunCheckSummaryRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETRUNCHECKSUMMARYREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetRunCheckSummaryRequest)
  })
_sym_db.RegisterMessage(GetRunCheckSummaryRequest)

LogRunHeartbeatRequest = _reflection.GeneratedProtocolMessageType('LogRunHeartbeatRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOGRUNHEARTBEATREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.LogRunHeartbeatRequest)
  })
_sym_db.RegisterMessage(LogRunHeartbeatRequest)

GetMetricRequest = _reflection.GeneratedProtocolMessageType('GetMetricRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETMETRICREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetMetricRequest)
  })
_sym_db.RegisterMessage(GetMetricRequest)

BatchGetMetricsRequest = _reflection.GeneratedProtocolMessageType('BatchGetMetricsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETMETRICSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetMetricsRequest)
  })
_sym_db.RegisterMessage(BatchGetMetricsRequest)

BatchGetMetricsResponse = _reflection.GeneratedProtocolMessageType('BatchGetMetricsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETMETRICSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetMetricsResponse)
  })
_sym_db.RegisterMessage(BatchGetMetricsResponse)

ListMetricsRequest = _reflection.GeneratedProtocolMessageType('ListMetricsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTMETRICSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListMetricsRequest)
  })
_sym_db.RegisterMessage(ListMetricsRequest)

ListMetricsResponse = _reflection.GeneratedProtocolMessageType('ListMetricsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTMETRICSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListMetricsResponse)
  })
_sym_db.RegisterMessage(ListMetricsResponse)

CreateMetricRequest = _reflection.GeneratedProtocolMessageType('CreateMetricRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEMETRICREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateMetricRequest)
  })
_sym_db.RegisterMessage(CreateMetricRequest)

LogMetricRequest = _reflection.GeneratedProtocolMessageType('LogMetricRequest', (_message.Message,), {
  'DESCRIPTOR' : _LOGMETRICREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.LogMetricRequest)
  })
_sym_db.RegisterMessage(LogMetricRequest)

UpdateMetricRequest = _reflection.GeneratedProtocolMessageType('UpdateMetricRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEMETRICREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UpdateMetricRequest)
  })
_sym_db.RegisterMessage(UpdateMetricRequest)

GetMetadataRequest = _reflection.GeneratedProtocolMessageType('GetMetadataRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETMETADATAREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetMetadataRequest)
  })
_sym_db.RegisterMessage(GetMetadataRequest)

BatchGetMetadataRequest = _reflection.GeneratedProtocolMessageType('BatchGetMetadataRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETMETADATAREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetMetadataRequest)
  })
_sym_db.RegisterMessage(BatchGetMetadataRequest)

BatchGetMetadataResponse = _reflection.GeneratedProtocolMessageType('BatchGetMetadataResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETMETADATARESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetMetadataResponse)
  })
_sym_db.RegisterMessage(BatchGetMetadataResponse)

ListMetadataRequest = _reflection.GeneratedProtocolMessageType('ListMetadataRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTMETADATAREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListMetadataRequest)
  })
_sym_db.RegisterMessage(ListMetadataRequest)

ListMetadataResponse = _reflection.GeneratedProtocolMessageType('ListMetadataResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTMETADATARESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListMetadataResponse)
  })
_sym_db.RegisterMessage(ListMetadataResponse)

CreateMetadataRequest = _reflection.GeneratedProtocolMessageType('CreateMetadataRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEMETADATAREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateMetadataRequest)
  })
_sym_db.RegisterMessage(CreateMetadataRequest)

UpdateMetadataRequest = _reflection.GeneratedProtocolMessageType('UpdateMetadataRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEMETADATAREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UpdateMetadataRequest)
  })
_sym_db.RegisterMessage(UpdateMetadataRequest)

DeleteMetadataRequest = _reflection.GeneratedProtocolMessageType('DeleteMetadataRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEMETADATAREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteMetadataRequest)
  })
_sym_db.RegisterMessage(DeleteMetadataRequest)

GetNotificationPolicyRequest = _reflection.GeneratedProtocolMessageType('GetNotificationPolicyRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETNOTIFICATIONPOLICYREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetNotificationPolicyRequest)
  })
_sym_db.RegisterMessage(GetNotificationPolicyRequest)

BatchGetNotificationPoliciesRequest = _reflection.GeneratedProtocolMessageType('BatchGetNotificationPoliciesRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETNOTIFICATIONPOLICIESREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetNotificationPoliciesRequest)
  })
_sym_db.RegisterMessage(BatchGetNotificationPoliciesRequest)

BatchGetNotificationPoliciesResponse = _reflection.GeneratedProtocolMessageType('BatchGetNotificationPoliciesResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETNOTIFICATIONPOLICIESRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetNotificationPoliciesResponse)
  })
_sym_db.RegisterMessage(BatchGetNotificationPoliciesResponse)

ListNotificationPoliciesRequest = _reflection.GeneratedProtocolMessageType('ListNotificationPoliciesRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTNOTIFICATIONPOLICIESREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListNotificationPoliciesRequest)
  })
_sym_db.RegisterMessage(ListNotificationPoliciesRequest)

ListNotificationPoliciesResponse = _reflection.GeneratedProtocolMessageType('ListNotificationPoliciesResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTNOTIFICATIONPOLICIESRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListNotificationPoliciesResponse)
  })
_sym_db.RegisterMessage(ListNotificationPoliciesResponse)

CreateNotificationPolicyRequest = _reflection.GeneratedProtocolMessageType('CreateNotificationPolicyRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATENOTIFICATIONPOLICYREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateNotificationPolicyRequest)
  })
_sym_db.RegisterMessage(CreateNotificationPolicyRequest)

UpdateNotificationPolicyRequest = _reflection.GeneratedProtocolMessageType('UpdateNotificationPolicyRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATENOTIFICATIONPOLICYREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UpdateNotificationPolicyRequest)
  })
_sym_db.RegisterMessage(UpdateNotificationPolicyRequest)

DeleteNotificationPolicyRequest = _reflection.GeneratedProtocolMessageType('DeleteNotificationPolicyRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETENOTIFICATIONPOLICYREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteNotificationPolicyRequest)
  })
_sym_db.RegisterMessage(DeleteNotificationPolicyRequest)

GetTagRequest = _reflection.GeneratedProtocolMessageType('GetTagRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETTAGREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetTagRequest)
  })
_sym_db.RegisterMessage(GetTagRequest)

BatchGetTagsRequest = _reflection.GeneratedProtocolMessageType('BatchGetTagsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETTAGSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetTagsRequest)
  })
_sym_db.RegisterMessage(BatchGetTagsRequest)

BatchGetTagsResponse = _reflection.GeneratedProtocolMessageType('BatchGetTagsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETTAGSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetTagsResponse)
  })
_sym_db.RegisterMessage(BatchGetTagsResponse)

ListTagsRequest = _reflection.GeneratedProtocolMessageType('ListTagsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTTAGSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListTagsRequest)
  })
_sym_db.RegisterMessage(ListTagsRequest)

ListTagsResponse = _reflection.GeneratedProtocolMessageType('ListTagsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTTAGSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListTagsResponse)
  })
_sym_db.RegisterMessage(ListTagsResponse)

CreateTagRequest = _reflection.GeneratedProtocolMessageType('CreateTagRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATETAGREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateTagRequest)
  })
_sym_db.RegisterMessage(CreateTagRequest)

UpdateTagRequest = _reflection.GeneratedProtocolMessageType('UpdateTagRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATETAGREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UpdateTagRequest)
  })
_sym_db.RegisterMessage(UpdateTagRequest)

DeleteTagRequest = _reflection.GeneratedProtocolMessageType('DeleteTagRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETETAGREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteTagRequest)
  })
_sym_db.RegisterMessage(DeleteTagRequest)

GetDataProfileRequest = _reflection.GeneratedProtocolMessageType('GetDataProfileRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETDATAPROFILEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetDataProfileRequest)
  })
_sym_db.RegisterMessage(GetDataProfileRequest)

CreateDataProfileRequest = _reflection.GeneratedProtocolMessageType('CreateDataProfileRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEDATAPROFILEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateDataProfileRequest)
  })
_sym_db.RegisterMessage(CreateDataProfileRequest)

ListDataProfilesRequest = _reflection.GeneratedProtocolMessageType('ListDataProfilesRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTDATAPROFILESREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListDataProfilesRequest)
  })
_sym_db.RegisterMessage(ListDataProfilesRequest)

ListDataProfilesResponse = _reflection.GeneratedProtocolMessageType('ListDataProfilesResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTDATAPROFILESRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListDataProfilesResponse)
  })
_sym_db.RegisterMessage(ListDataProfilesResponse)

UpdateDataProfileRequest = _reflection.GeneratedProtocolMessageType('UpdateDataProfileRequest', (_message.Message,), {
  'DESCRIPTOR' : _UPDATEDATAPROFILEREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.UpdateDataProfileRequest)
  })
_sym_db.RegisterMessage(UpdateDataProfileRequest)

GetResourceCheckRequest = _reflection.GeneratedProtocolMessageType('GetResourceCheckRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETRESOURCECHECKREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetResourceCheckRequest)
  })
_sym_db.RegisterMessage(GetResourceCheckRequest)

ListResourceChecksRequest = _reflection.GeneratedProtocolMessageType('ListResourceChecksRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTRESOURCECHECKSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListResourceChecksRequest)
  })
_sym_db.RegisterMessage(ListResourceChecksRequest)

ListResourceChecksResponse = _reflection.GeneratedProtocolMessageType('ListResourceChecksResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTRESOURCECHECKSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListResourceChecksResponse)
  })
_sym_db.RegisterMessage(ListResourceChecksResponse)

CreateResourceCheckRequest = _reflection.GeneratedProtocolMessageType('CreateResourceCheckRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATERESOURCECHECKREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateResourceCheckRequest)
  })
_sym_db.RegisterMessage(CreateResourceCheckRequest)

CreateDatasetRequest = _reflection.GeneratedProtocolMessageType('CreateDatasetRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEDATASETREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateDatasetRequest)
  })
_sym_db.RegisterMessage(CreateDatasetRequest)

GetDatasetRequest = _reflection.GeneratedProtocolMessageType('GetDatasetRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETDATASETREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetDatasetRequest)
  })
_sym_db.RegisterMessage(GetDatasetRequest)

BatchGetDatasetsRequest = _reflection.GeneratedProtocolMessageType('BatchGetDatasetsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETDATASETSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetDatasetsRequest)
  })
_sym_db.RegisterMessage(BatchGetDatasetsRequest)

BatchGetDatasetsResponse = _reflection.GeneratedProtocolMessageType('BatchGetDatasetsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETDATASETSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetDatasetsResponse)
  })
_sym_db.RegisterMessage(BatchGetDatasetsResponse)

ListDatasetsRequest = _reflection.GeneratedProtocolMessageType('ListDatasetsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTDATASETSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListDatasetsRequest)
  })
_sym_db.RegisterMessage(ListDatasetsRequest)

ListDatasetsResponse = _reflection.GeneratedProtocolMessageType('ListDatasetsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTDATASETSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListDatasetsResponse)
  })
_sym_db.RegisterMessage(ListDatasetsResponse)

DeleteDatasetRequest = _reflection.GeneratedProtocolMessageType('DeleteDatasetRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEDATASETREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteDatasetRequest)
  })
_sym_db.RegisterMessage(DeleteDatasetRequest)

CreateDatasetVersionRequest = _reflection.GeneratedProtocolMessageType('CreateDatasetVersionRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEDATASETVERSIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateDatasetVersionRequest)
  })
_sym_db.RegisterMessage(CreateDatasetVersionRequest)

GetDatasetVersionRequest = _reflection.GeneratedProtocolMessageType('GetDatasetVersionRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETDATASETVERSIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetDatasetVersionRequest)
  })
_sym_db.RegisterMessage(GetDatasetVersionRequest)

BatchGetDatasetVersionsRequest = _reflection.GeneratedProtocolMessageType('BatchGetDatasetVersionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETDATASETVERSIONSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetDatasetVersionsRequest)
  })
_sym_db.RegisterMessage(BatchGetDatasetVersionsRequest)

BatchGetDatasetVersionsResponse = _reflection.GeneratedProtocolMessageType('BatchGetDatasetVersionsResponse', (_message.Message,), {
  'DESCRIPTOR' : _BATCHGETDATASETVERSIONSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.BatchGetDatasetVersionsResponse)
  })
_sym_db.RegisterMessage(BatchGetDatasetVersionsResponse)

ListDatasetVersionsRequest = _reflection.GeneratedProtocolMessageType('ListDatasetVersionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTDATASETVERSIONSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListDatasetVersionsRequest)
  })
_sym_db.RegisterMessage(ListDatasetVersionsRequest)

ListDatasetVersionsResponse = _reflection.GeneratedProtocolMessageType('ListDatasetVersionsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTDATASETVERSIONSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListDatasetVersionsResponse)
  })
_sym_db.RegisterMessage(ListDatasetVersionsResponse)

CreateDatasetVersionAssignmentRequest = _reflection.GeneratedProtocolMessageType('CreateDatasetVersionAssignmentRequest', (_message.Message,), {
  'DESCRIPTOR' : _CREATEDATASETVERSIONASSIGNMENTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.CreateDatasetVersionAssignmentRequest)
  })
_sym_db.RegisterMessage(CreateDatasetVersionAssignmentRequest)

GetDatasetVersionAssignmentRequest = _reflection.GeneratedProtocolMessageType('GetDatasetVersionAssignmentRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETDATASETVERSIONASSIGNMENTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetDatasetVersionAssignmentRequest)
  })
_sym_db.RegisterMessage(GetDatasetVersionAssignmentRequest)

DeleteDatasetVersionAssignmentRequest = _reflection.GeneratedProtocolMessageType('DeleteDatasetVersionAssignmentRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEDATASETVERSIONASSIGNMENTREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.DeleteDatasetVersionAssignmentRequest)
  })
_sym_db.RegisterMessage(DeleteDatasetVersionAssignmentRequest)

ListDatasetVersionAssignmentsRequest = _reflection.GeneratedProtocolMessageType('ListDatasetVersionAssignmentsRequest', (_message.Message,), {
  'DESCRIPTOR' : _LISTDATASETVERSIONASSIGNMENTSREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListDatasetVersionAssignmentsRequest)
  })
_sym_db.RegisterMessage(ListDatasetVersionAssignmentsRequest)

ListDatasetVersionAssignmentsResponse = _reflection.GeneratedProtocolMessageType('ListDatasetVersionAssignmentsResponse', (_message.Message,), {
  'DESCRIPTOR' : _LISTDATASETVERSIONASSIGNMENTSRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.ListDatasetVersionAssignmentsResponse)
  })
_sym_db.RegisterMessage(ListDatasetVersionAssignmentsResponse)

GetServerVersionRequest = _reflection.GeneratedProtocolMessageType('GetServerVersionRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETSERVERVERSIONREQUEST,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetServerVersionRequest)
  })
_sym_db.RegisterMessage(GetServerVersionRequest)

GetServerVersionResponse = _reflection.GeneratedProtocolMessageType('GetServerVersionResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETSERVERVERSIONRESPONSE,
  '__module__' : 'continual.rpc.management.v1.management_pb2'
  # @@protoc_insertion_point(class_scope:continual.rpc.management.v1.GetServerVersionResponse)
  })
_sym_db.RegisterMessage(GetServerVersionResponse)


DESCRIPTOR._options = None
_GETORGANIZATIONFEATUREFLAGSRESPONSE_FEATUREFLAGSENTRY._options = None

_MANAGEMENTAPI = _descriptor.ServiceDescriptor(
  name='ManagementAPI',
  full_name='continual.rpc.management.v1.ManagementAPI',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=23051,
  serialized_end=48485,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetServerVersion',
    full_name='continual.rpc.management.v1.ManagementAPI.GetServerVersion',
    index=0,
    containing_service=None,
    input_type=_GETSERVERVERSIONREQUEST,
    output_type=_GETSERVERVERSIONRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetUser',
    full_name='continual.rpc.management.v1.ManagementAPI.GetUser',
    index=1,
    containing_service=None,
    input_type=_GETUSERREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._USER,
    serialized_options=b'\202\323\344\223\002\024\022\022/v1/{name=users/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListUsers',
    full_name='continual.rpc.management.v1.ManagementAPI.ListUsers',
    index=2,
    containing_service=None,
    input_type=_LISTUSERSREQUEST,
    output_type=_LISTUSERSRESPONSE,
    serialized_options=b'\202\323\344\223\002\013\022\t/v1/users',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetUsers',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetUsers',
    index=3,
    containing_service=None,
    input_type=_BATCHGETUSERSREQUEST,
    output_type=_BATCHGETUSERSRESPONSE,
    serialized_options=b'\202\323\344\223\002\024\022\022/v1/users:batchGet',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateUser',
    full_name='continual.rpc.management.v1.ManagementAPI.UpdateUser',
    index=4,
    containing_service=None,
    input_type=_UPDATEUSERREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._USER,
    serialized_options=b'\202\323\344\223\002\0372\027/v1/{user.name=users/*}:\004user',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteUser',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteUser',
    index=5,
    containing_service=None,
    input_type=_DELETEUSERREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\024*\022/v1/{name=users/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetViewer',
    full_name='continual.rpc.management.v1.ManagementAPI.GetViewer',
    index=6,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._USER,
    serialized_options=b'\202\323\344\223\002\014\022\n/v1/viewer',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Register',
    full_name='continual.rpc.management.v1.ManagementAPI.Register',
    index=7,
    containing_service=None,
    input_type=_REGISTERREQUEST,
    output_type=_REGISTERRESPONSE,
    serialized_options=b'\202\323\344\223\002\024\"\022/v1/users:register',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Login',
    full_name='continual.rpc.management.v1.ManagementAPI.Login',
    index=8,
    containing_service=None,
    input_type=_LOGINREQUEST,
    output_type=_LOGINRESPONSE,
    serialized_options=b'\202\323\344\223\002\021\"\017/v1/users:login',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='Logout',
    full_name='continual.rpc.management.v1.ManagementAPI.Logout',
    index=9,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\022\"\020/v1/users:logout',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CheckOrganizationAccess',
    full_name='continual.rpc.management.v1.ManagementAPI.CheckOrganizationAccess',
    index=10,
    containing_service=None,
    input_type=_CHECKORGANIZATIONACCESSREQUEST,
    output_type=_CHECKORGANIZATIONACCESSRESPONSE,
    serialized_options=b'\202\323\344\223\002#\"!/v1/users:checkOrganizationAccess',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetPrincipal',
    full_name='continual.rpc.management.v1.ManagementAPI.GetPrincipal',
    index=11,
    containing_service=None,
    input_type=_GETPRINCIPALREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PRINCIPAL,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CheckViewer',
    full_name='continual.rpc.management.v1.ManagementAPI.CheckViewer',
    index=12,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetApiKey',
    full_name='continual.rpc.management.v1.ManagementAPI.GetApiKey',
    index=13,
    containing_service=None,
    input_type=_GETAPIKEYREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._APIKEY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetApiKeys',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetApiKeys',
    index=14,
    containing_service=None,
    input_type=_BATCHGETAPIKEYSREQUEST,
    output_type=_BATCHGETAPIKEYSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListApiKeys',
    full_name='continual.rpc.management.v1.ManagementAPI.ListApiKeys',
    index=15,
    containing_service=None,
    input_type=_LISTAPIKEYSREQUEST,
    output_type=_LISTAPIKEYSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateApiKey',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateApiKey',
    index=16,
    containing_service=None,
    input_type=_CREATEAPIKEYREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._APIKEY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateApiKey',
    full_name='continual.rpc.management.v1.ManagementAPI.UpdateApiKey',
    index=17,
    containing_service=None,
    input_type=_UPDATEAPIKEYREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._APIKEY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='RevokeApiKey',
    full_name='continual.rpc.management.v1.ManagementAPI.RevokeApiKey',
    index=18,
    containing_service=None,
    input_type=_REVOKEAPIKEYREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetApiKeyProject',
    full_name='continual.rpc.management.v1.ManagementAPI.GetApiKeyProject',
    index=19,
    containing_service=None,
    input_type=_GETAPIKEYPROJECTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PROJECT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='SendEmailVerification',
    full_name='continual.rpc.management.v1.ManagementAPI.SendEmailVerification',
    index=20,
    containing_service=None,
    input_type=_SENDEMAILVERIFICATIONREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002!\"\037/v1/users:sendEmailVerification',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='VerifyEmail',
    full_name='continual.rpc.management.v1.ManagementAPI.VerifyEmail',
    index=21,
    containing_service=None,
    input_type=_VERIFYEMAILREQUEST,
    output_type=_VERIFYEMAILRESPONSE,
    serialized_options=b'\202\323\344\223\002\027\"\025/v1/users:verifyEmail',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='SendPasswordReset',
    full_name='continual.rpc.management.v1.ManagementAPI.SendPasswordReset',
    index=22,
    containing_service=None,
    input_type=_SENDPASSWORDRESETREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\035\"\033/v1/users:sendPasswordReset',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ResetPassword',
    full_name='continual.rpc.management.v1.ManagementAPI.ResetPassword',
    index=23,
    containing_service=None,
    input_type=_RESETPASSWORDREQUEST,
    output_type=_RESETPASSWORDRESPONSE,
    serialized_options=b'\202\323\344\223\002\032\"\030/v1/:users:resetPassword',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ChangePassword',
    full_name='continual.rpc.management.v1.ManagementAPI.ChangePassword',
    index=24,
    containing_service=None,
    input_type=_CHANGEPASSWORDREQUEST,
    output_type=_CHANGEPASSWORDRESPONSE,
    serialized_options=b'\202\323\344\223\002\033\"\031/v1/:users:changePassword',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetSession',
    full_name='continual.rpc.management.v1.ManagementAPI.GetSession',
    index=25,
    containing_service=None,
    input_type=_GETSESSIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._SESSION,
    serialized_options=b'\202\323\344\223\002\037\022\035/v1/{name=users/*/sessions/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListSessions',
    full_name='continual.rpc.management.v1.ManagementAPI.ListSessions',
    index=26,
    containing_service=None,
    input_type=_LISTSESSIONSREQUEST,
    output_type=_LISTSESSIONSRESPONSE,
    serialized_options=b'\202\323\344\223\002\037\022\035/v1/{parent=users/*}/sessions\312E\010sessions\322E\004User',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteSession',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteSession',
    index=27,
    containing_service=None,
    input_type=_DELETESESSIONREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\037*\035/v1/{name=users/*/sessions/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteAllSessions',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteAllSessions',
    index=28,
    containing_service=None,
    input_type=_DELETEALLSESSIONSREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\037*\035/v1/{parent=users/*}/sessions',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetOrganization',
    full_name='continual.rpc.management.v1.ManagementAPI.GetOrganization',
    index=29,
    containing_service=None,
    input_type=_GETORGANIZATIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ORGANIZATION,
    serialized_options=b'\202\323\344\223\002\034\022\032/v1/{name=organizations/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetOrganizations',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetOrganizations',
    index=30,
    containing_service=None,
    input_type=_BATCHGETORGANIZATIONSREQUEST,
    output_type=_BATCHGETORGANIZATIONSRESPONSE,
    serialized_options=b'\202\323\344\223\002\034\022\032/v1/organizations:batchGet',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListOrganizations',
    full_name='continual.rpc.management.v1.ManagementAPI.ListOrganizations',
    index=31,
    containing_service=None,
    input_type=_LISTORGANIZATIONSREQUEST,
    output_type=_LISTORGANIZATIONSRESPONSE,
    serialized_options=b'\202\323\344\223\002\023\022\021/v1/organizations',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateOrganization',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateOrganization',
    index=32,
    containing_service=None,
    input_type=_CREATEORGANIZATIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ORGANIZATION,
    serialized_options=b'\202\323\344\223\002!\"\021/v1/organizations:\014organization',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateOrganization',
    full_name='continual.rpc.management.v1.ManagementAPI.UpdateOrganization',
    index=33,
    containing_service=None,
    input_type=_UPDATEORGANIZATIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ORGANIZATION,
    serialized_options=b'\202\323\344\223\00272\'/v1/{organization.name=organizations/*}:\014organization',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteOrganization',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteOrganization',
    index=34,
    containing_service=None,
    input_type=_DELETEORGANIZATIONREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\034*\032/v1/{name=organizations/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetOnboarding',
    full_name='continual.rpc.management.v1.ManagementAPI.GetOnboarding',
    index=35,
    containing_service=None,
    input_type=_GETONBOARDINGREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ONBOARDING,
    serialized_options=b'\202\323\344\223\002\'\022%/v1/{name=organizations/*/onboarding}\322E\014Organization\312E\nonboarding\332E\026name:name:$/onboarding',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateOnboarding',
    full_name='continual.rpc.management.v1.ManagementAPI.UpdateOnboarding',
    index=36,
    containing_service=None,
    input_type=_UPDATEONBOARDINGREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ONBOARDING,
    serialized_options=b'\202\323\344\223\002>20/v1/{onboarding.name=organizations/*/onboarding}:\nonboarding',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListRoles',
    full_name='continual.rpc.management.v1.ManagementAPI.ListRoles',
    index=37,
    containing_service=None,
    input_type=_LISTROLESREQUEST,
    output_type=_LISTROLESRESPONSE,
    serialized_options=b'\202\323\344\223\002\013\022\t/v1/roles',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetRole',
    full_name='continual.rpc.management.v1.ManagementAPI.GetRole',
    index=38,
    containing_service=None,
    input_type=_GETROLEREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ROLE,
    serialized_options=b'\202\323\344\223\002\024\022\022/v1/{name=roles/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetRoles',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetRoles',
    index=39,
    containing_service=None,
    input_type=_BATCHGETROLESREQUEST,
    output_type=_BATCHGETROLESRESPONSE,
    serialized_options=b'\202\323\344\223\002\024\022\022/v1/roles:batchGet',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateAccessPolicy',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateAccessPolicy',
    index=40,
    containing_service=None,
    input_type=_CREATEACCESSPOLICYREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ACCESSPOLICY,
    serialized_options=b'\202\323\344\223\002/\"\036/v1/{parent=**}/accessPolicies:\raccess_policy',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetAccessPolicy',
    full_name='continual.rpc.management.v1.ManagementAPI.GetAccessPolicy',
    index=41,
    containing_service=None,
    input_type=_GETACCESSPOLICYREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ACCESSPOLICY,
    serialized_options=b'\202\323\344\223\002 \022\036/v1/{name=**/accessPolicies/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListAccessPolicies',
    full_name='continual.rpc.management.v1.ManagementAPI.ListAccessPolicies',
    index=42,
    containing_service=None,
    input_type=_LISTACCESSPOLICIESREQUEST,
    output_type=_LISTACCESSPOLICIESRESPONSE,
    serialized_options=b'\202\323\344\223\002 \022\036/v1/{parent=**}/accessPolicies\322E:Organization,Project,Dataset,Model,ModelVersion,Experiment\312E\016accessPolicies\332E\013name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BulkDeleteAccessPolicies',
    full_name='continual.rpc.management.v1.ManagementAPI.BulkDeleteAccessPolicies',
    index=43,
    containing_service=None,
    input_type=_BULKDELETEACCESSPOLICIESREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteAccessPolicy',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteAccessPolicy',
    index=44,
    containing_service=None,
    input_type=_DELETEACCESSPOLICYREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002 *\036/v1/{name=**/accessPolicies/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateAccessPolicy',
    full_name='continual.rpc.management.v1.ManagementAPI.UpdateAccessPolicy',
    index=45,
    containing_service=None,
    input_type=_UPDATEACCESSPOLICYREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ACCESSPOLICY,
    serialized_options=b'\202\323\344\223\002=2,/v1/{access_policy.name=**/accessPolicies/*}:\raccess_policy',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='TestPermissions',
    full_name='continual.rpc.management.v1.ManagementAPI.TestPermissions',
    index=46,
    containing_service=None,
    input_type=_TESTPERMISSIONSREQUEST,
    output_type=_TESTPERMISSIONSRESPONSE,
    serialized_options=b'\202\323\344\223\002&\"!/v1/{resource=**}:testPermissions:\001*\322E:Organization,Project,Dataset,Model,ModelVersion,Experiment\312E\013permissions\332E\rname:resource',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateInvite',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateInvite',
    index=47,
    containing_service=None,
    input_type=_CREATEINVITEREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._INVITE,
    serialized_options=b'\202\323\344\223\002\031\"\027/v1/{parent=**}/invites',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetInvite',
    full_name='continual.rpc.management.v1.ManagementAPI.GetInvite',
    index=48,
    containing_service=None,
    input_type=_GETINVITEREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._INVITE,
    serialized_options=b'\202\323\344\223\002\031\022\027/v1/{name=**/invites/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListInvites',
    full_name='continual.rpc.management.v1.ManagementAPI.ListInvites',
    index=49,
    containing_service=None,
    input_type=_LISTINVITESREQUEST,
    output_type=_LISTINVITESRESPONSE,
    serialized_options=b'\202\323\344\223\002\031\022\027/v1/{parent=**}/invites',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateInvite',
    full_name='continual.rpc.management.v1.ManagementAPI.UpdateInvite',
    index=50,
    containing_service=None,
    input_type=_UPDATEINVITEREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._INVITE,
    serialized_options=b'\202\323\344\223\002(2\036/v1/{invite.name=**/invites/*}:\006invite',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ResendInvite',
    full_name='continual.rpc.management.v1.ManagementAPI.ResendInvite',
    index=51,
    containing_service=None,
    input_type=_RESENDINVITEREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._INVITE,
    serialized_options=b'\202\323\344\223\002 \"\036/v1/{name=**/invites/*}:resend',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteInvite',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteInvite',
    index=52,
    containing_service=None,
    input_type=_DELETEINVITEREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\031*\027/v1/{name=**/invites/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='AcceptInvite',
    full_name='continual.rpc.management.v1.ManagementAPI.AcceptInvite',
    index=53,
    containing_service=None,
    input_type=_ACCEPTINVITEREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\026\"\024/v1/confirm/{code=*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetProject',
    full_name='continual.rpc.management.v1.ManagementAPI.GetProject',
    index=54,
    containing_service=None,
    input_type=_GETPROJECTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PROJECT,
    serialized_options=b'\202\323\344\223\002\027\022\025/v1/{name=projects/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetProjects',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetProjects',
    index=55,
    containing_service=None,
    input_type=_BATCHGETPROJECTSREQUEST,
    output_type=_BATCHGETPROJECTSRESPONSE,
    serialized_options=b'\202\323\344\223\002\027\022\025/v1/projects:batchGet',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListProjects',
    full_name='continual.rpc.management.v1.ManagementAPI.ListProjects',
    index=56,
    containing_service=None,
    input_type=_LISTPROJECTSREQUEST,
    output_type=_LISTPROJECTSRESPONSE,
    serialized_options=b'\202\323\344\223\0027\022\014/v1/projectsZ\'\022%/v1/{parent=organizations/*}/projects\322E\014Organization\312E\010projects\332E\013name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateProject',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateProject',
    index=57,
    containing_service=None,
    input_type=_CREATEPROJECTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PROJECT,
    serialized_options=b'\202\323\344\223\002\027\"\014/v1/projects:\007project',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateProject',
    full_name='continual.rpc.management.v1.ManagementAPI.UpdateProject',
    index=58,
    containing_service=None,
    input_type=_UPDATEPROJECTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PROJECT,
    serialized_options=b'\202\323\344\223\002(2\035/v1/{project.name=projects/*}:\007project',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteProject',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteProject',
    index=59,
    containing_service=None,
    input_type=_DELETEPROJECTREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002\027*\025/v1/{name=projects/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreatePromotion',
    full_name='continual.rpc.management.v1.ManagementAPI.CreatePromotion',
    index=60,
    containing_service=None,
    input_type=_CREATEPROMOTIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PROMOTION,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListPromotions',
    full_name='continual.rpc.management.v1.ManagementAPI.ListPromotions',
    index=61,
    containing_service=None,
    input_type=_LISTPROMOTIONSREQUEST,
    output_type=_LISTPROMOTIONSRESPONSE,
    serialized_options=b'\202\323\344\223\002g\022+/v1/{parent=projects/*/models/*}/promotionsZ8\0226/v1/{parent=projects/*/models/*/versions/*}/promotions\312E\npromotions\322E\022ModelVersion,Model\340E\000\332E\013name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetPromotion',
    full_name='continual.rpc.management.v1.ManagementAPI.GetPromotion',
    index=62,
    containing_service=None,
    input_type=_GETPROMOTIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PROMOTION,
    serialized_options=b'\202\323\344\223\002-\022+/v1/{name=projects/*/models/*/promotions/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetPromotions',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetPromotions',
    index=63,
    containing_service=None,
    input_type=_BATCHGETPROMOTIONSREQUEST,
    output_type=_BATCHGETPROMOTIONSRESPONSE,
    serialized_options=b'\202\323\344\223\0026\0224/v1/{parent=projects/*/models/*}/promotions:batchGet',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListEvents',
    full_name='continual.rpc.management.v1.ManagementAPI.ListEvents',
    index=64,
    containing_service=None,
    input_type=_LISTEVENTSREQUEST,
    output_type=_LISTEVENTSRESPONSE,
    serialized_options=b'\202\323\344\223\002 \022\036/v1/{parent=projects/*}:events',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetEvent',
    full_name='continual.rpc.management.v1.ManagementAPI.GetEvent',
    index=65,
    containing_service=None,
    input_type=_GETEVENTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._EVENT,
    serialized_options=b'\202\323\344\223\002 \022\036/v1/{name=projects/*/events/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetEvents',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetEvents',
    index=66,
    containing_service=None,
    input_type=_BATCHGETEVENTSREQUEST,
    output_type=_BATCHGETEVENTSRESPONSE,
    serialized_options=b'\202\323\344\223\002)\022\'/v1/{parent=projects/*}/events:batchGet',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateEvent',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateEvent',
    index=67,
    containing_service=None,
    input_type=_CREATEEVENTREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetSubscription',
    full_name='continual.rpc.management.v1.ManagementAPI.GetSubscription',
    index=68,
    containing_service=None,
    input_type=_GETSUBSCRIPTIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._SUBSCRIPTION,
    serialized_options=b'\202\323\344\223\002\'\022%/v1/{name=organizations/subscription}\322E\014Organization\312E\014subscription\332E\030name:name:$/subscription',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetBillingAccount',
    full_name='continual.rpc.management.v1.ManagementAPI.GetBillingAccount',
    index=69,
    containing_service=None,
    input_type=_GETBILLINGACCOUNTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._BILLINGACCOUNT,
    serialized_options=b'\202\323\344\223\002+\022)/v1/{name=organizations/*/billingAccount}\322E\014Organization\312E\016billingAccount\332E\032name:name:$/billingAccount',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetUsage',
    full_name='continual.rpc.management.v1.ManagementAPI.GetUsage',
    index=70,
    containing_service=None,
    input_type=_GETUSAGEREQUEST,
    output_type=_GETUSAGERESPONSE,
    serialized_options=b'\202\323\344\223\002$\"\"/v1/{parent=organizations/*}:usage\322E\014Organization\312E\005usage\332E\013name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListInvoices',
    full_name='continual.rpc.management.v1.ManagementAPI.ListInvoices',
    index=71,
    containing_service=None,
    input_type=_LISTINVOICESREQUEST,
    output_type=_LISTINVOICESRESPONSE,
    serialized_options=b'\202\323\344\223\002\'\022%/v1/{parent=organizations/*}/invoices\322E\014Organization\312E\010invoices\332E\013name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateBillingAccount',
    full_name='continual.rpc.management.v1.ManagementAPI.UpdateBillingAccount',
    index=72,
    containing_service=None,
    input_type=_UPDATEBILLINGACCOUNTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._BILLINGACCOUNT,
    serialized_options=b'\202\323\344\223\002?2,/v1/{billing_account.name=**/billingAccount}:\017billing_account',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateSubscription',
    full_name='continual.rpc.management.v1.ManagementAPI.UpdateSubscription',
    index=73,
    containing_service=None,
    input_type=_UPDATESUBSCRIPTIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._SUBSCRIPTION,
    serialized_options=b'\202\323\344\223\00272\'/v1/{subscription.name=**/subscription}:\014subscription',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListPrices',
    full_name='continual.rpc.management.v1.ManagementAPI.ListPrices',
    index=74,
    containing_service=None,
    input_type=_LISTPRICESREQUEST,
    output_type=_LISTPRICESRESPONSE,
    serialized_options=b'\202\323\344\223\002%\022#/v1/{parent=organizations/*}/prices\322E\014Organization\312E\006prices\332E\013name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetPrice',
    full_name='continual.rpc.management.v1.ManagementAPI.GetPrice',
    index=75,
    containing_service=None,
    input_type=_GETPRICEREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PRICE,
    serialized_options=b'\202\323\344\223\002%\022#/v1/{name=organizations/*/prices/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetPaymentMethod',
    full_name='continual.rpc.management.v1.ManagementAPI.GetPaymentMethod',
    index=76,
    containing_service=None,
    input_type=_GETPAYMENTMETHODREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PAYMENTMETHOD,
    serialized_options=b'\202\323\344\223\002-\022+/v1/{name=organizations/*/paymentMethods/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetUpcomingInvoice',
    full_name='continual.rpc.management.v1.ManagementAPI.GetUpcomingInvoice',
    index=77,
    containing_service=None,
    input_type=_GETUPCOMINGINVOICEREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._INVOICE,
    serialized_options=b'\202\323\344\223\002,\022*/v1/{name=organizations/*/upcomingInvoice}\322E\014Organization\312E\017upcomingInvoice\332E\033name:name:$/upcomingInvoice',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateModel',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateModel',
    index=78,
    containing_service=None,
    input_type=_CREATEMODELREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._MODEL,
    serialized_options=b'\202\323\344\223\002 \022\036/v1/{name=projects/*/models/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetModel',
    full_name='continual.rpc.management.v1.ManagementAPI.GetModel',
    index=79,
    containing_service=None,
    input_type=_GETMODELREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._MODEL,
    serialized_options=b'\202\323\344\223\002 \022\036/v1/{name=projects/*/models/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetModels',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetModels',
    index=80,
    containing_service=None,
    input_type=_BATCHGETMODELSREQUEST,
    output_type=_BATCHGETMODELSRESPONSE,
    serialized_options=b'\202\323\344\223\002)\022\'/v1/{parent=projects/*}/models:batchGet',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListModels',
    full_name='continual.rpc.management.v1.ManagementAPI.ListModels',
    index=81,
    containing_service=None,
    input_type=_LISTMODELSREQUEST,
    output_type=_LISTMODELSRESPONSE,
    serialized_options=b'\202\323\344\223\002 \022\036/v1/{parent=projects/*}/models\322E\023Project,Environment\312E\006models\332E\027name:parent;name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteModel',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteModel',
    index=82,
    containing_service=None,
    input_type=_DELETEMODELREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=b'\202\323\344\223\002 *\036/v1/{name=projects/*/models/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateModelVersion',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateModelVersion',
    index=83,
    containing_service=None,
    input_type=_CREATEMODELVERSIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._MODELVERSION,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetModelVersion',
    full_name='continual.rpc.management.v1.ManagementAPI.GetModelVersion',
    index=84,
    containing_service=None,
    input_type=_GETMODELVERSIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._MODELVERSION,
    serialized_options=b'\202\323\344\223\002+\022)/v1/{name=projects/*/models/*/versions/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetModelVersions',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetModelVersions',
    index=85,
    containing_service=None,
    input_type=_BATCHGETMODELVERSIONSREQUEST,
    output_type=_BATCHGETMODELVERSIONSRESPONSE,
    serialized_options=b'\202\323\344\223\0024\0222/v1/{parent=projects/*/models/*}/versions:batchGet',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListModelVersions',
    full_name='continual.rpc.management.v1.ManagementAPI.ListModelVersions',
    index=86,
    containing_service=None,
    input_type=_LISTMODELVERSIONSREQUEST,
    output_type=_LISTMODELVERSIONSRESPONSE,
    serialized_options=b'\202\323\344\223\002+\022)/v1/{parent=projects/*/models/*}/versions\312E\rmodelVersions\322E\031Model,Project,Environment\340E\000\332E9name:parent;name:parent:$/models/-;name:parent:$/models/-',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateBatchPrediction',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateBatchPrediction',
    index=87,
    containing_service=None,
    input_type=_CREATEBATCHPREDICTIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._BATCHPREDICTION,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetBatchPrediction',
    full_name='continual.rpc.management.v1.ManagementAPI.GetBatchPrediction',
    index=88,
    containing_service=None,
    input_type=_BATCHPREDICTIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._BATCHPREDICTION,
    serialized_options=b'\202\323\344\223\0023\0221/v1/{name=projects/*/models/*/batchPredictions/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetBatchPredictions',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetBatchPredictions',
    index=89,
    containing_service=None,
    input_type=_BATCHGETBATCHPREDICTIONSREQUEST,
    output_type=_BATCHGETBATCHPREDICTIONSRESPONSE,
    serialized_options=b'\202\323\344\223\002<\022:/v1/{parent=projects/*/models/*}/batchPredictions:batchGet',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListBatchPredictions',
    full_name='continual.rpc.management.v1.ManagementAPI.ListBatchPredictions',
    index=90,
    containing_service=None,
    input_type=_LISTBATCHPREDICTIONSREQUEST,
    output_type=_LISTBATCHPREDICTIONSRESPONSE,
    serialized_options=b'\202\323\344\223\0023\0221/v1/{parent=projects/*/models/*}/batchPredictions\312E\020batchPredictions\322E\031Project,Model,Environment\340E\000\332E#name:parent;name:parent;name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateExperiment',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateExperiment',
    index=91,
    containing_service=None,
    input_type=_CREATEEXPERIMENTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._EXPERIMENT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetExperiment',
    full_name='continual.rpc.management.v1.ManagementAPI.GetExperiment',
    index=92,
    containing_service=None,
    input_type=_GETEXPERIMENTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._EXPERIMENT,
    serialized_options=b'\202\323\344\223\0029\0227/v1/{name=projects/*/models/*/versions/*/experiments/*}',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetExperiments',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetExperiments',
    index=93,
    containing_service=None,
    input_type=_BATCHGETEXPERIMENTSREQUEST,
    output_type=_BATCHGETEXPERIMENTSRESPONSE,
    serialized_options=b'\202\323\344\223\002B\022@/v1/{parent=projects/*/models/*/versions/*}/experiments:batchGet',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListExperiments',
    full_name='continual.rpc.management.v1.ManagementAPI.ListExperiments',
    index=94,
    containing_service=None,
    input_type=_LISTEXPERIMENTSREQUEST,
    output_type=_LISTEXPERIMENTSRESPONSE,
    serialized_options=b'\202\323\344\223\0029\0227/v1/{parent=projects/*/models/*/versions/*}/experiments\312E\013experiments\322E\014ModelVersion\340E\000\332E\013name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetEnvironment',
    full_name='continual.rpc.management.v1.ManagementAPI.GetEnvironment',
    index=95,
    containing_service=None,
    input_type=_GETENVIRONMENTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ENVIRONMENT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetEnvironments',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetEnvironments',
    index=96,
    containing_service=None,
    input_type=_BATCHGETENVIRONMENTSREQUEST,
    output_type=_BATCHGETENVIRONMENTSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListEnvironments',
    full_name='continual.rpc.management.v1.ManagementAPI.ListEnvironments',
    index=97,
    containing_service=None,
    input_type=_LISTENVIRONMENTSREQUEST,
    output_type=_LISTENVIRONMENTSRESPONSE,
    serialized_options=b'\322E\007Project\312E\014environments\332E\013name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateEnvironment',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateEnvironment',
    index=98,
    containing_service=None,
    input_type=_CREATEENVIRONMENTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ENVIRONMENT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateEnvironment',
    full_name='continual.rpc.management.v1.ManagementAPI.UpdateEnvironment',
    index=99,
    containing_service=None,
    input_type=_UPDATEENVIRONMENTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ENVIRONMENT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteEnvironment',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteEnvironment',
    index=100,
    containing_service=None,
    input_type=_DELETEENVIRONMENTREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetSystemMessage',
    full_name='continual.rpc.management.v1.ManagementAPI.GetSystemMessage',
    index=101,
    containing_service=None,
    input_type=_GETSYSTEMMESSAGEREQUEST,
    output_type=_GETSYSTEMMESSAGERESPONSE,
    serialized_options=b'\202\323\344\223\002\023\"\021/v1/systemMessage',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetFeatureFlag',
    full_name='continual.rpc.management.v1.ManagementAPI.GetFeatureFlag',
    index=102,
    containing_service=None,
    input_type=_GETFEATUREFLAGREQUEST,
    output_type=_GETFEATUREFLAGRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetOrganizationFeatureFlags',
    full_name='continual.rpc.management.v1.ManagementAPI.GetOrganizationFeatureFlags',
    index=103,
    containing_service=None,
    input_type=_GETORGANIZATIONFEATUREFLAGSREQUEST,
    output_type=_GETORGANIZATIONFEATUREFLAGSRESPONSE,
    serialized_options=b'\322E\014Organization\312E\014featureFlags\332E\rname:resource',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateArtifact',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateArtifact',
    index=104,
    containing_service=None,
    input_type=_CREATEARTIFACTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ARTIFACT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetArtifact',
    full_name='continual.rpc.management.v1.ManagementAPI.GetArtifact',
    index=105,
    containing_service=None,
    input_type=_GETARTIFACTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._ARTIFACT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetArtifacts',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetArtifacts',
    index=106,
    containing_service=None,
    input_type=_BATCHGETARTIFACTSREQUEST,
    output_type=_BATCHGETARTIFACTSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteArtifact',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteArtifact',
    index=107,
    containing_service=None,
    input_type=_DELETEARTIFACTREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GenerateArtifactUploadURL',
    full_name='continual.rpc.management.v1.ManagementAPI.GenerateArtifactUploadURL',
    index=108,
    containing_service=None,
    input_type=_GENERATEARTIFACTUPLOADURLREQUEST,
    output_type=_GENERATEARTIFACTUPLOADURLRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListArtifacts',
    full_name='continual.rpc.management.v1.ManagementAPI.ListArtifacts',
    index=109,
    containing_service=None,
    input_type=_LISTARTIFACTSREQUEST,
    output_type=_LISTARTIFACTSRESPONSE,
    serialized_options=b'\312E\tartifacts\322E\034ModelVersion,BatchPrediction\340E\000\332E\027name:parent;name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListUserIntegrationProviders',
    full_name='continual.rpc.management.v1.ManagementAPI.ListUserIntegrationProviders',
    index=110,
    containing_service=None,
    input_type=_LISTUSERINTEGRATIONPROVIDERSREQUEST,
    output_type=_LISTUSERINTEGRATIONPROVIDERSRESPONSE,
    serialized_options=b'\312E\tproviders',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListUserIntegrations',
    full_name='continual.rpc.management.v1.ManagementAPI.ListUserIntegrations',
    index=111,
    containing_service=None,
    input_type=_LISTUSERINTEGRATIONSREQUEST,
    output_type=_LISTUSERINTEGRATIONSRESPONSE,
    serialized_options=b'\312E\014integrations\322E\004User',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DisconnectUserIntegration',
    full_name='continual.rpc.management.v1.ManagementAPI.DisconnectUserIntegration',
    index=112,
    containing_service=None,
    input_type=_DISCONNECTUSERINTEGRATIONREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GenerateSsoAdminPortalURL',
    full_name='continual.rpc.management.v1.ManagementAPI.GenerateSsoAdminPortalURL',
    index=113,
    containing_service=None,
    input_type=_GENERATESSOADMINPORTALURLREQUEST,
    output_type=_GENERATESSOADMINPORTALURLRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GenerateDSyncAdminPortalURL',
    full_name='continual.rpc.management.v1.ManagementAPI.GenerateDSyncAdminPortalURL',
    index=114,
    containing_service=None,
    input_type=_GENERATEDSYNCADMINPORTALURLREQUEST,
    output_type=_GENERATEDSYNCADMINPORTALURLRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ExchangeToken',
    full_name='continual.rpc.management.v1.ManagementAPI.ExchangeToken',
    index=115,
    containing_service=None,
    input_type=_EXCHANGETOKENREQUEST,
    output_type=_EXCHANGETOKENRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListGitProviderNamespaces',
    full_name='continual.rpc.management.v1.ManagementAPI.ListGitProviderNamespaces',
    index=116,
    containing_service=None,
    input_type=_LISTGITPROVIDERNAMESPACESREQUEST,
    output_type=_LISTGITPROVIDERNAMESPACESRESPONSE,
    serialized_options=b'\312E\nnamespaces',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListGitProviderRepositories',
    full_name='continual.rpc.management.v1.ManagementAPI.ListGitProviderRepositories',
    index=117,
    containing_service=None,
    input_type=_LISTGITPROVIDERREPOSITORIESREQUEST,
    output_type=_LISTGITPROVIDERREPOSITORIESRESPONSE,
    serialized_options=b'\312E\014repositories',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ConnectProjectToGitRepository',
    full_name='continual.rpc.management.v1.ManagementAPI.ConnectProjectToGitRepository',
    index=118,
    containing_service=None,
    input_type=_CONNECTPROJECTTOGITREPOSITORYREQUEST,
    output_type=_CONNECTPROJECTTOGITREPOSITORYRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetProjectGitIntegration',
    full_name='continual.rpc.management.v1.ManagementAPI.GetProjectGitIntegration',
    index=119,
    containing_service=None,
    input_type=_GETPROJECTGITINTEGRATIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._PROJECTGITINTEGRATION,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='SeedProjectGitRepository',
    full_name='continual.rpc.management.v1.ManagementAPI.SeedProjectGitRepository',
    index=120,
    containing_service=None,
    input_type=_SEEDPROJECTGITREPOSITORYREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetRun',
    full_name='continual.rpc.management.v1.ManagementAPI.GetRun',
    index=121,
    containing_service=None,
    input_type=_GETRUNREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._RUN,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetRuns',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetRuns',
    index=122,
    containing_service=None,
    input_type=_BATCHGETRUNSREQUEST,
    output_type=_BATCHGETRUNSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListRuns',
    full_name='continual.rpc.management.v1.ManagementAPI.ListRuns',
    index=123,
    containing_service=None,
    input_type=_LISTRUNSREQUEST,
    output_type=_LISTRUNSRESPONSE,
    serialized_options=b'\312E\004runs\322E\013Environment\340E\000\332E\013name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateRun',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateRun',
    index=124,
    containing_service=None,
    input_type=_CREATERUNREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._RUN,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateRun',
    full_name='continual.rpc.management.v1.ManagementAPI.UpdateRun',
    index=125,
    containing_service=None,
    input_type=_UPDATERUNREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._RUN,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteRun',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteRun',
    index=126,
    containing_service=None,
    input_type=_DELETERUNREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetRunSummary',
    full_name='continual.rpc.management.v1.ManagementAPI.GetRunSummary',
    index=127,
    containing_service=None,
    input_type=_GETRUNSUMMARYREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._RUNSUMMARY,
    serialized_options=b'\322E\003Run\312E\007summary\332E\tname:name',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetRunCheckSummary',
    full_name='continual.rpc.management.v1.ManagementAPI.GetRunCheckSummary',
    index=128,
    containing_service=None,
    input_type=_GETRUNCHECKSUMMARYREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._RUNCHECKSUMMARY,
    serialized_options=b'\322E\003Run\312E\rresultSummary\332E\tname:name',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='LogRunHeartbeat',
    full_name='continual.rpc.management.v1.ManagementAPI.LogRunHeartbeat',
    index=129,
    containing_service=None,
    input_type=_LOGRUNHEARTBEATREQUEST,
    output_type=google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetMetric',
    full_name='continual.rpc.management.v1.ManagementAPI.GetMetric',
    index=130,
    containing_service=None,
    input_type=_GETMETRICREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._METRIC,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetMetrics',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetMetrics',
    index=131,
    containing_service=None,
    input_type=_BATCHGETMETRICSREQUEST,
    output_type=_BATCHGETMETRICSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListMetrics',
    full_name='continual.rpc.management.v1.ManagementAPI.ListMetrics',
    index=132,
    containing_service=None,
    input_type=_LISTMETRICSREQUEST,
    output_type=_LISTMETRICSRESPONSE,
    serialized_options=b'\322E\027ModelVersion,Experiment\312E\007metrics\332E\013name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateMetric',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateMetric',
    index=133,
    containing_service=None,
    input_type=_CREATEMETRICREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._METRIC,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='LogMetric',
    full_name='continual.rpc.management.v1.ManagementAPI.LogMetric',
    index=134,
    containing_service=None,
    input_type=_LOGMETRICREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateMetric',
    full_name='continual.rpc.management.v1.ManagementAPI.UpdateMetric',
    index=135,
    containing_service=None,
    input_type=_UPDATEMETRICREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._METRIC,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetMetadata',
    full_name='continual.rpc.management.v1.ManagementAPI.GetMetadata',
    index=136,
    containing_service=None,
    input_type=_GETMETADATAREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._METADATA,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetMetadata',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetMetadata',
    index=137,
    containing_service=None,
    input_type=_BATCHGETMETADATAREQUEST,
    output_type=_BATCHGETMETADATARESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListMetadata',
    full_name='continual.rpc.management.v1.ManagementAPI.ListMetadata',
    index=138,
    containing_service=None,
    input_type=_LISTMETADATAREQUEST,
    output_type=_LISTMETADATARESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateMetadata',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateMetadata',
    index=139,
    containing_service=None,
    input_type=_CREATEMETADATAREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._METADATA,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateMetadata',
    full_name='continual.rpc.management.v1.ManagementAPI.UpdateMetadata',
    index=140,
    containing_service=None,
    input_type=_UPDATEMETADATAREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._METADATA,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteMetadata',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteMetadata',
    index=141,
    containing_service=None,
    input_type=_DELETEMETADATAREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetNotificationPolicy',
    full_name='continual.rpc.management.v1.ManagementAPI.GetNotificationPolicy',
    index=142,
    containing_service=None,
    input_type=_GETNOTIFICATIONPOLICYREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._NOTIFICATIONPOLICY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetNotificationPolicies',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetNotificationPolicies',
    index=143,
    containing_service=None,
    input_type=_BATCHGETNOTIFICATIONPOLICIESREQUEST,
    output_type=_BATCHGETNOTIFICATIONPOLICIESRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListNotificationPolicies',
    full_name='continual.rpc.management.v1.ManagementAPI.ListNotificationPolicies',
    index=144,
    containing_service=None,
    input_type=_LISTNOTIFICATIONPOLICIESREQUEST,
    output_type=_LISTNOTIFICATIONPOLICIESRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateNotificationPolicy',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateNotificationPolicy',
    index=145,
    containing_service=None,
    input_type=_CREATENOTIFICATIONPOLICYREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._NOTIFICATIONPOLICY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateNotificationPolicy',
    full_name='continual.rpc.management.v1.ManagementAPI.UpdateNotificationPolicy',
    index=146,
    containing_service=None,
    input_type=_UPDATENOTIFICATIONPOLICYREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._NOTIFICATIONPOLICY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteNotificationPolicy',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteNotificationPolicy',
    index=147,
    containing_service=None,
    input_type=_DELETENOTIFICATIONPOLICYREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetTag',
    full_name='continual.rpc.management.v1.ManagementAPI.GetTag',
    index=148,
    containing_service=None,
    input_type=_GETTAGREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._TAG,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetTags',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetTags',
    index=149,
    containing_service=None,
    input_type=_BATCHGETTAGSREQUEST,
    output_type=_BATCHGETTAGSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListTags',
    full_name='continual.rpc.management.v1.ManagementAPI.ListTags',
    index=150,
    containing_service=None,
    input_type=_LISTTAGSREQUEST,
    output_type=_LISTTAGSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateTag',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateTag',
    index=151,
    containing_service=None,
    input_type=_CREATETAGREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._TAG,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateTag',
    full_name='continual.rpc.management.v1.ManagementAPI.UpdateTag',
    index=152,
    containing_service=None,
    input_type=_UPDATETAGREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._TAG,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteTag',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteTag',
    index=153,
    containing_service=None,
    input_type=_DELETETAGREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetResourceCheck',
    full_name='continual.rpc.management.v1.ManagementAPI.GetResourceCheck',
    index=154,
    containing_service=None,
    input_type=_GETRESOURCECHECKREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._RESOURCECHECK,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListResourceChecks',
    full_name='continual.rpc.management.v1.ManagementAPI.ListResourceChecks',
    index=155,
    containing_service=None,
    input_type=_LISTRESOURCECHECKSREQUEST,
    output_type=_LISTRESOURCECHECKSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateResourceCheck',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateResourceCheck',
    index=156,
    containing_service=None,
    input_type=_CREATERESOURCECHECKREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._RESOURCECHECK,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetDataProfile',
    full_name='continual.rpc.management.v1.ManagementAPI.GetDataProfile',
    index=157,
    containing_service=None,
    input_type=_GETDATAPROFILEREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATAPROFILE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListDataProfiles',
    full_name='continual.rpc.management.v1.ManagementAPI.ListDataProfiles',
    index=158,
    containing_service=None,
    input_type=_LISTDATAPROFILESREQUEST,
    output_type=_LISTDATAPROFILESRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateDataProfile',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateDataProfile',
    index=159,
    containing_service=None,
    input_type=_CREATEDATAPROFILEREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATAPROFILE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='UpdateDataProfile',
    full_name='continual.rpc.management.v1.ManagementAPI.UpdateDataProfile',
    index=160,
    containing_service=None,
    input_type=_UPDATEDATAPROFILEREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATAPROFILE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateDataset',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateDataset',
    index=161,
    containing_service=None,
    input_type=_CREATEDATASETREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATASET,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetDataset',
    full_name='continual.rpc.management.v1.ManagementAPI.GetDataset',
    index=162,
    containing_service=None,
    input_type=_GETDATASETREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATASET,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetDatasets',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetDatasets',
    index=163,
    containing_service=None,
    input_type=_BATCHGETDATASETSREQUEST,
    output_type=_BATCHGETDATASETSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListDatasets',
    full_name='continual.rpc.management.v1.ManagementAPI.ListDatasets',
    index=164,
    containing_service=None,
    input_type=_LISTDATASETSREQUEST,
    output_type=_LISTDATASETSRESPONSE,
    serialized_options=b'\322E\023Project,Environment\312E\010datasets\332E\027name:parent;name:parent',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteDataset',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteDataset',
    index=165,
    containing_service=None,
    input_type=_DELETEDATASETREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateDatasetVersion',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateDatasetVersion',
    index=166,
    containing_service=None,
    input_type=_CREATEDATASETVERSIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATASETVERSION,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetDatasetVersion',
    full_name='continual.rpc.management.v1.ManagementAPI.GetDatasetVersion',
    index=167,
    containing_service=None,
    input_type=_GETDATASETVERSIONREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATASETVERSION,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='BatchGetDatasetVersions',
    full_name='continual.rpc.management.v1.ManagementAPI.BatchGetDatasetVersions',
    index=168,
    containing_service=None,
    input_type=_BATCHGETDATASETVERSIONSREQUEST,
    output_type=_BATCHGETDATASETVERSIONSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListDatasetVersions',
    full_name='continual.rpc.management.v1.ManagementAPI.ListDatasetVersions',
    index=169,
    containing_service=None,
    input_type=_LISTDATASETVERSIONSREQUEST,
    output_type=_LISTDATASETVERSIONSRESPONSE,
    serialized_options=b'\312E\017datasetVersions\322E\033Dataset,Project,Environment\340E\000\332E=name:parent;name:parent:$/datasets/-;name:parent:$/datasets/-',
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CreateDatasetVersionAssignment',
    full_name='continual.rpc.management.v1.ManagementAPI.CreateDatasetVersionAssignment',
    index=170,
    containing_service=None,
    input_type=_CREATEDATASETVERSIONASSIGNMENTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATASETVERSIONASSIGNMENT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='GetDatasetVersionAssignment',
    full_name='continual.rpc.management.v1.ManagementAPI.GetDatasetVersionAssignment',
    index=171,
    containing_service=None,
    input_type=_GETDATASETVERSIONASSIGNMENTREQUEST,
    output_type=continual_dot_rpc_dot_management_dot_v1_dot_management__types__pb2._DATASETVERSIONASSIGNMENT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteDatasetVersionAssignment',
    full_name='continual.rpc.management.v1.ManagementAPI.DeleteDatasetVersionAssignment',
    index=172,
    containing_service=None,
    input_type=_DELETEDATASETVERSIONASSIGNMENTREQUEST,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ListDatasetVersionAssignments',
    full_name='continual.rpc.management.v1.ManagementAPI.ListDatasetVersionAssignments',
    index=173,
    containing_service=None,
    input_type=_LISTDATASETVERSIONASSIGNMENTSREQUEST,
    output_type=_LISTDATASETVERSIONASSIGNMENTSRESPONSE,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_MANAGEMENTAPI)

DESCRIPTOR.services_by_name['ManagementAPI'] = _MANAGEMENTAPI

# @@protoc_insertion_point(module_scope)
